/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

//==============================================================================
//
//  $Id: BasisFunctions.h,v 1.5 2006/06/29 15:06:39 jochen Exp $
//
//==============================================================================
//-----------------------------------------------------------------------------
template <typename TYPE, DIMENSION dim_, int numberGaussianPoints_plus1, BasisFunctionType BFnTYPE, 
          int numberSet1_, int numberSet2_ = 0, int numberOfBasisFunctionSets = (0 < numberSet2_)? 2:1>
class BasisFunctionAndValueClass;  
//-----------------------------------------------------------------------------
template <typename TYPE, DIMENSION dim_, int numberGaussianPoints_plus1, 
          int numberSet1_, int numberSet2_, int numberOfBasisFunctionSets>
class BasisFunctionAndValueClass <TYPE,dim_,numberGaussianPoints_plus1,oneDimensional,
                                  numberSet1_, numberSet2_, numberOfBasisFunctionSets>
   {
     protected:
      const int numberOfBasisFunctionSets_, sizeOfValues;
      int numberSet[numberOfBasisFunctionSets];
      typedef std::vector<Base**> BasisFnVector[numberOfBasisFunctionSets];
      BasisFnVector Basis_Fn;
      TYPE* basisFunctionsAtGaussianPoints, ***derivativeValues;
     public:
      inline BasisFunctionAndValueClass (); 
      inline ~BasisFunctionAndValueClass ( ); 
      template <typename baseTYPE>
      inline void eval(baseTYPE** gaussianPoints_);
      template <class Expression, typename baseTYPE>
      inline void 
       Set_BFN (BasisFunctionSet setNumber,  const FunctionExpr<Expression>& a_, baseTYPE** gaussianPoints_); 
    /** 
     Methods for returning the amounts of basis functions
     */ 
      inline unsigned int getNumberOfBasisFunctions(BasisFunctionSet setNumber) const;
      int size_Set1() const { return numberSet[BasisSet_1];} 
      int size_Set2() const { return numberSet[BasisSet_2];} 
      int numberFunctions_Set1() const { return numberSet[BasisSet_1];} 
      int numberFunctions_Set2() const { return numberSet[BasisSet_2];} 
      inline TYPE*** getDerivativeValues() const ;
      inline const TYPE* getBasisFunctionsAtGaussianPoints_(int basisFunctionSetNumber,int gaussLevel,int actualBasisFunction)const;

      template <BasisFunctionSet basisFunctionSetNumber,unsigned int functionId,unsigned int iterLength>
      inline const TYPE getBasisFunctionsAtGaussianPointsFn(const int (&iterator)[iterLength]) const ;

      template <BasisFunctionSet basisFunctionSetNumber, unsigned int functionId, what derivativeType, unsigned int iterLength>
      inline const TYPE getElementMappingDerivativeValues(const int (&iterator)[iterLength]) const ;

      template <class Element>
      inline void precomputeDerivatives(Element& element, BasisFunctionSet number);
      template <class Element>
      inline void precomputeDerivativesAdditional(Element& element, BasisFunctionSet number);
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE, DIMENSION dim_, int numberGaussianPoints_plus1,
          int numberSet1_, int numberSet2_, int numberOfBasisFunctionSets>
class BasisFunctionAndValueClass <TYPE,dim_,numberGaussianPoints_plus1,vectorBasisFunctions,
                                  numberSet1_, numberSet2_, numberOfBasisFunctionSets>
   {
     protected:
      const int numberOfBasisFunctionSets_, sizeOfValues;
      int numberSet[numberOfBasisFunctionSets];
      typedef std::vector<Base**> BasisFnVector[numberOfBasisFunctionSets];
      BasisFnVector Basis_Fn;
      std::vector<std::pair<unsigned int, unsigned int > > edges;
      TYPE* basisFunctionsAtGaussianPoints, ***derivativeValues, **value;
      TYPE* vector_basisFunctionsAtGaussianPoints, ***vector_derivativeValues;
     public:
      inline BasisFunctionAndValueClass (); 
      inline ~BasisFunctionAndValueClass ( ); 
      template <typename baseTYPE>
      inline void eval(baseTYPE** gaussianPoints_);
      template <class Expr1, class Expr2, typename baseTYPE>
      inline void 
      Set_BFN (BasisFunctionSet setNumber, const FunctionExprVector2D<Expr1,Expr2>& expr,
               const unsigned cnr_1, const unsigned cnr_2, 
               baseTYPE **_gaussianPoints);
      template <class Expr1, class Expr2, class Expr3, typename baseTYPE>
      inline void 
      Set_BFN (BasisFunctionSet setNumber, const FunctionExprVector3D<Expr1,Expr2,Expr3>& expr, 
               const unsigned cnr_1, const unsigned cnr_2, 
               baseTYPE **_gaussianPoints);
    /** 
     Methods for returning the amounts of basis functions
     */ 
      inline unsigned int getNumberOfBasisFunctions(BasisFunctionSet setNumber) const;
      int size_Set1() const { return numberSet[BasisSet_1];} 
      int size_Set2() const { return numberSet[BasisSet_2];} 
      int numberFunctions_Set1() const { return numberSet[BasisSet_1]*dim_;} 
      int numberFunctions_Set2() const { return numberSet[BasisSet_2]*dim_;} 
      inline TYPE*** getDerivativeValues() const ;
      inline const TYPE& getElementMappingDerivativeValues(const int derivPos, const int iterPos, const int gaussLevel) const ;

      inline const TYPE* getBasisFunctionsAtGaussianPoints_(int basisFunctionSetNumber,int gaussLevel,int actualBasisFunction)const;

      template <BasisFunctionSet basisFunctionSetNumber,unsigned int functionId,unsigned int iterLength>
      inline const TYPE getBasisFunctionsAtGaussianPointsFn(const int (&iterator)[iterLength]) const ;

      template <BasisFunctionSet basisFunctionSetNumber, unsigned int functionId, what derivativeType, unsigned int iterLength>
      inline const TYPE getElementMappingDerivativeValues(const int (&iterator)[iterLength]) const ;

      template <BasisFunctionSet basisFunctionSetNumber,SpaceDirection direction,unsigned int functionId,unsigned int iterLength>
      inline const TYPE getBasisFunctionsAtGaussianPointsFn_Vector(const int (&iterator)[iterLength]) const ;

      template <BasisFunctionSet basisFunctionSetNumber,SpaceDirection direction,
                unsigned int functionId, what derivativeType, unsigned int iterLength>
      inline const TYPE getElementMappingDerivativeValues_Vector(const int (&iterator)[iterLength]) const ;
      template <class Element>
      inline void precomputeDerivatives(Element& element, BasisFunctionSet number);
      template <class Element>
      inline void precomputeDerivativesAdditional(Element& element, BasisFunctionSet number);
   } ;
//==============================================================================

