/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

//==============================================================================
//
//  $Id: CT_Manip_ET.h,v 1.11 2006/07/14 10:51:29 jochen Exp $
//
//==============================================================================
//-----------------------------------------------------------------------------
#define Define_Integrand(A, var)  \
		        __typeof__(A) var (A)
//-----------------------------------------------------------------------------
template <class A>
inline void 
 print(const A& a); 
//-----------------------------------------------------------------------------
// Data structures for rearranging the expressions. This is only applied for 
// the derivatives so far. But there might be a huge potential for optimizing, 
// e.g. the multiplication of a constant with a vector, where we should not
// immediately multiply every compontent with the constant, but only apply the 
// product to the result!
//-----------------------------------------------------------------------------
// Constructor functions to build the corresponding expression objects!
//-----------------------------------------------------------------------------
typedef BasisFunction<BasisSet_1,0> v_;
typedef BasisFunction<BasisSet_1,1> w_;
typedef BasisFunction<BasisSet_2,0> p_;
//-----------------------------------------------------------------------------
  template <class T>
  inline VertexVector<T> 
   vec(const T(&d));
//-----------------------------------------------------------------------------
  template <class T>
  inline FUNC1_<T> 
   func(const typename FunctionInterfaces<T>::functionInterfaceOne& f_);
//-----------------------------------------------------------------------------
  template <class T>
  inline FUNC2_<T> 
   func(const typename FunctionInterfaces<T>::functionInterfaceTwo& f_);
//-----------------------------------------------------------------------------
 template <class T>
 inline FUNC3_<T> 
  func(const typename FunctionInterfaces<T>::functionInterfaceThree& f_);
//-----------------------------------------------------------------------------
 template <class A> 
 inline Conjugate<A> 
  conj(const BasisFunctionExpr<A>& a);
//-----------------------------------------------------------------------------
 template <class A, class B> inline Add_<A,B>
 operator + (const BasisFunctionExpr<A>& a, const BasisFunctionExpr<B>& b);
//-----------------------------------------------------------------------------
 template <class A, class B> inline Sub_<A,B>
 operator - (const BasisFunctionExpr<A>& a, const BasisFunctionExpr<B>& b);
//-----------------------------------------------------------------------------
 template <class A> inline Min_<A>
 operator - (const BasisFunctionExpr<A>& a);
//-----------------------------------------------------------------------------
 template <class A, class B> inline Mult_<A,B>
 operator * (const BasisFunctionExpr<A>& a, const BasisFunctionExpr<B>& b);
//-----------------------------------------------------------------------------
 template <class A, class B> inline Div_<A,B>
 operator / (const BasisFunctionExpr<A>& a, const BasisFunctionExpr<B>& b);
//-----------------------------------------------------------------------------
 template <class A> 
 inline SQRT_<A>
  Sqrt (const BasisFunctionExpr<A>& a);
//-----------------------------------------------------------------------------
 template <class A, typename eTYPE> 
 inline POW_<A,eTYPE>
  Pow (const BasisFunctionExpr<A>& a, eTYPE exponent);
//-----------------------------------------------------------------------------
 template <class A> 
 inline EXP_<A>
  Exp (const BasisFunctionExpr<A>& a);
//-----------------------------------------------------------------------------
 template<class A> 
 inline Deri_<A,dx> 
  d_dx(const BasisFunctionExpr<A>& a);
//-----------------------------------------------------------------------------
 template<class A> 
 inline Deri_<A,dy>
  d_dy(const BasisFunctionExpr<A>& a);
//-----------------------------------------------------------------------------
 template<class A> 
 inline Deri_<A,dz> 
  d_dz(const BasisFunctionExpr<A>& a);
//-----------------------------------------------------------------------------
/* A vector is a type list of its components, namely a vector of two  
 * components or a component and vector (which is a component itself)!
 * The evaluation of the operator '&' is from left to right (hopefully),
 * and thereby, we build a template type of the structure
 *   VECTOR<VECTOR<VECTOR<....>,.>,.> 
 * This means, we can only access the VECTOR from the end to the top! 
 * A vector (so far) is always a line or a column, not differentiated 
 * between them! 
*/
template <class A>
struct BasisFunctionVectorExpr{};
//-----------------------------------------------------------------------------
template <class A, class B>
class VECTOR : public BasisFunctionVectorExpr<VECTOR<A,B> >
   {
     private:
      const A& a;
      const B& b;
     public:
      inline VECTOR(const A& a_, const B& b_) : a(a_), b(b_) {}
       enum {vContainment = _MAX_(A::vContainment,B::vContainment),
             wContainment = _MAX_(A::wContainment,B::wContainment), 
             pContainment = _MAX_(A::pContainment,B::pContainment)} ;
       enum {length = A::length+1, size = 1};
       inline void 
        print(std::ostream &os) const;
        inline const A& getHeadOfVector() const { return a;}
        const A& getA() const {return a;}
        const B& getB() const {return b;}
        template<class Element, unsigned int iterLength>
        inline typename Element::TYPE 
         evalLastComponent(const Element& element, const int (&iterator)[iterLength]) const;
};
//-----------------------------------------------------------------------------
template <class A1, class A2, class B>
class VECTOR<VECTOR<A1,A2>,B > : public BasisFunctionVectorExpr<VECTOR<VECTOR<A1,A2>,B > >
   {
     private:
      const VECTOR<A1,A2>& a;
      const B& b;
     public:
     inline VECTOR(const VECTOR<A1,A2>& a_, const B& b_) : a(a_), b(b_) {}
     enum {vContainment = _MAX_((VECTOR<A1,A2>::vContainment),B::vContainment),
           wContainment = _MAX_((VECTOR<A1,A2>::wContainment),B::wContainment), 
           pContainment = _MAX_((VECTOR<A1,A2>::pContainment),B::pContainment)} ;
     enum {length = VECTOR<A1,A2>::length+1, size = 1};
     inline void 
      print(std::ostream &os) const;
     inline const VECTOR<A1,A2>& getHeadOfVector() const { return a;}
     const A1& getA() const {return a.getA();}
     const A2& getB() const {return a.getB();}
     const B&  getC() const {return b;}
     template<class Element, unsigned int iterLength>
     inline typename Element::TYPE 
      evalLastComponent(const Element& element, const int (&iterator)[iterLength]) const;
};
//-----------------------------------------------------------------------------
template <class A, class B>
inline VECTOR<A,B> 
operator &(const BasisFunctionExpr<A>& a, const BasisFunctionExpr<B>& b);
//-----------------------------------------------------------------------------
template <class A, class B>
inline VECTOR<A,B> 
operator &(const BasisFunctionVectorExpr<A>& a, const BasisFunctionExpr<B>& b);
//-----------------------------------------------------------------------------
template <typename TYPE, class B>
class ConstVectorMult_;
//-----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
template<class A, class B, class Element, unsigned int iterLength>
inline typename Element::TYPE 
 evaluate(const BasisFunctionVectorExpr<A>& a, const BasisFunctionVectorExpr<B>& b,
          const Element& element, const int (&iterator)[iterLength]); 
//-----------------------------------------------------------------------------
template<class A, class B, class Element, unsigned int iterLength>
inline typename Element::TYPE 
 evaluate(const BasisFunctionExpr<A>& a, const BasisFunctionExpr<B>& b,
          const Element& element, const int (&iterator)[iterLength]); 
//-----------------------------------------------------------------------------
template <class A, class B>
class VectorVectorMult_ : public BasisFunctionExpr<VectorVectorMult_<A,B> >
   {
     private: 
      const A& a;
      const B& b;
     public:
     inline VectorVectorMult_(const A& a_, const B& b_) : a(a_), b(b_) {}
     enum {vContainment = _MAX_(A::vContainment,B::vContainment),
           wContainment = _MAX_(A::wContainment,B::wContainment), 
           pContainment = _MAX_(A::pContainment,B::pContainment)} ;
     enum {length = A::length, size = 1};
     inline void 
      print(std::ostream &os) const
        {
            os << " ("; a.print(os); os << ") * (" ; b.print(os) ; os << ") ";
        }
     template<class Element, unsigned int iterLength>
     inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]) const
        {
          return evaluate(a,b,element,iterator);
        } 
   };
//-----------------------------------------------------------------------------
template <typename TYPE, class B1, class B2>
class ConstVectorMult_<TYPE,VECTOR<B1,B2> > 
   : public BasisFunctionVectorExpr<ConstVectorMult_<TYPE,VECTOR<B1,B2> > >
   {
     private: 
      TYPE a;
      const VECTOR<B1,B2> & b;
     public:
     inline ConstVectorMult_(const TYPE& a_, const VECTOR<B1,B2>& b_) : a(a_), b(b_) {}
     enum {vContainment = VECTOR<B1,B2>::vContainment,
           wContainment = VECTOR<B1,B2>::wContainment, 
           pContainment = VECTOR<B1,B2>::pContainment} ;
     enum {length = VECTOR<B1,B2>::length, size = 1};
     inline void 
      print(std::ostream &os) const
        {
            os << " " << a << " * (" ; b.print(os) ; os << ") ";
        }
     inline const CMult_<TYPE,B1> getHeadOfVector() const { return a * b.getHeadOfVector();}
     template<class Element, unsigned int iterLength>
     inline typename Element::TYPE 
      evalLastComponent(const Element& element, const int (&iterator)[iterLength]) const
        {
          return a*b.evalLastComponent(element,iterator);
        } 
   };
//-----------------------------------------------------------------------------
template <typename TYPE, class B1, class B2, class C>
class ConstVectorMult_<TYPE,VECTOR<VECTOR<B1,B2>,C> > 
    : public BasisFunctionVectorExpr<ConstVectorMult_<TYPE,VECTOR<VECTOR<B1,B2>,C> > >
   {
     private: 
      TYPE a;
      const VECTOR<VECTOR<B1,B2>,C>& b;
     public:
     inline ConstVectorMult_(const TYPE& a_, const VECTOR<VECTOR<B1,B2>,C>& b_) : a(a_), b(b_) {}
     enum {vContainment = VECTOR<VECTOR<B1,B2>,C>::vContainment,
           wContainment = VECTOR<VECTOR<B1,B2>,C>::wContainment, 
           pContainment = VECTOR<VECTOR<B1,B2>,C>::pContainment} ;
     enum {length = VECTOR<VECTOR<B1,B2>,C>::length, size = 1};
     inline void 
      print(std::ostream &os) const
        {
            os << " " << a << " * (" ; b.print(os) ; os << ") ";
        }
     inline const ConstVectorMult_<TYPE,VECTOR<B1,B2> > getHeadOfVector() const { return a*b.getHeadOfVector();}
     template<class Element, unsigned int iterLength>
     inline typename Element::TYPE 
      evalLastComponent(const Element& element, const int (&iterator)[iterLength]) const
        {
          return a*b.evalLastComponent(element,iterator);
        } 
   };
//-----------------------------------------------------------------------------
template <class A, class B>
inline VectorVectorMult_<A,B>
 operator*(const BasisFunctionVectorExpr<A>& a, const BasisFunctionVectorExpr<B>& b)
   {
     return VectorVectorMult_<A,B>(static_cast<const A&>(a),static_cast<const B&>(b));
   } 
//-----------------------------------------------------------------------------
template <class B>
inline ConstVectorMult_<int,B>
 operator*(int a, const BasisFunctionVectorExpr<B>& b)
   {
     return ConstVectorMult_<int,B>(a,static_cast<const B&>(b));
   } 
//-----------------------------------------------------------------------------
template <class B>
inline ConstVectorMult_<float,B>
 operator*(float a, const BasisFunctionVectorExpr<B>& b)
   {
     return  ConstVectorMult_<float,B>(a,static_cast<const B&>(b));
   } 
//-----------------------------------------------------------------------------
template <class B>
inline ConstVectorMult_<double,B>
 operator*(double a, const BasisFunctionVectorExpr<B>& b)
   {
     return ConstVectorMult_<double,B>(a,static_cast<const B&>(b));
   } 
//-----------------------------------------------------------------------------
template <typename TYPE, class B>
inline ConstVectorMult_<typename std::complex<TYPE>,B>
 operator*(typename std::complex<TYPE> a, const BasisFunctionVectorExpr<B>& b)
   {
     return  ConstVectorMult_<typename std::complex<TYPE>,B>(a,static_cast<const B&>(b));
   } 
//-----------------------------------------------------------------------------
template <class B>
inline ConstVectorMult_<int,B>
 operator*(const BasisFunctionVectorExpr<B>& b, int a)
   {
     return ConstVectorMult_<int,B>(a,static_cast<const B&>(b));
   } 
//-----------------------------------------------------------------------------
template <class B>
inline ConstVectorMult_<float,B>
 operator*(const BasisFunctionVectorExpr<B>& b, float a)
   {
     return  ConstVectorMult_<float,B>(a,static_cast<const B&>(b));
   } 
//-----------------------------------------------------------------------------
template <class B>
inline ConstVectorMult_<double,B>
 operator*(const BasisFunctionVectorExpr<B>& b, double a)
   {
     return ConstVectorMult_<double,B>(a,static_cast<const B&>(b));
   } 
//-----------------------------------------------------------------------------
template <typename TYPE, class B>
inline ConstVectorMult_<typename std::complex<TYPE>,B>
 operator*(const BasisFunctionVectorExpr<B>& b, typename std::complex<TYPE> a)
   {
     return  ConstVectorMult_<typename std::complex<TYPE>,B>(a,static_cast<const B&>(b));
   } 
//-----------------------------------------------------------------------------
/* A matrix is a concatenation of two vectors or an vector and a matrix. 
 * The result of the operator '|' is from left to right. Thus, the template 
 * type we achieve is something of the type 
 * MATRIX< MATRIX < MATRIX < ..., VECTOR <..> >, VECTOR <...> >, VECTOR <...> >
*/
#if 0
template <class A, class B>
class MATRIX ;
//-----------------------------------------------------------------------------
template <class A1, class A2, class B1, class B2>
class MATRIX<VECTOR<A1,A2>,VECTOR<B1,B2> > : public BasisFunctionVectorExpr<MATRIX<VECTOR<A1,A2>, VECTOR<B1,B2> > >
   {
     private:
      const VECTOR<A1,A2>& a;
      const VECTOR<B1,B2>& b;
     public:
      MATRIX(const VECTOR<A1,A2>& a_, const VECTOR<B1,B2>& b_) : a(a_), b(b_) {}
     enum {vContainment = _MAX_(VECTOR<A1,A2>::vContainment,VECTOR<B1,B2>::vContainment),
           wContainment = _MAX_(VECTOR<A1,A2>::wContainment,VECTOR<B1,B2>::wContainment), 
           pContainment = _MAX_(VECTOR<A1,A2>::pContainment,VECTOR<B1,B2>::pContainment)} ;
     enum {length = VECTOR<B1,B2>::length ,size = VECTOR<A1,A2>::size + 1 };
     inline static void 
      print(std::ostream& os);
   };
//-----------------------------------------------------------------------------
template <class A1, class A2, class B1, class B2>
class MATRIX<MATRIX<A1,A2>,VECTOR<B1,B2> > : public BasisFunctionVectorExpr<MATRIX<VECTOR<A1,A2>, VECTOR<B1,B2> > >
   {
     private:
      const MATRIX<A1,A2>& a;
      const VECTOR<B1,B2>& b;
     public:
      MATRIX(const MATRIX<A1,A2>& a_, const VECTOR<B1,B2>& b_) : a(a_), b(b_) {}
     enum {vContainment = _MAX_(MATRIX<A1,A2>::vContainment,VECTOR<B1,B2>::vContainment),
           wContainment = _MAX_(MATRIX<A1,A2>::wContainment,VECTOR<B1,B2>::wContainment), 
           pContainment = _MAX_(MATRIX<A1,A2>::pContainment,VECTOR<B1,B2>::pContainment)} ;
     enum {length = VECTOR<B1,B2>::length ,size = MATRIX<A1,A2>::size + 1 };
     inline static void 
      print(std::ostream& os);
   };
//-----------------------------------------------------------------------------
template <class A, class B>
MATRIX<A,B> 
 operator | (const BasisFunctionVectorExpr<A>& a, const BasisFunctionVectorExpr<B>& b)
   {
      return MATRIX<A,B>(static_cast<const A&>(a),static_cast<const B&>(b));
   }
//-----------------------------------------------------------------------------
/* This struct describes the multiplication of a vector with a one dimensional
 * value. The result is the vector, muliplied with the value in each component.
*/
template <class A, class B>
struct BasisFunctionExpr_VECTOR_MULT : public BasisFunctionExpr<BasisFunctionExpr_VECTOR_MULT<A,B> >
   {
      private: 
        const A& a;
        const B& b;
      public:
       BasisFunctionExpr_VECTOR_MULT(const A& a_, const B& b_) : a(a_), b(b_) 
          {
             assert (A::length == B::length);
             assert (A::length == 2);
          }
       enum {vContainment = _MAX_(A::vContainment,B::vContainment),
             wContainment = _MAX_(A::wContainment,B::wContainment), 
             pContainment = _MAX_(A::pContainment,B::pContainment)} ;
       enum {length = 1 ,size = 1 };
     inline static void 
      print(std::ostream& os)
        {
          os << " (";   
          a.print(os);
          os << ") * (";   
          b.print(os);
          os << ") ";   
        }
   }; 
//-----------------------------------------------------------------------------
template <class A, class B, class C>
struct BasisFunctionExpr_VECTOR_MULT<VECTOR<A,B>,C> : public BasisFunctionExpr<BasisFunctionExpr_VECTOR_MULT<VECTOR<A,B>,C> >
   {
      private: 
        const VECTOR<A,B>& a;
        const B& b;
      public:
       BasisFunctionExpr_VECTOR_MULT(const VECTOR<A,B>& a_, const B& b_) : a(a_), b(b_) 
          {
             assert (VECTOR<A,B>::length == B::length);
             assert (VECTOR<A,B>::length == B::length);
          }
       enum {vContainment = _MAX_(VECTOR<A,B>::vContainment,B::vContainment),
             wContainment = _MAX_(VECTOR<A,B>::wContainment,B::wContainment), 
             pContainment = _MAX_(VECTOR<A,B>::pContainment,B::pContainment)} ;
       enum {length = 1 ,size = 1 };
     inline static void 
      print(std::ostream& os)
        {
          os << " (";   
          a.print(os);
          os << ") * (";   
          b.print(os);
          os << ") ";   
        }
   }; 
//-----------------------------------------------------------------------------
template <class A, class V1, class V2>
typename BasisFunctionExpr_VECTOR_MULT< A, VECTOR<V1,V2> >
 operator * (const BasisFunctionExpr<A>& a, const VECTOR<V1,V2>& b);
//-----------------------------------------------------------------------------
template <class A, class V1, class V2>
typename BasisFunctionExpr_VECTOR_MULT< A, VECTOR<V1,V2> > 
 operator * (const VECTOR<V1,V2>& a, const BasisFunctionExpr<A>& b);
//-----------------------------------------------------------------------------
/* This struct describes the addition of two vectors. The result is the vector, 
 * that holds the sum of the vector components. 
*/
template <class A, class B>
struct CT_VECTOR_Add;  
//-----------------------------------------------------------------------------
template <class A1, class A2, class B1, class B2>
typename CT_VECTOR_Add<VECTOR<A1,A2>,VECTOR<B1,B2> >
 operator+(const VECTOR<A1,A2>& a, const VECTOR<B1,B2>& b);
//-----------------------------------------------------------------------------
template <class A, class B>
struct CT_VECTOR_Sub;  
//-----------------------------------------------------------------------------
template <class A1, class A2, class B1, class B2>
typename CT_VECTOR_Sub<VECTOR<A1,A2>,VECTOR<B1,B2> >
 operator-(const VECTOR<A1,A2>& a, const VECTOR<B1,B2>& b);
//-----------------------------------------------------------------------------
template <class A>
struct CT_VECTOR_Min; 
//-----------------------------------------------------------------------------
template <class A, class B>
typename CT_VECTOR_Min<VECTOR<A,B> >
 operator-(const VECTOR<A,B>& a);
//-----------------------------------------------------------------------------
/* This struct describes the multiplication of two vectors. The result is  
 * the sum that holds the products of the vectors components. 
*/
template <class A, class B>
struct VECTOR_VECTOR_MULT; 
//-----------------------------------------------------------------------------
template <class A1, class A2, class B1, class B2>
typename VECTOR_VECTOR_MULT<VECTOR<A1,A2>,VECTOR<B1,B2> >
 operator*(const VECTOR<A1,A2>& a, const VECTOR<B1,B2>& b);
//-----------------------------------------------------------------------------
/* This struct describes the multiplication of a vector with a matrix!  
 * The result is a vector of the line sums . 
*/
template <class A, class B>
struct VECTOR_MATRIX_MULT;
//-----------------------------------------------------------------------------
template <class A1, class A2, class B1, class B2>
typename VECTOR_MATRIX_MULT<VECTOR<A1,A2>,MATRIX<B1,B2> >
 operator*(const VECTOR<A1,A2>& a, const MATRIX<B1,B2>& b);
//-----------------------------------------------------------------------------
/* This struct describes the multiplication of a matrix with a vector!  
 * The result is a vector of the line sums . 
*/
template <class A, class B>
struct MATRIX_VECTOR_MULT;
//-----------------------------------------------------------------------------
template <class A1, class A2, class B1, class B2>
typename MATRIX_VECTOR_MULT<MATRIX<A1,A2>,VECTOR<B1,B2> >
 operator*(const MATRIX<A1,A2>& a, const VECTOR<B1,B2>& b); 
#endif
//-----------------------------------------------------------------------------
template <class A>
class GRADIENT2D : public BasisFunctionVectorExpr<GRADIENT2D<A> >
   {
     private:
        const A& a;
     public:
        GRADIENT2D(const A& a_) : a(a_) {}
        enum {vContainment = A::vContainment,
              wContainment = A::wContainment, 
              pContainment = A::pContainment} ;
        enum {length = 2, size = 1};
        inline Deri_<A,dx> getHeadOfVector() const { return d_dx(a);}
        template<class Element, unsigned int iterLength>
        inline typename Element::TYPE 
         evalLastComponent(const Element& element, const int (&iterator)[iterLength]) const
           { 
              return (d_dy(a)).eval(element,iterator);
           }
        
   };
//-----------------------------------------------------------------------------
template <class A>
class GRADIENT3D : public BasisFunctionVectorExpr<GRADIENT3D<A> >
   {
     private:
        const A& a;
     public:
        GRADIENT3D(const A& a_) : a(a_) {}
        enum {vContainment = A::vContainment,
              wContainment = A::wContainment, 
              pContainment = A::pContainment} ;
        enum {length = 3, size = 1};
        inline GRADIENT2D<A> getHeadOfVector() const { return GRADIENT2D<A>(a);}
        template<class Element, unsigned int iterLength>
        inline typename Element::TYPE 
         evalLastComponent(const Element& element, const int (&iterator)[iterLength]) const
           { 
              return (d_dz(a)).eval(element,iterator);
           }
        
   };
//-----------------------------------------------------------------------------
 template <class A> 
 inline GRADIENT3D<A>
  grad(const BasisFunctionExpr<A>& a)
     {
        return GRADIENT3D<A>(static_cast<const A&>(a));
     } 
//-----------------------------------------------------------------------------
 template <class A> 
 inline GRADIENT2D<A>
  grad2(const BasisFunctionExpr<A>& a)
     {
        return GRADIENT2D<A>(static_cast<const A&>(a));
     } 
//-----------------------------------------------------------------------------
 template <class A, class B, class C> 
 inline Add_<Add_<Deri_<A,dx> ,Deri_<A,dy> >, Deri_<A,dz>  >
  div(const VECTOR< VECTOR <A,B>, C >& a); 
// ----------------------------------------------------------------------------
 inline VECTOR<VECTOR< NormalComponent<0>,NormalComponent<1> >, NormalComponent<2> >
  N(); 
// ----------------------------------------------------------------------------
 inline VECTOR<VECTOR< UnitNormalComponent<0>,UnitNormalComponent<1> >, UnitNormalComponent<2> >
  N_Unit(); 
// ----------------------------------------------------------------------------
template <BasisFunctionSet number, unsigned int functionId>
struct BF_VECTOR
   {
     typedef VECTOR<Vector_BasisFunction<number,functionId,dirX>,
                    Vector_BasisFunction<number,functionId,dirY> >  
                                                                      VEC_FUNC_2D;
     typedef VECTOR<VECTOR<Vector_BasisFunction<number,functionId,dirX>,
                           Vector_BasisFunction<number,functionId,dirY> >,
                    Vector_BasisFunction<number,functionId,dirZ> >         VEC_FUNC_3D;

   };
// ----------------------------------------------------------------------------
 inline BF_VECTOR<BasisSet_1,0>::VEC_FUNC_2D
  v_vec2D ()
    {
      return ( Vector_BasisFunction<BasisSet_1,0,dirX>() & Vector_BasisFunction<BasisSet_1,0,dirY>() );
    }

// ----------------------------------------------------------------------------
 inline BF_VECTOR<BasisSet_1,1>::VEC_FUNC_2D
  w_vec2D ()
    {
      return ( Vector_BasisFunction<BasisSet_1,1,dirX>() & Vector_BasisFunction<BasisSet_1,1,dirY>() );
    }

// ----------------------------------------------------------------------------
 inline BF_VECTOR<BasisSet_1,0>::VEC_FUNC_3D
  v_vec3D ()
    {
      return Vector_BasisFunction<BasisSet_1,0,dirX>() & 
             Vector_BasisFunction<BasisSet_1,0,dirY>() &
             Vector_BasisFunction<BasisSet_1,0,dirZ>();

    }
// ----------------------------------------------------------------------------
 inline BF_VECTOR<BasisSet_1,1>::VEC_FUNC_3D
  w_vec3D ()
    {
      return Vector_BasisFunction<BasisSet_1,1,dirX>() & 
             Vector_BasisFunction<BasisSet_1,1,dirY>() &
             Vector_BasisFunction<BasisSet_1,1,dirZ>();

    }
// ----------------------------------------------------------------------------
template <class A, class B>
class CURL2D : public BasisFunctionExpr<CURL2D<A,B> >
   {
     private:
       const A& a;
       const B& b;
     public:
        enum {vContainment = _MAX_(A::vContainment,B::vContainment),
              wContainment = _MAX_(A::wContainment,B::wContainment),
              pContainment = _MAX_(A::pContainment,B::pContainment)} ;
        enum {length = 2, size = 1};
       CURL2D(const VECTOR<A,B>& a_) : a(a_.getA()), b(a_.getB()) {}
       CURL2D(const A& a_, const B& b_) : a(a_), b(b_) {}
       template<class Element, unsigned int iterLength>
       inline typename Element::TYPE 
         eval(const Element& element, const int (&iterator)[iterLength]) const
           {
             return d_dx(b).eval(element,iterator) - d_dy(a).eval(element,iterator);
           }
   };
// ----------------------------------------------------------------------------
 template <class A, class B>
 inline CURL2D<A,B>
  curl(const VECTOR<A,B>& a)
    {
      return CURL2D<A,B>(a);
    }
// ----------------------------------------------------------------------------
template <class B, class C>
class CURL3D_Help_1 : public BasisFunctionExpr<CURL3D_Help_1<B,C> >
   {
     private:
       const B& b;
       const C& c;
     public:
        enum {vContainment = _MAX_(C::vContainment,B::vContainment),
              wContainment = _MAX_(C::wContainment,B::wContainment),
              pContainment = _MAX_(C::pContainment,B::pContainment)} ;
        enum {length = 2, size = 1};
       CURL3D_Help_1(const B& b_, const C& c_) : b(b_), c(c_) {}
       template<class Element, unsigned int iterLength>
       inline typename Element::TYPE 
        eval(const Element& element, const int (&iterator)[iterLength]) const
           { 
             return d_dy(c).eval(element,iterator) - d_dz(b).eval(element,iterator);
           }
   };
// ----------------------------------------------------------------------------
template <class A, class B, class C>
class CURL3D_Help : public BasisFunctionVectorExpr<CURL3D_Help<A,B,C> >
   {
     private:
       const A& a;
       const B& b;
       const C& c;
     public:
        enum {vContainment = _MAX_(_MAX_(A::vContainment,B::vContainment),C::vContainment),
              wContainment = _MAX_(_MAX_(A::wContainment,B::wContainment),C::wContainment),
              pContainment = _MAX_(_MAX_(A::pContainment,B::pContainment),C::pContainment)} ;
        enum {length = 2, size = 1};
       CURL3D_Help(const A& a_, const B& b_, const C& c_) : a(a_), b(b_), c(c_) {}
       inline CURL3D_Help_1<B,C> getHeadOfVector() const { return CURL3D_Help_1<B,C>(b,c);}
       template<class Element, unsigned int iterLength>
       inline typename Element::TYPE 
        evalLastComponent(const Element& element, const int (&iterator)[iterLength]) const
           { 
             return d_dx(b).eval(element,iterator) - d_dy(a).eval(element,iterator);
           }
   };
// ----------------------------------------------------------------------------
template <class A, class B, class C>
class CURL3D : public BasisFunctionVectorExpr<CURL3D<A,B,C> >
   {
     private:
       const A& a;
       const B& b;
       const C& c;
     public:
       enum {vContainment = _MAX_(_MAX_(A::vContainment,B::vContainment),C::vContainment),
              wContainment = _MAX_(_MAX_(A::wContainment,B::wContainment),C::wContainment),
              pContainment = _MAX_(_MAX_(A::pContainment,B::pContainment),C::pContainment)} ;
       enum {length = 3, size = 1};
       CURL3D(const VECTOR<VECTOR<A,B>,C>& a_) : a(a_.getA()), b(a_.getB()) , c(a_.getC()) {}
       inline CURL3D_Help<A,B,C> getHeadOfVector() const { return CURL3D_Help<A,B,C>(a,b,c);}
       template<class Element, unsigned int iterLength>
       inline typename Element::TYPE 
        evalLastComponent(const Element& element, const int (&iterator)[iterLength]) const
           { 
             return d_dz(a).eval(element,iterator) - d_dx(c).eval(element,iterator);
           }
   };
// ----------------------------------------------------------------------------
 template <class A, class B, class C>
 inline CURL3D<A,B,C>
  curl(const VECTOR<VECTOR<A,B>,C>& a)
    {
      return CURL3D<A,B,C>(a);
    }
// ----------------------------------------------------------------------------
struct EDGE2D : public BasisFunctionVectorExpr<EDGE2D>
   {
       enum {vContainment = 0, wContainment = 0, pContainment = 0} ;
       enum {length = 2, size = 1};
       inline Edge_Component<dirX> getHeadOfVector() const { return Edge_Component<dirX>();}
       template<class Element, unsigned int iterLength>
       inline typename Element::TYPE 
        evalLastComponent(const Element& element, const int (&iterator)[iterLength]) const
           { 
             return element.edge_(dirY);
           }
   };
#if 0
// ----------------------------------------------------------------------------
template <class A, class A2, class B1, class B2>
class CrossProduct2D : public BasisFunctionExpr<CrossProduct2D<A1,A2,B1,B2> >
   {
     private:
       const A& a;
       const B& b;
     public:
        enum {vContainment = _MAX_(A::vContainment,B::vContainment),
              wContainment = _MAX_(A::wContainment,B::wContainment),
              pContainment = _MAX_(A::pContainment,B::pContainment)} ;
        enum {length = 2, size = 1};
       CrossProduct2D(const VECTOR<A,B>& a_) : a(a_.getA()), b(a_.getB()) {}
       CrossProduct2D(const A& a_, const B& b_) : a(a_), b(b_) {}
       template<class Element, unsigned int iterLength>
       inline typename Element::TYPE 
         eval(const Element& element, const int (&iterator)[iterLength]) const
           {
             return d_dx(b).eval(element,iterator) - d_dy(a).eval(element,iterator);
           }
   };
#endif
// ----------------------------------------------------------------------------
/* Concluding comments
 * At this point there is still more to introduce. A compile time data structure
 * which holds tensor data. Therefore we would need an additional operator to be 
 * overloaded, whicht has less priority than operator '&'. 
*/

//==============================================================================

