/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

//==============================================================================
//
//  $Id: Domain.h,v 1.57 2006/07/06 09:26:13 jochen Exp $
//
//==============================================================================

/* There is no typeof with win32 compilers (as of 2007-10) so these macros
 * cannot be used on win32.                                                   */
#ifndef _WIN32
#define TRANSFORMATION( B ) typedef __typeof__(B)  TRAFO
#define Define_Element_Transformation( B ) typedef __typeof__(B) 
#endif /* _WIN32 */
//-----------------------------------------------------------------------------
#define TEMPLATES_OF_CLASS_DOMAIN_DECL \
          template < unsigned int cornersOfElement, \
                     DIMENSION init_dimension,  \
                     class TRAFO, \
                     domainSecification interiorIboundary, \
                     unsigned int numberOfAnsatzFunctions, \
                     unsigned int numberOfSecondAnsatzFunctions, \
                     ElementType elType, \
                     IntegrationAccuracy intAcc, \
                     typename ETYPE = double, \
                     stencilType STENCIL = STLVector,\
                     typename eTYPE = double, \
                     class BasisFunctions  = \
                       BasisFunctionAndValueClass<ETYPE,init_dimension,\
                                                  Gauss<eTYPE,intAcc,elType>::numberGaussianPoints+1, \
                                                  oneDimensional, \
                                                  numberOfAnsatzFunctions,numberOfSecondAnsatzFunctions> >
//-----------------------------------------------------------------------------
#define TEMPLATES_OF_CLASS_DOMAIN \
          template < unsigned int cornersOfElement, \
                     DIMENSION init_dimension,  \
                     class TRAFO, \
                     domainSecification interiorIboundary, \
                     unsigned int numberOfAnsatzFunctions, \
                     unsigned int numberOfSecondAnsatzFunctions, \
                     ElementType elType, \
                     IntegrationAccuracy intAcc, \
                     typename ETYPE, \
                     stencilType STENCIL,\
                     typename eTYPE, \
                     class BasisFunctions >
//-----------------------------------------------------------------------------
#define DOMAIN_WITH_TEMPLATES \
          _Domain_ < cornersOfElement,\
                     init_dimension,\
                     TRAFO,\
                     interiorIboundary,\
                     numberOfAnsatzFunctions, \
                     numberOfSecondAnsatzFunctions,\
                     elType,\
                     intAcc,\
                     ETYPE,\
                     STENCIL,\
                     eTYPE,\
                     BasisFunctions >
//-----------------------------------------------------------------------------
#define STENCIL_TYPE \
    typename ResultType< \
       typename StencilTypeInitialization<ETYPE,STENCIL,numberOfAnsatzFunctions,numberOfSecondAnsatzFunctions>::StencilType,\
        Integrand>::DimType
//-----------------------------------------------------------------------------
#define VEC_STENCIL_TYPE \
    typename ResultType< \
       typename StencilTypeInitialization<ETYPE,STENCIL,numberOfAnsatzFunctions,numberOfSecondAnsatzFunctions>::StencilType,\
        Integrand>::Vec_DimType
//=============================================================================
// Introducing the additional template type, in order to steer the type of the
// resulting stencil, i.e. nested stl vectors or an array
//=============================================================================
/**
 The class Domain serves as center within the Colsamm coding! However, only 
 inherited classes of Domain are used and get allocated. 
*/
TEMPLATES_OF_CLASS_DOMAIN_DECL
class _Domain_
  : public Gauss<eTYPE,intAcc,elType>,
    public BasisFunctions,
    public Mapping_<ETYPE,init_dimension,interiorIboundary,TRAFO,Gauss<eTYPE,intAcc,elType>::numberGaussianPoints>,
    public CornerClasses<eTYPE,init_dimension,cornersOfElement,elType>,
    public StencilTypeInitialization<ETYPE,STENCIL,numberOfAnsatzFunctions,numberOfSecondAnsatzFunctions>
   {
  // -----------------------------------------------------------------------
     public:
      typedef ETYPE TYPE;
      typedef eTYPE bTYPE;
      using CornerClasses<eTYPE,init_dimension,cornersOfElement,elType>::numberOfCorners;
  // -----------------------------------------------------------------------
    /**
      Method for evaluating the Dirac functions at particular points  
     */
      template < class Integrand >
      inline const STENCIL_TYPE
       value ( const BasisFunctionExpr<Integrand> & a_ );
    /**
      Method for evaluating the Dirac functions at particular points  
     */
      template < class Integrand >
      inline const VEC_STENCIL_TYPE
       value ( const BasisFunctionVectorExpr<Integrand> & a_ );

    /**
      Method for integrating an integrand of a weak formulation,
      returning the local stencil
     */
      template < class Integrand >
      inline const STENCIL_TYPE&
       integrate ( const BasisFunctionExpr < Integrand > & a_ );
    /**
      Method for integrating an integrand of a weak formulation,
      computing the local stencil in to the first component
     */
      template < typename PointArray, class Integrand >
      inline void
       integrate ( PointArray& array, const BasisFunctionExpr < Integrand > & a_ );

    /**
     Method for setting the basis functions into set 1
     */
      template < class FunctionExpression>
      inline void
       Set ( const FunctionExpression & a_ );
    /** 
     Method for setting the basis functions into any other set 
     by specifying the set Number
     */
      template < class FunctionExpression >
      inline void
       Set (BasisFunctionSet setNumber, const FunctionExpression & a_ );

    /**
     Method for setting the basis functions into set 1
     */
      template < class FunctionExpression>
      inline void
       Set ( const FunctionExpression & a_, const unsigned int cnr_1, const unsigned int cnr_2);


    /**
     Method for setting a special point for evaluation
     */
      template < class PointArrayTYPE >
      void
       Point_ ( const PointArrayTYPE & point );

    /**
     Method for setting a special point for evaluation including
     the test whether the point lies inside or outside the actual
     element.
     */ 
      template < class PointArrayTYPE >
      bool
       Point_Test ( const PointArrayTYPE & point );

    /**
     Method for passing the vertices of the actual element
     */
      template < class PointArrayTYPE >
      inline DOMAIN_WITH_TEMPLATES&
       operator ( ) ( const PointArrayTYPE & cc , const int numberPoints = cornersOfElement);
   };
//==============================================================================

