/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

//==============================================================================
//
//  $Id: Gauss_Points.h,v 1.24 2006/07/14 10:51:30 jochen Exp $
//
//==============================================================================
template <int p1, int p2, int p3, int p4, int p5>
struct PyramidVolume
   {
     template <class Element>
     static inline typename Element::bTYPE
      compute(const Element& element);
   };
//-----------------------------------------------------------------------------
template <int p1, int p2, int p3, int p4>
struct TetrahedronVolume
   {
     template <class Element>
     static inline typename Element::bTYPE
      compute(const Element& element);
   };
//-----------------------------------------------------------------------------
template <int p1, int p2, int p3>
struct TriangleVolume
   {
     template <class Element>
     static inline typename Element::bTYPE
      compute(const Element& element);
   };
//-----------------------------------------------------------------------------
template <typename TYPE, IntegrationAccuracy t_, ElementType et>
class Gauss;
//-----------------------------------------------------------------------------
template <typename TYPE, int numberOfGaussianPoints>
struct Gauss_Init
   {
     protected:
       TYPE **gaussianPoints,
             *stress;
     public: 
     enum {numberGaussianPoints = numberOfGaussianPoints} ;
     Gauss_Init();
     ~Gauss_Init();
     inline TYPE** getGaussianPoints() const;
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss<TYPE, Gauss3, cuboid>
  : public Gauss_Init <TYPE,27>
   {
     Gauss ();
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss2, cuboid>
  : public Gauss_Init <TYPE,8>
   {
     Gauss ();
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss1, cuboid>
  : public Gauss_Init <TYPE,1>
   {
     Gauss ();
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss2, pyramid>
  : public Gauss_Init <TYPE,8>
   {
     Gauss ();
   };
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss1, pyramid>
  : public Gauss_Init <TYPE,2>
   {
     Gauss ();
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss3, prism>
  : public Gauss_Init <TYPE,21>
   {
     Gauss ();
   };
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss2, prism>
  : public Gauss_Init <TYPE,6>
   {
     Gauss();
   };
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss1, prism >
  : public Gauss_Init <TYPE,1>
   {
     Gauss();
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss2, tetrahedron>
  : public Gauss_Init <TYPE,4>
   {
     Gauss();
   };
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss1, tetrahedron>
  : public Gauss_Init <TYPE,1>
   {
     Gauss();
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss3, quadrangle>
  : public Gauss_Init <TYPE,9>
   {
     Gauss();
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss2, quadrangle>
  : public Gauss_Init <TYPE,4>
   {
     Gauss ();
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss1, quadrangle>
  : public Gauss_Init <TYPE,1>
   {
     Gauss();
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss3, triangle>
  : public Gauss_Init <TYPE,7>
   {
     Gauss();
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss2, triangle>
  : public Gauss_Init <TYPE,3>
   {
     Gauss();
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss1, triangle>
  : public Gauss_Init <TYPE,1>
   {
     Gauss();
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss3, interval>
  : public Gauss_Init <TYPE,3>
   {
     Gauss();
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss2 , interval>
  : public Gauss_Init <TYPE,2>
   {
     Gauss();
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE>
struct Gauss <TYPE, Gauss1, interval>
  : public Gauss_Init <TYPE,1>
   {
     Gauss();
   } ;
//-----------------------------------------------------------------------------
template <typename TYPE, DIMENSION dimension, int numberCorners>
class CornerManagement 
   {
     protected:
       TYPE *barys,
            *coordinatesOfVertices;
     public:
       CornerManagement();
       ~CornerManagement(); 
       int getNumberOfCorners() const {return numberCorners;}
       inline TYPE* getCornerList() const;
       inline TYPE* getArrayOfBarycentricCoordinates() const;
       inline TYPE edge_length (const unsigned int cnr_1, const unsigned int cnr_2) const;
       inline TYPE edge_(SpaceDirection direction) const;
   };
//-----------------------------------------------------------------------------
template <typename TYPE, DIMENSION dimension, int numberCorners, ElementType elType>
struct CornerClasses;
//-----------------------------------------------------------------------------
template <typename TYPE, int numberCorners>
struct CornerClasses <TYPE,D3,numberCorners,cuboid> : CornerManagement<TYPE,D3,numberCorners>
   {
     enum {numberOfCorners = numberCorners};
     template <class Element>
     inline void
     compute_Barycentric_Coordinates (const Element& element);
   };
//-----------------------------------------------------------------------------
template <typename TYPE, int numberCorners>
struct CornerClasses <TYPE,D3,numberCorners,pyramid> : CornerManagement<TYPE,D3,numberCorners>
   {
     enum {numberOfCorners = numberCorners};
     template <class Element>
     inline void
     compute_Barycentric_Coordinates (const Element& element);
   };
//-----------------------------------------------------------------------------
template <typename TYPE, int numberCorners>
struct CornerClasses <TYPE,D3,numberCorners,prism> : CornerManagement<TYPE,D3,numberCorners>
   {
     enum {numberOfCorners = numberCorners};
     template <class Element>
     inline void
     compute_Barycentric_Coordinates (const Element& element);
   };
//-----------------------------------------------------------------------------
template <typename TYPE, int numberCorners>
struct CornerClasses <TYPE,D3,numberCorners,tetrahedron> : CornerManagement<TYPE,D3,numberCorners>
   {
     enum {numberOfCorners = numberCorners};
     template <class Element>
     inline void
     compute_Barycentric_Coordinates (const Element& element);
   };
//-----------------------------------------------------------------------------
template <typename TYPE, DIMENSION dimension, int numberCorners>
struct CornerClasses <TYPE,dimension,numberCorners,quadrangle> : CornerManagement<TYPE,dimension,numberCorners>
   {
     enum {numberOfCorners = numberCorners};
     template <class Element>
     inline void
     compute_Barycentric_Coordinates (const Element& element);
   };
//-----------------------------------------------------------------------------
template <typename TYPE, DIMENSION dimension, int numberCorners>
struct CornerClasses <TYPE,dimension,numberCorners,triangle> : CornerManagement<TYPE,dimension,numberCorners>
   {
     enum {numberOfCorners = numberCorners};
     template <class Element>
     inline void
     compute_Barycentric_Coordinates (const Element& element);
   };
//-----------------------------------------------------------------------------
template <typename TYPE, DIMENSION dimension, int numberCorners>
struct CornerClasses <TYPE,dimension,numberCorners,interval> : CornerManagement<TYPE,dimension,numberCorners>
   {
     enum {numberOfCorners = numberCorners};
     template <class Element>
     inline void
     compute_Barycentric_Coordinates (const Element& element);
   };
//-----------------------------------------------------------------------------
//==============================================================================
