/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

//==============================================================================
//
//  $Id: Mapping.h,v 1.43 2006/07/06 09:26:13 jochen Exp $
//
//==============================================================================
//-----------------------------------------------------------------------------
template <typename TYPE, DIMENSION dimension_, int _numberGaussianPoints>
class MappingManagement
   {
      public:
        enum{dimensionOfElement = dimension_};
      protected:
        enum{dimMap = dimension_ +1 , 
             dimVal = dimensionOfElement*dimensionOfElement, 
             dimExt = _extensionSize};
        TYPE ***mappingValues,
             *jacobianMatrixEntries,
             **extensionValues;
      public:
        MappingManagement();
        ~MappingManagement();
        inline int dimension() const {return (int)dimensionOfElement;}
        inline TYPE** getMappingValues(int gaussLevel) const ;
        inline TYPE* getValuesOfJacobianMatrix(int gaussLevel) const ;
        inline const TYPE& getDeterminant(int gaussLevel) const ;
        inline const TYPE& getDeterminantTimesWeight(int gaussLevel) const ;
   };
//-----------------------------------------------------------------------------
template <typename _TYPE, DIMENSION d, domainSecification int_bound, class TRAFO, int _numberGaussianPoints>
class Mapping_ ;
//-----------------------------------------------------------------------------
template <typename TYPE,class TRAFO, int _numberGaussianPoints>
class Mapping_ <TYPE,D3,interior,TRAFO,_numberGaussianPoints> : public MappingManagement<TYPE,D3,_numberGaussianPoints>
    {
      private: 
       const TRAFO trafo;
      public:
       Mapping_() : trafo(TRAFO()) {} 
       template <class Element>
        inline void point ( const Element& element, const unsigned int gaussLevel );
   };
//-----------------------------------------------------------------------------
template <typename TYPE,class TRAFO, int _numberGaussianPoints>
class Mapping_<TYPE,D2,interior,TRAFO,_numberGaussianPoints>  : public MappingManagement<TYPE,D2,_numberGaussianPoints> 
     {
      private:
       const TRAFO trafo;
      public:
       Mapping_() : trafo(TRAFO()) {} 
       template <class Element>
        inline void point(const Element& element, const unsigned int gaussLevel );
   };
//-----------------------------------------------------------------------------
template <typename TYPE,class TRAFO, int _numberGaussianPoints>
class Mapping_<TYPE,D1,interior,TRAFO,_numberGaussianPoints> : public MappingManagement<TYPE,D1,_numberGaussianPoints>
     {
      private:
       const TRAFO trafo;
      public:
       Mapping_() : trafo(TRAFO()) {} 
       template <class Element>
        inline void point(const Element& element, const unsigned int gaussLevel );
     };
// -----------------------------------------------------------------------------
template <typename TYPE,class TRAFO, int _numberGaussianPoints>
class Mapping_<TYPE, D2,boundary,TRAFO,_numberGaussianPoints> : public MappingManagement<TYPE,D2,_numberGaussianPoints> 
     {
      private: 
       const TRAFO trafo;
      public:
       Mapping_() : trafo(TRAFO()) {} 
       template <class Element>
        inline void point ( const Element& element, const unsigned int gaussLevel );
     };
// -----------------------------------------------------------------------------
template <typename TYPE,class TRAFO, int _numberGaussianPoints>
class Mapping_<TYPE,D3,boundary,TRAFO,_numberGaussianPoints> : public MappingManagement<TYPE,D3,_numberGaussianPoints> 
     {
      private: 
       const TRAFO trafo;
      public:
       Mapping_() : trafo(TRAFO()) {} 
       template <class Element>
        inline void point ( const Element& element, const unsigned int gaussLevel );
      };
// -----------------------------------------------------------------------------
template <typename TYPE,class TRAFO, int _numberGaussianPoints>
class Mapping_<TYPE,D3,boundary_with_derivatives,TRAFO,_numberGaussianPoints> : public MappingManagement<TYPE,D3,_numberGaussianPoints> 
    {
      private: 
       const TRAFO trafo;
      public:
       Mapping_() : trafo(TRAFO()) {} 
       template <class Element>
        inline void point(const Element& element, const unsigned int gaussLevel);
    };
//==============================================================================
