/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

//==============================================================================
//
//  $Id: Stencil_ET.h,v 1.6 2006/07/14 10:51:30 jochen Exp $
//
//==============================================================================
//-----------------------------------------------------------------------------
template <class TYPE> 
struct FunctionInterfaces
   {
     typedef TYPE (*functionInterfaceOne)(TYPE x);
     typedef TYPE (*functionInterfaceTwo)(TYPE x, TYPE y);
     typedef TYPE (*functionInterfaceThree)(TYPE x, TYPE y, TYPE z);
   };
//-----------------------------------------------------------------------------
template <class A> 
struct BasisFunctionExpr
   {
     enum{length = 1, size = 1};
     template<class Element, unsigned int iterLength> 
     inline const typename Element::TYPE
      eval(const Element& element,const int (&iterator)[iterLength]) const; 
   };
//-----------------------------------------------------------------------------
template <typename TYPE>
class Constant : BasisFunctionExpr<Constant<TYPE> >
   {
      private: 
        TYPE constant;
      public:
       enum{length = 1, size = 1};
       Constant (const TYPE c_) : constant(c_) {}
       template<class Element, unsigned int iterLength> 
       inline const typename Element::TYPE
        eval(const Element& element,const int (&iterator)[iterLength]) const
          {
            return constant;
          } 
   };
//-----------------------------------------------------------------------------
template <BasisFunctionSet number, unsigned int functionId>
struct BasisFunction 
  : public BasisFunctionExpr<BasisFunction<number,functionId> >
    {
      enum { vContainment = ((functionId == 0 && number == BasisSet_1)?1:0), 
             wContainment = ((functionId == 1 && number == BasisSet_1)?1:0), 
             pContainment = ((functionId == 0 && number == BasisSet_2)?1:0)};
      inline void print(std::ostream &os) const; 
      template<class Element, unsigned int iterLength> 
      inline const typename Element::TYPE
       eval(const Element& element,const int (&iterator)[iterLength]) const; 
      template<what type, class Element, unsigned int iterLength> 
      inline const typename Element::TYPE
       evaluateDerivative(const Element& element,const int (&iterator)[iterLength]) const; 
};
typedef BasisFunction<BasisSet_1,0> Ansatz_Function;
typedef BasisFunction<BasisSet_1,1> Testing_Function;
typedef BasisFunction<BasisSet_2,0> Mixed_Function;
//-----------------------------------------------------------------------------
template <BasisFunctionSet number, unsigned int functionId, SpaceDirection direction>
struct Vector_BasisFunction 
  : public BasisFunctionExpr<Vector_BasisFunction<number,functionId,direction> >
    {
      enum { vContainment = ((functionId == 0 && number == BasisSet_1)?1:0), 
             wContainment = ((functionId == 1 && number == BasisSet_1)?1:0), 
             pContainment = ((functionId == 0 && number == BasisSet_2)?1:0)};
      inline void print(std::ostream &os) const; 
      template<class Element, unsigned int iterLength> 
      inline const typename Element::TYPE
       eval(const Element& element,const int (&iterator)[iterLength]) const; 
      template<what type, class Element, unsigned int iterLength> 
      inline const typename Element::TYPE
       evaluateDerivative(const Element& element,const int (&iterator)[iterLength]) const; 
};
//-----------------------------------------------------------------------------
template <SpaceDirection direction>
struct Edge_Component 
  : public BasisFunctionExpr<Edge_Component<direction> >
    {
      enum { vContainment = 0, wContainment = 0,  pContainment = 0};
      inline void print(std::ostream &os) const { os << "EDGE <" << direction << "> ";} 
      template<class Element, unsigned int iterLength> 
      inline const typename Element::TYPE
       eval(const Element& element,const int (&iterator)[iterLength]) const
         {
           return element.edge_(direction);
         } 
};
//-----------------------------------------------------------------------------
template <class T>
class FUNC1_ : public BasisFunctionExpr<FUNC1_<T> >
   {
     private:
     typename FunctionInterfaces<T>::functionInterfaceOne func_;
     public:
     enum {vContainment = 0, wContainment =0, pContainment = 0};
     inline void 
      print(std::ostream &os) const; 
     inline FUNC1_(const typename FunctionInterfaces<T>::functionInterfaceOne& f); 
     template<class Element, unsigned int iterLength> 
     inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]) const; 
   };
//-----------------------------------------------------------------------------
template <class T>
class FUNC2_ : public BasisFunctionExpr<FUNC2_<T> >
   {
     private:
     typename FunctionInterfaces<T>::functionInterfaceTwo func_;
     public:
     enum {vContainment = 0, wContainment =0, pContainment = 0};
     inline void 
      print(std::ostream &os) const; 
     inline FUNC2_(const typename FunctionInterfaces<T>::functionInterfaceTwo& f); 
     template<class Element, unsigned int iterLength> 
     inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]) const; 
   };
//-----------------------------------------------------------------------------
template <class T>
class FUNC3_ : public BasisFunctionExpr<FUNC3_<T> >
   {
     private:
     typename FunctionInterfaces<T>::functionInterfaceThree func_;
     public:
     enum {vContainment = 0, wContainment =0, pContainment = 0};
     inline void 
      print(std::ostream &os) const; 
     inline FUNC3_(const typename FunctionInterfaces<T>::functionInterfaceThree& f); 
     template<class Element, unsigned int iterLength> 
     inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]) const; 
   };
//-----------------------------------------------------------------------------
template <int ex, SpaceDirection direction>
struct BasisMonom
  : public BasisFunctionExpr<BasisMonom<ex,direction> >
    {
      enum {vContainment = 0, wContainment = 0, pContainment = 0};
      inline void 
       print(std::ostream &os) const; 
      template<class Element, unsigned int iterLength> 
      inline typename Element::TYPE 
       eval(const Element& element,const int (&iterator)[iterLength]) const; 
    };
//-----------------------------------------------------------------------------
namespace Interfaces 
   {
     template <int p>
     inline BasisMonom<p,dirX> x_(); 
     template <int p>
     inline BasisMonom<p,dirY> y_(); 
     template <int p>
     inline BasisMonom<p,dirZ> z_(); 
   }
//-----------------------------------------------------------------------------
template <int position>
struct NormalComponent 
  : public BasisFunctionExpr< NormalComponent<position> >
    {
      enum {vContainment = 0, wContainment =0, pContainment = 0};
      inline void 
       print(std::ostream &os) const; 
      template<class Element, unsigned int iterLength>
      inline typename Element::TYPE 
       eval(const Element& element,const int (&iterator)[iterLength]) const; 
    };
//-----------------------------------------------------------------------------
template <int position>
struct UnitNormalComponent 
  : public BasisFunctionExpr< UnitNormalComponent<position> >
    {
      enum {vContainment = 0, wContainment =0, pContainment = 0};
      inline void 
       print(std::ostream &os) const; 
      template<class Element, unsigned int iterLength> 
      inline typename Element::TYPE 
       eval(const Element& element,const int (&iterator)[iterLength]) const; 
    };
//-----------------------------------------------------------------------------
template <class T>
class VertexVector : public BasisFunctionExpr<VertexVector<T> >
   {
     private:
       T  data;
     public:
     enum {vContainment = 1, wContainment =0, pContainment = 0};
     inline void print(std::ostream &os) const; 
     inline VertexVector(const T& d); 
     template<class Element, unsigned int iterLength> 
     inline typename Element::TYPE 
      eval(const Element& element,const int (&iterator)[iterLength]) const; 
   };
//-----------------------------------------------------------------------------
template <class A> 
class Conjugate : public BasisFunctionExpr<Conjugate<A> >
   {
     private: 
       const A& a;
     public:
       inline Conjugate(const A& a_) : a(a_) {};
     enum {vContainment = A::vContainment,
           wContainment = A::wContainment,
           pContainment = A::pContainment};
    inline void 
     print(std::ostream &os) const; 
    template<class Element, unsigned int iterLength>
    inline typename Element::TYPE 
     eval(const Element& element,const int (&iterator)[iterLength]) const;
   };
//-----------------------------------------------------------------------------
template<class A> 
class Min_ : public BasisFunctionExpr<Min_<A> >
   {
     private: 
       const A& a;
     public:
       inline Min_(const A& a_) : a(a_) {};
     enum {vContainment = A::vContainment,
           wContainment = A::wContainment,
           pContainment = A::pContainment};
     inline void 
      print(std::ostream &os) const; 
     template<class Element, unsigned int iterLength>
     inline typename Element::TYPE
      eval(const Element& element,const int (&iterator)[iterLength]) const; 
   };
//-----------------------------------------------------------------------------
template<class A, class B> 
class Add_ : public BasisFunctionExpr<Add_<A,B> >
   {
     private: 
       const A& a;
       const B& b;
     public:
       inline Add_(const A& a_, const B& b_) : a(a_), b(b_) {};
     enum {vContainment = _MAX_(A::vContainment,B::vContainment), 
           wContainment = _MAX_(A::wContainment,B::wContainment),
           pContainment = _MAX_(A::pContainment,B::pContainment)} ;
    inline void 
     print(std::ostream & os) const; 
    template<class Element, unsigned int iterLength>
    inline typename Element::TYPE
     eval(const Element& element,const int (&iterator)[iterLength]) const;
};
//-----------------------------------------------------------------------------
template<class TYPE, typename B> 
class CAdd_ : public BasisFunctionExpr<CAdd_<TYPE,B> >
   {
     private: 
       const TYPE a;
       const B& b;
     public:
       inline CAdd_(const TYPE& a_, const B& b_) : a(a_), b(b_) {};
     enum {vContainment = B::vContainment, 
           wContainment = B::wContainment,
           pContainment = B::pContainment} ;
    inline void 
     print(std::ostream & os) const; 
    template<class Element, unsigned int iterLength>
    inline typename Element::TYPE
     eval(const Element& element,const int (&iterator)[iterLength]) const;
};
//-----------------------------------------------------------------------------
template<class A, class B>
class Sub_ : public BasisFunctionExpr<Sub_<A,B> >
   {
     private: 
       const A& a;
       const B& b;
     public:
     inline Sub_(const A& a_, const B& b_) : a(a_), b(b_) {};
     enum {vContainment = _MAX_(A::vContainment,B::vContainment), 
           wContainment = _MAX_(A::wContainment,B::wContainment),
           pContainment = _MAX_(A::pContainment,B::pContainment)} ;
     inline void
      print(std::ostream& os) const; 
     template<class Element, unsigned int iterLength>
     inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]) const; 
};
//-----------------------------------------------------------------------------
template<class TYPE, class B>
class CSub_ : public BasisFunctionExpr<CSub_<TYPE,B> >
   {
     private: 
       const TYPE a;
       const B& b;
     public:
     inline CSub_(const TYPE& a_, const B& b_) : a(a_), b(b_) {};
     enum {vContainment = B::vContainment, 
           wContainment = B::wContainment,
           pContainment = B::pContainment} ;
     inline void
      print(std::ostream& os) const; 
     template<class Element, unsigned int iterLength>
     inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]) const; 
};
//-----------------------------------------------------------------------------
template<class A, class B> 
class Mult_: public BasisFunctionExpr<Mult_<A,B> >
   {
     private: 
       const A& a;
       const B& b;
     public:
     inline Mult_(const A& a_, const B& b_) : a(a_), b(b_) {};
     enum {vContainment = _MAX_(A::vContainment,B::vContainment), 
           wContainment = _MAX_(A::wContainment,B::wContainment),
           pContainment = _MAX_(A::pContainment,B::pContainment)} ;
     inline void 
      print(std::ostream &os) const; 
     template<class Element, unsigned int iterLength>
     inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]) const; 
   };
//-----------------------------------------------------------------------------
template<class TYPE, class B> 
class CMult_: public BasisFunctionExpr<CMult_<TYPE,B> >
   {
     private: 
       const TYPE a;
       const B& b;
     public:
     inline CMult_(const TYPE& a_, const B& b_) : a(a_), b(b_) {};
     enum {vContainment = B::vContainment, 
           wContainment = B::wContainment,
           pContainment = B::pContainment} ;
     inline void 
      print(std::ostream &os) const; 
     template<class Element, unsigned int iterLength>
     inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]) const; 
   };
//-----------------------------------------------------------------------------
template<class A, class B> 
struct Div_: public BasisFunctionExpr<Div_<A,B> >
   {
     private: 
       const A& a;
       const B& b;
     public:
     inline Div_(const A& a_, const B& b_) : a(a_), b(b_) {};
     enum {vContainment = _MAX_(A::vContainment,B::vContainment), 
           wContainment = _MAX_(A::wContainment,B::wContainment),
           pContainment = _MAX_(A::pContainment,B::pContainment)} ;
     inline void 
      print(std::ostream &os) const; 
     template<class Element, unsigned int iterLength>
     inline typename Element::TYPE 
      eval(const Element& element,const int (&iterator)[iterLength]) const; 
   };
//-----------------------------------------------------------------------------
template<class A> 
class SQRT_: public BasisFunctionExpr<SQRT_<A> >
   {
     private: 
       const A& a;
     public:
       inline SQRT_(const A& a_) : a(a_) {};
     enum {vContainment = A::vContainment,
           wContainment = A::wContainment,
           pContainment = A::pContainment};
     inline void 
      print(std::ostream &os) const; 
     template<class Element, unsigned int iterLength>
     inline typename Element::TYPE 
      eval(const Element& element,const int (&iterator)[iterLength]) const; 
   };
//-----------------------------------------------------------------------------
template<class A, typename eTYPE> 
class POW_: public BasisFunctionExpr<POW_<A,eTYPE> >
   {
     private: 
       const A& a;
       eTYPE exponent;
     public:
       inline POW_(const A& a_, eTYPE expo_) : a(a_), exponent(expo_) {};
     enum {vContainment = A::vContainment,
           wContainment = A::wContainment,
           pContainment = A::pContainment};
     inline POW_ (eTYPE e);
     inline void 
      print(std::ostream &os) const; 
     template<class Element, unsigned int iterLength>
     inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]) const; 
   };
//-----------------------------------------------------------------------------
template<class A> 
class EXP_: public BasisFunctionExpr<EXP_<A> >
   {
     private: 
       const A& a;
     public:
       inline EXP_(const A& a_) : a(a_) {};
     enum {vContainment = A::vContainment,
           wContainment = A::wContainment,
           pContainment = A::pContainment};
     inline void 
      print(std::ostream &os) const; 
     template<class Element, unsigned int iterLength>
     inline typename Element::TYPE 
      eval(const Element& element, const int (&iterator)[iterLength]) const; 
   };
//-----------------------------------------------------------------------------
template <class A, what typ> 
class Deri_ : public BasisFunctionExpr<Deri_<A,typ> >
   {
     private: 
       const A& a;
     public:
       inline Deri_(const A& a_) : a(a_) {};
     inline void 
      print(std::ostream &os) const; 
     enum {vContainment = A::vContainment,
           wContainment = A::wContainment,
           pContainment = A::pContainment};
     template<class Element, unsigned int iterLength>
     inline const typename Element::TYPE
      eval(const Element& element, const int (&iterator)[iterLength]) const; 
   };
//==============================================================================

