/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/


//////////////////////////////////////////////////////////
//                  abbbreviations
//////////////////////////////////////////////////////////

#ifndef ABBREVICOMMON_H_
#define ABBREVICOMMON_H_

#ifdef _MPI_PARALLEL
#include <mpi.h>
#endif 

#include <iostream>
#include <fstream>
using std::ofstream;
using std::cout;
using std::endl;

#include <complex>
#include <cmath>

#include <map>
#include <cassert>



#ifndef M_PI
#define M_PI 3.1415927
#endif

enum elementTyp { pointEl, edgeEl,  rectangleEl, triangleEl, quadrangleEl, hexahedronEl };


// 1D
//-------------
enum dir1D { Ldir1D,  Rdir1D };
enum Ort1D { LOrt, MOrt, ROrt };

// 2D
//-------------
// directions in 2D
enum dir2D { Wdir2D, Edir2D, Sdir2D, Ndir2D };
// direction of sons in 2D
enum dir2D_sons { SWdir2D, SEdir2D, NWdir2D, NEdir2D };
// local directions of hexahedra with fixed SD-direction
enum local_dir_of_hex2D{ ND, ST, NT };

// 3D
//-------------
// directions in 3D
enum dir3D { Wdir3D, Edir3D, Sdir3D, Ndir3D, Ddir3D, Tdir3D };
// directions fuer Soehne, oder Ecken einer Zellen
enum dir3D_sons { WSDdir3D, ESDdir3D, WNDdir3D, ENDdir3D, WSTdir3D, ESTdir3D, WNTdir3D, ENTdir3D };
// local directions of hexahedra with fixed SWD-direction
enum local_dir_of_hex3D{ SED, NWD, NED, SWT, SET, NWT, NET };
// main_directions in 3D
enum main_dir_3D { WE_dir, SN_dir, DT_dir };

// edges of a cell
enum Edges_cell { SDed, NDed, STed, NTed, WDed, EDed, WTed, ETed,
                  SWed, SEed, NWed, NEed };





enum marker_typ { no_mark, yes_mark, partial_mark };


enum stencil_typ { no_stencil, yes_stencil };


enum sten_part_typ { point_sten_part,    edge_sten_part,  quadrangle_in_sten_part, quadrangle_ex_sten_part, 
                     hexahedron_sten_part };


enum control_typ { not_thread_save=0, thread_save=1};

#endif // ABBREVICOMMON_H_
