//    solid laser for solid state laser simulation:
//    Copyright (C) 2017  ASLD GmbH
//    This program is not a free software;
//    do not redistribute it and/or modify it
/////////////////////////////////////////////////

/*
 * fouriertransform.h
 *
 *     Author: zhabiz
 */

#ifndef FOURIERTRANSFORM_H_
#define FOURIERTRANSFORM_H_

#include <complex>
using namespace std;



class FourierTransform {

public:
	FourierTransform(int n_);
        
	//recursive implementation
	static void FFT(complex<double> *v, int n);
	static void iFFT(complex<double> *v, int n);
	
	void FFT_efficient(int dir,int m, double *x,double *y);
	static void FFTShift(complex<double> *v, int n);

	static bool isPowerOfTwo(unsigned int x);
	static bool isEven(int x);

	static void scalarProduct(complex<double> *x,  double scalar);
	static void copy(complex<double> *x, complex<double> *y);
	static void calcIntensity(complex<double> *x, complex<double> *intensity);

	static void calcIntensityField(const char* filename, complex<double> *x, complex<double> *intensity, double *xAxis, int n);

	static void print_vector(const char *title, complex<double> *x);
	static void print_vector_to_file(const char* filename, complex<double> *x, double *xAxis);

	static int nk;

private:
  static void FFT(complex<double> *v,  int n, complex<double> *tmp);
  static void iFFT(complex<double> *v, int n, complex<double> *tmp);
  
  static void resizeTemp(int n);
  static complex<double> *temp;
  static int size_temp;


};



#endif /* FOURIERTRANSFORM_H_ */
