/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/


// ------------------------------------------------------------
// math_lib_common.h
//
// ------------------------------------------------------------

#ifndef MATHLIBCOMMON_H_
#define MATHLIBCOMMON_H_

#include<complex>

#define eps_for_tet_test 0.000001

//////////////////////////////////////////////////////////////////////
// 0. basic constants
// 1. Konjugation Konstruktion   für  MPI data typ   
// 2. trigonometrie, complex, norm
// 3. MAX, MIN
// 4. gerade, zweipotenz
//////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////
// 0. basic constants
//////////////////////////////////////////////////////////////////////
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

//////////////////////////////////////////////////////////////////////
// 1. Konjugation 
//////////////////////////////////////////////////////////////////////

// Code for getting MPI data typ
template <class DTyp>
struct  MyConj {
	static DTyp Give ( DTyp x );
};

//----------------------------------------------------------------------------


template <>
inline double MyConj<double>::Give ( double x )
{
	return x;
}

template <>
inline bool MyConj<bool>::Give ( bool x )
{
	return x;
}

template <>
inline std::complex<double>  MyConj<std::complex<double> >::Give ( std::complex<double> x )
{
	return conj ( x );
}


//////////////////////////////////////////////////////////////////////
// 2. trigonometrie, complex, norm
//////////////////////////////////////////////////////////////////////



inline double arccos(const double y) {
   return atan2(sqrt(1-y*y),y);
   // if(y>=0) return atan(sqrt(1-y*y)/y);
   // else    return atan(sqrt(1-y*y)/y) + M_PI;
}

inline double my_tan(const double y) {
  return tan(y);
}

inline double my_atan(const double y) {
  return atan(y);
}

inline double my_sqrt(const double y) {
  return sqrt(y);
}

inline double myConj(double x) {
	return x;
}

inline std::complex<double> myConj(std::complex<double> x) {
	return conj(x);
}

inline std::complex<double> IMAGFROMREAL ( double x)
{
    return std::complex<double>(x,0);
}

inline double myReal ( std::complex<double>   x )
{
	return real ( x );
}

inline double myImag (std::complex<double>   x )
 {
	return imag ( x );
}

inline double myNorm(std::complex<double> x)
{
	return norm(x);
}


//////////////////////////////////////////////////////////////////////
// 3. MAX, MIN


/* Maximum */
inline double MAX(double a, double b) {
  if(a < b) return b;
  else return a;
}

inline double MAX(double a, double b, double c) {
  return MAX(MAX(a,b),c);
}

inline int MAX(int a, int b) {
  if(a < b) return b;
  else return a;
}

inline int MAX(int a, int b, int c) {
  static int m;
  if(a < b) m = b;
  else m = a;
  if(c < m) return m;
  else return c;
}

/* Minimum */
inline double MIN(double a, double b) {
  if(a < b) return a;
  else return b;
}

/* Absolute value */
inline double ABS(double a) {
  if(a < 0.0) return -a;
  else return a;
}

/* Absolute value */
inline double ABS(std::complex<double> a) {
  return abs(a);
}

/* Absolute value */
inline int ABS(int a) {
  if(a < 0) return -a;
  else return a;
}


/* real value */
inline double myREAL(double a) {
  return a;
}

/* real value */
inline double myREAL(std::complex<double> a) {
  return real(a);
}

/* Absolute value */
inline int myREAL(int a) {
  return a;
}

/* real value */
inline std::complex<double> myCOMPLEX(double a) {
  return std::complex<double>(a,0.0);
}

/* real value */
inline std::complex<double> myCOMPLEX(std::complex<double> a) {
  return a;
}




//////////////////////////////////////////////////////////////////////
// 4. gerade, zweipotenz


/* Test ob Zahl gerade */
inline int gerade(int N) {
  return (~N)&1;
}

// Berechnung der Zweierpotenz 
inline int Zweipotenz (int k) {
  return 1<<k;
}

// ganzzahliger Anteil
inline int Ganzzahliger_Anteil (double k) {
  if(k<0) return (int)k-1;
  return (int)k;
}






#endif // MATHLIBCOMMON_H_


