/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// invertMatrix.h
//
// ------------------------------------------------------------

#ifndef INVERT_MATRIX_H
#define INVERT_MATRIX_H



///////////////////////////
// inverse matrix
///////////////////////////



template <class DTyp>
class InversMatrix {		
  public:	
    InversMatrix(DTyp* data_, int sizeI_, int sizeJ_) {
      data = data_;   sizeI = sizeI_;  sizeJ = sizeJ_;
    }
    DTyp* data;
    int sizeI;
    int sizeJ;  
};

template <class DTyp>
InversMatrix<DTyp> InvertMatrix(MatrixIndex<DTyp>& Mat) {
  return InversMatrix<DTyp>(Mat.getCopyOfData(),Mat.getSizeI(),Mat.getSizeJ());
}

template <class DTyp>
InversMatrix<DTyp> InvertMatrix(MatrixAlg<DTyp>& Mat) {
  return InversMatrix<DTyp>(Mat.CopyOfgetData(),Mat.getSizeI(),Mat.getSizeJ());
}


template<class DTyp>
void MatrixIndex<DTyp>::operator=(const InversMatrix<DTyp>& matrixA) {

  DTyp* dataA = matrixA.data;
  int sizeI_A = matrixA.sizeI;
  int sizeJ_A = matrixA.sizeJ;
  
  assert(sizeI==sizeJ);
  assert(sizeI_A==sizeJ_A);
  assert(sizeI==sizeI_A);
  
  int size = sizeI;
  
  // 1. Schritt: setze Einheitsmatrix:
  for(int I=0;I<size;++I) for(int J=0;J<size;++J) {
    if(I==J) data[I*sizeJ+J] = 1.0;
    else     data[I*sizeJ+J] = 0.0;
  }

  
  
  // 2. Schritt: L-Matrizen
  DTyp factor;
  DTyp AIJ;
  for(int J=0;J<=size-2;++J) {
      for(int I=J+1;I<=size-1;++I) {
	  AIJ = dataA[I*sizeJ+J];	  
//cout << " a[" << I << "," << J << "]=" << AIJ << endl;	  
	  // if(ABS(AIJ) > eps) ...
	  factor = AIJ * Inverse(dataA[J*sizeJ+J]);
          //for(int S=0;S<=J;++S) {
	  for(int S=0;S<=size-1;++S) {
	      dataA[I*sizeJ+S] = dataA[I*sizeJ+S] - factor * dataA[J*sizeJ+S];
	      data[I*sizeJ+S]  = data[I*sizeJ+S]  - factor * data[J*sizeJ+S];
	  }
      }
  }
  
  // 3. Schritt: Diagonale
  DTyp AII;
  for(int I=0;I<=size-1;++I) {
      AII = dataA[I*sizeJ+I];
      // if(ABS(AIJ) > eps) ...
      factor = Inverse(AII);
      //for(int S=0;S<=I;++S) {
      for(int S=0;S<=size-1;++S) {
	  dataA[I*sizeJ+S] = factor * dataA[I*sizeJ+S];
	  data[I*sizeJ+S]  = factor * data[I*sizeJ+S];
      }
  }

  // 4. Schritt: R-Matrizen
  for(int J=size-1;J>=1;--J) {
      for(int I=J-1;I>=0;--I) {
	  AIJ = dataA[I*sizeJ+J];
	  // if(ABS(AIJ) > eps) ...
	  factor = AIJ;
          for(int S=0;S<=size-1;++S) {
	      dataA[I*sizeJ+S] = dataA[I*sizeJ+S] - factor * dataA[J*sizeJ+S]; // notwendig??
	      data[I*sizeJ+S]  = data[I*sizeJ+S]  - factor * data[J*sizeJ+S];
	  }
      }
  }
  
  
}





#endif 
