/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// matrix.h
//
// ------------------------------------------------------------

#ifndef MATRIX_ALG_H
#define MATRIX_ALG_H

template <class DTyp>
class InversMatrix;

///////////////////////////
// 1. matrix - vector
///////////////////////////



template <class A, class V>
class MatrixVectorIndex : public ExprAlg<MatrixVectorIndex< A, V > > {		
	public:	
	   typedef typename V::Result Result;
	   MatrixVectorIndex( const A& a, const V& v ) : a_ ( a ), v_ ( v ) {}
	   void startI_S(int max_size) const {
		     a_.startI_S(max_size);
	   };
	
	   inline Result getValueI(int Iglobal) const { 
		    Result sum = Result(0.0);
		    int sizeJ = a_.getSizeJ();
		    int sizeV = v_.getSize();
		    if(sizeJ==sizeV) {  // hoffentlich passt alles zusammen!!
		       for(int s=0;s<sizeJ;++s)
		           sum = sum + a_.getValueIS(Iglobal, s) * v_.getValueS(s); 
		    }
		    else {
		       a_.startJ(sizeJ);
		       v_.startI_S(sizeV);
		       for(int s=0;s<sizeJ;++s) {
			   int Jglobal = a_.getIndexJ(s);
		           sum = sum + a_.getValueII(Iglobal, Jglobal) * v_.getValueI(Jglobal);
		       }
		    }
		    return sum;  
	   }
	   inline Result getValueS(int Is) const { 
		    Result sum = 0.0;
		    int sizeJ = a_.getSizeJ();
		    int sizeV = v_.getSize();
		    assert(sizeJ==sizeV);  // soll erst mal zusammen passen

		    for(int s=0;s<sizeJ;++s)
		       sum = sum + a_.getValueSS(Is, s) * v_.getValueS(s); 
		    return sum;  
	   }
	private:
	  const A& a_;
	  const V& v_;
};


///////////////////////////
// 2. matrix
///////////////////////////

class MatrixAuxilliaryData
{
  public:
  static int   sizeAux;
  static char* dataAux;
};


template <class DTyp>
class MatrixAuxilliary
 {
  protected:
    DTyp* getCopyDataAux(int size, DTyp* data) { 
      if(size*sizeof(DTyp) > MatrixAuxilliaryData::sizeAux) {
	if(MatrixAuxilliaryData::dataAux!=NULL) delete[] MatrixAuxilliaryData::dataAux;
	MatrixAuxilliaryData::dataAux = (char*)new DTyp[size];
      }
      for(int s=0;s<size;++s) ((DTyp*)MatrixAuxilliaryData::dataAux)[s] = data[s];
      return (DTyp*)MatrixAuxilliaryData::dataAux;
    }

};
template <class DTyp>
class MatrixAlg : public ExprAlg<MatrixAlg< DTyp > >, public MatrixAuxilliary<DTyp> {		
	public:			
                /**
		 * erzeigt Einheits-Matrix der Groesse sizeI_ , sizeJ_
		 */	  
		MatrixAlg(int sizeI_,int sizeJ_);
		typedef DTyp Result;
		
		template <class A>
		void operator=(const ExprAlg<A>& a);
		void operator=(DTyp x);
		void operator=(const InversMatrix<DTyp>* matrixA);
		void operator=(const MatrixAlg<DTyp>& matrixR);
				
		void startI_S(int max_size) const { 
		  assert(sizeI<=max_size);
		}
		void startJ(int max_size) const { 
		  assert(sizeJ<=max_size);
		}

		DTyp getValueSS(int Is, int Js) const { 
		     return data[Is*sizeJ + Js]; 
		}
		int getSizeJ() { return sizeJ; }
		int getSizeI()  const { return sizeI; }		
		DTyp* getCopyOfData() { return MatrixAuxilliary<DTyp>::getCopyDataAux(sizeI*sizeJ, data); }

		void Print();	
	private:		
		// own data
		DTyp* data;     
		int sizeI;
		int sizeJ;
};





template <class DTyp>
class MatrixIndex : public ExprAlg<MatrixIndex< DTyp > >, public MatrixAuxilliary<DTyp> {		
	public:			
	         /**
		 * erzeugt Einheits-Matrix
		 * der Groesse indexI.getSize(),  indexJ.getSize()
		 */	  
	        template <class IndexClass>
		MatrixIndex(const IndexClass& indexI,const IndexClass& indexJ);

		typedef DTyp Result;

                template <class DTypVector>
                MatrixVectorIndex<MatrixIndex<DTyp>,
		                  VectorIndex<DTypVector> > operator() (const VectorIndex<DTypVector>& v) const {
		  return MatrixVectorIndex<MatrixIndex<DTyp>,
		                           VectorIndex<DTypVector> >(*this,v);
		}

		template <class A>
		void operator=(const ExprAlg<A>& a);
		void operator=(DTyp x);
		void operator=(const InversMatrix<DTyp>& matrixA);
		void operator=(const MatrixIndex<DTyp>& matrixR);
				
		
		void startI_S(int max_size) const  { 
		  (*i) = 0;  
		  if(sizeI>0) (*Imy) = In[(*i)];
		  else (*Imy) = -1;
		}
		void startJ(int max_size) const  { 
		  (*j) = 0;  
		  if(sizeJ>0) (*Jmy) = Jn[(*j)];
		  else (*Jmy) = -1;
		}
		
		void setValueSS(int Is, int Js, DTyp x); 
		
		DTyp getValueII(int Iglobal, int Jglobal) const; 
		DTyp getValueIS(int Iglobal, int Js     ) const; 
		DTyp getValueSI(int Is,      int Jglobal) const; 
		DTyp getValueSS(int Is,      int Js     ) const; 

		int getSizeJ()  const { return sizeJ; }		
		int getSizeI()  const { return sizeI; }		
		DTyp* getCopyOfData() { return MatrixAuxilliary<DTyp>::getCopyDataAux(sizeI*sizeJ, data); }

		int getIndexJ(int Js) const { assert(Js<sizeJ); return Jn[Js]; }	
		void Print();	
		void Test();
	private:		
		// own data
		DTyp* data;    // zuerst Spalten j, dann Zeilen i 
		
		// Speicher Indizes
		int sizeI;
		int *In;
		int sizeJ;
		int *Jn;
		
		// für die Auswertung
		int *Imy;
		int *i;
		int *Jmy;
		int *j;
};


///////////////////////////
// 3. Other functions, max, ...
///////////////////////////



//////////////////////////////////////////
// and it's implementation



//////////////////////////////////////////
// 4. Implementation of inline functions
// 4. 1. MatrixAlg
//////////////////////////////////////////


template <class DTyp>
MatrixAlg<DTyp>::MatrixAlg(int sizeI_,int sizeJ_) { 
   sizeI = sizeI_;
   sizeJ = sizeJ_;
   data = new DTyp[sizeI*sizeJ];
   for(int sI=0;sI<sizeI;++sI) for(int sJ=0;sJ<sizeJ;++sJ) {
       if(sI==sJ) data[sI*sizeJ+sJ] = 1.0;
       else       data[sI*sizeJ+sJ] = 0.0;
   }
}


template <class DTyp>
void MatrixAlg<DTyp>::Print() {
     cout << " Index Matrix of sizeI: " << sizeI << " sizeJ: " << sizeJ << endl;
     for(int sI = 0;sI < sizeI;++sI ) {
         cout << "i: " << sI;
	 for(int sJ = 0;sJ < sizeJ;++sJ ) {
             cout << " j: " << sJ << " value: " << data[sI*sizeJ+sJ];
	 }
	 cout << endl;
     } 
}


template <class DTyp>
template <class A>
void MatrixAlg<DTyp>::operator=(const ExprAlg<A>& a) {
	const A& ao ( a );
	ao.startI_S(sizeI);
	ao.startJ(sizeJ);

	for(int sI = 0;sI < sizeI;++sI ) 
	  for(int sJ = 0;sJ < sizeJ;++sJ ) {
            data[sI*sizeJ+sJ] =  ao.getValueSS(sI,sJ);
	  }
}


template <class DTyp>
void MatrixAlg<DTyp>::operator=(DTyp x) {  
      for(int sI = 0;sI < sizeI;++sI ) 
	for(int sJ = 0;sJ < sizeJ;++sJ ) {
            data[sI*sizeJ+sJ] =  x;
	}  
}

///////////////////////////////////////////////////////////////
//  4.2  MatrixIndex
///////////////////////////////////////////////////////////////

template<class DTyp>
template <class IndexClass>
MatrixIndex<DTyp>::MatrixIndex(const IndexClass& indexI,const IndexClass& indexJ) {
  sizeI = indexI.getSize(); 
  In    = indexI.getIndices(); 
  sizeJ = indexJ.getSize(); 
  Jn    = indexJ.getIndices(); 

  data = new DTyp[sizeI*sizeJ];

  i    = new int;
  Imy  = new int;
  j    = new int;
  Jmy  = new int;
  
  for(int sI=0;sI<sizeI;++sI) for(int sJ=0;sJ<sizeJ;++sJ) {
      if(sI==sJ) data[sI*sizeJ+sJ] = 1.0;
      else       data[sI*sizeJ+sJ] = 0.0;
  }
}
	
template<class DTyp>
DTyp MatrixIndex<DTyp>::getValueII(int Iglobal, int Jglobal) const {  
   if(   Iglobal > (*Imy)) {        // Da zuerst Spalten j, dann Zeilen i 
     (*j) = 0; (*Jmy) = Jn[(*j)]; 
   }
   while(Iglobal > (*Imy) && (*i) < sizeI) {
      ++(*i);
      (*Imy) = In[(*i)];
   }
   if((*Imy) > Iglobal || (*i)>=sizeI) return DTyp(0);
   
   while(Jglobal > (*Jmy) && (*j) < sizeJ) {
      ++(*j);
      (*Jmy) = Jn[(*j)];
   }
   if((*Jmy) > Jglobal || (*j)>=sizeJ) return DTyp(0);
   

   return data[(*i)*sizeJ + (*j)]; 
}

template<class DTyp>
DTyp MatrixIndex<DTyp>::getValueIS(int Iglobal, int Js) const {  
   while(Iglobal > (*Imy) && (*i) < sizeI) {
      ++(*i);
      (*Imy) = In[(*i)];
   }
   if((*Imy) > Iglobal || (*i)>=sizeI) return DTyp(0);

   return data[(*i)*sizeJ + (Js)]; 
}

template<class DTyp>
DTyp MatrixIndex<DTyp>::getValueSI(int Is, int Jglobal) const {
   while(Jglobal > (*Jmy) && (*j) < sizeJ) {
      ++(*j);
      (*Jmy) = Jn[(*j)];
   }
   if((*Jmy) > Jglobal || (*j)>=sizeJ) return DTyp(0);
   

   return data[Is*sizeJ + (*j)]; 
}

template<class DTyp>
DTyp MatrixIndex<DTyp>::getValueSS(int Is, int Js) const {
   return data[Is*sizeJ + Js]; 
}

template<class DTyp>
void MatrixIndex<DTyp>::setValueSS(int Is, int Js, DTyp x) {
  assert(Is<sizeI);
  assert(Js<sizeJ);
  
  data[Is*sizeJ + Js] = x;
}


template <class DTyp>
template <class A>
void MatrixIndex<DTyp>::operator=(const ExprAlg<A>& a) {
	const A& ao ( a );
 	
	ao.startI_S(sizeI);

	for(int sI = 0;sI < sizeI;++sI ) {
	    ao.startJ(sizeJ);	  
	    for(int sJ = 0;sJ < sizeJ;++sJ ) {
                data[sI*sizeJ+sJ] =  ao.getValueII(In[sI],Jn[sJ]);
	    }
	}
}

template <class DTyp>
void MatrixIndex<DTyp>::operator=(const MatrixIndex<DTyp>& matrixR) {
    matrixR.startI_S(sizeI);
    
    for(int sI = 0;sI < sizeI;++sI ) {
        matrixR.startJ(sizeJ);
	for(int sJ = 0;sJ < sizeJ;++sJ ) {
            data[sI*sizeJ+sJ] =  matrixR.getValueII(In[sI],Jn[sJ]);
	}
    }
}


template <class DTyp>
void MatrixIndex<DTyp>::operator=(DTyp x) {
    for(int sI = 0;sI < sizeI;++sI ) 
	for(int sJ = 0;sJ < sizeJ;++sJ ) {
            data[sI*sizeJ+sJ] =  x;
	}  
}

template <class DTyp>
void MatrixIndex<DTyp>::Test() {
  /*
  cout << " Test: Index Vector of size: " << size << endl;
  for(int ss = 0;ss < size;++ss ) {
      cout << "i: " << ss << " index: " << Sn[ss] << endl;
  }
  */
}


template <class DTyp>
void MatrixIndex<DTyp>::Print() {
     cout << " Index Matrix of sizeI: " << sizeI << " sizeJ: " << sizeJ << " values at (i,j): " << endl;
     for(int sI = 0;sI < sizeI;++sI ) {
         //cout << "(" << sI;
	 for(int sJ = 0;sJ < sizeJ;++sJ ) {
             cout << " (" << sI << "," << sJ << "): " << data[sI*sizeJ+sJ];
	 }
	 cout << endl;
     }
}


#endif 
