// ------------------------------------------------------------
// main.cc
//
// ------------------------------------------------------------

#include "source/ugblock.h"



/** Kern_matrix
*** 6 x 6 matrix for rigid body modes
**/
Kern_matrix::Kern_matrix() {
  matrix = new double*[6];
  for(int i=0;i<6;++i)
    matrix[i] = new double[6];

  aux_matrix = new double*[6];
  for(int i=0;i<6;++i)
    aux_matrix[i] = new double[6];
  perm = new int[6];
  b = new double[6];
  x = new double[6];
}

Kern_matrix::~Kern_matrix() {
  for(int i=0;i<6;++i) {
    delete[](matrix[i]);
    delete[](aux_matrix[i]);
  }
  delete[](matrix);
  delete[](aux_matrix);
  delete[](perm);
  delete[](b);
  delete[](x);
}


void Calc_weights(Variable<double>& One,
		  Variable<double>& Xcoord, Variable<double>& Ycoord, Variable<double>& Zcoord,
		  Kern_matrix& M) {
			  double weight_const;
  One          = 1.0;
  weight_const = sqrt(product(One,One));
  One          = One / weight_const;

  // erste Zeile
  M.matrix[0][0] = 1.0;
  M.matrix[0][1] = 0.0;
  M.matrix[0][2] = 0.0;

  M.matrix[0][3] = 0.0;
  M.matrix[0][4] =   product(One,Zcoord);
  M.matrix[0][5] = - product(One,Ycoord);

  // zweite Zeile
  M.matrix[1][0] = 0.0;
  M.matrix[1][1] = 1.0;
  M.matrix[1][2] = 0.0;

  M.matrix[1][3] = - product(One,Zcoord);
  M.matrix[1][4] = 0.0;
  M.matrix[1][5] =   product(One,Xcoord);

  // dritte Zeile


  M.matrix[2][0] = 0.0;
  M.matrix[2][1] = 0.0;
  M.matrix[2][2] = 1.0;

  M.matrix[2][3] =   product(One,Ycoord);
  M.matrix[2][4] = - product(One,Xcoord);
  M.matrix[2][5] = 0.0;

  // vierte Zeile
  M.matrix[3][0] = 0.0;
  M.matrix[3][1] = - product(One,Zcoord);
  M.matrix[3][2] =   product(One,Ycoord);

  M.matrix[3][3] =   product(Zcoord,Zcoord) + product(Ycoord,Ycoord);
  M.matrix[3][4] = - product(Ycoord,Xcoord);
  M.matrix[3][5] = - product(Xcoord,Zcoord);

  // fuenfte Zeile
  M.matrix[4][0] =   product(One,Zcoord);
  M.matrix[4][1] = 0.0;
  M.matrix[4][2] = - product(One,Xcoord);

  M.matrix[4][3] = - product(Xcoord,Ycoord);
  M.matrix[4][4] =   product(Zcoord,Zcoord) + product(Xcoord,Xcoord);
  M.matrix[4][5] = - product(Ycoord,Zcoord);

  // sechste Zeile
  M.matrix[5][0] = - product(One,Ycoord);
  M.matrix[5][1] =   product(One,Xcoord);
  M.matrix[5][2] = 0.0;

  M.matrix[5][3] = - product(Xcoord,Zcoord);
  M.matrix[5][4] = - product(Ycoord,Zcoord);
  M.matrix[5][5] =   product(Ycoord,Ycoord) + product(Xcoord,Xcoord);

  // Calc LR-decomposition
  int result;
  result = gauss(1,6,M.matrix,M.aux_matrix,M.perm,M.b,M.x,&(M.signd));
  if(result!=0)std::cout << " error in gauss: " << result <<std::endl;
}



double sinus (double x){
  return sin(x);
}

double cosinus (double x){
  return cos(x);
}

double sinush (double x){
  return sinh(x);
}



int main(int argc, char** argv)
{
  cout.precision(10);
  cout.setf(std::ios::fixed,std::ios::floatfield);

  std::ifstream PARAMETER;
  int iteration, n, iter_relax;
  int level;

  double normi;  

  ofstream DATEI;

  PARAMETER.open("para.dat",std::ios::in);
  PARAMETER >> n >> iteration >> iter_relax;
  PARAMETER.close();

  cout << "\n Solving Poisson's equation on a cube \n";
  cout << " with Dirichlet boundary conditions by multigrid: ";
  cout << "\n Depth of grid: " << n;
  cout << "\n Maximal number of iterations: " << iteration;
  cout << "\n Number of smoothing steps: " 
       << iter_relax << endl;     
       
       
      
  // examples
  Hexahedron hexa(1.0, 1.0, 1.0); //funktioniert
  //Cylinder hexa(1.0, 0.2, 1.0, true); //funktioniert
  //HexahedronXYList hexa(ly,ly,ly) ; // paraview error weil NaN, aber Form und Grid identisch. zahlenbeispiel bloed?
  //HexahedronList hexa(1.0, 1.0, ly, 0.2, 0.1, 0.1, true) ; // paraview error weil NaN, aber Form und Grid identisch. zahlenbeispiel bloed?
  //HexahedronXYListTest hexa(1.0, 1.2, 1.1, -0.3, -0.2, 0.5 ); //funktioniert
  //AdaptiveHexahedronList hexa(3,2,1,3,lx, 2,2,2 ) ; //funktioniert
  //CylinderList hexa( 3,1,ly,1,1,1,true ); //funktioniert
  //SpecialPrismList hexa(1.0, 0.2, 1.4, 0.6, 1.9, ly, 0.3,0.3,1.0); //funktioniert
	 
  Boundary_Marker boundary(&hexa);
  Unstructured_grid_Marker inner_points(&hexa);
  inner_points.complement(&boundary);

  DATEI.open("hex.dx",std::ios :: out);
  hexa.Print_Dx(&DATEI);
  DATEI.close();

  cout << " Die Anzahl der Freiheitsgrade sind "
       << hexa.degree_of_freedom() << endl;


  //  block_grids
  ////////////////
  int m = 2; 
  for(int i=0;i<n;++i) {
    m = m * 2;
  }
  Blockgrid blockgrid(&hexa,m);
  cout << " Blockgrid hexa erstellt! " << endl;

 
  //  variables
  ////////////////
  Variable<double>  u(blockgrid);
  Variable<double>  interpol(blockgrid);
  Variable<double>  g(blockgrid);

  Cell_variable<double>  u_cell(blockgrid);

  X_coordinate X(blockgrid);
  Y_coordinate Y(blockgrid);
  Z_coordinate Z(blockgrid);

  Function1<double,double> Sin(sinus);
  Function1<double,double> Cos(cosinus);
  Function1<double,double> Sinh(sinush);

  u_cell = Cos(X*M_PI)*Cos(Y*M_PI)*Cos(Z*M_PI);
  u      = Cos(X*M_PI)*Cos(Y*M_PI)*Cos(Z*M_PI);
  
  
  /*
  Interpolate_from_cell_to_point<Stencil_constant_z> interpolGrid(iteration);
  interpolGrid.init(&blockgrid); 
  interpolGrid.doInterpolation(&u_cell,&interpol);
*/

  FastInterpolate_from_cell_to_point interpolGrid;
  interpolGrid.init(&blockgrid); 
  interpolGrid.doInterpolation(&u_cell,&interpol);
  
  
    std::ofstream DATEIA;

    DATEIA.open("interpol.vtk");
    interpol.Print_VTK(DATEIA);
    DATEIA.close();     
     
//     cout << " Maximum: " << Maximum(interpol) << " Minimum: " << Minimum(interpol) << endl;  
  
  cout << " Maximum : " << L_infty(u) << endl; 
  cout << " Error : " << L_infty(interpol-u) << endl;
  
  cout << " Variable ausgegeben! " << endl;
}


