/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/


//////////////////////////////////////////////////////////
//                  abbbreviations
//////////////////////////////////////////////////////////

#ifndef ABBREVI_H_
#define ABBREVI_H_

#ifdef _MPI_PARALLEL
#include <mpi.h>
#endif 

#include "../../common_source/abbrevi_common.h"



#define params_in  int id, int i, int j, int k, int Nx, int Ny
#define params_out id, i, j, k, Nx, Ny

//#define params_in_loc_matrix   int ind_loc_matrix, int ind_cell, int i, int j, int k, int Nx, int Ny
//#define params_out_loc_matrix      ind_loc_matrix, ind_cell, i, j, k, Nx, Ny


#define params_in_loc_matrix   int id_hex, int ind_loc_matrix, int ind_cell, int i, int j, int k, int Nx, int Ny
#define params_out_loc_matrix      id_hex,     ind_loc_matrix,     ind_cell,     i,     j,     k,     Nx,     Ny

#define params_in_cell   int id_hex, int ind_cell, int i, int j, int k, int Nx, int Ny
#define params_out_cell      id_hex,     ind_cell,     i,     j,     k,     Nx,     Ny


// params for stencil
#define params_in_sten   int id, int num, int i
#define params_out_sten  id,     num,     i


#define Ind_data_hexahedra(i,j,k,Nnx,Nny) data_hexahedra[id][(i)+(Nnx+1)*((j)+(Nny+1)*(k))] 

#define Ind_data_quadrangle(i,j,Nnx) data_quadrangles[id][(i)+(Nnx+1)*(j)] 

#define Ind_loc_matrix_hexahedra(i,j,k) (((Ny)*(k)+(j))*(Nx)+(i))
#define Ind_loc_matrix_hexahedra_fine(i,j,k) (((Ny_fine)*(k)+(j))*(Nx_fine)+(i))

//#define Ind_loc_matrix_hexahedra(i,j,k) ((Nk)*(k)+(Nj)*(j)+(Ni)*(i))
#define Ind_loc_matrix_quadrangles(i,j) ((j)*(Nx)+(i))

#define Ind_loc_variable_hexahedra(i,j,k) (((Ny+1)*(k)+(j))*(Nx+1)+(i))
//#define Ind_loc_variable_hexahedra(i,j,k) ((Nk)*(k)+(Nj)*(j)+(Ni)*(i))
#define Ind_loc_variable_quadrangles(i,j) ((j)*(Nx+1)+(i))

#ifndef M_PI
#define M_PI 3.1415927
#endif

#ifndef DIR2D_DEFINES
#define DIR2D_DEFINES


// integer-value of a point ds in a neighbour cell dz
inline int trans_cellP_9(dir2D_sons dz, dir2D_sons ds) {
  return ((ds&1) + (dz&1)) +
    3*(((ds>>1)&1) + ((dz>>1)&1));
}

inline void Print(dir2D_sons d) {
  if(d==SWdir2D)  cout << "SWdir2D";
  if(d==SEdir2D)  cout << "SEdir2D";
  if(d==NWdir2D)  cout << "NWdir2D";
  if(d==NEdir2D)  cout << "NEdir2D";
}

inline void Print(dir2D d) {
  if(d==Wdir2D) cout << "Wdir2D";
  if(d==Edir2D) cout << "Edir2D";
  if(d==Sdir2D) cout << "Sdir2D";
  if(d==Ndir2D) cout << "Ndir2D";
}

#endif


/////////////////////////////////////////////////////////
//  funktionen for transformation
/////////////////////////////////////////////////////////

inline int transform_part(dir3D d) {
  if(d>Ndir3D) return (d-4)<<2;
  if(d<Sdir3D) return d;
  return (d-2)<<1;
}

// main direction of 3D-direction
//-----------------------------
inline main_dir_3D Main_direction(dir3D d) {
  if(d>Ndir3D) return DT_dir;
  if(d<Sdir3D) return WE_dir;
  return SN_dir;
}


// 1D-part of 3D-direction
//-----------------------------
inline dir1D OneDpart(dir3D d) {
  if(d>Ndir3D) return (dir1D)(d-4);
  if(d<Sdir3D) return (dir1D)d;
  return (dir1D)(d-2);
}

// 3D-direction by 1D-part and main direction
//--------------------------------------------
/*inline dir3D Add(dir1D lr, main_dir_3D md) {
  if(md==WE_dir) return (dir3D)(lr);
  if(md==SN_dir) return (dir3D)(2+lr);
  return (dir3D)(4+lr);
}*/


// corners of the face of a cell
//------------------------------
inline dir3D_sons corner_sur(dir3D f, dir2D_sons cor) {
  if(f < Sdir3D)
    return (dir3D_sons)(f + 2*(cor&1) + 4*(cor>>1));
  if(f > Ndir3D)
    return (dir3D_sons)(4*(f-4) + (cor&1) + 2*(cor>>1));
  return (dir3D_sons)(2*(f-2) + (cor&1) + 4*(cor>>1));
}

// direction from one corner to another
//-------------------------------------
inline dir3D direction_from_to(dir3D_sons from, dir3D_sons to) {
  if((from&1) != (to&1)) {
    if((from&1) < (to&1)) return Edir3D;
    else return Wdir3D;
  }
  if((from&2) != (to&2)) {
    if((from&2) < (to&2)) return Ndir3D;
    else return Sdir3D;
  }
  if((from&4) < (to&4)) return Tdir3D;
  else return Ddir3D;
}

// corners of an edge of a cell 
//------------------------------
inline dir3D_sons Transform(Edges_cell ec, dir1D d) {
  enum dir3D dir;
  dir = (dir3D) ((ec >> 2) << 1);
  if(dir<Sdir3D) return (dir3D_sons)(d+4*((ec>>1)&1)+2*(ec&1));
  if(dir>Ndir3D) return (dir3D_sons)(2*((ec>>1)&1)+(ec&1)+4*d);
  return (dir3D_sons)(4*((ec>>1)&1)+2*d+(ec&1));
}

// corners of an edge of a face 
//------------------------------
inline dir2D_sons Transform(dir2D ed, dir1D d) {
  if(ed == Wdir2D && d == Rdir1D) return NWdir2D;
  if(ed == Wdir2D && d == Ldir1D) return SWdir2D;

  if(ed == Edir2D && d == Rdir1D) return NEdir2D;
  if(ed == Edir2D && d == Ldir1D) return SEdir2D;

  if(ed == Ndir2D && d == Ldir1D) return NWdir2D;
  if(ed == Ndir2D && d == Rdir1D) return NEdir2D;

  if(ed == Sdir2D && d == Ldir1D) return SWdir2D;
  //  if(ed == Sdir2D && d == Rdir1D)
  return SEdir2D;
}

// direction of the edge of a cell
//---------------------------------
inline dir3D Trans_dir(Edges_cell ec, dir1D d) {
  return (dir3D) (((ec >> 2) << 1) + d);
}

// next corner of a cell in a certain direction
//--------------------------------------------------
inline dir3D_sons Next_corner(dir3D_sons co, dir3D dir) {
  if(dir<Sdir3D) {
    return (dir3D_sons)((co&(~1)) + ((~co)&1));
  }
  if(dir>Ndir3D) {
    return (dir3D_sons)((co&(~(1<<2))) + (((~(co>>2))&1)<<2));
  }
  return (dir3D_sons)((co&(~(1<<1))) + (((~(co>>1))&1)<<1));
}

inline dir3D_sons Next_corner_x(dir3D_sons co) {
  return (dir3D_sons)((co&(~1)) + ((~co)&1));
}
inline dir3D_sons Next_corner_y(dir3D_sons co) {
  return (dir3D_sons)((co&(~(1<<1))) + (((~(co>>1))&1)<<1));
}
inline dir3D_sons Next_corner_z(dir3D_sons co) {
    return (dir3D_sons)((co&(~(1<<2))) + (((~(co>>2))&1)<<2));
}


// dir3D coordinate of the direction of "dir3D_sons"
//-------------------------------------------------
inline dir3D xCoord(dir3D_sons d) {
  return (dir3D)(d&1);
}
inline dir3D yCoord(dir3D_sons d) {
  return (dir3D)(((d>>1)&1)+2);
}
inline dir3D zCoord(dir3D_sons d) {
  return (dir3D)(((d>>2)&1)+4);
}

// D3vector unit-vector of the corners of a cell
inline double Unit_vector_x(dir3D_sons d) {
  return double(d&1);
}

inline double Unit_vector_y(dir3D_sons d) {
  return double((d&2)>>1);
}

inline double Unit_vector_z(dir3D_sons d) {
  return double((d&4)>>2);
}


// Sign (-1 falls links, 1 falls rechts) of a dir3D direction
//-------------------------------------------------------------------
inline int sign(dir3D d) {
  return (2*(d&1)-1);
}


// dir1D coordinate of direction "dir3D_sons"
//---------------------------------------------
inline dir1D x1DCoord(dir3D_sons d) {
  return (dir1D)(d&1);
}
inline dir1D y1DCoord(dir3D_sons d) {
  return (dir1D)((d>>1)&1);
}
inline dir1D z1DCoord(dir3D_sons d) {
  return (dir1D)((d>>2)&1);
}


// number of equal coordinates x,y,z of two corners (dir3D_sons)
//---------------------------------------------------------------
inline int How_Equal(dir3D_sons da, dir3D_sons db) {
  return (int)((da&1) == (db&1)) +
    (int)((da&2) == (db&2)) +
    (int)((da&4) == (db&4));
}

// opposite directions in 3D
//-----------------------------------
inline dir3D opposite3D(dir3D dir) {
  if(dir<Sdir3D) return (dir3D)((~dir)&1);
  if(dir>Ndir3D) return (dir3D)(((~(dir-4))&1)+4);
  return (dir3D)(((~(dir-2))&1)+2);
}

inline dir3D_sons opposite3D(dir3D_sons dir) {
  return (dir3D_sons)((~dir)&7);
}

inline Edges_cell opposite3D(Edges_cell ed) {
  if(ed==SDed) return NTed;
  if(ed==NDed) return STed;
  if(ed==STed) return NDed;
  if(ed==NTed) return SDed;

  if(ed==EDed) return WTed;
  if(ed==WDed) return ETed;
  if(ed==ETed) return WDed;
  if(ed==WTed) return EDed;

  if(ed==NEed) return SWed;
  if(ed==NWed) return SEed;
  if(ed==SEed) return NWed;
  return NEed;
}

// edge of a cell
// d direction of the edge, sons direction of the orthogonal face
//--------------------------------------------------------------------
inline Edges_cell CalcEdge_cells(dir3D d, dir2D_sons sons) {
  return (Edges_cell)(((d >> 1) << 2) + sons);
}


// edge of the face of a cell
//-----------------------------
inline Edges_cell Edge_surface(dir3D surf, dir2D dir) {
  if(surf<Sdir3D) {
    return (Edges_cell)(surf+2*(dir>>1)+4*((dir&1)+1));
  }
  if(surf==Ddir3D) {
    if((dir>>1)==0)
      return (Edges_cell)(4*((dir&1)));
    return (Edges_cell)(1+4*((dir&1)));
  }
  if(surf==Tdir3D) {
    if((dir>>1)==0)
      return (Edges_cell)(2+4*((dir&1)));
    return (Edges_cell)(3+4*((dir&1)));
  }
  if((dir&1)==0)
    return (Edges_cell)((surf-2)+2*(dir>>1)+4*(2*(dir&1)));
  return (Edges_cell)(2*(surf-2)+(dir>>1)+4*(2*(dir&1)));

}

//  orthogonal directions of an edge
//  ----------------------------------
inline dir3D OrthoA(Edges_cell ec) {
  static dir3D dir;
  dir = (dir3D) ((ec >> 2) << 1);
  if(dir<Sdir3D) return (dir3D)((ec&1)+2);
  if(dir>Ndir3D) return (dir3D)(ec&1);
  return (dir3D)(ec&1);
}


inline dir3D OrthoB(Edges_cell ec) {
  static dir3D dir;
  dir = (dir3D) ((ec >> 2) << 1);
  if(dir<Sdir3D) return (dir3D) (((ec>>1)&1)+4);
  if(dir>Ndir3D) return (dir3D) (((ec>>1)&1)+2);
  return (dir3D) (((ec>>1)&1)+4);
}

// new: OrthoC is always the left, OrthoD is always the right
inline dir3D OrthoC(Edges_cell ec) {
  if(ec == SDed) return Ddir3D;
  if(ec == NDed) return Ndir3D;
  if(ec == STed) return Sdir3D;
  if(ec == NTed) return Tdir3D;
  if(ec == WDed) return Wdir3D;
  if(ec == EDed) return Ddir3D;
  if(ec == WTed) return Tdir3D;
  if(ec == ETed) return Edir3D;
  if(ec == SWed) return Sdir3D;
  if(ec == SEed) return Edir3D;
  if(ec == NWed) return Wdir3D;
  //if(ec == NEed)
                  return Ndir3D;
}


inline dir3D OrthoD(Edges_cell ec) {
  if(ec == SDed) return Sdir3D;
  if(ec == NDed) return Ddir3D;
  if(ec == STed) return Tdir3D;
  if(ec == NTed) return Ndir3D;
  if(ec == WDed) return Ddir3D;
  if(ec == EDed) return Edir3D;
  if(ec == WTed) return Wdir3D;
  if(ec == ETed) return Tdir3D;
  if(ec == SWed) return Wdir3D;
  if(ec == SEed) return Sdir3D;
  if(ec == NWed) return Ndir3D;
  //if(ec == NEed)
                  return Edir3D;
}

// formula for integer-value: i+3*j+9*k
inline dir3D_sons dir_sons_from_three_LR(dir1D i, dir1D j, dir1D k) {
  return (dir3D_sons)(i+2*j+4*k);
}


///////
// describtion of a 27-stencil by
// - (Ort1D)i, (Ort1D)j, (Ort1D)k       or
// - the integer-value i+3*j+9*k
// here are some transformatiuon formulas
//////


// formula for integer-value: i+3*j+9*k
inline int trans27(Ort1D i, Ort1D j,Ort1D k) {
  return i+3*j+9*k;
}

// integer-value of a point ds in a neighbour cell dz
inline int trans_cellP_27(dir3D_sons dz, dir3D_sons ds) {
  return ((ds&1) + (dz&1)) +
    3*(((ds>>1)&1) + ((dz>>1)&1)) +
    9*(((ds>>2)&1) + ((dz>>2)&1));
}

inline Ort1D trans_cellP_27_x(dir3D_sons dz, dir3D_sons ds) {
  return (Ort1D)((ds&1) + (dz&1));
}
inline Ort1D trans_cellP_27_y(dir3D_sons dz, dir3D_sons ds) {
  return (Ort1D)(((ds>>1)&1) + ((dz>>1)&1));
}
inline Ort1D trans_cellP_27_z(dir3D_sons dz, dir3D_sons ds) {
  return (Ort1D)(((ds>>2)&1) + ((dz>>2)&1));
}

inline Ort1D trans_cellP_3(dir1D dz, dir1D ds) {
  return (Ort1D)((ds&1) + (dz&1));
}





// print-functions for testing
// ----------------------------
inline void Print(dir3D_sons d) {
  if(d==WSDdir3D) cout << "WSDdir3D";
  if(d==ESDdir3D) cout << "ESDdir3D";
  if(d==WNDdir3D) cout << "WNDdir3D";
  if(d==ENDdir3D) cout << "ENDdir3D";
  if(d==WSTdir3D) cout << "WSTdir3D";
  if(d==ESTdir3D) cout << "ESTdir3D";
  if(d==WNTdir3D) cout << "WNTdir3D";
  if(d==ENTdir3D) cout << "ENTdir3D";
}

inline void Print(dir3D d) {
  if(d==Wdir3D) cout << "Wdir3D";
  if(d==Edir3D) cout << "Edir3D";
  if(d==Sdir3D) cout << "Sdir3D";
  if(d==Ndir3D) cout << "Ndir3D";
  if(d==Ddir3D) cout << "Ddir3D";
  if(d==Tdir3D) cout << "Tdir3D";
}

inline void Print(Edges_cell ed) {
  if(ed==SDed) cout << "SDed";
  if(ed==NDed) cout << "NDed";
  if(ed==STed) cout << "STed";
  if(ed==NTed) cout << "NTed";
  if(ed==WDed) cout << "WDed";
  if(ed==EDed) cout << "EDed";
  if(ed==WTed) cout << "WTed";
  if(ed==ETed) cout << "ETed";
  if(ed==SWed) cout << "SWed";
  if(ed==SEed) cout << "SEed";
  if(ed==NWed) cout << "NWed";
  if(ed==NEed) cout << "NEed";
}

inline void Print(dir1D d) {
  if(d==Ldir1D) cout << "Ldir1D";
  if(d==Rdir1D) cout << "Rdir1D";
}



#endif // ABBREVI_H_
