/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#ifndef BASIC_H
#define BASIC_H

template <int i, class Container>
class Topf;

struct sorted_pair
   {
      int first, second;
      sorted_pair(){}

      sorted_pair(int a, int b) : first (a), second (b) {
        if(developer_version)
           if(b <= a)
              std::cout << "Indices of edge in wrong order ... " << a << " <-> " << b << std::endl;
        assert ( a < b );
      }

      sorted_pair( const sorted_pair &c ) : first(c.first), second(c.second) {}

      const sorted_pair& operator= ( const sorted_pair &c )
        {
          first = c.first;
          second = c.second;
          return (*this);
        }
      bool operator < (const sorted_pair& b) const 
         {
            if (first == b.first)
              {
                 return second < b.second;
              }
            else
              {
                return first < b.first;
              }
         }
   };


struct unsorted_pair
   {
      int first, second;
      unsorted_pair(){} 
      unsorted_pair(int a, int b) : first (a), second (b) 
       {}
      unsorted_pair( const unsorted_pair &c ) : first(c.first), second(c.second) {}

      const unsorted_pair& operator= ( const sorted_pair &c )
        {
          first = c.first;
          second = c.second;
          return (*this);
        }
      bool operator < (const unsorted_pair& b) const 
         {
            if (first == b.first)
              {
                 return second < b.second;
              }
            else
              {
                return first < b.first;
              }
         }
   };


struct sorted_quad
   {
      int first, second, third, forth;
      sorted_quad(){} 
      sorted_quad(int a, int b, int c, int d) : 
          first(a), second(b), third(c), forth(d) {} 
      sorted_quad( const sorted_quad &c ) : 
          first(c.first), second(c.second), third(c.third), forth(c.forth) {} 
      const sorted_quad& operator= ( const sorted_quad &c )
        {
          first = c.first;
          second = c.second;
          third = c.third;
          forth = c.forth;
          return (*this);
        }
      bool operator < (const sorted_quad& b) const 
         {
        // ordering the entries ...
           int array_A[4] = {first,second,third,forth};
           for (int i= 0; i< 3; ++i)
              {
                for (int j= 0; j< 3; ++j)
                if (array_A[j] > array_A[j+1]) 
                   {
                     int temp = array_A[j];
                     array_A[j] = array_A[j+1];
                     array_A[j+1] = temp;
                   }  
              }
           int array_B[4] = {b.first,b.second,b.third,b.forth};
           for (int i= 0; i< 3; ++i)
              {
                for (int j= 0; j< 3; ++j)
                if (array_B[j] > array_B[j+1]) 
                   {
                     int temp = array_B[j];
                     array_B[j] = array_B[j+1];
                     array_B[j+1] = temp;
                   }  
              }
        // comparing ...
           if (array_A[0] == array_B[0])
              {
                 if (array_A[1] == array_B[1])
                   {
                     if(array_A[2] == array_B[2])
                       {
                         return array_A[3] < array_B[3];
                       }
                     else
                       {
                         return array_A[2] < array_B[2];
                       }
                   }
                 else
                   {
                     return array_A[1] < array_B[1];
                   }
              }
            else
              {
                return array_A[0] < array_B[0];
              }
         }
   };


struct unsorted_quad
   {
      int first, second, third, forth;
      unsorted_quad(){} 
      unsorted_quad(int a, int b, int c, int d) : 
          first(a), second(b), third(c), forth(d) {} 
      unsorted_quad( const sorted_quad &c ) : 
          first(c.first), second(c.second), third(c.third), forth(c.forth) {} 
      const unsorted_quad& operator= ( const unsorted_quad &c )
        {
          first = c.first;
          second = c.second;
          third = c.third;
          forth = c.forth;
          return (*this);
        }
      bool operator < (const unsorted_quad& b) const 
         {
        // comparing ...
           if (first == b.first)
              {
                 if (second == b.second)
                   {
                     if(third == b.third)
                       {
                         return forth < b.forth;
                       }
                     else
                       {
                         return third < b.third;
                       }
                   }
                 else
                   {
                     return second < b.second;
                   }
              }
            else
              {
                return first < b.first;
              }
         }
   };


template <class Container>
class Topf<2,Container> {
   private:
    typename std::map<Container,int> data_list;
   public:
  Topf()
    {
      data_list.clear();
    }
  void put(int a, int b)
     {
//        assert (a != b);
       Container init(a,b);
       if (data_list.find(init) == data_list.end())
          {
             int number = data_list.size();
             data_list[init] = number;
          }
     }
  int Give_id(int a, int b)
     {
       Container init(a,b);
       assert (data_list.find(init) != data_list.end());
       return data_list[init];
     }
  int size() const 
     {
       return data_list.size();
     }
  const Container& find( int index ) const 
     {
       typename std::map<Container,int>::const_iterator iter = data_list.begin();
       while (iter->second != index && iter != data_list.end())
          {
             iter++;
          }
       assert ( iter != data_list.end() );
       return iter->first;
     }
};

template <class Container>
class Topf<4,Container> {
   private:
    typename std::map<Container,int> data_list;
   public:
  Topf()
    {
      data_list.clear();
    }
  void put(int a, int b, int c, int d)
     {
       Container init(a,b,c,d);
       if (data_list.find(init) == data_list.end())
          {
             int number = data_list.size();
             data_list[init] = number;
          }
     }
  int Give_id(int a, int b, int c, int d)
     {
       Container init(a,b,c,d);
       assert (data_list.find(init) != data_list.end());
       return data_list[init];
     }
  int size() const 
     {
       return data_list.size();
     }
  const Container& find( int index ) const 
     {
       typename std::map<Container,int>::const_iterator iter = data_list.begin();
       while (iter->second != index && iter != data_list.end())
          {
             iter++;
          }
       assert ( iter != data_list.end() );
       return iter->first;
     }
};

#endif // BASIC_H

