/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// co_fu.h
//
// ------------------------------------------------------------

#ifndef CO_FU_H
#define CO_FU_H

//////////////////////////////////////////////////////////////
// 1. Coordinate
// 2. Meshsize x,y,z
// 3. Functions
//////////////////////////////////////////////////////////////

// Gitter:
// * boundary point
// + interior point

/*
Ny    *  *  *  *  *  *
Ny-1  *  +  +  +  +  *
3     *  +  +  +  +  *
2     *  +  +  +  +  *
1     *  +  +  +  +  *
0     *  *  *  *  *  *
      0  1  2  3 Nx-1 Nx
*/


///////////////////////////
// 1. Coordinate
///////////////////////////

/** \defgroup ExpressionTemplates  ''Expression Template Operators in 3D (excluding +,-,... ) ''    
 **/
/* @{  */ 
class X_coordinate : public Expr<X_coordinate> {

	public:
		X_coordinate ( Blockgrid& bg ) {
			blockgrid = &bg;
		}
		
		Blockgrid* Give_blockgrid() const { return blockgrid; };
		
		stencil_typ Give_stencil_typ() const { return no_stencil; }
                bool totalCalcNotPossible()    const { return true; }	
		control_typ Give_control_typ() const { return thread_save; }
		
		typedef double Result;
		
		template <elementTyp TYP_EL>
		inline double Give_data ( params_in ) const;
		
		inline double Give_cell_hexahedra ( params_in_cell ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {}
  
                double Give_fromTotal(int i) const  { assert(false); return 0.0; }
	private:
		Blockgrid*     blockgrid;
};
/* @} */

inline double X_coordinate::Give_cell_hexahedra ( params_in_cell ) const {
	//  return  blockgrid->Give_coord_hexahedron(id_hex, i, j, k).x;
	
	return 0.125* ( blockgrid->Give_coord_hexahedron ( id_hex, i  , j  , k ).x +
	                blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j  , k ).x +
	                blockgrid->Give_coord_hexahedron ( id_hex, i  , j + 1, k ).x +
	                blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j + 1, k ).x +
	                blockgrid->Give_coord_hexahedron ( id_hex, i  , j  , k + 1 ).x +
	                blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j  , k + 1 ).x +
	                blockgrid->Give_coord_hexahedron ( id_hex, i  , j + 1, k + 1 ).x +
	                blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j + 1, k + 1 ).x );
}

template <>
inline double X_coordinate::Give_data<hexahedronEl> ( params_in ) const {
	return blockgrid->Give_coord_hexahedron ( id, i, j, k ).x;
}

template <>
inline double X_coordinate::Give_data<quadrangleEl> ( params_in ) const {
	return blockgrid->Give_coord_quadrangle ( id, i, j ).x;
}

template <>
inline double X_coordinate::Give_data<edgeEl> ( params_in ) const {
	return blockgrid->Give_coord_edge ( id, i ).x;
}

template <>
inline double X_coordinate::Give_data<pointEl> ( params_in ) const {
	return blockgrid->Give_coord_point ( id ).x;
}

/** \addtogroup ExpressionTemplates **/
/* @{ */  
class Y_coordinate : public Expr<Y_coordinate> {

	public:
		Y_coordinate ( Blockgrid& bg ) {
			blockgrid = &bg;
		}
		
		Blockgrid* Give_blockgrid() const { return blockgrid; };
		
		stencil_typ Give_stencil_typ() const { return no_stencil; }
                bool totalCalcNotPossible()    const { return true; }			
		control_typ Give_control_typ() const { return thread_save; }
		
		typedef double Result;
		
		template <elementTyp TYP_EL>
		inline double Give_data ( params_in ) const;
		
		inline double Give_cell_hexahedra ( params_in_cell ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {}
                double Give_fromTotal(int i) const  { assert(false); return 0.0; }		
	private:
		Blockgrid*     blockgrid;
};
/* @} */

inline double Y_coordinate::Give_cell_hexahedra ( params_in_cell ) const {
	return 0.125* ( blockgrid->Give_coord_hexahedron ( id_hex, i  , j  , k ).y +
	                blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j  , k ).y +
	                blockgrid->Give_coord_hexahedron ( id_hex, i  , j + 1, k ).y +
	                blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j + 1, k ).y +
	                blockgrid->Give_coord_hexahedron ( id_hex, i  , j  , k + 1 ).y +
	                blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j  , k + 1 ).y +
	                blockgrid->Give_coord_hexahedron ( id_hex, i  , j + 1, k + 1 ).y +
	                blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j + 1, k + 1 ).y );
}

template <>
inline double Y_coordinate::Give_data<hexahedronEl> ( params_in ) const {
	return blockgrid->Give_coord_hexahedron ( id, i, j, k ).y;
}

template <>
inline double Y_coordinate::Give_data<quadrangleEl> ( params_in ) const {
	return blockgrid->Give_coord_quadrangle ( id, i, j ).y;
}

template <>
inline double Y_coordinate::Give_data<edgeEl> ( params_in ) const {
	return blockgrid->Give_coord_edge ( id, i ).y;
}

template <>
inline double Y_coordinate::Give_data<pointEl> ( params_in ) const {
	return blockgrid->Give_coord_point ( id ).y;
}

//////////////////////////////////////////////////////////////////////

/** \addtogroup ExpressionTemplates **/
/* @{ */  
class Z_coordinate : public Expr<Z_coordinate> {

	public:
		Z_coordinate ( Blockgrid& bg ) {
			blockgrid = &bg;
		}
		
		Blockgrid* Give_blockgrid() const { return blockgrid; };
		
		stencil_typ Give_stencil_typ() const { return no_stencil; }
                bool totalCalcNotPossible()    const { return true; }			
		control_typ Give_control_typ() const { return thread_save; }
		
		typedef double Result;
		
		template <elementTyp TYP_EL>
		inline double Give_data ( params_in ) const;
		
		inline double Give_cell_hexahedra ( params_in_cell ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {}
                double Give_fromTotal(int i) const  { assert(false); return 0.0; }		
	private:
		Blockgrid*     blockgrid;
};
/* @} */

inline double Z_coordinate::Give_cell_hexahedra ( params_in_cell ) const {
	return 0.125* ( blockgrid->Give_coord_hexahedron ( id_hex, i  , j  , k ).z +
	                blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j  , k ).z +
	                blockgrid->Give_coord_hexahedron ( id_hex, i  , j + 1, k ).z +
	                blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j + 1, k ).z +
	                blockgrid->Give_coord_hexahedron ( id_hex, i  , j  , k + 1 ).z +
	                blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j  , k + 1 ).z +
	                blockgrid->Give_coord_hexahedron ( id_hex, i  , j + 1, k + 1 ).z +
	                blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j + 1, k + 1 ).z );
}

template <>
inline double Z_coordinate::Give_data<hexahedronEl> ( params_in ) const {
	return blockgrid->Give_coord_hexahedron ( id, i, j, k ).z;
}

template <>
inline double Z_coordinate::Give_data<quadrangleEl> ( params_in ) const {
	return blockgrid->Give_coord_quadrangle ( id, i, j ).z;
}

template <>
inline double Z_coordinate::Give_data<edgeEl> ( params_in ) const {
	return blockgrid->Give_coord_edge ( id, i ).z;
}

template <>
inline double Z_coordinate::Give_data<pointEl> ( params_in ) const {
	return blockgrid->Give_coord_point ( id ).z;
}


/////////////////////////////////////////////////////////////////////////

///////////////////////////
// 2. Meshsize_x,y,z
///////////////////////////

/** \addtogroup ExpressionTemplates **/
/* @{ */  
///////// x-direction
class X_meshsize : public Expr<X_meshsize> {

	public:
		X_meshsize ( Blockgrid& bg ) {
			blockgrid = &bg;
		}
		
		Blockgrid* Give_blockgrid() const { return blockgrid; };
		
		stencil_typ Give_stencil_typ() const { return no_stencil; }
		control_typ Give_control_typ() const { return thread_save; }
		
		typedef double Result;
		
		
		//  template <elementTyp TYP_EL>
		//  inline double Give_data(params_in) const;
		
		inline double Give_cell_hexahedra ( params_in_cell ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {}
		
	private:
		Blockgrid*     blockgrid;
};
/* @} */

inline double X_meshsize::Give_cell_hexahedra ( params_in_cell ) const {
	return 0.25 * ABS ( ( blockgrid->Give_coord_hexahedron ( id_hex, i + 1 , j     , k ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i + 1 , j + 1 , k ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i + 1 , j     , k + 1 ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i + 1 , j + 1 , k + 1 ).z ) -
	                    ( blockgrid->Give_coord_hexahedron ( id_hex, i,      j     , k ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i,      j + 1 , k ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i,      j     , k + 1 ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i,      j + 1 , k + 1).z ) );
}

/** \addtogroup ExpressionTemplates **/
/* @{ */  
///////// y-direction
class Y_meshsize : public Expr<Y_meshsize> {

	public:
		Y_meshsize ( Blockgrid& bg ) {
			blockgrid = &bg;
		}
		
		Blockgrid* Give_blockgrid() const { return blockgrid; };
		
		stencil_typ Give_stencil_typ() const { return no_stencil; }
		control_typ Give_control_typ() const { return thread_save; }
		
		typedef double Result;
		
		
		//  template <elementTyp TYP_EL>
		//  inline double Give_data(params_in) const;
		
		inline double Give_cell_hexahedra ( params_in_cell ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {}
		
	private:
		Blockgrid*     blockgrid;
};
/* @} */

inline double Y_meshsize::Give_cell_hexahedra ( params_in_cell ) const {
  	return 0.25 * ABS ( ( blockgrid->Give_coord_hexahedron ( id_hex, i    , j + 1, k     ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j + 1, k     ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i    , j + 1, k + 1 ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j + 1, k + 1 ).z ) -
	                    ( blockgrid->Give_coord_hexahedron ( id_hex, i    , j    , k     ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j    , k     ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i    , j    , k + 1 ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j    , k + 1 ).z ) );
}

/** \addtogroup ExpressionTemplates **/
/* @{ */  
///////// z-direction
class Z_meshsize : public Expr<Z_meshsize> {

	public:
		Z_meshsize ( Blockgrid& bg ) {
			blockgrid = &bg;
		}
		
		Blockgrid* Give_blockgrid() const { return blockgrid; };
		
		stencil_typ Give_stencil_typ() const { return no_stencil; }
		control_typ Give_control_typ() const { return thread_save; }
		
		typedef double Result;
		
		
		//  template <elementTyp TYP_EL>
		//  inline double Give_data(params_in) const;
		
		inline double Give_cell_hexahedra ( params_in_cell ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {}
		
	private:
		Blockgrid*     blockgrid;
};
/* @} */

inline double Z_meshsize::Give_cell_hexahedra ( params_in_cell ) const {
	return 0.25 * ABS ( ( blockgrid->Give_coord_hexahedron ( id_hex, i  , j  , k + 1 ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j  , k + 1 ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i  , j + 1, k + 1 ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j + 1, k + 1 ).z ) -
	                    ( blockgrid->Give_coord_hexahedron ( id_hex, i  , j  , k ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j  , k ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i  , j + 1, k ).z +
	                      blockgrid->Give_coord_hexahedron ( id_hex, i + 1, j + 1, k ).z ) );
}












/*
template <>
inline double Z_meshsize::Give_data<hexahedronEl>(params_in) const {
  return blockgrid->Give_coord_hexahedron(id, i, j, k).z;
}

template <>
inline double Z_meshsize::Give_data<quadrangleEl>(params_in) const {
    return blockgrid->Give_coord_quadrangle(id, i, j).z;
}

template <>
inline double Z_meshsize::Give_data<edgeEl>(params_in) const {
    return blockgrid->Give_coord_edge(id, i).z;
}

template <>
inline double Z_meshsize::Give_data<pointEl>(params_in) const {
    return blockgrid->Give_coord_point(id).z;
}
*/

///////////////////////////
// 3. Functions
///////////////////////////
template <class VType>
class Variable2D;

//  1 argument

template <class A, class DTyp, class DTyp_Result>
class Exp_Function1 : public Expr<Exp_Function1<A, DTyp, DTyp_Result> > {
		const A& a_;
		DTyp_Result ( *Formula_ ) ( DTyp x );
		control_typ funCtrTyp_;
		
	public:
		inline Exp_Function1 ( const A& a, DTyp_Result ( *Formula ) ( DTyp x ), control_typ funCtrTyp ) :
				a_ ( a ), Formula_ ( Formula ), funCtrTyp_ ( funCtrTyp ) {}
				
		stencil_typ Give_stencil_typ() const { return a_.Give_stencil_typ();     }
		bool totalCalcNotPossible()    const { return a_.totalCalcNotPossible(); }
		control_typ Give_control_typ() const { return funCtrTyp_; }
		
		typedef DTyp_Result Result;
		
		template <elementTyp TYP_EL>
		inline DTyp_Result Give_data ( params_in ) const;
		
		inline DTyp_Result Give_cell_hexahedra ( params_in_cell ) const;
		inline DTyp_Result Give_matrix_hexahedra(params_in_loc_matrix) const;
		
		inline Result Give_fromTotal(int i) const;		
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
		}

		template <class VType>
                friend class Variable2D;
};

template <class A, class DTyp, class DTyp_Result>
template <elementTyp TYP_EL>
inline DTyp_Result Exp_Function1<A, DTyp, DTyp_Result>::Give_data ( params_in ) const {
	return Formula_ ( a_.template  Give_data<TYP_EL> ( params_out ) );
}

template <class A, class DTyp, class DTyp_Result>
inline DTyp_Result Exp_Function1<A, DTyp, DTyp_Result>::Give_cell_hexahedra ( params_in_cell ) const {
	return Formula_ ( a_.Give_cell_hexahedra ( params_out_cell ) );
}

template <class A, class DTyp, class DTyp_Result>
inline DTyp_Result Exp_Function1<A, DTyp, DTyp_Result>::Give_matrix_hexahedra ( params_in_loc_matrix ) const {
	return Formula_ ( a_.Give_matrix_hexahedra ( params_out_loc_matrix ) );
}

template <class A, class DTyp, class DTyp_Result>
inline DTyp_Result Exp_Function1<A, DTyp, DTyp_Result>::Give_fromTotal(int i) const {
	return Formula_ ( a_.Give_fromTotal(i));
}

/** \addtogroup ExpressionTemplates **/
/* @{ */  
template <class DTyp_Result, class DTyp>
class Function1 {

	public:
		Function1 ( DTyp_Result ( *formula ) ( DTyp x ), control_typ funCtrTyp = thread_save ) : Formula ( formula ), funCtrTyp_ ( funCtrTyp ) {}
		
		template <class A>
		inline Exp_Function1<A, DTyp, DTyp_Result> operator() ( const Expr<A>& a ) const
		          { return Exp_Function1<A, DTyp, DTyp_Result> ( a, Formula, funCtrTyp_ ); }
		
	private:
		DTyp_Result ( *Formula ) ( DTyp x );
		control_typ funCtrTyp_;
};
/* @} */

//  2 arguments

template <class A, class B, class DTyp>

class Exp_Function2 : public Expr<Exp_Function2<A, B, DTyp> > {
		const A& a_;
		const B& b_;
		DTyp ( *Formula_ ) ( DTyp x, DTyp y );
		control_typ funCtrTyp_;
		
	public:
		inline Exp_Function2 ( const A& a, const B& b, DTyp ( *Formula ) ( DTyp x, DTyp y ), control_typ funCtrTyp ) :
				a_ ( a ), b_ ( b ), Formula_ ( Formula ), funCtrTyp_ ( funCtrTyp ) {}
				
		stencil_typ Give_stencil_typ() const {
			if(a_.Give_stencil_typ() == yes_stencil ) return yes_stencil;
			return b_.Give_stencil_typ();
		};		
		
		bool totalCalcNotPossible() const {
			if(a_.totalCalcNotPossible()) return true;
			return b_.totalCalcNotPossible();
		};		
		
		control_typ Give_control_typ() const { return funCtrTyp_; }
		
		typedef DTyp Result;
		
		template <elementTyp TYP_EL>
		inline DTyp Give_data ( params_in ) const;
		
		inline DTyp Give_cell_hexahedra ( params_in_cell ) const;

		inline Result Give_fromTotal(int i) const;		
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
                        b_.template Update<TYP_EL> ( id );
		}
};

template <class A, class B, class DTyp>
template <elementTyp TYP_EL>
inline DTyp Exp_Function2<A, B, DTyp>::Give_data ( params_in ) const {
	return Formula_ ( a_.template  Give_data<TYP_EL> ( params_out ), b_.template  Give_data<TYP_EL> ( params_out ) );
}

template <class A, class B, class DTyp>
inline DTyp Exp_Function2<A, B, DTyp>::Give_cell_hexahedra ( params_in_cell ) const {
	return Formula_ ( a_.Give_cell_hexahedra ( params_out_cell ),
	                  b_.Give_cell_hexahedra ( params_out_cell ) );
}

template <class A, class B, class DTyp>
inline DTyp Exp_Function2<A, B, DTyp>::Give_fromTotal(int i) const {
	return Formula_ ( a_.Give_fromTotal(i),
	                  b_.Give_fromTotal(i) );
}

/** \addtogroup ExpressionTemplates **/
/* @{ */  
template <class DTyp>

class Function2 {

	public:
		Function2 ( DTyp ( *formula ) ( DTyp x, DTyp y ), control_typ funCtrTyp = thread_save ) : Formula ( formula ), funCtrTyp_ ( funCtrTyp ) {};
		
		template <class A, class B>
		inline Exp_Function2<A, B, DTyp> operator() ( const Expr<A>& a,
		        const Expr<B>& b ) const
		{ return Exp_Function2<A, B, DTyp> ( a, b, Formula, funCtrTyp_ ); }
		
		
	private:
		DTyp ( *Formula ) ( DTyp x, DTyp y );
		control_typ funCtrTyp_;
};
/* @} */

//  3 arguments

template <class A, class B, class C, class DTyp>

class Exp_Function3 : public Expr<Exp_Function3<A, B, C, DTyp> > {
		const A& a_;
		const B& b_;
		const C& c_;
		DTyp ( *Formula_ ) ( DTyp x, DTyp y, DTyp z );
		control_typ funCtrTyp_;
		
	public:
		inline Exp_Function3 ( const A& a, const B& b, const C& c, DTyp ( *Formula ) ( DTyp x, DTyp y, DTyp z ), control_typ funCtrTyp ) :
				a_ ( a ), b_ ( b ), c_ ( c ), Formula_ ( Formula ), funCtrTyp_ ( funCtrTyp ) {}
				
		stencil_typ Give_stencil_typ() const {
			if(a_.Give_stencil_typ() == yes_stencil ) return yes_stencil;
			if(b_.Give_stencil_typ() == yes_stencil ) return yes_stencil;			
			return c_.Give_stencil_typ();
		};		
		
		bool totalCalcNotPossible() const {
			if(a_.totalCalcNotPossible()) return true;
			if(b_.totalCalcNotPossible()) return true;			
			return c_.totalCalcNotPossible();
		};		
		
		
		control_typ Give_control_typ() const { return funCtrTyp_; }
		
		typedef DTyp Result;
		
		template <elementTyp TYP_EL>
		inline DTyp Give_data ( params_in ) const;

		inline DTyp Give_cell_hexahedra ( params_in_cell ) const;		

		inline Result Give_fromTotal(int i) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			            a_.template Update<TYP_EL> ( id );
                        b_.template Update<TYP_EL> ( id );
                        c_.template Update<TYP_EL> ( id );
		}
};

template <class A, class B, class C, class DTyp>
template <elementTyp TYP_EL>
inline DTyp Exp_Function3<A, B, C, DTyp>::Give_data ( params_in ) const {
	return Formula_ ( a_.template Give_data<TYP_EL> ( params_out ),
	                  b_.template Give_data<TYP_EL> ( params_out ),
	                  c_.template Give_data<TYP_EL> ( params_out ) );
}

template <class A, class B, class C, class DTyp>
inline DTyp Exp_Function3<A, B, C, DTyp>::Give_cell_hexahedra ( params_in_cell ) const {
	return Formula_ ( a_.Give_cell_hexahedra ( params_out_cell ),
	                  b_.Give_cell_hexahedra ( params_out_cell ),
	                  c_.Give_cell_hexahedra ( params_out_cell ) );
}

template <class A, class B, class C, class DTyp>
inline DTyp Exp_Function3<A, B, C, DTyp>::Give_fromTotal(int i) const {
	return Formula_ ( a_.Give_fromTotal(i),
	                  b_.Give_fromTotal(i),
	                  c_.Give_fromTotal(i) );
}

/** \addtogroup ExpressionTemplates **/
/* @{ */  
template <class DTyp>
class Function3 {

	public:
		Function3 ( DTyp ( *formula ) ( DTyp x, DTyp y, DTyp z ), control_typ funCtrTyp = thread_save ) : Formula ( formula ), funCtrTyp_ ( funCtrTyp ) {};
		
		template <class A, class B, class C>
		inline Exp_Function3<A, B, C, DTyp> operator() ( const Expr<A>& a,
		        const Expr<B>& b,
		        const Expr<C>& c ) const
		{ return Exp_Function3<A, B, C, DTyp> ( a, b, c, Formula, funCtrTyp_ ); }
		
		
		
	private:
		DTyp ( *Formula ) ( DTyp x, DTyp y, DTyp z );
		control_typ funCtrTyp_;
};//  4 arguments
/* @} */

template <class A, class B, class C, class D, class DTyp, class DTyp_Result>

class Exp_Function4 : public Expr<Exp_Function4<A, B, C, D, DTyp,DTyp_Result> > {
                const A& a_;
                const B& b_;
                const C& c_;
                const D& d_;
                DTyp_Result ( *Formula_ ) ( DTyp x, DTyp y, DTyp z, DTyp z2 );
                control_typ funCtrTyp_;

        public:
                inline Exp_Function4 ( const A& a, const B& b, const C& c, const D& d, DTyp_Result ( *Formula ) ( DTyp x, DTyp y, DTyp z, DTyp z2 ), control_typ funCtrTyp ) :
                                a_ ( a ), b_ ( b ), c_ ( c ), d_ ( d ), Formula_ ( Formula ), funCtrTyp_ ( funCtrTyp ) {}

		stencil_typ Give_stencil_typ() const {
			if(a_.Give_stencil_typ() == yes_stencil ) return yes_stencil;
			if(b_.Give_stencil_typ() == yes_stencil ) return yes_stencil;	
			if(c_.Give_stencil_typ() == yes_stencil ) return yes_stencil;				
			return d_.Give_stencil_typ();
		};		
		
		bool totalCalcNotPossible() const {
			if(a_.totalCalcNotPossible()) return true;
			if(b_.totalCalcNotPossible()) return true;
			if(c_.totalCalcNotPossible()) return true;						
			return d_.totalCalcNotPossible();
		};
                control_typ Give_control_typ() const { return funCtrTyp_; }

                typedef DTyp Result;

                template <elementTyp TYP_EL>
                inline DTyp_Result Give_data ( params_in ) const;

                inline DTyp_Result Give_cell_hexahedra ( params_in_cell ) const;

		inline DTyp_Result Give_fromTotal(int i) const;		

                template <elementTyp TYP_EL>
                void Update ( int id ) const {
                        a_.template Update<TYP_EL> ( id );
                        b_.template Update<TYP_EL> ( id );
                        c_.template Update<TYP_EL> ( id );
                        d_.template Update<TYP_EL> ( id );
                }

                template <class VType>
                friend class Variable2D;
};

template <class A, class B, class C, class D, class DTyp, class DTyp_Result>
template <elementTyp TYP_EL>
inline DTyp_Result Exp_Function4<A, B, C, D, DTyp, DTyp_Result>::Give_data ( params_in ) const {
        return Formula_ ( a_.template  Give_data<TYP_EL> ( params_out ),
                          b_.template  Give_data<TYP_EL> ( params_out ),
                          c_.template  Give_data<TYP_EL> ( params_out ),
                          d_.template  Give_data<TYP_EL> ( params_out ) );
}

template <class A, class B, class C, class D, class DTyp, class DTyp_Result>
inline DTyp_Result Exp_Function4<A, B, C, D, DTyp, DTyp_Result>::Give_cell_hexahedra ( params_in_cell ) const {
        return Formula_ ( a_.Give_cell_hexahedra ( params_out_cell ),
                          b_.Give_cell_hexahedra ( params_out_cell ),
                          c_.Give_cell_hexahedra ( params_out_cell ),
                          d_.Give_cell_hexahedra ( params_out_cell ) );
}

template <class A, class B, class C, class D, class DTyp, class DTyp_Result>
inline DTyp_Result Exp_Function4<A, B, C, D, DTyp, DTyp_Result>::Give_fromTotal(int i) const {
	return Formula_ ( a_.Give_fromTotal(i),
			  b_.Give_fromTotal(i),
	                  c_.Give_fromTotal(i),
	                  d_.Give_fromTotal(i) );
}

/** \addtogroup ExpressionTemplates **/
/* @{ */  
template <class DTyp, class DTyp_Result>
class Function4 {

        public:
                Function4 ( DTyp_Result ( *formula ) ( DTyp x, DTyp y, DTyp z, DTyp z2 ), control_typ funCtrTyp = thread_save ) : Formula ( formula ), funCtrTyp_ ( funCtrTyp ) {};

                template <class A, class B, class C, class D>
                inline Exp_Function4<A, B, C, D, DTyp, DTyp_Result> operator() ( const Expr<A>& a,
                        const Expr<B>& b,
                        const Expr<C>& c,
                        const Expr<D>& d ) const
                { return Exp_Function4<A, B, C, D, DTyp, DTyp_Result> ( a, b, c, d, Formula, funCtrTyp_ ); }



        private:
                DTyp_Result ( *Formula ) ( DTyp x, DTyp y, DTyp z, DTyp z2 );
                control_typ funCtrTyp_;
};
/* @} */


//////////////////////////////////////////////////////////////
// fuer Thomas

//  3 arguments

template <class A, class B, class C, class DTyp, class DTyp_Result>
class Exp_Function3R : public Expr<Exp_Function3R<A, B, C, DTyp, DTyp_Result> > {
		const A& a_;
		const B& b_;
		const C& c_;
		DTyp_Result ( *Formula_ ) ( DTyp x, DTyp y, DTyp z );
		control_typ funCtrTyp_;
		
	public:
		inline Exp_Function3R ( const A& a, const B& b, const C& c, DTyp_Result ( *Formula ) ( DTyp x, DTyp y, DTyp z ), control_typ funCtrTyp ) :
				a_ ( a ), b_ ( b ), c_ ( c ), Formula_ ( Formula ), funCtrTyp_ ( funCtrTyp ) {}
				
		stencil_typ Give_stencil_typ() const {
			if(a_.Give_stencil_typ() == yes_stencil ) return yes_stencil;
			if(b_.Give_stencil_typ() == yes_stencil ) return yes_stencil;			
			return c_.Give_stencil_typ();
		};		
		
		bool totalCalcNotPossible() const {
			if(a_.totalCalcNotPossible()) return true;
			if(b_.totalCalcNotPossible()) return true;			
			return c_.totalCalcNotPossible();
		};
		
		control_typ Give_control_typ() const { return funCtrTyp_; }
		
		typedef DTyp Result;
		
		template <elementTyp TYP_EL>
		inline DTyp_Result Give_data ( params_in ) const;
		
		inline DTyp_Result Give_cell_hexahedra ( params_in_cell ) const;
		
		inline DTyp_Result Give_fromTotal(int i) const;		
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
                        b_.template Update<TYP_EL> ( id );
                        c_.template Update<TYP_EL> ( id );
		}
};

template <class A, class B, class C, class DTyp, class DTyp_Result>
template <elementTyp TYP_EL>
inline DTyp_Result Exp_Function3R<A, B, C, DTyp, DTyp_Result>::Give_data ( params_in ) const {
	return Formula_ ( a_.template Give_data<TYP_EL> ( params_out ),
	                  b_.template Give_data<TYP_EL> ( params_out ),
	                  c_.template  Give_data<TYP_EL> ( params_out ) );
}

template <class A, class B, class C, class DTyp, class DTyp_Result>
inline DTyp_Result Exp_Function3R<A, B, C, DTyp, DTyp_Result>::Give_cell_hexahedra ( params_in_cell ) const {
	return Formula_ ( a_.Give_cell_hexahedra ( params_out_cell ),
	                  b_.Give_cell_hexahedra ( params_out_cell ),
	                  c_.Give_cell_hexahedra ( params_out_cell ) );
}

template <class A, class B, class C, class DTyp, class DTyp_Result>
inline DTyp_Result Exp_Function3R<A, B, C, DTyp, DTyp_Result>::Give_fromTotal(int i) const {
	return Formula_ ( a_.Give_fromTotal(i),
			  b_.Give_fromTotal(i),
	                  c_.Give_fromTotal(i) );
}

/** \addtogroup ExpressionTemplates **/
/* @{ */  
template <class DTyp_Result, class DTyp>
class Function3R {

	public:
		Function3R ( DTyp_Result ( *formula ) ( DTyp x, DTyp y, DTyp z ), control_typ funCtrTyp = thread_save ) : Formula ( formula ), funCtrTyp_ ( funCtrTyp ) {};
		
		template <class A, class B, class C>
		inline Exp_Function3R<A, B, C, DTyp, DTyp_Result> operator() ( const Expr<A>& a,
		        const Expr<B>& b,
		        const Expr<C>& c ) const
		{ return Exp_Function3R<A, B, C, DTyp, DTyp_Result> ( a, b, c, Formula, funCtrTyp_ ); }
		
		
		
	private:
		DTyp_Result ( *Formula ) ( DTyp x, DTyp y, DTyp z );
		control_typ funCtrTyp_;
};
/* @} */


#endif // CO_FU_H
