/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// copy.h
//
// ------------------------------------------------------------

#ifndef COPY_H
#define COPY_H

//////////////////////////////////////////////////////////////
// 1. copy function of Variable 
//////////////////////////////////////////////////////////////

#include "variable.h"

// copy of data from hex_a to hex_b
template <typename DTyp> 
void Variable<DTyp>::Copy_invert_z(int hex_b, Variable<DTyp>& a, int hex_a) {
  int Nx, Ny, Nz;
  int i,j,k;
  int max_num_hex = ug->Give_number_hexahedra();
  
  if(hex_a<0 || hex_b<0 || hex_a>=max_num_hex || hex_b>=max_num_hex) {
    throw(std::runtime_error(std::string(" error 1 in  Variable<DTyp>::Copy_invert_z ")));
  }

  Update<hexahedronEl>(hex_a);

  /** @todo: alle daten kopieren oder rück-update implementieren 
  *** @todo: MPI
  **/
  
  Nx = a.Give_blockgrid()->Give_Nx_hexahedron(hex_b);
  Ny = a.Give_blockgrid()->Give_Ny_hexahedron(hex_b);
  Nz = a.Give_blockgrid()->Give_Nz_hexahedron(hex_b);
  
  if(Nx!=a.Give_blockgrid()->Give_Nx_hexahedron(hex_a) ||
     Ny!=a.Give_blockgrid()->Give_Ny_hexahedron(hex_a) ||
     Nz!=a.Give_blockgrid()->Give_Nz_hexahedron(hex_a)) {
     throw(std::runtime_error(std::string(" error 2 in  Variable<DTyp>::Copy_invert_z ")));
  }


  /// todo dieses parallel for kann man vielleicht noch reinbauen  
  #pragma omp parallel for private(j,i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
  for(k=0;k<=Nz;++k) {
    for(j=0;j<=Ny;++j)
      for(i=0;i<=Nx;++i)
        data_hexahedra[hex_b][i+(Nx+1)*(j+(Ny+1)*k)] = a.data_hexahedra[hex_a][i+(Nx+1)*(j+(Ny+1)*(Nz-k))];
  }
}



// copy of data from hex_a to hex_b
template <typename DTyp> 
void Cell_variable<DTyp>::Copy_invert_z(int hex_b, Cell_variable<DTyp>& a, int hex_a) {
  int Nx, Ny, Nz;
  int i,j,k;
  int max_num_hex = ug->Give_number_hexahedra();
  
  if(hex_a<0 || hex_b<0 || hex_a>=max_num_hex || hex_b>=max_num_hex) {
    throw(std::runtime_error(std::string(" error 3 in  Cell_Variable<DTyp>::Copy_invert_z ")));
  }
  
  /// @todo: MPI
  
  Nx = a.Give_blockgrid()->Give_Nx_hexahedron(hex_b);
  Ny = a.Give_blockgrid()->Give_Ny_hexahedron(hex_b);
  Nz = a.Give_blockgrid()->Give_Nz_hexahedron(hex_b);
  
  if(Nx!=a.Give_blockgrid()->Give_Nx_hexahedron(hex_a) ||
     Ny!=a.Give_blockgrid()->Give_Ny_hexahedron(hex_a) ||
     Nz!=a.Give_blockgrid()->Give_Nz_hexahedron(hex_a)) {
     throw(std::runtime_error(std::string(" error 4 in  Cell_Variable<DTyp>::Copy_invert_z ")));
  }

  
  #pragma omp parallel for private(k,j,i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
  for(k=0;k<Nz;++k) {
      for(j=0;j<Ny;++j)
          for(i=0;i<Nx;++i)
	      data_cell[hex_b][Ind_loc_matrix_hexahedra(i,j,k)] =
	          a.data_cell[hex_a][Ind_loc_matrix_hexahedra(i,j,Nz-1-k)];
  }
}


// copy of data from hex_a to hex_b
template <typename DTyp> 
void Cell_variable<DTyp>::Add_invert_z(int hex_b, Cell_variable<DTyp>& a, int hex_a) {
  int Nx, Ny, Nz;
  int i,j,k;
  int max_num_hex = ug->Give_number_hexahedra();
  
  if(hex_a<0 || hex_b<0 || hex_a>=max_num_hex || hex_b>=max_num_hex) {
    throw(std::runtime_error(std::string("error 5 in  Cell_Variable<DTyp>::Copy_invert_z ")));
  }

  /// @todo: MPI
  
  Nx = a.Give_blockgrid()->Give_Nx_hexahedron(hex_b);
  Ny = a.Give_blockgrid()->Give_Ny_hexahedron(hex_b);
  Nz = a.Give_blockgrid()->Give_Nz_hexahedron(hex_b);
  
  if(Nx!=a.Give_blockgrid()->Give_Nx_hexahedron(hex_a) ||
     Ny!=a.Give_blockgrid()->Give_Ny_hexahedron(hex_a) ||
     Nz!=a.Give_blockgrid()->Give_Nz_hexahedron(hex_a)) {
    throw(std::runtime_error(std::string(" error 6 in  Cell_Variable<DTyp>::Copy_invert_z ")));
  }

  /// todo dieses parallel for kann man vielleicht noch reinbauen
  #pragma omp parallel for private(j,i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
  for(k=0;k<Nz;++k) {
    for(j=0;j<Ny;++j)
      for(i=0;i<Nx;++i)
	  data_cell[hex_b][Ind_loc_matrix_hexahedra(i,j,k)] +=
	      a.data_cell[hex_a][Ind_loc_matrix_hexahedra(i,j,Nz-1-k)];
  }
}

template <typename DTyp> 
void Cell_variable<DTyp>::Add_invert_z(int hex_b_start, Cell_variable<DTyp>& a, int hex_a_start, int numBlocks) {
  for (int i  = 0; i < numBlocks; ++i) {
    Add_invert_z(hex_b_start+i,a,hex_a_start+(numBlocks-1)-i); 
  }
}

template <typename DTyp> 
void Cell_variable<DTyp>::Copy_invert_z(int hex_b_start, Cell_variable<DTyp>& a, int hex_a_start, int numBlocks) {
  for (int i  = 0; i < numBlocks; ++i) {
    Copy_invert_z(hex_b_start+i,a,hex_a_start+(numBlocks-1)-i); 
  }
}


#endif // COPY_H
