/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/


#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "../grid/elements.h"
#include "../grid/parti.h"
#include "../grid/ug.h"
#include "../grid/blockgrid.h"
#include "../grid/marker.h"
#include "extemp.h"
#include "parallel.h"
#include "variable.h"

#include <QtGui/QtGui>

#ifndef _VTK_GEHT

class vtkUnstructuredGrid {
  
};

template<>
vtkUnstructuredGrid* Variable<double>::GetVtkOutput(double stretch_z, QString title,  Unstructured_grid_Marker * marker) {
   return NULL;
}

#else


#include <vtkUnstructuredGrid.h>
#include <vtkPolyDataMapper.h>
#include "vtkHexahedron.h"
#include "vtkDoubleArray.h"
#include "vtkPointData.h"
#include "vtkMath.h"

template<>
vtkUnstructuredGrid* Variable<double>::GetVtkOutput(double stretch_z, QString title,  Unstructured_grid_Marker * marker) {
  D3vector v;
  int Nx, Ny, Nz;
  int num_total = 0;
  int cell_total = 0;

  
  vtkUnstructuredGrid *dataSet;
    
  // for parallel
  double* data_coord;
  int size, rank;
  MPI_Status status;

//  Datei.precision(12);
//  Datei.setf(std::ios::fixed,std::ios::floatfield);

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {

     Update<hexahedronEl>(id_hex);

    num_total += blockgrid->Give_N_total_hexahedron(id_hex);
    cell_total +=        (blockgrid->Give_Nx_hexahedron(id_hex))
                        *(blockgrid->Give_Ny_hexahedron(id_hex))
                        *(blockgrid->Give_Nz_hexahedron(id_hex));
    }
  }

  // Teil 0: Write information
  vtkSmartPointer<vtkPoints> hexahedronPoints;
  int countPoint=0;
  if(my_rank==0) {
     hexahedronPoints = vtkSmartPointer<vtkPoints>::New();
     hexahedronPoints->SetNumberOfPoints(num_total);
  }
  
  // Teil 2: Koordinaten der Punkte ausgeben
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1) * 3;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int k=0;k<=Nz;++k)
          for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      hexahedronPoints->InsertPoint(countPoint, v.x, v.y, v.z);
	      countPoint++;
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0] = v.x;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1] = v.y;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] = v.z;
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,500+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,500+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
              hexahedronPoints->InsertPoint(countPoint, v.x, v.y, v.z);
	      countPoint++;
	  }
    }
   }
  }

  // Teil 3: Zellen ausgeben
  vtkSmartPointer<vtkHexahedron> aHexahedron = vtkSmartPointer<vtkHexahedron>::New();  
  dataSet =  vtkUnstructuredGrid::New();
  dataSet->Allocate(cell_total, cell_total);
  if(my_rank==0) {
    //Datei << "\nCELLS " << cell_total << " " <<  cell_total*9  << "\n";

    int id_start;
    id_start = 0;
    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
     if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
	  for(int k=0;k<Nz;++k) {  
              aHexahedron->GetPointIds()->SetId(0, id_start + i   + (Nx+1)*(j+1+(k+1)*(Ny+1) ));  // WNT
              aHexahedron->GetPointIds()->SetId(1, id_start + i   + (Nx+1)*(j  +(k+1)*(Ny+1) ));  // WST
              aHexahedron->GetPointIds()->SetId(2, id_start + i+1 + (Nx+1)*(j  +(k+1)*(Ny+1) ));  // EST
              aHexahedron->GetPointIds()->SetId(3, id_start + i+1 + (Nx+1)*(j+1+(k+1)*(Ny+1) ));  // ENT
              aHexahedron->GetPointIds()->SetId(4, id_start + i   + (Nx+1)*(j+1+  k * (Ny+1) ));  // WND
              aHexahedron->GetPointIds()->SetId(5, id_start + i   + (Nx+1)*(j  +  k * (Ny+1) ));  // WSD
              aHexahedron->GetPointIds()->SetId(6, id_start + i+1 + (Nx+1)*(j  +  k * (Ny+1) ));  // ESD
              aHexahedron->GetPointIds()->SetId(7, id_start + i+1 + (Nx+1)*(j+1+  k * (Ny+1) ));  // END
	    
	      
	      dataSet->InsertNextCell(aHexahedron->GetCellType(), aHexahedron->GetPointIds());
	  }
      id_start = id_start +  blockgrid->Give_N_total_hexahedron(id_hex);
     }
    }


    dataSet->SetPoints(hexahedronPoints);
  }

   // Teil 6: Ausgabe der Werte der Variable
   countPoint = 0;    
   vtkSmartPointer<vtkDoubleArray> valueVariable = vtkSmartPointer<vtkDoubleArray>::New();
   valueVariable->SetName("valueVariable");
   valueVariable->SetNumberOfTuples(num_total);   
   
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1);
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int k=0;k<=Nz;++k) 
          for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	        valueVariable->SetTuple1(countPoint, data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]);
	        //Datei << convert(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)])
		//    << "\n";
	        countPoint++;
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      
	      
	      data_coord[i + (Nx+1) * (j + (Ny+1) * k )] = data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)];
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,600+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,600+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	      // Datei << data_coord[i + (Nx+1) * (j + (Ny+1) * k )] << "\n";
	      valueVariable->SetTuple1(countPoint, data_coord[i + (Nx+1) * (j + (Ny+1) * k )]);
	      countPoint++;	    
	  }
    }
   }
  }
   
   if(my_rank==0) {
      dataSet->GetPointData()->AddArray(valueVariable);
      dataSet->GetPointData()->SetActiveScalars("valueVariable");  
      return dataSet;
   }
   else return NULL;
}

#endif
