/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#include "../mympi.h"
#include <vector>
#include <complex>

#include "../parameter.h"
#include "parallel.h"


////////////////////////////////////////

template <>
int Give_MPI_data_typ<double>::Give() { return MPI_DOUBLE; }

template <>
int Give_MPI_data_typ<std::complex<double> >::Give() { return MPI_DOUBLE_COMPLEX; }

template <>
int Give_MPI_data_typ<bool>::Give() { return MPI_BYTE; }


template <>
double Make_MPI_All_Sum<double>::Do(double x, MPI_Comm comm) { 
#ifndef _MPI_PARALLEL
   return x; 
#endif

#ifdef _MPI_PARALLEL
   double sum;
   MPI_Allreduce(&x,&sum,1,MPI_DOUBLE,MPI_SUM,comm);
   return sum; 
#endif
}


template <>
std::complex<double> Make_MPI_All_Sum<std::complex<double> >::Do(std::complex<double> x,
								 MPI_Comm comm) {
#ifndef _MPI_PARALLEL
   return x; 
#endif

#ifdef _MPI_PARALLEL
   double sum_re, sum_im;
   double   x_re,   x_im;

   x_re = x.real();
   x_im = x.imag();

   MPI_Allreduce(&x_re,&sum_re,1,MPI_DOUBLE,MPI_SUM,comm);
   MPI_Allreduce(&x_im,&sum_im,1,MPI_DOUBLE,MPI_SUM,comm);

   return std::complex<double>(sum_re,sum_im); 
#endif
}



////////////////////////////////////////
// double

template <> int     Parallel_buffer<double> ::size_send = 0;
template <> double* Parallel_buffer<double> ::send = 0;

template <> int     Parallel_buffer<double> ::size_receive = 0;
template <> double* Parallel_buffer<double> ::receive = 0;


////////////////////////////////////////
// std::complex<double>

template <> int                   Parallel_buffer<std::complex<double> > ::size_send = 0;
template <> std::complex<double>* Parallel_buffer<std::complex<double> > ::send = 0;

template <> int                   Parallel_buffer<std::complex<double> > ::size_receive = 0;
template <> std::complex<double>* Parallel_buffer<std::complex<double> > ::receive = 0;


