/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "../grid/elements.h"
#include "../grid/parti.h"
#include "../grid/ug.h"
#include "../grid/blockgrid.h"
#include "../grid/marker.h"
#include "extemp.h"
#include "parallel.h"
#include "variable.h"
#include "cellvar.h"
#include "print_var_vtk_cc.h"

#include <QtGui/QtGui>
//#include <QDialog>

// Print Variables

/*
void Variable<double>::QPrint_VTK(QString DateiName, 
				 double stretch_z = 1.,
				 Unstructured_grid_Marker * marker = NULL ) {
  QPrint_VTK(DateiName, stretch_z, QString("myData"),marker);
}
*/

template<>
void Cell_variable<std::complex<double> >::QPrint_VTK(QString DateiName, double stretch_z, QString title, Unstructured_grid_Marker * marker) {
  QPrint_VTK(DateiName,myCOMPLEX ,stretch_z,title,marker);
}

template<>
void Variable<double>::QPrint_VTK(QString DateiName, double stretch_z, QString title, Unstructured_grid_Marker * marker) {
   QPrint_VTK(DateiName,myREAL,stretch_z,title, marker);
}

template<>
void Cell_variable<double>::QPrint_VTK(QString DateiName, double stretch_z, QString title, Unstructured_grid_Marker * marker) {
  QPrint_VTK(DateiName,myREAL,stretch_z,title,marker);
}


template<>
void Variable<std::complex<double> >::QPrint_VTK(QString DateiName, double stretch_z, QString title, Unstructured_grid_Marker * marker) {
  QPrint_VTK(DateiName,myCOMPLEX ,stretch_z,title,marker);
}


// gibt Querschnitt fuer z = z_cut aus;
// z_cut = Nz, falls z_cut zu gross gewaehlt
template <>
void Variable<std::complex<double> >::QPrint_VTK(QString DateiName, int z_cut, QString title ) {

  D3vector v;
  int Nx, Ny, Nz;
  int num_total = 0;
  int cell_total = 0;
  
   // for parallel
  double* data_coord;
  int size, rank;
  MPI_Status status;

  QFile file(DateiName);
  if(!file.open(QIODevice::WriteOnly)) {
      std::cerr << "error open file for writing in print_var_Qt_vtk.cc:" 
                << qPrintable(file.errorString()) << endl;
   }
  QTextStream Datei(&file);
  
  // Datei.precision(12);
  // Datei.setf(std::ios::fixed,std::ios::floatfield);

  for(int id=0;id<ug->Give_number_hexahedra();++id) {
    Update<hexahedronEl>(id);
    num_total+=		(blockgrid->Give_Nx_hexahedron(id)+1)
			*(blockgrid->Give_Ny_hexahedron(id)+1);
    cell_total +=        (blockgrid->Give_Nx_hexahedron(id))
                        *(blockgrid->Give_Ny_hexahedron(id));
  }


  // Teil 0: Write information
  if(my_rank==0) {
    Datei << "# vtk DataFile Version 2.0\n"
          <<  title << endl
  	  << "ASCII\n"
  	  << "DATASET UNSTRUCTURED_GRID\n";

    // Teil 1: Kopfzeile schreiben
    Datei << "POINTS " << num_total << " float\n";
  }

  // Teil 2: Koordinaten der Punkte ausgeben
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {

    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);

    size = (Ny+1) * (Nx+1) * 3;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
          for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,z_cut);
	      Datei << v.x
		    << " " << v.y
		    << " " << v.z
		    << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,z_cut);
	      data_coord[3*(i + (Nx+1) * j) + 0] = v.x;
	      data_coord[3*(i + (Nx+1) * j) + 1] = v.y;
	      data_coord[3*(i + (Nx+1) * j) + 2] = v.z;
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,500+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,500+rank,ug->Give_MPI_comm(),&status);
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    Datei <<	data_coord[3*(i + (Nx+1) * j) + 0]
		   <<	"  " << data_coord[3*(i + (Nx+1) * j) + 1]
		   <<	"  " << data_coord[3*(i + (Nx+1) * j) + 2] << "\n";
	  }
    }
  }

  // Teil 3: Zellen ausgeben
  // todo: 9 mal cell_total ?? 
  if(my_rank==0) {
    Datei << "\nCELLS " << cell_total << " " <<  cell_total*5  << "\n";

    int id_start;
    id_start = 0;
    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);

      // z_cut ggf. verkleinern
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      if (z_cut > Nz) z_cut = Nz;

      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
	  {
	    Datei <<  "4 "
		  << id_start + i   + (Nx+1)*(j+1+  z_cut * (Ny+1) ) << " " // WND
		  << id_start + i   + (Nx+1)*(j  +  z_cut * (Ny+1) ) << " " // WSD
		  << id_start + i+1 + (Nx+1)*(j  +  z_cut * (Ny+1) ) << " " // ESD
		  << id_start + i+1 + (Nx+1)*(j+1+  z_cut * (Ny+1) ) // END
		  << "\n";
	  }
      id_start = id_start +  blockgrid->Give_N_total_hexahedron(id_hex);
    }

    Datei << "\nCELL_TYPES " << cell_total << "\n";

    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
	  {
	    Datei << "9"
		  << "\n";
	  }
    }

    // Teil 4:
    Datei << "\nPOINT_DATA " << num_total << "\n";
    Datei << "SCALARS " << title << " float 1\n";

    Datei << "LOOKUP_TABLE default\n";
  }

  // Teil 6: Ausgabe der Werte der Variable

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);

    size = (Ny+1) * (Nx+1);
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
          for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      Datei << abs(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*z_cut)])
		    << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      data_coord[i + (Nx+1) * j] = abs(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*z_cut)]);
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,600+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,600+rank,ug->Give_MPI_comm(),&status);
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    Datei << data_coord[i + (Nx+1) * j] << "\n";
	  }
    }
  }
  file.close();
}

// gibt Querschnitt fuer z = z_cut aus;
// z_cut = Nz, falls z_cut zu gross gewaehlt
template <>
void Variable<double>::QPrint_VTK(QString DateiName, int z_cut, QString title ) {
  D3vector v;
  int id, Nx, Ny, Nz;
  int num_total = 0;
  int cell_total = 0;

  
   // for parallel
  double* data_coord;
  int size, rank;
  MPI_Status status;
   
  QFile fileIn(DateiName);
  if(!fileIn.open(QIODevice::WriteOnly)) {
      std::cerr << "error open fileIn for writing QPrint_VTK:" 
                  << qPrintable(fileIn.errorString()) << endl;
  }
  QTextStream Datei(&fileIn);


  for(id=0;id<ug->Give_number_hexahedra();++id) {
    Update<hexahedronEl>(id);
    num_total+=		(blockgrid->Give_Nx_hexahedron(id)+1)
			*(blockgrid->Give_Ny_hexahedron(id)+1);
    cell_total +=        (blockgrid->Give_Nx_hexahedron(id))
                        *(blockgrid->Give_Ny_hexahedron(id));
  }


  // Teil 0: Write information
  if(my_rank==0) {
    Datei << "# vtk DataFile Version 2.0\n"
	  << title << endl
  	  << "ASCII\n"
  	  << "DATASET UNSTRUCTURED_GRID\n";

    // Teil 1: Kopfzeile schreiben
    Datei << "POINTS " << num_total << " float\n";
  }

  // Teil 2: Koordinaten der Punkte ausgeben
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {

    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);

    size = (Ny+1) * (Nx+1) * 3;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
          for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,z_cut);
	      Datei << v.x
		    << " " << v.y
		    << " " << v.z
		    << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,z_cut);
	      data_coord[3*(i + (Nx+1) * j) + 0] = v.x;
	      data_coord[3*(i + (Nx+1) * j) + 1] = v.y;
	      data_coord[3*(i + (Nx+1) * j) + 2] = v.z;
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,500+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,500+rank,ug->Give_MPI_comm(),&status);
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    Datei <<	data_coord[3*(i + (Nx+1) * j) + 0]
		   <<	"  " << data_coord[3*(i + (Nx+1) * j) + 1]
		   <<	"  " << data_coord[3*(i + (Nx+1) * j) + 2] << "\n";
	  }
    }
  }

  // Teil 3: Zellen ausgeben
  // todo: 9 mal cell_total ?? 
  if(my_rank==0) {
    Datei << "\nCELLS " << cell_total << " " <<  cell_total*5  << "\n";

    int id_start;
    id_start = 0;
    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);

      // z_cut ggf. verkleinern
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      if (z_cut > Nz) z_cut = Nz;

      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
	  {
	    Datei <<  "4 "
		  << id_start + i   + (Nx+1)*(j+1+  z_cut * (Ny+1) ) << " " // WND
		  << id_start + i   + (Nx+1)*(j  +  z_cut * (Ny+1) ) << " " // WSD
		  << id_start + i+1 + (Nx+1)*(j  +  z_cut * (Ny+1) ) << " " // ESD
		  << id_start + i+1 + (Nx+1)*(j+1+  z_cut * (Ny+1) ) // END
		  << "\n";
	  }
      id_start = id_start +  blockgrid->Give_N_total_hexahedron(id_hex);
    }

    Datei << "\nCELL_TYPES " << cell_total << "\n";

    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
	  {
	    Datei << "9"
		  << "\n";
	  }
    }

    // Teil 4:
    Datei << "\nPOINT_DATA " << num_total << "\n";
    Datei << "SCALARS " << title << " float 1\n";

    Datei << "LOOKUP_TABLE default\n";
  }

  // Teil 6: Ausgabe der Werte der Variable

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);

    size = (Ny+1) * (Nx+1);
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
          for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      Datei << data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*z_cut)]
		    << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      data_coord[i + (Nx+1) * j] = data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*z_cut)];
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,600+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,600+rank,ug->Give_MPI_comm(),&status);
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    Datei <<	data_coord[i + (Nx+1) * j] << "\n";
	  }
    }
  }
}




