/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#ifndef PRINTVARVTKCCH
#define PRINTVARVTKCCH

#include "variable.h"
#include "cellvar.h"
//#include "../vtk/output.h"
// Print VTK Template Funktionen
#include <QtGui/QtGui>
//#include <QMessageBox>

template <class DTyp>
void Variable<DTyp>::Print_VTK(std::ostream& Datei, double (*convert)(DTyp x),double stretch_z, std::string title,  Unstructured_grid_Marker * marker) {
  D3vector v;
  int Nx, Ny, Nz;
  int num_total = 0;
  int cell_total = 0;
  
  //QWidget* ww = new QWidget();
  //QMessageBox::information(ww, QString("1 Variable<DTyp>::Print_VTK C"), QString("ASLD test"));

  // for parallel
  double* data_coord;
  int size, rank;
  MPI_Status status;

  Datei.precision(12);
  Datei.setf(std::ios::fixed,std::ios::floatfield);
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
 /*  Variable<std::complex<double> > * shift;
   shift = blockgrid->Give_phase_shift();
   blockgrid->Set_phase_shift(NULL);
*/
     Update<hexahedronEl>(id_hex);
    /*
    blockgrid->Set_phase_shift(shift);
 */   
    num_total += blockgrid->Give_N_total_hexahedron(id_hex);
    cell_total +=        (blockgrid->Give_Nx_hexahedron(id_hex))
                        *(blockgrid->Give_Ny_hexahedron(id_hex))
                        *(blockgrid->Give_Nz_hexahedron(id_hex));
    }
  }

  // Teil 0: Write information
  if(my_rank==0) {
    Datei << "# vtk DataFile Version 2.0\n"
          << title << endl
  	  << "ASCII\n"
  	  << "DATASET UNSTRUCTURED_GRID\n";

    // Teil 1: Kopfzeile schreiben
    Datei << "POINTS " << num_total << " float\n";
  }

  // Teil 2: Koordinaten der Punkte ausgeben
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1) * 3;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int k=0;k<=Nz;++k)
          for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      Datei << v.x
		    << " " << v.y
		    << " " << v.z * stretch_z
		    << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0] = v.x;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1] = v.y;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] = v.z;
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,500+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,500+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    Datei <<	data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] << "\n";
	  }
    }
     }
  }

  // Teil 3: Zellen ausgeben
  if(my_rank==0) {
    Datei << "\nCELLS " << cell_total << " " <<  cell_total*9  << "\n";

    int id_start;
    id_start = 0;
    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
     if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
	  for(int k=0;k<Nz;++k) {
	    Datei << "8 "
		  << id_start + i   + (Nx+1)*(j+1+(k+1)*(Ny+1) ) << " " // WNT
		  << id_start + i   + (Nx+1)*(j  +(k+1)*(Ny+1) ) << " " // WST
		  << id_start + i+1 + (Nx+1)*(j  +(k+1)*(Ny+1) ) << " " // EST
		  << id_start + i+1 + (Nx+1)*(j+1+(k+1)*(Ny+1) ) << " " // ENT
		  << id_start + i   + (Nx+1)*(j+1+  k * (Ny+1) ) << " " // WND
		  << id_start + i   + (Nx+1)*(j  +  k * (Ny+1) ) << " " // WSD
		  << id_start + i+1 + (Nx+1)*(j  +  k * (Ny+1) ) << " " // ESD
		  << id_start + i+1 + (Nx+1)*(j+1+  k * (Ny+1) ) // END
		  << "\n";
	  }
      id_start = id_start +  blockgrid->Give_N_total_hexahedron(id_hex);
     }
    }

    
    Datei << "\nCELL_TYPES " << cell_total << "\n";

    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
     if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
	  for(int k=0;k<Nz;++k) {
	    Datei << "12"
		  << "\n";
	  }
     }
    }


    // Teil 4:
    Datei << "\nPOINT_DATA " << num_total << "\n";
    Datei << "SCALARS " << title << " float 1\n";

    Datei << "LOOKUP_TABLE default\r\n";
  }

  // Teil 6: Ausgabe der Werte der Variable

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1);
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int k=0;k<=Nz;++k) 
          for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      Datei << convert(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)])
		    << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      data_coord[i + (Nx+1) * (j + (Ny+1) * k )] = convert(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]);
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,600+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,600+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    Datei << data_coord[i + (Nx+1) * (j + (Ny+1) * k )] << "\n";
	  }
    }
   }
  }
}


template <class DTyp>
void Variable<DTyp>::QPrint_VTK(QString DateiName, double (*convert)(DTyp x),double stretch_z, QString title,  
				Unstructured_grid_Marker * marker, bool cutSmallData) {

    if (this==NULL)
    {
        std::cout << "could not write file, var == NULL" << std::endl;
        return;
    }

  D3vector v;
  int Nx, Ny, Nz;
  int num_total = 0;
  int cell_total = 0;  
  
 // QWidget* ww = new QWidget();
 // QMessageBox::information(ww, QString("2 Variable<DTyp>::QPrint_VTK B"), QString("ASLD test"));

  QFile file(DateiName);
  if(!file.open(QIODevice::WriteOnly)) {
      std::cerr << "error open file for writing in print_var_vtk_cc.h:" 
                << qPrintable(file.errorString()) 
		<< " DateiName: " << DateiName.toStdString()  << endl;
   }
  QTextStream Datei(&file);
  
  Datei.setRealNumberPrecision(10);
  // for parallel
  double* data_coord;
  int size, rank;
  MPI_Status status;

//  Datei.precision(12);
//  Datei.setf(std::ios::fixed,std::ios::floatfield);

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
 /*  Variable<std::complex<double> > * shift;
   shift = blockgrid->Give_phase_shift();
   blockgrid->Set_phase_shift(NULL);
*/
     Update<hexahedronEl>(id_hex);
    /*
    blockgrid->Set_phase_shift(shift);
 */   
    num_total += blockgrid->Give_N_total_hexahedron(id_hex);
    cell_total +=        (blockgrid->Give_Nx_hexahedron(id_hex))
                        *(blockgrid->Give_Ny_hexahedron(id_hex))
                        *(blockgrid->Give_Nz_hexahedron(id_hex));
    }
  }

  // Teil 0: Write information

  if(my_rank==0) {
    
    Datei << "# vtk DataFile Version 2.0 created by UGBlocks"
          << "# typeUg " << ug->getUgTyp();
	  
    Datei << " # degrees " << ug->degree_of_freedom() << " ";
    for(int d=0;d<ug->degree_of_freedom();++d)
        Datei << blockgrid->getNumberPointsDegree(d) << " " ;	 
    
    
    std::vector<double> constructionParameters = ug->getConstructionParameters();
    Datei << "# constructParameters " << constructionParameters.size() << " ";
    for(int nc=0;nc<constructionParameters.size();++nc)
        Datei << constructionParameters.at(nc) << " ";	
    
    std::vector<double> relativeCoordZ = ug->getRelativeCoord();
    if (!relativeCoordZ.empty()){     
        Datei << "# relativeCoordZ " << relativeCoordZ.size() << " ";
            for(int nc=0;nc<relativeCoordZ.size();++nc)
                Datei << relativeCoordZ.at(nc) << " ";	 
    }
    std::vector<double> lx = ug->getConstructionParametersLx();
    std::vector<double> ly = ug->getConstructionParametersLy();
    std::vector<double> lz = ug->getConstructionParametersLz();
    if (!lx.empty())
    {
    Datei << "# l_xyz " << lx.size() << " ";
        for(int nc=0;nc<lx.size();++nc)
            Datei << lx.at(nc) << " ";	 
	for(int nc=0;nc<ly.size();++nc)
            Datei << ly.at(nc) << " ";	 
	for(int nc=0;nc<lz.size();++nc)
            Datei << lz.at(nc) << " ";	 
    }
 
    Datei << "# inner_grid_arched ";
    Datei << (ug->getConstructionBoolArched()? "1": "0") << " ";
    Datei << "# periodic ";
    Datei << (ug->getConstructionBoolPeriodic()? "1": "0") << " ";
     
   
    Datei << " # Filename " << endl << title << endl
  	  << "ASCII\n"
  	  << "DATASET UNSTRUCTURED_GRID\n";

    // Teil 1: Kopfzeile schreiben
    Datei << "POINTS " << num_total << " float\n";
  }

  // Teil 2: Koordinaten der Punkte ausgeben
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1) * 3;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int k=0;k<=Nz;++k)
          for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      Datei << v.x
		    << " " << v.y
		    << " " << v.z * stretch_z
		    << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0] = v.x;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1] = v.y;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] = v.z;
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,500+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,500+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    Datei <<	data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] << "\n";
	  }
    }
     }
  }

  // Teil 3: Zellen ausgeben
  if(my_rank==0) {
    Datei << "\nCELLS " << cell_total << " " <<  cell_total*9  << "\n";

    int id_start;
    id_start = 0;
    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
     if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
	  for(int k=0;k<Nz;++k) {
	    Datei << "8 "
		  << id_start + i   + (Nx+1)*(j+1+(k+1)*(Ny+1) ) << " " // WNT
		  << id_start + i   + (Nx+1)*(j  +(k+1)*(Ny+1) ) << " " // WST
		  << id_start + i+1 + (Nx+1)*(j  +(k+1)*(Ny+1) ) << " " // EST
		  << id_start + i+1 + (Nx+1)*(j+1+(k+1)*(Ny+1) ) << " " // ENT
		  << id_start + i   + (Nx+1)*(j+1+  k * (Ny+1) ) << " " // WND
		  << id_start + i   + (Nx+1)*(j  +  k * (Ny+1) ) << " " // WSD
		  << id_start + i+1 + (Nx+1)*(j  +  k * (Ny+1) ) << " " // ESD
		  << id_start + i+1 + (Nx+1)*(j+1+  k * (Ny+1) ) // END
		  << "\n";
	  }
      id_start = id_start +  blockgrid->Give_N_total_hexahedron(id_hex);
     }
    }

    Datei << "\nCELL_TYPES " << cell_total << "\n";

    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
     if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
	  for(int k=0;k<Nz;++k) {
	    Datei << "12"
		  << "\n";
	  }
     }
    }


    // Teil 4:
    Datei << "\nPOINT_DATA " << num_total << "\n";
    Datei << "SCALARS " << title << " float 1\n";

    Datei << "LOOKUP_TABLE default\r\n";
  }

  // Teil 6: Ausgabe der Werte der Variable

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1);
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int k=0;k<=Nz;++k) 
          for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
			if (cutSmallData) {
				double wert = convert(data_hexahedra[id_hex][i + (Nx + 1)*(j + (Ny + 1)*k)]);
				if (ABS(wert) < 1.0e-40)  Datei << 0.0 << "\n";
				else	                Datei << wert << "\n";
			}
			else {
				Datei << convert(data_hexahedra[id_hex][i + (Nx + 1)*(j + (Ny + 1)*k)])
					  << "\n";
			} 
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      data_coord[i + (Nx+1) * (j + (Ny+1) * k )] = convert(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]);
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,600+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,600+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    if(cutSmallData) {
	       double wert = data_coord[i + (Nx+1) * (j + (Ny+1) * k )];
	       if(ABS(wert) < 1.0e-40)  Datei << 0.0 << "\n";
	       else	                Datei << wert << "\n";
	    }
	    else {
	       Datei << data_coord[i + (Nx+1) * (j + (Ny+1) * k )] << "\n";	      
	    }
	  }
    }
   }
  }
}

template <class DTyp>
void Variable<DTyp>::QPrint_VTK(QString DateiName, complex<double> (*convert)(DTyp x),double stretch_z, QString title,  Unstructured_grid_Marker * marker) {
  D3vector v;
  int Nx, Ny, Nz;
  int num_total = 0;
  int cell_total = 0;

  //QWidget* ww = new QWidget();
  //QMessageBox::information(ww,  QString("3 Variable<DTyp>::QPrint_VTK A"), QString("ASLD test"));

  QFile file(DateiName);
  if(!file.open(QIODevice::WriteOnly)) {
      std::cerr << "error open file for writing in print_var_vtk_cc.h:"
                << qPrintable(file.errorString())
                << " DateiName: " << DateiName.toStdString()  << endl;
   }
  QTextStream Datei(&file);


  // for parallel
  double* data_coord;
  complex<double>* cdata_coord;
  int size, rank;
  MPI_Status status;

//  Datei.precision(12);
//  Datei.setf(std::ios::fixed,std::ios::floatfield);

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
 /*  Variable<std::complex<double> > * shift;
   shift = blockgrid->Give_phase_shift();
   blockgrid->Set_phase_shift(NULL);
*/
     Update<hexahedronEl>(id_hex);
    /*
    blockgrid->Set_phase_shift(shift);
 */
    num_total += blockgrid->Give_N_total_hexahedron(id_hex);
    cell_total +=        (blockgrid->Give_Nx_hexahedron(id_hex))
                        *(blockgrid->Give_Ny_hexahedron(id_hex))
                        *(blockgrid->Give_Nz_hexahedron(id_hex));
    }
  }

  // Teil 0: Write information
  if(my_rank==0) {
    Datei << "# vtk DataFile Version 2.0\n"
          << title << endl
          << "ASCII\n"
          << "DATASET UNSTRUCTURED_GRID\n";

    // Teil 1: Kopfzeile schreiben
    Datei << "POINTS " << num_total << " float\n";
  }

  // Teil 2: Koordinaten der Punkte ausgeben
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1) * 3;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int k=0;k<=Nz;++k)
          for(int j=0;j<=Ny;++j)
            for(int i=0;i<=Nx;++i) {
              v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
              Datei << v.x
                    << " " << v.y
                    << " " << v.z * stretch_z
                    << "\n";
            }
      }
      else {
        data_coord = Parallel_buffer<double>::Give_send(size);
        for(int k=0;k<=Nz;++k)
          for(int j=0;j<=Ny;++j)
            for(int i=0;i<=Nx;++i) {
              v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
              data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0] = v.x;
              data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1] = v.y;
              data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] = v.z;
            }
        MPI_Send(data_coord,size,MPI_DOUBLE,0,500+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,500+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
        for(int j=0;j<=Ny;++j)
          for(int i=0;i<=Nx;++i) {
            Datei <<    data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0]
                   <<   "  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1]
                   <<   "  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] << "\n";
          }
    }
     }
  }

  // Teil 3: Zellen ausgeben
  if(my_rank==0) {
    Datei << "\nCELLS " << cell_total << " " <<  cell_total*9  << "\n";

    int id_start;
    id_start = 0;
    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
     if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
          for(int k=0;k<Nz;++k) {
            Datei << "8 "
                  << id_start + i   + (Nx+1)*(j+1+(k+1)*(Ny+1) ) << " " // WNT
                  << id_start + i   + (Nx+1)*(j  +(k+1)*(Ny+1) ) << " " // WST
                  << id_start + i+1 + (Nx+1)*(j  +(k+1)*(Ny+1) ) << " " // EST
                  << id_start + i+1 + (Nx+1)*(j+1+(k+1)*(Ny+1) ) << " " // ENT
                  << id_start + i   + (Nx+1)*(j+1+  k * (Ny+1) ) << " " // WND
                  << id_start + i   + (Nx+1)*(j  +  k * (Ny+1) ) << " " // WSD
                  << id_start + i+1 + (Nx+1)*(j  +  k * (Ny+1) ) << " " // ESD
                  << id_start + i+1 + (Nx+1)*(j+1+  k * (Ny+1) ) // END
                  << "\n";
          }
      id_start = id_start +  blockgrid->Give_N_total_hexahedron(id_hex);
     }
    }

    Datei << "\nCELL_TYPES " << cell_total << "\n";

    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
     if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
          for(int k=0;k<Nz;++k) {
            Datei << "12"
                  << "\n";
          }
     }
    }


    // Teil 4:
    Datei << "\nPOINT_DATA " << num_total << "\n";
    Datei << "SCALARS " << title << " float 2\n";

    Datei << "LOOKUP_TABLE default\r\n";
  }

  // Teil 6: Ausgabe der Werte der Variable

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1);
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int k=0;k<=Nz;++k)
          for(int j=0;j<=Ny;++j)
            for(int i=0;i<=Nx;++i) {
              Datei << convert(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]).real() << "  "
                    << convert(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]).imag() << "\n";
            }
      }
      else {
        cdata_coord = Parallel_buffer<complex<double> >::Give_send(size);
        for(int k=0;k<=Nz;++k)
          for(int j=0;j<=Ny;++j)
            for(int i=0;i<=Nx;++i) {
              cdata_coord[i + (Nx+1) * (j + (Ny+1) * k )] = convert(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]);
            }
        MPI_Send(cdata_coord,size,MPI_DOUBLE_COMPLEX,0,600+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      cdata_coord = Parallel_buffer<complex<double> >::Give_receive(size);
      MPI_Recv(cdata_coord,size,MPI_DOUBLE_COMPLEX,rank,600+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
        for(int j=0;j<=Ny;++j)
          for(int i=0;i<=Nx;++i) {
            Datei << cdata_coord[i + (Nx+1) * (j + (Ny+1) * k )].real() << "  "
                  << cdata_coord[i + (Nx+1) * (j + (Ny+1) * k )].imag() << "\n";
          }
    }
   }
  }
  if(my_rank==0) {
    Datei << "\nVECTORS vectors float\r\n";
  }
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1);
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int k=0;k<=Nz;++k)
          for(int j=0;j<=Ny;++j)
            for(int i=0;i<=Nx;++i) {
              Datei << convert(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]).real() << "  "
                    << convert(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]).imag() << "  0\n";
            }
      }
      else {
        cdata_coord = Parallel_buffer<complex<double> >::Give_send(size);
        for(int k=0;k<=Nz;++k)
          for(int j=0;j<=Ny;++j)
            for(int i=0;i<=Nx;++i) {
              cdata_coord[i + (Nx+1) * (j + (Ny+1) * k )] = convert(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]);
            }
        MPI_Send(cdata_coord,size,MPI_DOUBLE_COMPLEX,0,600+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      cdata_coord = Parallel_buffer<complex<double> >::Give_receive(size);
      MPI_Recv(cdata_coord,size,MPI_DOUBLE_COMPLEX,rank,600+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
        for(int j=0;j<=Ny;++j)
          for(int i=0;i<=Nx;++i) {
            Datei << cdata_coord[i + (Nx+1) * (j + (Ny+1) * k )].real() << "  "
                  << cdata_coord[i + (Nx+1) * (j + (Ny+1) * k )].imag() << "  0\n";
          }
    }
   }
  }
}

template<class DTyp>
void Cell_variable<DTyp>::Print_VTK(std::ostream& Datei,double (*convert)(DTyp x), double stretch_z, std::string title,  Unstructured_grid_Marker * marker) {
  D3vector v;
  int id, Nx, Ny, Nz;
  int num_total = 0;
  int cell_total = 0;

  // QWidget* ww = new QWidget();
  // QMessageBox::information(ww, QString("4 Cell_variable<DTyp>::Print_VTK C"), QString("ASLD test"));

  // for parallel
  double* data_coord;
  int size, rank;
  MPI_Status status;

  Datei.precision(12);
  Datei.setf(std::ios::fixed,std::ios::floatfield);

  for(id=0;id<ug->Give_number_hexahedra();++id) { 
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id)==yes_mark) {
    
//    Update<hexahedronEl>(id);
    num_total += blockgrid->Give_N_total_hexahedron(id);
    cell_total +=        (blockgrid->Give_Nx_hexahedron(id))
                        *(blockgrid->Give_Ny_hexahedron(id))
                        *(blockgrid->Give_Nz_hexahedron(id));
   }
  }

  // Teil 0: Write information
  if(my_rank==0) {
    Datei << "# vtk DataFile Version 2.0\n"
          << title << endl
  	  << "ASCII\n"
  	  << "DATASET UNSTRUCTURED_GRID\n";

    // Teil 1: Kopfzeile schreiben
    Datei << "POINTS " << num_total << " float\n";
  }

  // Teil 2: Koordinaten der Punkte ausgeben
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
  
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1) * 3;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int k=0;k<=Nz;++k) 
          for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      Datei << v.x
		    << " " << v.y
		    << " " << v.z * stretch_z
		    << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0] = v.x;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1] = v.y;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] = v.z;
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,500+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,500+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    Datei <<	data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] << "\n";
	  }
    }
   }
  }

  // Teil 3: Zellen ausgeben
  if(my_rank==0) {
    Datei << "\nCELLS " << cell_total << " " <<  cell_total*9  << "\n";

    int id_start;
    id_start = 0;
    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
     if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
  
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
	  for(int k=0;k<Nz;++k) {
	    Datei <<  "8 "
		  << id_start + i   + (Nx+1)*(j+1+(k+1)*(Ny+1) ) << " " // WNT
		  << id_start + i   + (Nx+1)*(j  +(k+1)*(Ny+1) ) << " " // WST
		  << id_start + i+1 + (Nx+1)*(j  +(k+1)*(Ny+1) ) << " " // EST
		  << id_start + i+1 + (Nx+1)*(j+1+(k+1)*(Ny+1) ) << " " // ENT
		  << id_start + i   + (Nx+1)*(j+1+  k * (Ny+1) ) << " " // WND
		  << id_start + i   + (Nx+1)*(j  +  k * (Ny+1) ) << " " // WSD
		  << id_start + i+1 + (Nx+1)*(j  +  k * (Ny+1) ) << " " // ESD
		  << id_start + i+1 + (Nx+1)*(j+1+  k * (Ny+1) ) // END
		  << "\n";
	  }
      id_start = id_start +  blockgrid->Give_N_total_hexahedron(id_hex);
     }
    }

    Datei << "\nCELL_TYPES " << cell_total << "\n";

    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
     if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
  
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
	  for(int k=0;k<Nz;++k) {
	    Datei << "12"
		  << "\n";
	  }
     }
    }


    // Teil 4:
    Datei << "\nCELL_DATA " << cell_total << "\n";
    Datei << "SCALARS " << title << " float 1\n";

    Datei << "LOOKUP_TABLE default\n";
  }

  // Teil 6: Ausgabe der Werte der Variable

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
  
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);

    size = Nz * Ny * Nx;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int i=0;i<Nx;++i) 
          for(int j=0;j<Ny;++j)
	    for(int k=0;k<Nz;++k) {
  	      Datei << convert(data_cell[id_hex][((Ny)*(k)+(j))*(Nx)+(i)])
		    << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<Nz;++k)
	  for(int j=0;j<Ny;++j)
	    for(int i=0;i<Nx;++i) {
  	    data_coord[((Ny)*(k)+(j))*(Nx)+(i)] = convert(data_cell[id_hex][((Ny)*(k)+(j))*(Nx)+(i)]);
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,600+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,600+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<Nz;++k)
	for(int j=0;j<Ny;++j)
	  for(int i=0;i<Nx;++i) {
	    Datei << data_coord[((Ny)*(k)+(j))*(Nx)+(i)] << "\n";
	  }
    }
   }
  }
}

template<class DTyp>
void Cell_variable<DTyp>::QPrint_VTK(QString DateiName,double (*convert)(DTyp x), double stretch_z, QString title,  Unstructured_grid_Marker * marker) {
  D3vector v;
  int id, Nx, Ny, Nz;
  int num_total = 0;
  int cell_total = 0;

  // QWidget* ww = new QWidget();
  // QMessageBox::information(ww, QString("5 Cell_variable<DTyp>::QPrint_VTK B"), QString("ASLD test"));

  QFile file(DateiName);
  if(!file.open(QIODevice::WriteOnly)) {
      std::cerr << "error open file for writing in print_var_Qt_cc.h:" 
                << qPrintable(file.errorString()) << endl;
   }
  QTextStream Datei(&file);
  
  
  // for parallel
  double* data_coord;
  int size, rank;
  MPI_Status status;

//  Datei.precision(12);
//  Datei.setf(std::ios::fixed,std::ios::floatfield);

  for(id=0;id<ug->Give_number_hexahedra();++id) { 
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id)==yes_mark) {
    
//    Update<hexahedronEl>(id);
    num_total += blockgrid->Give_N_total_hexahedron(id);
    cell_total +=        (blockgrid->Give_Nx_hexahedron(id))
                        *(blockgrid->Give_Ny_hexahedron(id))
                        *(blockgrid->Give_Nz_hexahedron(id));
   }
  }

  // Teil 0: Write information
  if(my_rank==0) {
    Datei << "# vtk DataFile Version 2.0\n"
          << title << endl
  	  << "ASCII\n"
  	  << "DATASET UNSTRUCTURED_GRID\n";

    // Teil 1: Kopfzeile schreiben
    Datei << "POINTS " << num_total << " float\n";
  }

  // Teil 2: Koordinaten der Punkte ausgeben
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
  
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1) * 3;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int k=0;k<=Nz;++k) 
          for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      Datei << v.x
		    << " " << v.y
		    << " " << v.z * stretch_z
		    << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<=Nz;++k)
	  for(int j=0;j<=Ny;++j)
	    for(int i=0;i<=Nx;++i) {
	      v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0] = v.x;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1] = v.y;
	      data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] = v.z;
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,500+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,500+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
	for(int j=0;j<=Ny;++j)
	  for(int i=0;i<=Nx;++i) {
	    Datei <<	data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1]
		   <<	"  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] << "\n";
	  }
    }
   }
  }

  // Teil 3: Zellen ausgeben
  if(my_rank==0) {
    Datei << "\nCELLS " << cell_total << " " <<  cell_total*9  << "\n";

    int id_start;
    id_start = 0;
    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
     if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
  
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
	  for(int k=0;k<Nz;++k) {
	    Datei <<  "8 "
		  << id_start + i   + (Nx+1)*(j+1+(k+1)*(Ny+1) ) << " " // WNT
		  << id_start + i   + (Nx+1)*(j  +(k+1)*(Ny+1) ) << " " // WST
		  << id_start + i+1 + (Nx+1)*(j  +(k+1)*(Ny+1) ) << " " // EST
		  << id_start + i+1 + (Nx+1)*(j+1+(k+1)*(Ny+1) ) << " " // ENT
		  << id_start + i   + (Nx+1)*(j+1+  k * (Ny+1) ) << " " // WND
		  << id_start + i   + (Nx+1)*(j  +  k * (Ny+1) ) << " " // WSD
		  << id_start + i+1 + (Nx+1)*(j  +  k * (Ny+1) ) << " " // ESD
		  << id_start + i+1 + (Nx+1)*(j+1+  k * (Ny+1) ) // END
		  << "\n";
	  }
      id_start = id_start +  blockgrid->Give_N_total_hexahedron(id_hex);
     }
    }

    Datei << "\nCELL_TYPES " << cell_total << "\n";

    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
     if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
  
      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
	  for(int k=0;k<Nz;++k) {
	    Datei << "12"
		  << "\n";
	  }
     }
    }


    // Teil 4:
    Datei << "\nCELL_DATA " << cell_total << "\n";
    Datei << "SCALARS " << title << " float 1\n";

    Datei << "LOOKUP_TABLE default\n";
  }

  // Teil 6: Ausgabe der Werte der Variable

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {
  
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);

    size = Nz * Ny * Nx;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int i=0;i<Nx;++i) 
          for(int j=0;j<Ny;++j)
	    for(int k=0;k<Nz;++k) {
  	      Datei << convert(data_cell[id_hex][((Ny)*(k)+(j))*(Nx)+(i)])
		    << "\n";
	    }
      }
      else {
	data_coord = Parallel_buffer<double>::Give_send(size);
	for(int k=0;k<Nz;++k)
	  for(int j=0;j<Ny;++j)
	    for(int i=0;i<Nx;++i) {
  	    data_coord[((Ny)*(k)+(j))*(Nx)+(i)] = convert(data_cell[id_hex][((Ny)*(k)+(j))*(Nx)+(i)]);
	    }
	MPI_Send(data_coord,size,MPI_DOUBLE,0,600+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,600+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<Nz;++k)
	for(int j=0;j<Ny;++j)
	  for(int i=0;i<Nx;++i) {
	    Datei << data_coord[((Ny)*(k)+(j))*(Nx)+(i)] << "\n";
	  }
    }
   }
  }
}

template<class DTyp>
void Cell_variable<DTyp>::QPrint_VTK(QString DateiName,complex<double> (*convert)(DTyp x), double stretch_z, QString title,  Unstructured_grid_Marker * marker) {
  D3vector v;
  int id, Nx, Ny, Nz;
  int num_total = 0;
  int cell_total = 0;

  // QWidget* ww = new QWidget();
  // QMessageBox::information(ww, QString("6 Cell_variable<DTyp>::QPrint_VTK A"), QString("ASLD test"));

  QFile file(DateiName);
  if(!file.open(QIODevice::WriteOnly)) {
      std::cerr << "error open file for writing in print_var_Qt_cc.h:"
                << qPrintable(file.errorString()) << endl;
   }
  QTextStream Datei(&file);


  // for parallel
  double* data_coord;
  complex<double>* cdata_coord;
  int size, rank;
  MPI_Status status;

//  Datei.precision(12);
//  Datei.setf(std::ios::fixed,std::ios::floatfield);

  for(id=0;id<ug->Give_number_hexahedra();++id) {
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id)==yes_mark) {

//    Update<hexahedronEl>(id);
    num_total += blockgrid->Give_N_total_hexahedron(id);
    cell_total +=        (blockgrid->Give_Nx_hexahedron(id))
                        *(blockgrid->Give_Ny_hexahedron(id))
                        *(blockgrid->Give_Nz_hexahedron(id));
   }
  }

  // Teil 0: Write information
  if(my_rank==0) {
    Datei << "# vtk DataFile Version 2.0\n"
          << title << endl
          << "ASCII\n"
          << "DATASET UNSTRUCTURED_GRID\n";

    // Teil 1: Kopfzeile schreiben
    Datei << "POINTS " << num_total << " float\n";
  }

  // Teil 2: Koordinaten der Punkte ausgeben
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {

    Nz = blockgrid->Give_Nz_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);

    size = (Nz+1) * (Ny+1) * (Nx+1) * 3;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int k=0;k<=Nz;++k)
          for(int j=0;j<=Ny;++j)
            for(int i=0;i<=Nx;++i) {
              v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
              Datei << v.x
                    << " " << v.y
                    << " " << v.z * stretch_z
                    << "\n";
            }
      }
      else {
        data_coord = Parallel_buffer<double>::Give_send(size);
        for(int k=0;k<=Nz;++k)
          for(int j=0;j<=Ny;++j)
            for(int i=0;i<=Nx;++i) {
              v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
              data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0] = v.x;
              data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1] = v.y;
              data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] = v.z;
            }
        MPI_Send(data_coord,size,MPI_DOUBLE,0,500+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      data_coord = Parallel_buffer<double>::Give_receive(size);
      MPI_Recv(data_coord,size,MPI_DOUBLE,rank,500+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<=Nz;++k)
        for(int j=0;j<=Ny;++j)
          for(int i=0;i<=Nx;++i) {
            Datei <<    data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 0]
                   <<   "  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 1]
                   <<   "  " << data_coord[3*(i + (Nx+1) * (j + (Ny+1) * k )) + 2] << "\n";
          }
    }
   }
  }

  // Teil 3: Zellen ausgeben
  if(my_rank==0) {
    Datei << "\nCELLS " << cell_total << " " <<  cell_total*9  << "\n";

    int id_start;
    id_start = 0;
    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
     if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {

      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
          for(int k=0;k<Nz;++k) {
            Datei <<  "8 "
                  << id_start + i   + (Nx+1)*(j+1+(k+1)*(Ny+1) ) << " " // WNT
                  << id_start + i   + (Nx+1)*(j  +(k+1)*(Ny+1) ) << " " // WST
                  << id_start + i+1 + (Nx+1)*(j  +(k+1)*(Ny+1) ) << " " // EST
                  << id_start + i+1 + (Nx+1)*(j+1+(k+1)*(Ny+1) ) << " " // ENT
                  << id_start + i   + (Nx+1)*(j+1+  k * (Ny+1) ) << " " // WND
                  << id_start + i   + (Nx+1)*(j  +  k * (Ny+1) ) << " " // WSD
                  << id_start + i+1 + (Nx+1)*(j  +  k * (Ny+1) ) << " " // ESD
                  << id_start + i+1 + (Nx+1)*(j+1+  k * (Ny+1) ) // END
                  << "\n";
          }
      id_start = id_start +  blockgrid->Give_N_total_hexahedron(id_hex);
     }
    }

    Datei << "\nCELL_TYPES " << cell_total << "\n";

    for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
     if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {

      Nx = blockgrid->Give_Nx_hexahedron(id_hex);
      Ny = blockgrid->Give_Ny_hexahedron(id_hex);
      Nz = blockgrid->Give_Nz_hexahedron(id_hex);
      for(int i=0;i<Nx;++i)
        for(int j=0;j<Ny;++j)
          for(int k=0;k<Nz;++k) {
            Datei << "12"
                  << "\n";
          }
     }
    }


    // Teil 4:
    Datei << "\nCELL_DATA " << cell_total << "\n";
    Datei << "SCALARS " << title << " float 2\n";

    Datei << "LOOKUP_TABLE default\n";
  }

  // Teil 6: Ausgabe der Werte der Variable

  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
   if(marker == NULL || marker->Give_marker<hexahedronEl>(id_hex)==yes_mark ) {

    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);

    size = Nz * Ny * Nx;
    rank =  ug->Give_hexahedron(id_hex)->Give_rank();

    if(ug->Give_hexahedron(id_hex)->my_object(my_rank)) {
      if(my_rank==0) {
        for(int i=0;i<Nx;++i)
          for(int j=0;j<Ny;++j)
            for(int k=0;k<Nz;++k) {
              Datei << convert(data_cell[id_hex][((Ny)*(k)+(j))*(Nx)+(i)]).real() << "  "
                    << convert(data_cell[id_hex][((Ny)*(k)+(j))*(Nx)+(i)]).imag() << "\n";
            }
      }
      else {
        cdata_coord = Parallel_buffer<complex<double> >::Give_send(size);
        for(int k=0;k<Nz;++k)
          for(int j=0;j<Ny;++j)
            for(int i=0;i<Nx;++i) {
            cdata_coord[((Ny)*(k)+(j))*(Nx)+(i)] = convert(data_cell[id_hex][((Ny)*(k)+(j))*(Nx)+(i)]);
            }
        MPI_Send(cdata_coord,size,MPI_DOUBLE_COMPLEX,0,600+rank,ug->Give_MPI_comm());
      }
    }
    else if(my_rank==0) {
      cdata_coord = Parallel_buffer<complex<double> >::Give_receive(size);
      MPI_Recv(cdata_coord,size,MPI_DOUBLE_COMPLEX,rank,600+rank,ug->Give_MPI_comm(),&status);
      for(int k=0;k<Nz;++k)
        for(int j=0;j<Ny;++j)
          for(int i=0;i<Nx;++i) {
            Datei << cdata_coord[((Ny)*(k)+(j))*(Nx)+(i)].real() << "  "
                  << cdata_coord[((Ny)*(k)+(j))*(Nx)+(i)].imag() << "\n";
          }
    }
   }
  }
}


#ifdef _VTK

template <class DTyp>
OutputVTK & Variable<DTyp>::Objects_VTK( std::string name , double stretch_z) {

  if(my_rank>0) {
    cout << "Attention, my rank is " << my_rank << ", but Objects_VTK does not work in the MPI-parallel-version! " << endl;
  }

  D3vector v;
  int id, Nx, Ny, Nz;
  int num_total = 0;
  int cell_total = 0;
  
///////////UPDATE
  for(id=0;id<ug->Give_number_hexahedra();++id) {
    Update<hexahedronEl>(id);
    num_total += blockgrid->Give_N_total_hexahedron(id);
    cell_total +=        (blockgrid->Give_Nx_hexahedron(id))
                        *(blockgrid->Give_Ny_hexahedron(id))
                        *(blockgrid->Give_Nz_hexahedron(id));
  }

std::vector<std::vector<double> > v_Points;
std::vector<std::vector<double> > v_Cells;
std::vector<DTyp> v_Values;
int celltype = 12;

///////////POINTS
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex)
    for(int k=0;k<=blockgrid->Give_Nz_hexahedron(id_hex);++k) 
      for(int j=0;j<=blockgrid->Give_Ny_hexahedron(id_hex);++j)
	for(int i=0;i<=blockgrid->Give_Nx_hexahedron(id_hex);++i) {
	  v = blockgrid->Give_coord_hexahedron(id_hex,i,j,k);
	  std::vector<double> v_Tmp;
	  v_Tmp.push_back(v.x);
	  v_Tmp.push_back(v.y);
	  v_Tmp.push_back(v.z*stretch_z);
	  v_Points.push_back(v_Tmp);
	}

///////////ZELLEN
  int id_start;
  id_start = 0;
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);
    for(int i=0;i<Nx;++i)
      for(int j=0;j<Ny;++j)
	for(int k=0;k<Nz;++k) {
	  std::vector<double> v_Tmp;
		 v_Tmp.push_back(id_start + i   + (Nx+1)*(j+1+(k+1)*(Ny+1) )); // WNT
		 v_Tmp.push_back(id_start + i   + (Nx+1)*(j  +(k+1)*(Ny+1) )); // WST
		 v_Tmp.push_back(id_start + i+1 + (Nx+1)*(j  +(k+1)*(Ny+1) )); // EST
		 v_Tmp.push_back(id_start + i+1 + (Nx+1)*(j+1+(k+1)*(Ny+1) )); // ENT
		 v_Tmp.push_back(id_start + i   + (Nx+1)*(j+1+  k * (Ny+1) )); // WND
		 v_Tmp.push_back(id_start + i   + (Nx+1)*(j  +  k * (Ny+1) )); // WSD
		 v_Tmp.push_back(id_start + i+1 + (Nx+1)*(j  +  k * (Ny+1) )); // ESD
		 v_Tmp.push_back(id_start + i+1 + (Nx+1)*(j+1+  k * (Ny+1) )); // END
	  v_Cells.push_back(v_Tmp);
	}
    id_start = id_start +  blockgrid->Give_N_total_hexahedron(id_hex);
  }

///////////WERTE
  for(int id_hex=0;id_hex<ug->Give_number_hexahedra();++id_hex) {
    Nx = blockgrid->Give_Nx_hexahedron(id_hex);
    Ny = blockgrid->Give_Ny_hexahedron(id_hex);
    Nz = blockgrid->Give_Nz_hexahedron(id_hex);
    for(int k=0;k<=Nz;++k) 
      for(int j=0;j<=Ny;++j)
	for(int i=0;i<=Nx;++i) {
	  v_Values.push_back(data_hexahedra[id_hex][i+(Nx+1)*(j+(Ny+1)*k)]);
	}
  }
  OutputVTK* out;
  out = new OutputVTK(v_Points,v_Cells,v_Values,celltype,name);
  return *out;
}


#endif



#endif //#ifndef PRINTVARVTKCCH
