/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#ifndef COMPOSE_GRID_H
#define COMPOSE_GRID_H
#include <vector>
#include "blockgrid.h"

/** \addtogroup GridGeneration **/
/* @{ */   
/**
 * Composition von unterschiedlichen Gittern
 * 
 * **/
class ComposeUg : public Unstructured_grid {
 public:
   /**
    * List of block grids and list of transformation of z-coordinate
    */
  ComposeUg(std::vector<Blockgrid*> grids_, std::vector<double> transZ_);

  ~ComposeUg(){};
  
  bool isComposeGrid() { return true; }

  int getStartHex(int thisIdGrid);
  int getEndHex(  int thisIdGrid);

  int getStartQuad(int thisIdGrid);
  int getEndQuad(  int thisIdGrid);

  int getStartEdge(int thisIdGrid);
  int getEndEdge(  int thisIdGrid);

  int getStartPoint(int thisIdGrid);
  int getEndPoint(  int thisIdGrid);
    
 private:
   std::vector<double> startNumberHex;
   std::vector<double> startNumberQuad;
   std::vector<double> startNumberEdge;
   std::vector<double> startNumberPoint;
   
   std::vector<double> idGrid;    
};



class ComposeBlockGrid : public Blockgrid {
 public:
   /**
    * List of block grids and list of transformation of z-coordinate
    */
  ComposeBlockGrid(std::vector<Blockgrid*> grids_, std::vector<double> transZ_);
  ComposeUg* Give_unstructured_compose_grid() const { return static_cast<ComposeUg*>(ug); };
  
  ~ComposeBlockGrid(){};
  
 private:
   std::vector<Blockgrid*> grids;
   std::vector<double> transZ; 
};

/* @} */






#endif // EXAMPLES_UG_H
