/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "elements.h"

D3vector transform_nothing(double x, double*);
D3vector transform_nothing(double x, double y);

// Point_el

Point_el::Point_el(){
  typ_el_neighbor    = NULL;
  id_el_neighbor     = NULL;
  loc_id_hex         = NULL;
  rel_neigh_locstiff = NULL;
}

Point_el::~Point_el(){
  if (typ_el_neighbor != NULL) delete [] typ_el_neighbor;
  typ_el_neighbor = NULL;
  if (id_el_neighbor != NULL) delete [] id_el_neighbor;
  id_el_neighbor  = NULL;
  if (loc_id_hex  != NULL) delete [] loc_id_hex;
  loc_id_hex    = NULL;
  if (rel_neigh_locstiff  != NULL) delete [] rel_neigh_locstiff;
  rel_neigh_locstiff = NULL;
}

void Point_el::Set_size_typ_el_neighbor(int size){
  if(developer_version)
     if (typ_el_neighbor != NULL) std::cerr << " error in Set_size_typ_el_neighbor! " << endl;
  typ_el_neighbor = new elementTyp[size];
}

void Point_el::Set_size_id_el_neighbor(int size){
  if(developer_version)
     if (id_el_neighbor != NULL) std::cerr << " error in Set_size_id_el_neighbor! " << endl;
  id_el_neighbor = new int[size];
}

void Point_el::Set_size_loc_id_hex (int size){
  if(developer_version)
     if (loc_id_hex != NULL) std::cerr << " error in Set_size_loc_id_hex! " << endl;
  loc_id_hex = new local_id_of_hex3D[size];
}

void Point_el::Set_size_rel_neigh_locstiff (int size){
  if(developer_version)
     if(rel_neigh_locstiff != NULL) std::cerr << " error in Set_size_rel_neigh_locstiff! "
                                               << endl;
  rel_neigh_locstiff = new int[size];
}

// Edge_el

Edge_el::Edge_el() {
  color_of_edge = -1; // means: no color
  transform     = transform_nothing;
  transformFace = NULL;
  typ_M_el      = NULL;
  typ_W_el      = NULL;
  typ_E_el      = NULL;

  id_M_el       = NULL;
  id_W_el       = NULL;
  id_E_el       = NULL;

  loc_id_hex    = NULL;

  rel_neigh_locstiff = NULL;

  shift_index_transform = 0;
}

Edge_el::~Edge_el(){
  if (typ_M_el != NULL) delete [] typ_M_el;
  typ_M_el      = NULL;
  if (typ_W_el != NULL) delete [] typ_W_el;
  typ_W_el      = NULL;
  if (typ_E_el != NULL) delete [] typ_E_el;
  typ_E_el      = NULL;
  if (id_M_el  != NULL) delete [] id_M_el;
  id_M_el       = NULL;
  if (id_W_el  != NULL) delete [] id_W_el;
  id_W_el       = NULL;
  if (id_E_el  != NULL) delete [] id_E_el;
  id_E_el       = NULL;
  if (loc_id_hex  != NULL) delete [] loc_id_hex;
  loc_id_hex    = NULL;
  if (rel_neigh_locstiff  != NULL) delete [] rel_neigh_locstiff;
  rel_neigh_locstiff = NULL;
}

void Edge_el::Set_size_typ_M_el(int size){
  if(developer_version)
     if (typ_M_el != NULL) std::cerr << " error in Set_size_typ_M_el! " << endl;
  typ_M_el = new elementTyp[size];
}

void Edge_el::Set_size_typ_W_el(int size){
  if(developer_version)
     if (typ_W_el != NULL) std::cerr << " error in Set_size_typ_W_el! " << endl;
  typ_W_el = new elementTyp[size];
}

void Edge_el::Set_size_typ_E_el(int size){
  if(developer_version)
     if (typ_E_el != NULL) std::cerr << " error in Set_size_typ_E_el! " << endl;
  typ_E_el = new elementTyp[size];
}

void Edge_el::Set_size_id_M_el(int size){
  if(developer_version)
     if (id_M_el != NULL) std::cerr << " error in Set_size_id_M_el! " << endl;
  id_M_el = new int[size];
}

void Edge_el::Set_size_id_W_el(int size){
  if(developer_version)
     if (id_W_el != NULL) std::cerr << " error in Set_size_id_W_el! " << endl;
  id_W_el = new int[size];
}

void Edge_el::Set_size_id_E_el(int size){
  if(developer_version)
     if (id_E_el != NULL) std::cerr << " error in Set_size_id_E_el! " << endl;
  id_E_el = new int[size];
}

void Edge_el::Set_size_loc_id_hex (int size){
  if(developer_version)
     if (loc_id_hex != NULL) std::cerr << " error in Set_size_loc_id_hex! " << endl;
  loc_id_hex = new local_id_of_hex2D[size];
}

void Edge_el::Set_size_rel_neigh_locstiff (int size){
  if(developer_version)
     if(rel_neigh_locstiff != NULL) std::cerr << " error in Set_size_rel_neigh_locstiff! "
                                               << endl;
  rel_neigh_locstiff = new int[size];
}

// Quadrangle_el

Quadrangle_el::Quadrangle_el() {
  //transform = transform_nothing;
  exists_exterior = true;
  transform = NULL;

  shift_index_transform = 0;
}

D3vector Quadrangle_el::Give_middle_coord() {
  D3vector middle = D3vector(0.0,0.0,0.0);
  for(int i=0;i<4;++i) 
  {
     // Give_coord((dir2D_sons)i).Print();cout << endl;
    // @todo : hab folgendes ersetzt durch
    // middle = middle + ug->Give_point(quad->Give_id_corner((dir2D_sons)i))->Give_coordinate();
     middle = middle + Give_coord((dir2D_sons)i);
  }
  //(middle/4).Print();cout << endl;
  return (middle / 4.0);
}

// Hexahedron_el

Hexahedron_el::Hexahedron_el() {
  for(int i=0;i<12;++i){
     transform[i]  = NULL;
     transform_hex = NULL;
     orientation[i] = true;
  }
}

D3vector Hexahedron_el::Give_middle_coord() {
  D3vector middle = D3vector(0.0,0.0,0.0);
  for(int i=0;i<8;++i){
      //Give_coord((dir3D_sons)i).Print();cout << endl;
      middle = middle + Give_coord((dir3D_sons)i);
  }
  //(middle/8).Print();cout << endl;
  return (middle / 8.0);
}
