/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#ifndef ELEMENTS_H
#define ELEMENTS_H

struct local_id_of_hex2D{
  int id_ND, id_ST, id_NT;
};

struct local_id_of_hex3D{
  int id_SED, id_NWD, id_NED, id_SWT, id_SET, id_NWT, id_NET;
};

class Geometric_Object{
  public:
  Geometric_Object() { rank_owner = 0; };
    virtual ~Geometric_Object(){};

    int  Give_rank()            { return rank_owner; }
    void set_rank(int rank)     { rank_owner = rank; }
    bool my_object(int my_rank) { return my_rank==rank_owner; }
      //      std::cout << " my_rank: " << my_rank << " rank_owner: " << rank_owner << std::endl;
    
 private:
    int rank_owner;
};

class Point_el : public Geometric_Object {
  public:
  Point_el();
  ~Point_el();

  void Set_coordinate ( D3vector coordinate_ ) { coordinate = coordinate_; }
  D3vector Give_coordinate() { return coordinate; }

  void Set_number_neighbors(int n) { number_neighbors = n; };
  int Give_number_neighbors() { return number_neighbors; };

  void Set_number_neighbors_hex(int n) { number_neighbors_hex = n; };
  int Give_number_neighbors_hex() { return number_neighbors_hex; };

  void Put_to_topf_neighbor(int et, int id)    { topf_neighbor.put(et, id); };
  unsorted_pair Find_in_topf_neighbor(int num) { return topf_neighbor.find(num); };
  int Give_size_of_topf_neighbor()             { return topf_neighbor.size(); };

  void Set_size_typ_el_neighbor(int size);
  void Set_size_id_el_neighbor (int size);

  void Set_typ_el_neighbor(int pos, elementTyp et){ typ_el_neighbor[pos] = et;    };
  void Set_id_el_neighbor (int pos, int et_id     ){ id_el_neighbor [pos] = et_id; };

  elementTyp Give_typ_el_neighbor(int pos){ return typ_el_neighbor[pos]; };
  int         Give_id_el_neighbor (int pos){ return id_el_neighbor [pos]; };

  void Set_size_loc_id_hex (int size);

  inline void Set_loc_id_hex  (int pos, local_dir_of_hex3D x, int loc_id);
  inline int  Give_loc_id_hex (int pos, local_dir_of_hex3D x);

  void Set_size_rel_neigh_locstiff (int size);

  void Set_rel_neigh_locstiff  (int neigh, int pos){ rel_neigh_locstiff[neigh] = pos;    };
  int  Give_rel_neigh_locstiff (int neigh)         { return rel_neigh_locstiff[neigh]; };

  private:
  D3vector coordinate;

  // unstructured part:
  // information about neighbor
  int number_neighbors, number_neighbors_hex;

  elementTyp* typ_el_neighbor; // [number_neighbors];
  int        *  id_el_neighbor; // [number_neighbors];

  Topf<2, unsorted_pair> topf_neighbor;           // map: global id    -> local id
                                                  //      (typ,id_num)

  local_id_of_hex3D* loc_id_hex;
  int* rel_neigh_locstiff;
};

class Edge_el : public Geometric_Object {
 public:
  Edge_el();
  ~Edge_el();

  int Give_id_corner_W() { return id_corner_W; }
  int Give_id_corner_E() { return id_corner_E; }

  inline int Give_id_corner(dir1D LR);
  inline void Set_id_corners(int id_W, int id_E);

  // color with respect to degree of freedom
  void Set_color_of_edge(int c) { color_of_edge = c; };
  int  Give_color_of_edge()     { return color_of_edge;};

  void Set_number_neighbors(int n) { number_neighbors = n; };
  int Give_number_neighbors() { return number_neighbors; };

  void Set_number_neighbors_hex(int n) { number_neighbors_hex = n; };
  int Give_number_neighbors_hex() { return number_neighbors_hex; };

  // information about element mapping [0,1] ->  R^3
  D3vector (*transform)(double, double*); // transformation of edge

  // information about element mapping [0,1]^2 ->  R^3
  D3vector (*transformFace)(double, double, double*); // transformation of edge from face


  double* shiftPointer(double* p) { return &(p[shift_index_transform]); }  ///> notwendig wegen compose grid 
  void setShiftPointer(int shiftPoi) { shift_index_transform = shiftPoi; } ///> notwendig wegen compose grid 
  int  getShiftPointer() { return shift_index_transform; } ///> notwendig wegen compose grid 
  
  
  void Put_to_topf_neighbor(int et, int id)    { topf_neighbor.put(et, id); };
  unsorted_pair Find_in_topf_neighbor(int num) { return topf_neighbor.find(num); };
  int Give_size_of_topf_neighbor()             { return topf_neighbor.size(); };

  void Set_size_typ_M_el(int size);
  void Set_size_id_M_el (int size);

  void Set_size_typ_W_el(int size);
  void Set_size_id_W_el (int size);

  void Set_size_typ_E_el(int size);
  void Set_size_id_E_el (int size);

  void Set_typ_M_el(int pos, elementTyp et){ typ_M_el[pos] = et;    };
  void Set_id_M_el (int pos, int et_id     ){ id_M_el [pos] = et_id; };

  void Set_typ_W_el(int pos, elementTyp et){ typ_W_el[pos] = et;    };
  void Set_id_W_el (int pos, int et_id     ){ id_W_el [pos] = et_id; };

  void Set_typ_E_el(int pos, elementTyp et){ typ_E_el[pos] = et;    };
  void Set_id_E_el (int pos, int et_id     ){ id_E_el [pos] = et_id; };

  elementTyp Give_typ_M_el(int pos){ return typ_M_el[pos]; };
  int         Give_id_M_el (int pos){ return id_M_el [pos]; };

  elementTyp Give_typ_W_el(int pos){ return typ_W_el[pos]; };
  int         Give_id_W_el (int pos){ return id_W_el [pos]; };

  elementTyp Give_typ_E_el(int pos){ return typ_E_el[pos]; };
  int         Give_id_E_el (int pos){ return id_E_el [pos]; };

  void Set_size_loc_id_hex (int size);

  inline void Set_loc_id_hex  (int pos, local_dir_of_hex2D x, int loc_id);
  inline int  Give_loc_id_hex (int pos, local_dir_of_hex2D x);

  void Set_size_rel_neigh_locstiff (int size);

  void Set_rel_neigh_locstiff  (int neigh, int pos){ rel_neigh_locstiff[neigh] = pos;    };
  int  Give_rel_neigh_locstiff (int neigh)         { return rel_neigh_locstiff[neigh]; };
  
  D3vector Give_coord(dir1D c)        { return coordinates[c]; }
  void Set_coord(dir1D c, D3vector v) { coordinates[c] = v; }
  
 private:
  // structured part:
  int id_corner_W;
  int id_corner_E;

  // unstructured part:
  // information about neighbor
  int number_neighbors;
  int number_neighbors_hex;

  Topf<2, unsorted_pair> topf_neighbor;           // map: global id    -> local id
                                                  //      (typ,id_num)
  int color_of_edge;  // with respect to degree of freedom

  elementTyp* typ_M_el;  // typ of corresponding middel element
  elementTyp* typ_W_el;  // typ of corresponding west  element
  elementTyp* typ_E_el;  // typ of corresponding east  element

  int* id_M_el;   // id number of corresponding middel element
  int* id_W_el;   // id number of corresponding west  element
  int* id_E_el;   // id number of corresponding east  element

  local_id_of_hex2D* loc_id_hex;
  int* rel_neigh_locstiff;
  
    // for local coordinates with respect to one adjecent hexahedron
  D3vector coordinates[2];
  
  int shift_index_transform; // notwendig wenn man composite grid macht
};

class Quadrangle_el : public Geometric_Object {
 public:
  Quadrangle_el();
  ~Quadrangle_el(){};

  // structured part:
  int Give_id_corner(dir2D_sons i)         { return id_corner[i]; }
  void Set_id_corner(dir2D_sons i, int id) { id_corner[i] = id; }

  int Give_id_edge(dir2D i)         { return id_edge[i]; }
  void Set_id_edge(dir2D i, int id) { id_edge[i] = id; }

  // unstructured part:
  // interior
  int Give_id_interior_edge(dir2D_sons i)         { return id_interior_edge[i]; }
  void Set_id_interior_edge(dir2D_sons i, int id) { id_interior_edge[i] = id; }

  int Give_id_interior_face(dir2D i)             { return id_interior_face[i]; }
  void Set_id_interior_face(dir2D i, int id)     { id_interior_face[i] = id; }

  elementTyp Give_typ_interior_face(dir2D i)           { return typ_interior_face[i]; }
  void Set_typ_interior_face(dir2D i, elementTyp typ)  { typ_interior_face[i] = typ; }

  int Give_id_interior_vol()                  { return id_interior_vol; }
  void Set_id_interior_vol(int id)            { id_interior_vol = id; }

  elementTyp Give_typ_interior_vol()         { return typ_interior_vol; }
  void Set_typ_interior_vol(elementTyp typ)  { typ_interior_vol = typ; }

  // exterior
  bool Give_exists_exterior()        { return exists_exterior; }
  void Set_exists_exterior(bool yn) { exists_exterior = yn;   }

  int Give_id_exterior_edge(dir2D_sons i)         { return id_exterior_edge[i]; }
  void Set_id_exterior_edge(dir2D_sons i, int id) { id_exterior_edge[i] = id; }

  int Give_id_exterior_face(dir2D i)             { return id_exterior_face[i]; }
  void Set_id_exterior_face(dir2D i, int id)     { id_exterior_face[i] = id; }

  elementTyp Give_typ_exterior_face(dir2D i)           { return typ_exterior_face[i]; }
  void Set_typ_exterior_face(dir2D i, elementTyp typ)  { typ_exterior_face[i] = typ; }

  int Give_id_exterior_vol()                  { return id_exterior_vol; }
  void Set_id_exterior_vol(int id)            { id_exterior_vol = id; }

  elementTyp Give_typ_exterior_vol()         { return typ_exterior_vol; }
  void Set_typ_exterior_vol(elementTyp typ)  { typ_exterior_vol = typ; }

  // information about element
  //D3vector (*transform)(double, double); // transformation of quad
  D3vector (*transform)(double, double, double*); // transformation of quad
  double* shiftPointer(double* p) { return &(p[shift_index_transform]); }  ///> notwendig wegen compose grid
  void setShiftPointer(int shiftPoi) { shift_index_transform = shiftPoi; } ///> notwendig wegen compose grid
  int  getShiftPointer() { return shift_index_transform; } ///> notwendig wegen compose grid

  // for local coordinates with respect to one adjecent hexahedron
  D3vector Give_coord(dir2D_sons c)        { return coordinates[c]; }
  void Set_coord(dir2D_sons c, D3vector v) { coordinates[c] = v; }

  // geometric middle of coordinates of corner points
  D3vector Give_middle_coord();
  
 private:
  int id_corner[4];
  int id_edge[4];

  int id_interior_edge[4];
  int id_interior_face[4];
  elementTyp typ_interior_face[4];
  int id_interior_vol;
  elementTyp typ_interior_vol;

  bool exists_exterior;

  int id_exterior_edge[4];
  int id_exterior_face[4];
  elementTyp typ_exterior_face[4];
  int id_exterior_vol;
  elementTyp typ_exterior_vol;
  
  
   // for local coordinates with respect to one adjecent hexahedron
  D3vector coordinates[4];

  int shift_index_transform; // notwendig wenn man composite grid macht
};



class Hexahedron_el : public Geometric_Object {
 public:
  Hexahedron_el();
  ~Hexahedron_el(){};

  // structured part:
  int Give_id_corner(dir3D_sons i)         { return id_corner[i]; }
  void Set_id_corner(dir3D_sons i, int id) { id_corner[i] = id; }

  int Give_id_edge(Edges_cell i)         { return id_edge[i]; }
  void Set_id_edge(Edges_cell i, int id) { id_edge[i] = id; }

  int Give_id_quadrangle(dir3D i)         { return id_quadrangle[i]; }
  void Set_id_quadrangle(dir3D i, int id) { id_quadrangle[i] = id; }

  D3vector Give_coord(dir3D_sons c)        { return coordinates[c]; }
  void Set_coord(dir3D_sons c, D3vector v) { coordinates[c] = v; }

  // information edges 12 times [0,1] -> R^3
  D3vector (*transform[12])(double, double*); // transformation of edges
  double* shiftPointer(double* p) { return &(p[shift_index_transform]); }  ///> notwendig wegen compose grid 
  void setShiftPointer(int shiftPoi) { shift_index_transform = shiftPoi; } ///> notwendig wegen compose grid 

  // information hex [0,1]^2 -> R^2
  D3vector (*transform_quad[6])(int, double,double,double*); // transformation of quadrangles
  // information hex [0,1]^3 -> R^3
  D3vector (*transform_hex)(int, double,double,double); // transformation of volumne

  bool     orientation[12];
  //bool     orientationQuad[6];

  // geometric middle of coordinates of corner points
  D3vector Give_middle_coord();
  
 private:
  int id_corner[8];
  int id_edge[12];
  int id_quadrangle[6];

  // for local coordinates in hexahedron
  D3vector coordinates[8];
  
  int shift_index_transform; // notwendig wenn man composite grid macht, wir nehmen zur Vereinfachung an, 
                             // dass der shift pro hex der gleiche ist
};

//////////////////////////////////////////
// Implementation of inline functions
//////////////////////////////////////////

inline int Edge_el::Give_id_corner(dir1D LR) {
  if (LR == Ldir1D) return id_corner_W;
  if(developer_version)
     if (LR != Rdir1D) std::cout << " error in Edge_el::Give_id_corner!" << std::endl;
  return id_corner_E;
}

inline void Edge_el::Set_id_corners(int id_W, int id_E) {
  id_corner_W = id_W;
  id_corner_E = id_E;
}

inline void Edge_el::Set_loc_id_hex(int pos, local_dir_of_hex2D x, int loc_id){
  if      (x==ND) loc_id_hex[pos].id_ND = loc_id;
  else if(x==ST) loc_id_hex[pos].id_ST = loc_id;
  else {
     if(developer_version)
        if (x != NT) std::cout << " error in Edge_el::Set_loc_id_hex!" << std::endl;
     loc_id_hex[pos].id_NT = loc_id;
  }
}

inline int  Edge_el::Give_loc_id_hex (int pos, local_dir_of_hex2D x){
  if(x==ND) return loc_id_hex[pos].id_ND;
  if(x==ST) return loc_id_hex[pos].id_ST;
  if(developer_version)
     if (x != NT) std::cout << " error in Edge_el::Give_loc_id_hex!" << std::endl;
  return loc_id_hex[pos].id_NT;
}

inline void Point_el::Set_loc_id_hex(int pos, local_dir_of_hex3D x, int loc_id){
  if      (x==SED) loc_id_hex[pos].id_SED = loc_id;
  else if(x==NWD) loc_id_hex[pos].id_NWD = loc_id;
  else if(x==NED) loc_id_hex[pos].id_NED = loc_id;
  else if(x==SWT) loc_id_hex[pos].id_SWT = loc_id;
  else if(x==SET) loc_id_hex[pos].id_SET = loc_id;
  else if(x==NWT) loc_id_hex[pos].id_NWT = loc_id;
  else {
     if(developer_version)
        if (x != NET) std::cout << " error in Point_el::Set_loc_id_hex!" << std::endl;
     loc_id_hex[pos].id_NET = loc_id;
  }
}

inline int  Point_el::Give_loc_id_hex (int pos, local_dir_of_hex3D x){
  if(x==SED) return loc_id_hex[pos].id_SED;
  if(x==NWD) return loc_id_hex[pos].id_NWD;
  if(x==NED) return loc_id_hex[pos].id_NED;
  if(x==SWT) return loc_id_hex[pos].id_SWT;
  if(x==SET) return loc_id_hex[pos].id_SET;
  if(x==NWT) return loc_id_hex[pos].id_NWT;
  if(developer_version)
     if (x != NET) std::cout << " error in Point_el::Give_loc_id_hex!" << std::endl;
  return loc_id_hex[pos].id_NET;
}

#endif // ELEMENTS_H


