/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "elements.h"
#include "parti.h"
#include "ug.h"
#include "examples_ug.h"


//double Radius_mirror_left_global;
//double z_left_global;
//double Rx_global;
//double Ry_global;
//double R_global;





D3vector global_transform_mirror_left ( int chosenBlock, double x,double y,double z )
{
//   if (x > 0) return 1.; else return 0.;
//   return x*y;
  //  return 0.0;
  /*
  return - ( 1.0 - (z-z_left_global)/Lz_global)
    * ( sqrt(Radius_mirror_left_global*Radius_mirror_left_global-x*x -y*y)
       -sqrt(Radius_mirror_left_global*Radius_mirror_left_global-
      (Rx_global* Rx_global + Ry_global*Ry_global)));
  */
//   /*x = x / Lx_global;
//   y = y / Ly_global;
//   */z /= Lz_global;
//
  x = x - 0.5;
  y = y - 0.5;

  //oben nun omputational von -0.5 bis 0.5

  /*  x = x * Lx_global;
    y = y * Ly_global;*/

  // nun  in physik. Koordinaten
  if ( x > 0.6 ) cout << "error1" << endl;
  if ( y > 0.6 ) cout << "error2" << endl;
  return D3vector(- ( 1.0 - z ) * ( 0.25-x*x ) * ( 0.25-y*y ));
//   return - (1.0 - z) * (0.25*Lx_global*Lx_global-x*x)*(0.25*Ly_global*Ly_global-y*y) ;
//   return - (1.0 - z) * (0.25*Lx_global*Lx_global-x*x)*(0.25*Ly_global*Ly_global-y*y) * ;
//   return - (1.0 - z) * (sqrt(Radius_mirror_left_global*Radius_mirror_left_global/(4.*Rx_global*Rx_global)-x*x -y*y)
//         -sqrt(Radius_mirror_left_global*Radius_mirror_left_global/(4.*Rx_global*Rx_global-0.5)));
}



Hexahedron::Hexahedron(double lx, double ly, double lz,
                       double rx, double ry, double rz ) : UG_WithZdirection(hexahedronUgTyp,2) { 
  constructionParameters.resize(6);			 
  constructionParameters[0] = lx;
  constructionParameters[1] = ly;
  constructionParameters[2] = lz;
  constructionParameters[3] = rx;
  constructionParameters[4] = ry;
  constructionParameters[5] = rz;

			 
  Lz.at(0) = lz;
  Lz.at(1) = rz;
  Set_number_points ( 8 ); // 8 an den Ecken

  Set_coordinate_point ( WSDdir3D, D3vector ( lx,ly,lz ) );
  Set_coordinate_point ( ESDdir3D, D3vector ( rx,ly,lz ) );
  Set_coordinate_point ( WNDdir3D, D3vector ( lx,ry,lz ) );
  Set_coordinate_point ( ENDdir3D, D3vector ( rx,ry,lz ) );
  Set_coordinate_point ( WSTdir3D, D3vector ( lx,ly,rz ) );
  Set_coordinate_point ( ESTdir3D, D3vector ( rx,ly,rz ) );
  Set_coordinate_point ( WNTdir3D, D3vector ( lx,ry,rz ) );
  Set_coordinate_point ( ENTdir3D, D3vector ( rx,ry,rz ) );

  Set_number_hexahedra ( 1 );
  Set_hexahedron ( 0, WSDdir3D, ESDdir3D, WNDdir3D, ENDdir3D, WSTdir3D, ESTdir3D, WNTdir3D, ENTdir3D );

  construction_hexahedron_points_done();
  // set all directions constant
  Give_not_constant_directions()->Mark_x_as_not_constant_at ( 0,false );
  Give_not_constant_directions()->Mark_y_as_not_constant_at ( 0,false );
  Give_not_constant_directions()->Mark_z_as_not_constant_at ( 0,false );

  // no transformations of the edges and faces
  construction_done();
}

HexahedronRotY::HexahedronRotY(double lx, double ly, double lz, double rx, double ry, double rz, double sa) 
                                  : UG_WithZdirection(hexahedronUgTyp,2){
  constructionParameters.resize(7);
  constructionParameters[0] = lx;
  constructionParameters[1] = ly;
  constructionParameters[2] = lz;
  constructionParameters[3] = rx;
  constructionParameters[4] = ry;
  constructionParameters[5] = rz;
  constructionParameters[6] = sa;
			 
  Lz.at(0) = lz;
  Lz.at(1) = rz;
  Set_number_points ( 8 ); // 8 an den Ecken

  assert(sa*sa <=1.0);
  double ca = sqrt(1.0 - sa*sa);
  
  double sx = 0.5*(rx-lx);
  double sz = 0.5*(rz-lz);
  
  double ETz =  ca * sz - sa * sx;
  double ETx =  sa * sz + ca * sx;
  
  double WTz =  ca * sz + sa * sx;
  double WTx =  sa * sz - ca * sx;

  double EDz = -ca * sz - sa * sx;
  double EDx = -sa * sz + ca * sx;
  
  double WDz = -ca * sz + sa * sx;
  double WDx = -sa * sz - ca * sx;

  double centerX = 0.5*(rx+lx);
  double centerZ = 0.5*(rz+lz);
  
  
  Set_coordinate_point ( WSDdir3D, D3vector ( centerX + WDx,ly, centerZ + WDz ) );
  Set_coordinate_point ( ESDdir3D, D3vector ( centerX + EDx,ly, centerZ + EDz ) );
  Set_coordinate_point ( WNDdir3D, D3vector ( centerX + WDx,ry, centerZ + WDz ) );
  Set_coordinate_point ( ENDdir3D, D3vector ( centerX + EDx,ry, centerZ + EDz ) );

  Set_coordinate_point ( WSTdir3D, D3vector ( centerX + WTx,ly, centerZ + WTz ) );
  Set_coordinate_point ( ESTdir3D, D3vector ( centerX + ETx,ly, centerZ + ETz ) );
  Set_coordinate_point ( WNTdir3D, D3vector ( centerX + WTx,ry, centerZ + WTz ) );
  Set_coordinate_point ( ENTdir3D, D3vector ( centerX + ETx,ry, centerZ + ETz ) );
  
  Set_number_hexahedra ( 1 );
  Set_hexahedron ( 0, WSDdir3D, ESDdir3D, WNDdir3D, ENDdir3D, WSTdir3D, ESTdir3D, WNTdir3D, ENTdir3D );

  construction_hexahedron_points_done();
  // set all directions constant
  Give_not_constant_directions()->Mark_x_as_not_constant_at ( 0,false );
  Give_not_constant_directions()->Mark_y_as_not_constant_at ( 0,false );
  Give_not_constant_directions()->Mark_z_as_not_constant_at ( 0,false );

  // no transformations of the edges and faces
  construction_done();
}

void HexahedronBasis::PrintLengthEdgesHere(std::vector<D3vector> pointsHere) {
   cout << " Print Hexaeder info of UGBlocks. " << endl;
   cout << " -------------------------------- " << endl;
   cout << " L1: " << D3VectorNorm(pointsHere.at(WSDdir3D) - pointsHere.at(ESDdir3D)) << endl;
   cout << " L1: " << D3VectorNorm(pointsHere.at(WSDdir3D) - pointsHere.at(WNDdir3D)) << endl;
   cout << " L1: " << D3VectorNorm(pointsHere.at(WSDdir3D) - pointsHere.at(WSTdir3D)) << endl;   
}

Hexahedron::Hexahedron ( double l_x, double l_y, double l_z ) : UG_WithZdirection(hexahedronUgTyp,2) {  
  Lz.at(0) = 0.0; 
  Lz.at(1) = l_z; 
  
  constructionParameters.resize(6);	
  constructionParameters[0] = 0.0;
  constructionParameters[1] = 0.0;
  constructionParameters[2] = 0.0;
  constructionParameters[3] = l_x;
  constructionParameters[4] = l_y;
  constructionParameters[5] = l_z;  
  
  Set_number_points ( 8 ); // 8 an den Ecken

  Set_coordinate_point ( WSDdir3D, D3vector ( 0.0,0.0,0.0 ) );
  Set_coordinate_point ( ESDdir3D, D3vector ( l_x,0.0,0.0 ) );
  Set_coordinate_point ( WNDdir3D, D3vector ( 0.0,l_y,0.0 ) );
  Set_coordinate_point ( ENDdir3D, D3vector ( l_x,l_y,0.0 ) );
  Set_coordinate_point ( WSTdir3D, D3vector ( 0.0,0.0,l_z ) );
  Set_coordinate_point ( ESTdir3D, D3vector ( l_x,0.0,l_z ) );
  Set_coordinate_point ( WNTdir3D, D3vector ( 0.0,l_y,l_z ) );
  Set_coordinate_point ( ENTdir3D, D3vector ( l_x,l_y,l_z ) );

  Set_number_hexahedra ( 1 );
  Set_hexahedron ( 0, WSDdir3D, ESDdir3D, WNDdir3D, ENDdir3D, WSTdir3D, ESTdir3D, WNTdir3D, ENTdir3D );

  construction_hexahedron_points_done();
  // set all directions constant
  Give_not_constant_directions()->Mark_x_as_not_constant_at ( 0,false );
  Give_not_constant_directions()->Mark_y_as_not_constant_at ( 0,false );
  Give_not_constant_directions()->Mark_z_as_not_constant_at ( 0,false );

  // no transformations of the edges and faces
  construction_done();
}

HexahedronXYList::HexahedronXYList(std::vector<double>& lx, std::vector<double>& ly,  std::vector<double>& lz) : UG_WithZdirection(hexahedronListUgTyp,2) {
  
  constructionParameters.resize(0);
  
  construction_lx = lx;
  construction_ly = ly;
  construction_lz = lz;
  

  
  Lz.at(0) = lz[0];
  Lz.at(1) = lz[lz.size()-1];
  n_x = lx.size()-1;
  n_y = ly.size()-1;
  n_z = lz.size()-1;

  Set_number_points ((n_x+1) * (n_y+1) * (n_z+1) );

  for(int i_z = 0; i_z <= 1; ++i_z ) {
    for(int i_y = 0; i_y <= n_y; ++i_y ) {
      for(int i_x = 0; i_x <= n_x; ++i_x ) {
        Set_coordinate_point((i_z*(n_x+1) * (n_y+1) + i_y * (n_x+1) + i_x),D3vector(lx[i_x],ly[i_y],lz[i_z]) );
     }
    }
  }

  Set_number_hexahedra ( n_x * n_y * n_z );

  for(int i_z = 0; i_z < n_z; ++i_z ) {
    for(int i_y = 0; i_y < n_y; ++i_y ) {
      for(int i_x = 0; i_x < n_x; ++i_x ) {
        Set_hexahedron((i_z*n_x*n_y + i_y * n_x + i_x),
          ((i_z)*  (n_x+1) * (n_y+1) + (i_y)   * (n_x+1) + (i_x)),   // WSDdir3D
          ((i_z)*  (n_x+1) * (n_y+1) + (i_y)   * (n_x+1) + (i_x+1)), // ESDdir3D
          ((i_z)*  (n_x+1) * (n_y+1) + (i_y+1) * (n_x+1) + (i_x)),   // WNDdir3D
          ((i_z)*  (n_x+1) * (n_y+1) + (i_y+1) * (n_x+1) + (i_x+1)), // ENDdir3D
          ((i_z+1)*(n_x+1) * (n_y+1) + (i_y)   * (n_x+1) + (i_x)),   // WSTdir3D
          ((i_z+1)*(n_x+1) * (n_y+1) + (i_y)   * (n_x+1) + (i_x+1)), // ESTdir3D
          ((i_z+1)*(n_x+1) * (n_y+1) + (i_y+1) * (n_x+1) + (i_x)),   // WNTdir3D
          ((i_z+1)*(n_x+1) * (n_y+1) + (i_y+1) * (n_x+1) + (i_x+1)));// ENTdir3D
      }
    }
  }

  construction_hexahedron_points_done();
  // set all directions constant
  Give_not_constant_directions()->Mark_x_as_not_constant_at ( 0,false );
  Give_not_constant_directions()->Mark_y_as_not_constant_at ( 0,false );
  Give_not_constant_directions()->Mark_z_as_not_constant_at ( 0,false );

  // no transformations of the edges and faces
  construction_done();
}

HexahedronXYListTest::HexahedronXYListTest(double l_x, double l_y, double l_z,
    double r_x, double r_y, double r_z ) : UG_WithZdirection(hexahedronListTestUgTyp,2) {
  Lz.at(0) = l_z;
  Lz.at(1) = r_z;
  n_x = 3;
  n_y = 3;
  n_z = 1;
  constructionParameters.resize(6);
  constructionParameters[0] = l_x;
  constructionParameters[1] = l_y;
  constructionParameters[2] = l_z;
  constructionParameters[3] = r_x;
  constructionParameters[4] = r_y;
  constructionParameters[5] = r_z;

  Set_number_points ((n_x+1) * (n_y+1) * (n_z+1) );

  for(int i_z = 0; i_z <= 1; ++i_z ) {
    for(int i_y = 0; i_y <= 3; ++i_y ) {
      for(int i_x = 0; i_x <= 3; ++i_x ) {
        double x;
        double y;

        if(i_x == 0)
          x = l_x;
        else if(i_x == 1)
          x = -0.2*fabs(1.5-i_y)*(r_y-l_y);
        else if(i_x == 2)
          x =  0.2*fabs(1.5-i_y)*(r_y-l_y);
        else if(i_x == 3)
          x = r_x;

        if(i_y == 0)
          y = l_y;
        else if(i_y == 1)
          y = -0.2*fabs(1.5-i_x)*(r_x-l_x);
        else if(i_y == 2)
          y =  0.2*fabs(1.5-i_x)*(r_x-l_x);
        else if(i_y == 3)
          y = r_y;
        Set_coordinate_point((i_z*(n_x+1) * (n_y+1) + i_y * (n_x+1) + i_x),D3vector(x,y,(i_z==0?l_z:r_z) ));
     }
    }
  }

  Set_number_hexahedra ( n_x * n_y * n_z );

  for(int i_z = 0; i_z < n_z; ++i_z ) {
    for(int i_y = 0; i_y < n_y; ++i_y ) {
      for(int i_x = 0; i_x < n_x; ++i_x ) {
        Set_hexahedron((i_z*n_x*n_y + i_y * n_x + i_x),
          ((i_z)*  (n_x+1) * (n_y+1) + (i_y)   * (n_x+1) + (i_x)),   // WSDdir3D
          ((i_z)*  (n_x+1) * (n_y+1) + (i_y)   * (n_x+1) + (i_x+1)), // ESDdir3D
          ((i_z)*  (n_x+1) * (n_y+1) + (i_y+1) * (n_x+1) + (i_x)),   // WNDdir3D
          ((i_z)*  (n_x+1) * (n_y+1) + (i_y+1) * (n_x+1) + (i_x+1)), // ENDdir3D
          ((i_z+1)*(n_x+1) * (n_y+1) + (i_y)   * (n_x+1) + (i_x)),   // WSTdir3D
          ((i_z+1)*(n_x+1) * (n_y+1) + (i_y)   * (n_x+1) + (i_x+1)), // ESTdir3D
          ((i_z+1)*(n_x+1) * (n_y+1) + (i_y+1) * (n_x+1) + (i_x)),   // WNTdir3D
          ((i_z+1)*(n_x+1) * (n_y+1) + (i_y+1) * (n_x+1) + (i_x+1)));// ENTdir3D
      }
    }
  }

  construction_hexahedron_points_done();
  // set all directions constant
  Give_not_constant_directions()->Mark_x_as_not_constant_at ( 0,false );
  Give_not_constant_directions()->Mark_y_as_not_constant_at ( 0,false );
  Give_not_constant_directions()->Mark_z_as_not_constant_at ( 0,false );

  // no transformations of the edges and faces
  construction_done();
}

Hexahedron_periodic::Hexahedron_periodic ( double l_x, double l_y, double l_z ) {
  //Set_number_points(8,12);
	Set_number_points ( 12 );

	Set_coordinate_point ( 0, D3vector ( 0.0,0.0,0.0 ) );
	Set_coordinate_point ( 1, D3vector ( l_x,0.0,0.0 ) );
	Set_coordinate_point ( 2, D3vector ( 0.0,l_y,0.0 ) );
	Set_coordinate_point ( 3, D3vector ( l_x,l_y,0.0 ) );
	Set_coordinate_point ( 4, D3vector ( 0.0,0.0,0.25*l_z ) );
	Set_coordinate_point ( 5, D3vector ( l_x,0.0,0.25*l_z ) );
	Set_coordinate_point ( 6, D3vector ( 0.0,l_y,0.25*l_z ) );
	Set_coordinate_point ( 7, D3vector ( l_x,l_y,0.25*l_z ) );
	Set_coordinate_point ( 8, D3vector ( 0.0,0.0,0.5*l_z ) );
	Set_coordinate_point ( 9, D3vector ( l_x,0.0,0.5*l_z ) );
	Set_coordinate_point ( 10, D3vector ( 0.0,l_y,0.5*l_z ) );
	Set_coordinate_point ( 11, D3vector ( l_x,l_y,0.5*l_z ) );
	Set_number_hexahedra ( 3 );
	Set_hexahedron ( 0, 0, 1, 2, 3, 4, 5, 6, 7 );
	Set_hexahedron ( 1, 4, 5, 6, 7, 8, 9, 10, 11 );
	Set_hexahedron ( 2, 8, 9, 10, 11, 0, 1, 2, 3,
									 D3vector ( 0.0,0.0,0.5*l_z ),
															D3vector ( l_x,0.0,0.5*l_z ),
																				 D3vector ( 0.0,l_y,0.5*l_z ),
																						 D3vector ( l_x,l_y,0.5*l_z ),
																								 D3vector ( 0.0,0.0,l_z ),
																										 D3vector ( l_x,0.0,l_z ),
																												 D3vector ( 0.0,l_y,l_z ),
																														 D3vector ( l_x,l_y,l_z ) );


	construction_hexahedron_points_done();
  // no transformations of the edges and faces
	construction_done();
}


HexahedronList::HexahedronList(double lx, double ly, std::vector<double>& relativeCoordZ,
			       double offsetX, double offsetY, double offsetZ, bool periodic) 
                : UG_WithZdirection(hexahedronUgTyp,relativeCoordZ.size()) {
        Lz = relativeCoordZ;
	
	constructionParameters.resize(5);
	constructionParameters[0] = lx;
	constructionParameters[1] = ly;
	constructionParameters[2] = offsetX;
	constructionParameters[3] = offsetY;
	constructionParameters[4] = offsetZ;
	
	relativeCoordVector = relativeCoordZ;
	
	constructionBoolArched = periodic;

	const int numBlocks(relativeCoordZ.size()-1);
	assert (numBlocks > 0);
	
	// no last interface for periodic boundary condiditions
	const int numInterfaces(relativeCoordZ.size() - ( int ) periodic);
	
	const int numpoints(numInterfaces *4);
	Set_number_points ( numpoints );

	// set points
	for(int i = 0; i < numInterfaces; ++i ) {
	        Set_coordinate_point ( i*4+0,
					 D3vector ( offsetX,offsetY,offsetZ+relativeCoordZ.at ( i ) ) );
		Set_coordinate_point ( i*4+1,
					 D3vector ( offsetX+lx,offsetY,offsetZ+relativeCoordZ.at ( i ) ) );
		Set_coordinate_point ( i*4+2,
				         D3vector ( offsetX,offsetY+ly,offsetZ+relativeCoordZ.at ( i ) ) );
		Set_coordinate_point ( i*4+3,
					 D3vector ( offsetX+lx,offsetY+ly,offsetZ+relativeCoordZ.at ( i ) ) );
	}
	
	// set hexahedra
	Set_number_hexahedra ( numBlocks );
  
	// alle bis auf letzten block
	for (int i = 0; i < numBlocks-1; ++i )
	{
		Set_hexahedron ( i, i*4 + 0, i*4 + 1, i*4 + 2, i*4 + 3,
										 i*4 + 4, i*4 + 5, i*4 + 6, i*4 + 7 );
	}
	if ( !periodic ) // set last block
	{
		int i = numBlocks-1;
		Set_hexahedron ( i, i*4 + 0, i*4 + 1, i*4 + 2, i*4 + 3,
										 i*4 + 4, i*4 + 5, i*4 + 6, i*4 + 7 );
	} else // set last block with periodic boundary conditions
	{
		unsigned int i = numBlocks-1;
		double lz = relativeCoordZ.at ( i+1 );
		double coordz = relativeCoordZ.at ( i );
    /*std::cout << "Periodic block infos: " << endl
		<< "z = " << lz << endl
		<< "coordz = " << coordz << endl
		<< "index of block = " << i << endl;
		*/
		Set_hexahedron ( i, i*4 + 0, i*4 + 1, i*4 + 2, i*4 + 3, 0, 1, 2, 3,
											D3vector ( offsetX,offsetY,coordz ), 
											D3vector ( offsetX+lx,offsetY,coordz ),
											D3vector ( offsetX,offsetY+ly,coordz ),	
											D3vector ( offsetX+lx,offsetY+ly,coordz ),
											D3vector ( offsetX,offsetY,lz ), 
											D3vector ( offsetX+lx,offsetY,lz ),
											D3vector ( offsetX,offsetY+ly,lz ), D3vector ( offsetX+lx,offsetY+ly,lz ) );
	}

	
	construction_hexahedron_points_done();
  // set all directions constant
  /*
	Give_not_constant_directions()->Mark_x_as_not_constant_at(0,false);
	Give_not_constant_directions()->Mark_y_as_not_constant_at(0,false);
	Give_not_constant_directions()->Mark_z_as_not_constant_at(0,false);
	*/

	construction_done();
}



////////////////////////////////////////////////////////////////////
// adaptive hexahedron list
////////////////////////////////////////////////////////////////////

AdaptiveHexahedronList::AdaptiveHexahedronList (
                         double Routx, double Routy,
			 double Rinx,  double Riny,
			 std::vector<double> & relativeCoordZ, double offsetX, double offsetY, double offsetZ ) 
                       : UG_WithZdirection(hexahedronAdaptiveUgTyp,relativeCoordZ.size()) { 
			  
	constructionParameters.resize(7);
	constructionParameters[0] = Routx;
	constructionParameters[1] = Routy;
	constructionParameters[2] = Rinx;
	constructionParameters[3] = Riny;
	constructionParameters[4] = offsetX;
	constructionParameters[5] = offsetY;
	constructionParameters[6] = offsetZ;
	relativeCoordVector = relativeCoordZ;
	
			 
			 
  Lz = relativeCoordZ;
//  pointer_global_data = new double[1];
//  pointer_global_data[0] = R;
  int num_blocks = relativeCoordZ.size() -1;
  assert(num_blocks>0);
  int num_blocks_total = 5* ( relativeCoordZ.size()-1 );
  int num_points_per_face = 8;
//   int num_faces = num_blocks +1;
  int numPoints = num_points_per_face*(num_blocks+1) ;

  Set_number_points ( numPoints );
  Set_number_hexahedra ( num_blocks_total );


  for ( int i = 0; i <= num_blocks; ++i ) {
    Set_coordinate_point ( i*num_points_per_face+0, D3vector ( offsetX+Rinx, offsetY+Riny,offsetZ+relativeCoordZ.at ( i ) ) );
    Set_coordinate_point ( i*num_points_per_face+1, D3vector ( offsetX+Routx,offsetY+Routy,offsetZ+relativeCoordZ.at ( i ) ) );

    Set_coordinate_point ( i*num_points_per_face+2, D3vector ( offsetX-Rinx, offsetY+Riny,offsetZ+relativeCoordZ.at ( i ) ) );
    Set_coordinate_point ( i*num_points_per_face+3, D3vector ( offsetX-Routx,offsetY+Routy,offsetZ+relativeCoordZ.at ( i ) ) );

    Set_coordinate_point ( i*num_points_per_face+4, D3vector ( offsetX-Rinx, offsetY-Riny,offsetZ+relativeCoordZ.at ( i ) ) );
    Set_coordinate_point ( i*num_points_per_face+5, D3vector ( offsetX-Routx,offsetY-Routy,offsetZ+relativeCoordZ.at ( i ) ) );

    Set_coordinate_point ( i*num_points_per_face+6, D3vector ( offsetX+Rinx, offsetY-Riny,offsetZ+relativeCoordZ.at ( i ) ) );
    Set_coordinate_point ( i*num_points_per_face+7, D3vector ( offsetX+Routx,offsetY-Routy,offsetZ+relativeCoordZ.at ( i ) ) );
    
    
  }

  for ( int i = 0; i < num_blocks; ++i ) {
    Set_hexahedron ( i, i*num_points_per_face + 0, i*num_points_per_face + 2,
                        i*num_points_per_face + 6, i*num_points_per_face + 4,
                        ( i+1 ) * num_points_per_face + 0, ( i+1 ) * num_points_per_face +2,  
                        ( i+1 ) * num_points_per_face + 6, ( i+1 ) * num_points_per_face +4 );
  } 

  for ( int j = 0; j < num_points_per_face; j=j+2 )
  {
    for ( int i = 0; i < num_blocks; ++i )
    {
      int myIndex = (j/2+1)*num_blocks +i;
      Set_hexahedron ( myIndex,
                       i*num_points_per_face     +  j+0,
                       i*num_points_per_face     +  j+1,
                       i*num_points_per_face     +((j+2) % num_points_per_face),
                       i*num_points_per_face     +((j+3) % num_points_per_face),
                       (i+1)*num_points_per_face +  j+0,
                       (i+1)*num_points_per_face +  j+1,
                       (i+1)*num_points_per_face +((j+2) % num_points_per_face),
                       (i+1)*num_points_per_face +((j+3) % num_points_per_face)
                     );
    }
  }

  construction_hexahedron_points_done();
  construction_done();
}



////////////////////////////////////////////////////////////////////
// cylinder
////////////////////////////////////////////////////////////////////


D3vector transform_cyl_NE ( double t, double* pointer_global_data) {
  double R_global_data = pointer_global_data[0];
  return R_global_data*D3vector ( cos ( t*0.5*M_PI )- ( 1-t ), sin ( t*0.5*M_PI )-t, 0.0 );
}

D3vector transform_cyl_NW ( double t, double* pointer_global_data ) {
  double R_global_data = pointer_global_data[0];
  return R_global_data*D3vector ( -sin ( t*0.5*M_PI ) +t, cos ( t*0.5*M_PI )- ( 1-t ), 0.0 );
}

D3vector transform_cyl_SW ( double t, double* pointer_global_data ) {
  double R_global_data = pointer_global_data[0];
  return R_global_data*D3vector ( -cos ( t*0.5*M_PI ) + ( 1-t ), -sin ( t*0.5*M_PI ) +t, 0.0 );
}

D3vector transform_cyl_SE ( double t, double* pointer_global_data ) {
  double R_global_data = pointer_global_data[0];
  return R_global_data*D3vector ( cos ( t*0.5*M_PI )- ( 1-t ), -sin ( t*0.5*M_PI ) +t, 0.0 );
}

D3vector transform_cyl_NE_inner ( double t, double* pointer_global_data) {
  double R_global_data = pointer_global_data[1];
  return R_global_data*D3vector ( cos ( t*0.5*M_PI )- ( 1-t ), sin ( t*0.5*M_PI )-t, 0.0 );
}

D3vector transform_cyl_NW_inner ( double t, double* pointer_global_data ) {
  double R_global_data = pointer_global_data[1];
  return R_global_data*D3vector ( -sin ( t*0.5*M_PI ) +t, cos ( t*0.5*M_PI )- ( 1-t ), 0.0 );
}

D3vector transform_cyl_SW_inner ( double t, double* pointer_global_data ) {
  double R_global_data = pointer_global_data[1];
  return R_global_data*D3vector ( -cos ( t*0.5*M_PI ) + ( 1-t ), -sin ( t*0.5*M_PI ) +t, 0.0 );
}

D3vector transform_cyl_SE_inner ( double t, double* pointer_global_data ) {
  double R_global_data = pointer_global_data[1];
  return R_global_data*D3vector ( cos ( t*0.5*M_PI )- ( 1-t ), -sin ( t*0.5*M_PI ) +t, 0.0 );
}

// Fuer DEformation der 4 inneren Kanten t=0,...,1
D3vector transform_cyl_NE_square ( double t, double* pointer_global_data) {
  double r = pointer_global_data[1];
  double rho = pointer_global_data[2];
  //return D3vector(0.0,0.0,0.0); //Keine Aenderung
  return D3vector (rho*(cos(M_PI/6.0*(1.0+t))-0.5)-r*(1.0-t), rho*(sin(M_PI/6.0*(1.0+t))-0.5)-r*t,0.0);
}

D3vector transform_cyl_NW_square ( double t, double* pointer_global_data ) {
  double r = pointer_global_data[1];
  double rho = pointer_global_data[2];
  //return D3vector(0.0,0.0,0.0); //Keine Aenderung
  return D3vector (rho*(-sin(M_PI/6.0*(1.0+t))+0.5)+r*t, rho*(cos(M_PI/6.0*(1.0+t))-0.5)-r*(1.0-t),0.0);
}

D3vector transform_cyl_SW_square ( double t, double* pointer_global_data ) {
  double r = pointer_global_data[1];
  double rho = pointer_global_data[2];
  //return D3vector(0.0,0.0,0.0); //Keine Aenderung
  return D3vector (-rho*(cos(M_PI/6.0*(1.0+t))-0.5)+r*(1.0-t), rho*(-sin(M_PI/6.0*(1.0+t))+0.5)+r*t,0.0);
}

D3vector transform_cyl_SE_square ( double t, double* pointer_global_data ) {
  double r = pointer_global_data[1];
  double rho = pointer_global_data[2];
  //return D3vector(0.0,0.0,0.0); //Keine Aenderung
  return D3vector (rho*(cos(M_PI/6.0*(1.0+t))-0.5)-r*(1.0-t), rho*(-sin(M_PI/6.0*(1.0+t))+0.5)+r*t,0.0);
}

D3vector transform_zero( double t1,double t2, double* pointer_global_data )
{
   return D3vector (0,0,0);
}



D3vector transform_edge_sphere_T( double t1,double t2, double* pointer_global_data ) {
  double r = pointer_global_data[0];
  double x = -r+2.0*t1*r;
  double y = -r+2.0*t2*r;
  double z = r;
  double xDash = x * sqrt(1-y*y/2-z*z/2+y*y*z*z/3);
  double yDash = y * sqrt(1-z*z/2-x*x/2+x*x*z*z/3);
  double zDash = z * sqrt(1-y*y/2-x*x/2+y*y*x*x/3);
  xDash = xDash * sqrt(3);
  yDash = yDash * sqrt(3);
  zDash = zDash * sqrt(3);
  return D3vector (xDash-x,yDash-y,zDash-z);
}




D3vector transform_quadrangle_sphere_T( double t1,double t2, double* pointer_global_data ) {
  double r = pointer_global_data[0];
  double x = -r+2.0*t1*r;
  double y = -r+2.0*t2*r;
  double z = r;
  double xDash = x * sqrt(1-y*y/2-z*z/2+y*y*z*z/3);
  double yDash = y * sqrt(1-z*z/2-x*x/2+x*x*z*z/3);
  double zDash = z * sqrt(1-y*y/2-x*x/2+y*y*x*x/3);
  xDash = xDash * sqrt(3);
  yDash = yDash * sqrt(3);
  zDash = zDash * sqrt(3);
  return D3vector (xDash-x,yDash-y,zDash-z);
}


// facetest face_transform_test transformTest = new face_transform_(transform_quadrangle_sphere_T);


D3vector transform_quadrangle_sphere_D( double t1,double t2, double* pointer_global_data ) {
  double r = pointer_global_data[0];
  double x = -r+2.0*t1*r;
  double y = -r+2.0*t2*r;
  double z = -r;
  double xDash = x * sqrt(1-y*y/2-z*z/2+y*y*z*z/3);
  double yDash = y * sqrt(1-z*z/2-x*x/2+x*x*z*z/3);
  double zDash = z * sqrt(1-y*y/2-x*x/2+y*y*x*x/3);
  xDash = xDash * sqrt(3);
  yDash = yDash * sqrt(3);
  zDash = zDash * sqrt(3);
  return D3vector (xDash-x,yDash-y,zDash-z);
}

D3vector transform_quadrangle_sphere_N( double t1,double t2, double* pointer_global_data ) {
  double r = pointer_global_data[0];
  double x = -r+2.0*t1*r;
  double y = r;
  double z = -r+2.0*t2*r;
  double xDash = x * sqrt(1-y*y/2-z*z/2+y*y*z*z/3);
  double yDash = y * sqrt(1-z*z/2-x*x/2+x*x*z*z/3);
  double zDash = z * sqrt(1-y*y/2-x*x/2+y*y*x*x/3);
  xDash = xDash * sqrt(3);
  yDash = yDash * sqrt(3);
  zDash = zDash * sqrt(3);
  return D3vector (xDash-x,yDash-y,zDash-z);
}

D3vector transform_quadrangle_sphere_S( double t1,double t2, double* pointer_global_data ) {
  double r = pointer_global_data[0];
  double x = -r+2.0*t1*r;
  double y = -r;
  double z = -r+2.0*t2*r;
  double xDash = x * sqrt(1-y*y/2-z*z/2+y*y*z*z/3);
  double yDash = y * sqrt(1-z*z/2-x*x/2+x*x*z*z/3);
  double zDash = z * sqrt(1-y*y/2-x*x/2+y*y*x*x/3);
  xDash = xDash * sqrt(3);
  yDash = yDash * sqrt(3);
  zDash = zDash * sqrt(3);
  return D3vector (xDash-x,yDash-y,zDash-z);
}

D3vector transform_quadrangle_sphere_E( double t1,double t2, double* pointer_global_data ) {
  double r = pointer_global_data[0];
  double x = r;
  double y = -r+2.0*t1*r;
  double z = -r+2.0*t2*r;
  double xDash = x * sqrt(1-y*y/2-z*z/2+y*y*z*z/3);
  double yDash = y * sqrt(1-z*z/2-x*x/2+x*x*z*z/3);
  double zDash = z * sqrt(1-y*y/2-x*x/2+y*y*x*x/3);
  xDash = xDash * sqrt(3);
  yDash = yDash * sqrt(3);
  zDash = zDash * sqrt(3);
  return D3vector (xDash-x,yDash-y,zDash-z);
}

D3vector transform_quadrangle_sphere_W( double t1,double t2, double* pointer_global_data ) {
  double r = pointer_global_data[0];
  double x = -r;
  double y = -r+2.0*t1*r;
  double z = -r+2.0*t2*r;
  double xDash = x * sqrt(1-y*y/2-z*z/2+y*y*z*z/3);
  double yDash = y * sqrt(1-z*z/2-x*x/2+x*x*z*z/3);
  double zDash = z * sqrt(1-y*y/2-x*x/2+y*y*x*x/3);
  xDash = xDash * sqrt(3);
  yDash = yDash * sqrt(3);
  zDash = zDash * sqrt(3);
  return D3vector (xDash-x,yDash-y,zDash-z);
}

bool Cylinder::getInnerGridArched() const {
  return inner_grid_arched_;
}
Cylinder::Cylinder ( double R, double r, double l , bool inner_grid_arched) : UG_WithZdirection(cylinderUgTyp,2), inner_grid_arched_(inner_grid_arched), Radius(R),radius(r),length(l) {
  Lz.at(0) = 0.0;
  Lz.at(1) = l;
  
  constructionParameters.resize(3);
  constructionParameters[0] = R;
  constructionParameters[1] = r;
  constructionParameters[2] = l;
  constructionBoolArched = inner_grid_arched;

  typFuerSlice = cylinderUgTyp;
  
  double offsetX = 0.0;
  double offsetY = 0.0;
  double offsetZ = 0.0;
  
  pointer_global_data = new double[3];
  size_pointer_global_data = 3;
   
  //pointer_global_data = new double[1];
  pointer_global_data[0] = R;
  pointer_global_data[1] = r;
  double R_innen = 2.0*r*sqrt(2.0)/(sqrt(6.0)-sqrt(2.0));
  pointer_global_data[2] = R_innen;
  int num_blocks = Lz.size() -1;
  assert(num_blocks>0);
  int num_blocks_total = 5* ( Lz.size()-1 );
  int num_points_per_face = 8;
//   int num_faces = num_blocks +1;
  int numPoints = num_points_per_face*(num_blocks+1) ;

  Set_number_points ( numPoints );
  Set_number_hexahedra ( num_blocks_total );


  for(int i = 0; i <= num_blocks; ++i ) {
      Set_coordinate_point( i*num_points_per_face+0, D3vector ( offsetX+r,offsetY,offsetZ+Lz.at ( i ) ) );
      Set_coordinate_point( i*num_points_per_face+1, D3vector ( offsetX+R,offsetY,offsetZ+Lz.at ( i ) ) );

      Set_coordinate_point( i*num_points_per_face+2, D3vector ( offsetX,offsetY+r,offsetZ+Lz.at ( i ) ) );
      Set_coordinate_point( i*num_points_per_face+3, D3vector ( offsetX,offsetY+R,offsetZ+Lz.at ( i ) ) );

      Set_coordinate_point( i*num_points_per_face+4, D3vector ( offsetX-r,offsetY,offsetZ+Lz.at ( i ) ) );
      Set_coordinate_point( i*num_points_per_face+5, D3vector ( offsetX-R,offsetY,offsetZ+Lz.at ( i ) ) );

      Set_coordinate_point( i*num_points_per_face+6, D3vector ( offsetX,offsetY-r,offsetZ+Lz.at ( i ) ) );
      Set_coordinate_point( i*num_points_per_face+7, D3vector ( offsetX,offsetY-R,offsetZ+Lz.at ( i ) ) );
  }

  for ( int i = 0; i < num_blocks; ++i ) {
    Set_hexahedron ( i, i*num_points_per_face + 0, i*num_points_per_face + 2,
                        i*num_points_per_face + 6, i*num_points_per_face + 4,
                        ( i+1 ) * num_points_per_face + 0, ( i+1 ) * num_points_per_face +2,  
                        ( i+1 ) * num_points_per_face + 6, ( i+1 ) * num_points_per_face +4 );
  } 

  for ( int j = 0; j < num_points_per_face; j=j+2) {
    for ( int i = 0; i < num_blocks; ++i) {
      int myIndex = (j/2+1)*num_blocks +i;
      Set_hexahedron ( myIndex,
                       i*num_points_per_face     +  j+0,
                       i*num_points_per_face     +  j+1,
                       i*num_points_per_face     +((j+2) % num_points_per_face),
                       i*num_points_per_face     +((j+3) % num_points_per_face),
                       (i+1)*num_points_per_face +  j+0,
                       (i+1)*num_points_per_face +  j+1,
                       (i+1)*num_points_per_face +((j+2) % num_points_per_face),
                       (i+1)*num_points_per_face +((j+3) % num_points_per_face)
                     );
//test      
//      std::cout<<"Set_hexahedron("<<myIndex<<","<<i*num_points_per_face     +  j+0<<","<<i*num_points_per_face     +  j+1<<","<<i*num_points_per_face     +((j+2) % num_points_per_face)<<","<<i*num_points_per_face     +((j+3) % num_points_per_face)
//    <<","<<(i+1)*num_points_per_face     +  j+0<<","<<(i+1)*num_points_per_face     +  j+1<<","<<(i+1)*num_points_per_face     +((j+2) % num_points_per_face)<<","<<(i+1)*num_points_per_face     +((j+3) % num_points_per_face)<<")"<<std::endl;
    }
  }

  construction_hexahedron_points_done();

  int shift;
  for ( int i = 0; i <= num_blocks; ++i ) {
    shift = i * num_points_per_face;
    Set_transformation_edge ( 1+shift, 3+shift, transform_cyl_NE );
    Set_transformation_edge ( 3+shift, 5+shift, transform_cyl_NW );
    Set_transformation_edge ( 5+shift, 7+shift, transform_cyl_SW );
    Set_transformation_edge ( 1+shift, 7+shift, transform_cyl_SE );
    
    if (inner_grid_arched == true) {
      Set_transformation_edge ( 0+shift, 2+shift, transform_cyl_NE_square );
      Set_transformation_edge ( 2+shift, 4+shift, transform_cyl_NW_square );
      Set_transformation_edge ( 4+shift, 6+shift, transform_cyl_SW_square );
      Set_transformation_edge ( 0+shift, 6+shift, transform_cyl_SE_square );
    }
  }
  construction_done();
  
  
  /*
  //alte Version
  pointer_global_data = new double[1];
  pointer_global_data[0] = R;
  Set_number_points ( 16 ); // 16an den Flaechen
  Set_coordinate_point ( 0, D3vector ( r,0.0,0.0 ) );          // D
  Set_coordinate_point ( 1, D3vector ( R,0.0,0.0 ) );          // D
  Set_coordinate_point ( 2, D3vector ( r,0.0,l ) );            // T
  Set_coordinate_point ( 3, D3vector ( R,0.0,l ) );            // T
  Set_coordinate_point ( 4, D3vector ( 0.0,r,0.0 ) );          // D
  Set_coordinate_point ( 5, D3vector ( 0.0,R,0.0 ) );          // D
  Set_coordinate_point ( 6, D3vector ( 0.0,r,l ) );            // T
  Set_coordinate_point ( 7, D3vector ( 0.0,R,l ) );            // T
  Set_coordinate_point ( 8, D3vector ( -r,0.0,0.0 ) );         // D
  Set_coordinate_point ( 9, D3vector ( -R,0.0,0.0 ) );         // D
  Set_coordinate_point ( 10, D3vector ( -r,0.0,l ) );          // T
  Set_coordinate_point ( 11, D3vector ( -R,0.0,l ) );          // T
  Set_coordinate_point ( 12, D3vector ( 0.0,-r,0.0 ) );        // D
  Set_coordinate_point ( 13, D3vector ( 0.0,-R,0.0 ) );        // D
  Set_coordinate_point ( 14, D3vector ( 0.0,-r,l ) );          // T
  Set_coordinate_point ( 15, D3vector ( 0.0,-R,l ) );          // T
  Set_number_hexahedra ( 5 );
  Set_hexahedron ( 0, 0, 1, 4, 5, 2, 3, 6, 7 );
  Set_hexahedron ( 1, 4, 5, 8, 9, 6, 7, 10, 11 );
  Set_hexahedron ( 2, 8, 9, 12, 13, 10, 11, 14, 15 );
  Set_hexahedron ( 3, 12, 13, 0, 1, 14, 15, 2, 3 );
  Set_hexahedron ( 4, 0, 4, 12, 8, 2, 6, 14, 10 );

  construction_hexahedron_points_done();

  Set_transformation_edge ( 1, 5, transform_cyl_NE );
  Set_transformation_edge ( 3, 7, transform_cyl_NE );
  Set_transformation_edge ( 5, 9, transform_cyl_NW );
  Set_transformation_edge ( 7, 11, transform_cyl_NW );
  Set_transformation_edge ( 9, 13, transform_cyl_SW );
  Set_transformation_edge ( 11, 15, transform_cyl_SW );
  Set_transformation_edge ( 1, 13, transform_cyl_SE );
  Set_transformation_edge ( 3, 15, transform_cyl_SE );

  construction_done();
  */
}


Ring::Ring ( double R, double r, double l ) {
  pointer_global_data = new double[2];
  size_pointer_global_data = 2;
    
  pointer_global_data[0] = R;
  pointer_global_data[1] = r;
  Set_number_points ( 16 ); // 16an den Flaechen
  Set_coordinate_point ( 0, D3vector ( r,0.0,0.0 ) );          // D
  Set_coordinate_point ( 1, D3vector ( R,0.0,0.0 ) );          // D
  Set_coordinate_point ( 2, D3vector ( r,0.0,l ) );            // T
  Set_coordinate_point ( 3, D3vector ( R,0.0,l ) );            // T
  Set_coordinate_point ( 4, D3vector ( 0.0,r,0.0 ) );          // D
  Set_coordinate_point ( 5, D3vector ( 0.0,R,0.0 ) );          // D
  Set_coordinate_point ( 6, D3vector ( 0.0,r,l ) );            // T
  Set_coordinate_point ( 7, D3vector ( 0.0,R,l ) );            // T
  Set_coordinate_point ( 8, D3vector ( -r,0.0,0.0 ) );         // D
  Set_coordinate_point ( 9, D3vector ( -R,0.0,0.0 ) );         // D
  Set_coordinate_point ( 10, D3vector ( -r,0.0,l ) );          // T
  Set_coordinate_point ( 11, D3vector ( -R,0.0,l ) );          // T
  Set_coordinate_point ( 12, D3vector ( 0.0,-r,0.0 ) );        // D
  Set_coordinate_point ( 13, D3vector ( 0.0,-R,0.0 ) );        // D
  Set_coordinate_point ( 14, D3vector ( 0.0,-r,l ) );          // T
  Set_coordinate_point ( 15, D3vector ( 0.0,-R,l ) );          // T
  Set_number_hexahedra ( 4 );
  //Set_number_hexahedra ( 5 );
  Set_hexahedron ( 0, 0, 1, 4, 5, 2, 3, 6, 7 );
  Set_hexahedron ( 1, 4, 5, 8, 9, 6, 7, 10, 11 );
  Set_hexahedron ( 2, 8, 9, 12, 13, 10, 11, 14, 15 );
  Set_hexahedron ( 3, 12, 13, 0, 1, 14, 15, 2, 3 );
  //Set_hexahedron ( 4, 0, 4, 12, 8, 2, 6, 14, 10 );

  construction_hexahedron_points_done();

  Set_transformation_edge ( 1, 5, transform_cyl_NE );
  Set_transformation_edge ( 3, 7, transform_cyl_NE );
  Set_transformation_edge ( 5, 9, transform_cyl_NW );
  Set_transformation_edge ( 7, 11, transform_cyl_NW );
  Set_transformation_edge ( 9, 13, transform_cyl_SW );
  Set_transformation_edge ( 11, 15, transform_cyl_SW );
  Set_transformation_edge ( 1, 13, transform_cyl_SE );
  Set_transformation_edge ( 3, 15, transform_cyl_SE );

  Set_transformation_edge ( 0, 4, transform_cyl_NE_inner );
  Set_transformation_edge ( 2, 6, transform_cyl_NE_inner );
  Set_transformation_edge ( 4, 8, transform_cyl_NW_inner );
  Set_transformation_edge ( 6, 10, transform_cyl_NW_inner );
  Set_transformation_edge ( 8, 12, transform_cyl_SW_inner );
  Set_transformation_edge ( 10, 14, transform_cyl_SW_inner );
  Set_transformation_edge ( 0, 12, transform_cyl_SE_inner );
  Set_transformation_edge ( 2, 14, transform_cyl_SE_inner );

  construction_done();
}



CylinderList::CylinderList ( double R, double r, std::vector<double> & relativeCoordZ, 
			     double offsetX, double offsetY, double offsetZ, bool inner_grid_arched )
             : Radius(R),radius(r), length(relativeCoordZ), inner_grid_arched_(inner_grid_arched), UG_WithZdirection(cylinderListUgTyp,relativeCoordZ.size()) {
  Lz = relativeCoordZ;
  
  
    constructionParameters.resize(5);
  constructionParameters[0] = R;
  constructionParameters[1] = r;
  constructionParameters[2] = offsetX;
  constructionParameters[3] = offsetY;
  constructionParameters[4] = offsetZ;
  constructionBoolArched = inner_grid_arched;
  relativeCoordVector = relativeCoordZ;
  

  typFuerSlice = cylinderListUgTyp;

  pointer_global_data = new double[3];
  size_pointer_global_data = 3;
    
  pointer_global_data[0] = R;
  pointer_global_data[1] = r;
  double R_innen = 2.0*r*sqrt(2.0)/(sqrt(6.0)-sqrt(2.0));
  pointer_global_data[2] = R_innen;
  int num_blocks = relativeCoordZ.size() -1;
  assert(num_blocks>0);
  int num_blocks_total = 5* ( relativeCoordZ.size()-1 );
  int num_points_per_face = 8;
//   int num_faces = num_blocks +1;
  int numPoints = num_points_per_face*(num_blocks+1) ;

  Set_number_points ( numPoints );
  Set_number_hexahedra ( num_blocks_total );


  for ( int i = 0; i <= num_blocks; ++i ) {
    Set_coordinate_point ( i*num_points_per_face+0, D3vector ( offsetX+r,offsetY,offsetZ+relativeCoordZ.at ( i ) ) );
    Set_coordinate_point ( i*num_points_per_face+1, D3vector ( offsetX+R,offsetY,offsetZ+relativeCoordZ.at ( i ) ) );

    Set_coordinate_point ( i*num_points_per_face+2, D3vector ( offsetX,offsetY+r,offsetZ+relativeCoordZ.at ( i ) ) );
    Set_coordinate_point ( i*num_points_per_face+3, D3vector ( offsetX,offsetY+R,offsetZ+relativeCoordZ.at ( i ) ) );

    Set_coordinate_point ( i*num_points_per_face+4, D3vector ( offsetX-r,offsetY,offsetZ+relativeCoordZ.at ( i ) ) );
    Set_coordinate_point ( i*num_points_per_face+5, D3vector ( offsetX-R,offsetY,offsetZ+relativeCoordZ.at ( i ) ) );

    Set_coordinate_point ( i*num_points_per_face+6, D3vector ( offsetX,offsetY-r,offsetZ+relativeCoordZ.at ( i ) ) );
    Set_coordinate_point ( i*num_points_per_face+7, D3vector ( offsetX,offsetY-R,offsetZ+relativeCoordZ.at ( i ) ) );
  }

  for ( int i = 0; i < num_blocks; ++i ) {
    Set_hexahedron ( i, i*num_points_per_face + 0, i*num_points_per_face + 2,
                        i*num_points_per_face + 6, i*num_points_per_face + 4,
                        ( i+1 ) * num_points_per_face + 0, ( i+1 ) * num_points_per_face +2,  
                        ( i+1 ) * num_points_per_face + 6, ( i+1 ) * num_points_per_face +4 );
  } 

  for ( int j = 0; j < num_points_per_face; j=j+2 )
  {
    for ( int i = 0; i < num_blocks; ++i )
    {
      int myIndex = (j/2+1)*num_blocks +i;
      Set_hexahedron ( myIndex,
                       i*num_points_per_face     +  j+0,
                       i*num_points_per_face     +  j+1,
                       i*num_points_per_face     +((j+2) % num_points_per_face),
                       i*num_points_per_face     +((j+3) % num_points_per_face),
                       (i+1)*num_points_per_face +  j+0,
                       (i+1)*num_points_per_face +  j+1,
                       (i+1)*num_points_per_face +((j+2) % num_points_per_face),
                       (i+1)*num_points_per_face +((j+3) % num_points_per_face)
                     );
    }
  }

  construction_hexahedron_points_done();

  int shift;
  for ( int i = 0; i <= num_blocks; ++i )
  {
    shift = i * num_points_per_face;
    Set_transformation_edge ( 1+shift, 3+shift, transform_cyl_NE );
    Set_transformation_edge ( 3+shift, 5+shift, transform_cyl_NW );
    Set_transformation_edge ( 5+shift, 7+shift, transform_cyl_SW );
    Set_transformation_edge ( 1+shift, 7+shift, transform_cyl_SE );
    
    if (inner_grid_arched == true) {
      Set_transformation_edge ( 0+shift, 2+shift, transform_cyl_NE_square );
      Set_transformation_edge ( 2+shift, 4+shift, transform_cyl_NW_square );
      Set_transformation_edge ( 4+shift, 6+shift, transform_cyl_SW_square );
      Set_transformation_edge ( 0+shift, 6+shift, transform_cyl_SE_square );
    }
  }
  construction_done();
}
bool CylinderList::getInnerGridArched() const {
  return inner_grid_arched_;
}


int CylinderList::getNz() const {
  return length.size()-1;
}

//--------------------------------------------



SpecialPrismList::SpecialPrismList(double yL_, double xL_, 
		                   double yS_, double xS_, 
		                   double yH_, std::vector<double> & relativeCoordZ, 
			     double offsetX, double offsetY, double offsetZ)
             : yL(yL_), xL(xL_), yS(yS_), xS(xS_), yH(yH_), 
               length(relativeCoordZ), UG_WithZdirection(cylinderListUgTyp,relativeCoordZ.size()) {
  Lz = relativeCoordZ;
  
  constructionParameters.resize(8);
  constructionParameters[0] = yL_;
  constructionParameters[1] = xL_;
  constructionParameters[2] = yS_;
  constructionParameters[3] = xS_;
  constructionParameters[4] = yH_;
  constructionParameters[5] = offsetX;
  constructionParameters[6] = offsetY;
  constructionParameters[7] = offsetZ;
  relativeCoordVector = relativeCoordZ;

  typFuerSlice = noWithZTyp;

  double rX = xL * 0.5;
  double rY = yL * 0.5;
  
  int num_blocks = relativeCoordZ.size() -1;
  assert(num_blocks>0);
  int num_blocks_total = 2* ( relativeCoordZ.size()-1 );
  int num_points_per_face = 6;
//   int num_faces = num_blocks +1;
  int numPoints = num_points_per_face*(num_blocks+1) ;

  Set_number_points ( numPoints );
  Set_number_hexahedra ( num_blocks_total );


  for ( int i = 0; i <= num_blocks; ++i ) {
    Set_coordinate_point ( i*num_points_per_face+0, D3vector ( offsetX-rX,offsetY-rY,offsetZ+relativeCoordZ.at ( i ) ) );
    Set_coordinate_point ( i*num_points_per_face+1, D3vector ( offsetX+rX,offsetY-rY,offsetZ+relativeCoordZ.at ( i ) ) );

    Set_coordinate_point ( i*num_points_per_face+2, D3vector ( offsetX+rX,offsetY-rY+yS,offsetZ+relativeCoordZ.at ( i ) ) );
    Set_coordinate_point ( i*num_points_per_face+3, D3vector ( offsetX-rX+xS,offsetY+rY,offsetZ+relativeCoordZ.at ( i ) ) );

    Set_coordinate_point ( i*num_points_per_face+4, D3vector ( offsetX-rX,offsetY+rY,offsetZ+relativeCoordZ.at ( i ) ) );
    Set_coordinate_point ( i*num_points_per_face+5, D3vector ( offsetX-rX,offsetY-rY+yH,offsetZ+relativeCoordZ.at ( i ) ) );
 }

  for ( int i = 0; i < num_blocks; ++i ) {
    Set_hexahedron ( i, i*num_points_per_face + 0, i*num_points_per_face + 1,
                        i*num_points_per_face + 5, i*num_points_per_face + 2,
                        ( i+1 ) * num_points_per_face + 0, ( i+1 ) * num_points_per_face +1,  
                        ( i+1 ) * num_points_per_face + 5, ( i+1 ) * num_points_per_face +2 );
  }
  
  
  for ( int i = 0; i < num_blocks; ++i ) {
    Set_hexahedron ( i+num_blocks, i*num_points_per_face + 5, i*num_points_per_face + 2,
                                   i*num_points_per_face + 4, i*num_points_per_face + 3,
                        ( i+1 ) * num_points_per_face + 5, ( i+1 ) * num_points_per_face +2,  
                        ( i+1 ) * num_points_per_face + 4, ( i+1 ) * num_points_per_face +3 );
  } 


  construction_hexahedron_points_done();

  construction_done();
}

int SpecialPrismList::getNz() const {
  return length.size()-1;
}


Sphere::Sphere(double radius)
{
    size_pointer_global_data = 1;
    pointer_global_data = new double[size_pointer_global_data];
    pointer_global_data[0] = radius;
    constructionParameters.resize(1);
    constructionParameters[0] = radius;

    Set_number_points ( 8 ); // 8 an den Ecken

    Set_coordinate_point ( WSDdir3D, D3vector ( -radius,-radius,-radius ) );
    Set_coordinate_point ( ESDdir3D, D3vector ( radius,-radius,-radius ) );
    Set_coordinate_point ( WNDdir3D, D3vector ( -radius,radius,-radius ) );
    Set_coordinate_point ( ENDdir3D, D3vector ( radius,radius,-radius ) );
    Set_coordinate_point ( WSTdir3D, D3vector ( -radius,-radius,radius ) );
    Set_coordinate_point ( ESTdir3D, D3vector ( radius,-radius,radius ) );
    Set_coordinate_point ( WNTdir3D, D3vector ( -radius,radius,radius ) );
    Set_coordinate_point ( ENTdir3D, D3vector ( radius,radius,radius ) );

    Set_number_hexahedra ( 1 );
    Set_hexahedron ( 0, WSDdir3D, ESDdir3D, WNDdir3D, ENDdir3D, WSTdir3D, ESTdir3D, WNTdir3D, ENTdir3D );

    construction_hexahedron_points_done();

    // transformations of the edges and faces

    //Set_transformation_face(0,1,2,3,transform_quadrangle_sphere_D);

    Set_transformation_face(0,1,2,3,transform_quadrangle_sphere_D);
    Set_transformation_face(4,5,6,7,transform_quadrangle_sphere_T);  //transform_quadrangle_test
    Set_transformation_face(0,2,4,6,transform_quadrangle_sphere_W);
    Set_transformation_face(1,3,5,7,transform_quadrangle_sphere_E);
    Set_transformation_face(0,1,4,5,transform_quadrangle_sphere_S);
    Set_transformation_face(2,3,6,7,transform_quadrangle_sphere_N);
//    Set_transformation_face(0,1,2,3,transform_zero);
//    Set_transformation_face(4,5,6,7,transform_zero);  //transform_quadrangle_test
//    Set_transformation_face(0,2,4,6,transform_zero);
//    Set_transformation_face(1,3,5,7,transform_zero);
//    Set_transformation_face(0,1,4,5,transform_zero);
//    Set_transformation_face(2,3,6,7,transform_zero);
//    Set_transformation_edge();


    construction_done();

}
// facetest
//face_transform_test::face_transform_test(D3vector (*Transform)(double, double, double *))
//{
//    transform = Transform;
//}
//D3vector face_transform_test::transformEast( double t1, double* pointer_global_data )
//{
//   return transform (t1,0,pointer_global_data);
//}

//D3vector face_transform_test::transformWest( double t1, double* pointer_global_data )
//{
//   return transform (t1,1,pointer_global_data);
//}

//D3vector face_transform_test::transformNorth( double t1, double* pointer_global_data )
//{
//   return transform (0,t1,pointer_global_data);
//}

//D3vector face_transform_test::transformSouth( double t1, double* pointer_global_data )
//{
//   return transform (1,t1,pointer_global_data);
//}

