/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "elements.h"
#include "parti.h"
#include "ug.h"


Partitioning::Partitioning()  {
 // 1. start
  my_rank = 0;
  max_rank_used = 1;
  p = 1;
  comm = 0;
  partitioning_done = false;
}

void Unstructured_grid::Make_Partitioning(MPI_Comm com) {
  if(partitioning_done) 
    cout << " Apply \n "
	 << "        Unstructured_grid::Make_Partitioning(MPI_Comm com)"
	 << "\n only one time! " << endl;

  partitioning_done = true;

  // 1. start
  comm = com;
  
  MPI_Comm_rank(comm, &my_rank);
  MPI_Comm_size(comm, &p);
  
  if(Give_number_hexahedra() < p)
    max_rank_used = Give_number_hexahedra();
  else
    max_rank_used = p;

  // hexahedron
  for(int id=0;id<Give_number_hexahedra();++id) {
    Give_hexahedron(id)->set_rank(id % p);
  }

  Quadrangle_el *quad;
  int id_hex;
  // quadrangle
  for(int id=0;id<Give_number_quadrangles();++id) {
    quad   = Give_quadrangle(id);
    id_hex = quad->Give_id_interior_vol();
    quad->set_rank(id_hex % p);

    for(int c=0;c<4;++c) {
      // edge
      Give_edge(quad->Give_id_edge((dir2D)c))->set_rank(id_hex % p);

      // corner
      Give_point(quad->Give_id_corner((dir2D_sons)c))->set_rank(id_hex % p);
    }
  }
}

