/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/


#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "elements.h"
#include "parti.h"
#include "ug.h"

void Unstructured_grid::Print_AVS(ofstream *Datei) {
  D3vector v;

  //  Datei->precision(3);
  //  Datei->setf(std::ios::fixed,std::ios::floatfield);

  // Teil 0: Write information
  *Datei << "# File created by 3DUGBlock " << endl;
  *Datei << "# UCD file format for AVS " << endl;
  *Datei << "# ug " << endl;

  // Teil 1: Kopfzeile schreiben
  *Datei << num_points << " "
         << num_hexahedra << " 1 0 0" << endl;

  // Teil 2: Koordinaten der Punkte ausgeben
  for(int id=0;id < num_points;++id) {
    v = points[id].Give_coordinate();
    *Datei << id
	   << "  " << v.x
	   << " " << v.y
	   << " " << v.z
	   << "\n";
  }

  // Teil 3: Zellen ausgeben
  for(int id=0;id < num_hexahedra;++id) {
    *Datei << id << "  1  hex  "
	   << hexahedra[id].Give_id_corner(WNTdir3D)  << " "
	   << hexahedra[id].Give_id_corner(WSTdir3D)  << " "
	   << hexahedra[id].Give_id_corner(ESTdir3D)  << " "
	   << hexahedra[id].Give_id_corner(ENTdir3D)  << " "
	   << hexahedra[id].Give_id_corner(WNDdir3D)  << " "
	   << hexahedra[id].Give_id_corner(WSDdir3D)  << " "
	   << hexahedra[id].Give_id_corner(ESDdir3D)  << " "
	   << hexahedra[id].Give_id_corner(ENDdir3D)  << " "
	   << "\n";
  }

  // Teil 4:
  *Datei << "1 1" << "\n";

  // Teil 5:
  *Datei << "variable " << "\n";

  // Teil 6: Ausgabe der id der Points
  for(int id=0;id < num_points;++id) {
    v = points[id].Give_coordinate();
    *Datei << id << "  " << id
	   << "\n";
  }
}
