/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/


#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "elements.h"
#include "parti.h"
#include "ug.h"

void Unstructured_grid::Print_Dx(ofstream *Datei) {
  D3vector v;
  int id_WSD, id_ESD, id_WND, id_END;
  int id_WST, id_EST, id_WNT, id_ENT;

  //  Datei->precision(3);
  //  Datei->setf(std::ios::fixed,std::ios::floatfield);

  // Teil 0: Write information
  *Datei << "# File created by 3DUGBlock " << endl;
  *Datei << "# dx file format for Opendx " << endl;
  *Datei << "# ug " << endl;

  // Teil 1: Kopfzeile schreiben
  *Datei << "object 1 class array type float rank 1 shape 3 items "
           << num_points << " data follows" << endl;

  // Teil 2: Koordinaten der Punkte ausgeben
  for(int id=0;id < num_points;++id) {
    v = points[id].Give_coordinate();
    *Datei << "  " << v.x
	   << " " << v.y
	   << " " << v.z
	   << "\n";
  }

  // Teil 3: Zellen ausgeben
  *Datei << "object 2 class array type int rank 1 shape 4 items "
           << 6*num_hexahedra << " data follows" << endl;
  for(int id=0;id < num_hexahedra;++id) {
    id_WSD = hexahedra[id].Give_id_corner(WSDdir3D);
    id_ESD = hexahedra[id].Give_id_corner(ESDdir3D);
    id_WND = hexahedra[id].Give_id_corner(WNDdir3D);
    id_END = hexahedra[id].Give_id_corner(ENDdir3D);
    id_WST = hexahedra[id].Give_id_corner(WSTdir3D);
    id_EST = hexahedra[id].Give_id_corner(ESTdir3D);
    id_WNT = hexahedra[id].Give_id_corner(WNTdir3D);
    id_ENT = hexahedra[id].Give_id_corner(ENTdir3D);

    // 0. Tet: WND, WNT, WST, EST
    *Datei << id_WNT  << "  ";
    *Datei << id_WND  << "  ";
    *Datei << id_WST  << "  ";
    *Datei << id_EST  << "  ";
    *Datei << "\n";

    // 1. Tet: EST, WND, WST, ESD
    *Datei << id_EST  << "  ";
    *Datei << id_WND  << "  ";
    *Datei << id_WST  << "  ";
    *Datei << id_ESD  << "  ";
    *Datei << "\n";

    // 2. Tet: WND, WSD, WST, ESD
    *Datei << id_WND  << "  ";
    *Datei << id_WSD  << "  ";
    *Datei << id_WST  << "  ";
    *Datei << id_ESD  << "  ";
    *Datei << "\n";

    // 3. Tet: EST, WND, ESD, END
    *Datei << id_EST  << "  ";
    *Datei << id_WND  << "  ";
    *Datei << id_ESD  << "  ";
    *Datei << id_END  << "  ";
    *Datei << "\n";

    // 4. Tet: ENT, WNT, EST, END
    *Datei << id_ENT  << "  ";
    *Datei << id_WNT  << "  ";
    *Datei << id_EST  << "  ";
    *Datei << id_END  << "  ";
    *Datei << "\n";

    // 5. Tet: WNT, WND, EST, END
    *Datei << id_WNT  << "  ";
    *Datei << id_WND  << "  ";
    *Datei << id_EST  << "  ";
    *Datei << id_END  << "  ";
    *Datei << "\n";
  }

  // Teil 4:
  *Datei << "attribute " << '"' << "element type"
	 << '"' << " string " << '"' << "tetrahedra" 
	 << '"' << "" << endl;
  *Datei << "attribute " << '"' << "ref" << '"'
	 << " string " << '"' << "positions" 
	 << '"' << "" << endl;

  // Teil 5: Ausgabe der id der Points
  *Datei << "object 3 class array type float rank 0 items "
           << num_points << " data follows" << endl;
  for(int id=0;id < num_points;++id) {
    v = points[id].Give_coordinate();
    *Datei << id << "\n";
  }


  // Teil 6: end variable
  *Datei << "attribute "
	 << '"' << "dep"
	 << '"' << " string " << '"' << "positions" << '"' << "" << endl;

  // Teil 7: end remarks
  *Datei << "object " << '"'
	 << "irregular positions irregular connections"
	 << '"' << " class field" << endl;
  *Datei << "component " << '"' << "positions" << '"' << " value 1" << endl;
  *Datei << "component "
	 << '"' << "connections" << '"' << " value 2" << endl;
  *Datei << "component " << '"' << "data" << '"' << " value 3" << endl;
  *Datei << "end" << endl; 
}
