/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "elements.h"
#include "parti.h"
#include "ug.h"
#include "examples_ug.h"



PyramidList::PyramidList(double lxLeft, double lyLeft,
	                 double lxRight, double lyRight,  
			 std::vector<double>& relativeCoordZ,												 
		         double offsetX, double offsetY, double offsetZ) {
	const int numBlocks(relativeCoordZ.size()-1);
	assert (numBlocks > 0);
	
	const int numInterfaces(relativeCoordZ.size());
	
	const int numpoints(numInterfaces *4);
	Set_number_points ( numpoints );

	// set points
	for ( int i = 0; i < numInterfaces; ++i )	{
	        double rx = lxLeft*0.5 + (lxRight-lxLeft) * 0.5 *  relativeCoordZ.at(i) / relativeCoordZ.at(relativeCoordZ.size()-1);
	        double ry = lyLeft*0.5 + (lyRight-lyLeft) * 0.5 *  relativeCoordZ.at(i) / relativeCoordZ.at(relativeCoordZ.size()-1);
		
		Set_coordinate_point ( i*4+0,  D3vector ( offsetX-rx, offsetY-ry, offsetZ+relativeCoordZ.at ( i ) ) );
		Set_coordinate_point ( i*4+1,  D3vector ( offsetX+rx, offsetY-ry, offsetZ+relativeCoordZ.at ( i ) ) );
		Set_coordinate_point ( i*4+2,  D3vector ( offsetX-rx, offsetY+ry, offsetZ+relativeCoordZ.at ( i ) ) );
		Set_coordinate_point ( i*4+3,  D3vector ( offsetX+rx, offsetY+ry, offsetZ+relativeCoordZ.at ( i ) ) );
	}
	
	// set hexahedra
	Set_number_hexahedra ( numBlocks );
  
	// alle bis auf letzten block
	for (int i = 0; i < numBlocks; ++i ) {
		Set_hexahedron ( i, i*4 + 0, i*4 + 1, i*4 + 2, i*4 + 3,  i*4 + 4, i*4 + 5, i*4 + 6, i*4 + 7 );
	}


	construction_hexahedron_points_done();
	construction_done();
}

