/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
#include "../../../Source_Colsamm/Colsamm.h"
using namespace ::_COLSAMM_;

#include "interpolCellVar.h"
#include "../extemp/variable_cc.h"
#include "../pde_op/diffop_cc.h"


#include "assert.h"


FastInterpolate_from_cell_to_point::FastInterpolate_from_cell_to_point() {
     weight = NULL;
}

void FastInterpolate_from_cell_to_point::deleteData() {
   if(weight!=NULL) delete weight;
   weight = NULL;
}

void FastInterpolate_from_cell_to_point::init(Blockgrid* blockgrid) {
     deleteData();
     weight = new Variable<double>(*blockgrid);
     weight->SetAllZero();
     weight->AddFromOne();
     weight->AddUpFromHex();
     weight->UpdateHexahedra();
     
/*     
    std::ofstream DATEIA;

    DATEIA.open("weight.vtk");
    weight->Print_VTK(DATEIA);
    DATEIA.close();     
     
     cout << " Maximum: " << Maximum(*weight) << " Minimum: " << Minimum(*weight) << endl;
     */
}

void FastInterpolate_from_cell_to_point::doInterpolation(Cell_variable<double>* varCell, Variable<double>* varPoint) {
     assert(weight!=NULL);
     varPoint->SetAllZero();
     varPoint->AddFromZell(*varCell);
     varPoint->AddUpFromHex();
     
     (*varPoint) = (*varPoint) / (*weight);
     varPoint->UpdateHexahedra();     
     
     
     
}


/////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////


template<>
void Interpolate_from_cell_to_point<Stencil_variable>::init(Blockgrid* blockgrid_) {
     if(blockgrid_ != blockgrid) {
        deleteData();
        blockgrid = blockgrid_;
        HelmSten  = new Sten_matrix<double, Stencil_variable>(*blockgrid);
        Helm      = new Local_stiffness_matrix<double>(*blockgrid);
        sp        = new Local_stiffness_matrix<double>(*blockgrid);	
        f         = new Variable<double>(*blockgrid);
	
	Helm->Calculate(v_() * w_());
	(*HelmSten) = (*Helm);
	
	
        (*f) = 1.0;
        (*d) = (*Helm)(*f);
		
     }
}
   
template<>
void Interpolate_from_cell_to_point<Stencil_constant_z>::init(Blockgrid* blockgrid_) {
     if(blockgrid_ != blockgrid) {
        deleteData();
        blockgrid = blockgrid_;
        HelmSten  = new Sten_matrix<double, Stencil_constant_z>(*blockgrid);
        Helm      = new Local_stiffness_matrix<double>(*blockgrid);
        sp        = new Local_stiffness_matrix<double>(*blockgrid);	
        f         = new Variable<double>(*blockgrid);
        d         = new Variable<double>(*blockgrid);
	
	Helm->Calculate(v_() * w_());
	(*HelmSten) = (*Helm);
	
        (*f) = 1.0;
        (*d) = (*Helm)(*f);
	
	
     }
}

