/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/


#ifndef MG_ARRAY_H_
#define MG_ARRAY_H_


template <class T>
class Set_MG_array;



// apply this->

template <class T>
class MG_array {
 public:
   MG_array(int l);
   template<class U> MG_array(MG_array<U> org)       { assert(false); }
   ~MG_array();
   Set_MG_array<T> operator() (int i);
   T& operator[] (int i);
   T* &get(int i);
 private:
   int L;
   bool* set;
   T** MG_obj;
};


//////////////////////////////////////////////

template <class T>
class Set_MG_array {
   friend class MG_array<T>;
 public:
   void operator=(T* ob) { MG_obj[i] = ob; };
 private:
   Set_MG_array(int ii, T** mg_obj) :  MG_obj(mg_obj),  i(ii)  {};
   T** MG_obj;
   int i;
};


template <class T>
MG_array<T>::MG_array(int l) {
  assert(l>=0);
  L = l;
  MG_obj = new T*[l];
  set    = new bool[l];
  for(int i=0;i<l;++i) {
    set[i] = false;
  }
}

template<class T>
MG_array<T>::~MG_array(){
    for(int i=0;i<L;++i)
    {
      delete this->get(i);
    }
    delete[] MG_obj;
    delete[] set;
}

template <class T>
T* &MG_array<T>::get(int i) {
  assert(i>=0 && i<L);
  assert(set[i]);
  return (MG_obj[i]);
}

template <class T>
T& MG_array<T>::operator[] (int i) {
  assert(i>=0 && i<L);
  assert(set[i]);
  return (*MG_obj[i]);
}

template <class T>
Set_MG_array<T> MG_array<T>::operator() (int i) {
  assert(i>=0 && i<L);
  set[i] = true;

  return Set_MG_array<T>(i,MG_obj);
}



#endif // MG_ARRAY_H_

