/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

// ------------------------------------------------------------
//
// mgop.h
//
// ------------------------------------------------------------

#ifndef MGOP_H
#define MGOP_H

void Print_Loc_stencil ( double* sten );

enum coarsening_direction { coarse_xyz,
                            coarse_xy,  coarse_xz,  coarse_yz,
                            coarse_x,   coarse_y,   coarse_z,
                            no_coarse
                          };
                          
                          
#define hexWSD  v.Give_data<hexahedronEl>(id,i-1,j-1,k-1,Nxf,Nyf)
#define hexSD  v.Give_data<hexahedronEl>(id,i,j-1,k-1,Nxf,Nyf)
#define hexESD  v.Give_data<hexahedronEl>(id,i+1,j-1,k-1,Nxf,Nyf)
                          
#define hexWD  v.Give_data<hexahedronEl>(id,i-1,j,k-1,Nxf,Nyf)
#define hexD  v.Give_data<hexahedronEl>(id,i,j,k-1,Nxf,Nyf)
#define hexED  v.Give_data<hexahedronEl>(id,i+1,j,k-1,Nxf,Nyf)
                          
#define hexWND  v.Give_data<hexahedronEl>(id,i-1,j+1,k-1,Nxf,Nyf)
#define hexND  v.Give_data<hexahedronEl>(id,i,j+1,k-1,Nxf,Nyf)
#define hexEND  v.Give_data<hexahedronEl>(id,i+1,j+1,k-1,Nxf,Nyf)
                          
                          
#define hexWS  v.Give_data<hexahedronEl>(id,i-1,j-1,k,Nxf,Nyf)
#define hexS  v.Give_data<hexahedronEl>(id,i,j-1,k,Nxf,Nyf)
#define hexES  v.Give_data<hexahedronEl>(id,i+1,j-1,k,Nxf,Nyf)
                          
#define hexW  v.Give_data<hexahedronEl>(id,i-1,j,k,Nxf,Nyf)
#define hexM  v.Give_data<hexahedronEl>(id,i,j,k,Nxf,Nyf)
#define hexE  v.Give_data<hexahedronEl>(id,i+1,j,k,Nxf,Nyf)
                          
#define hexWN  v.Give_data<hexahedronEl>(id,i-1,j+1,k,Nxf,Nyf)
#define hexN  v.Give_data<hexahedronEl>(id,i,j+1,k,Nxf,Nyf)
#define hexEN  v.Give_data<hexahedronEl>(id,i+1,j+1,k,Nxf,Nyf)
                          
                          
#define hexWST  v.Give_data<hexahedronEl>(id,i-1,j-1,k+1,Nxf,Nyf)
#define hexST  v.Give_data<hexahedronEl>(id,i,j-1,k+1,Nxf,Nyf)
#define hexEST  v.Give_data<hexahedronEl>(id,i+1,j-1,k+1,Nxf,Nyf)
                          
#define hexWT  v.Give_data<hexahedronEl>(id,i-1,j,k+1,Nxf,Nyf)
#define hexT  v.Give_data<hexahedronEl>(id,i,j,k+1,Nxf,Nyf)
#define hexET  v.Give_data<hexahedronEl>(id,i+1,j,k+1,Nxf,Nyf)
                          
#define hexWNT  v.Give_data<hexahedronEl>(id,i-1,j+1,k+1,Nxf,Nyf)
#define hexNT  v.Give_data<hexahedronEl>(id,i,j+1,k+1,Nxf,Nyf)
#define hexENT  v.Give_data<hexahedronEl>(id,i+1,j+1,k+1,Nxf,Nyf)
                          
                          
/////

#define quadWS  v.Give_data_quadrangle_middle(id,i-1,j-1,Nxf)
#define quadS   v.Give_data_quadrangle_middle(id,i,  j-1,Nxf)
#define quadES  v.Give_data_quadrangle_middle(id,i+1,j-1,Nxf)

#define quadW  v.Give_data_quadrangle_middle(id,i-1,j,Nxf)
#define quadM  v.Give_data_quadrangle_middle(id,i,  j,Nxf)
#define quadE  v.Give_data_quadrangle_middle(id,i+1,j,Nxf)

#define quadWN  v.Give_data_quadrangle_middle(id,i-1,j+1,Nxf)
#define quadN   v.Give_data_quadrangle_middle(id,i,  j+1,Nxf)
#define quadEN  v.Give_data_quadrangle_middle(id,i+1,j+1,Nxf)

#define quad_ex_WN  v.Give_data_quadrangle_exterior(id,i-1,j+1,Nxf)
#define quad_ex_N   v.Give_data_quadrangle_exterior(id,i,  j+1,Nxf)
#define quad_ex_EN  v.Give_data_quadrangle_exterior(id,i+1,j+1,Nxf)

#define quad_ex_WS  v.Give_data_quadrangle_exterior(id,i-1,j-1,Nxf)
#define quad_ex_S   v.Give_data_quadrangle_exterior(id,i,  j-1,Nxf)
#define quad_ex_ES  v.Give_data_quadrangle_exterior(id,i+1,j-1,Nxf)

#define quad_ex_W  v.Give_data_quadrangle_exterior(id,i-1,j,Nxf)
#define quad_ex_M  v.Give_data_quadrangle_exterior(id,i,  j,Nxf)
#define quad_ex_E  v.Give_data_quadrangle_exterior(id,i+1,j,Nxf)


#define quad_in_WN  v.Give_data_quadrangle_interior(id,i-1,j+1,Nxf)
#define quad_in_N   v.Give_data_quadrangle_interior(id,i,  j+1,Nxf)
#define quad_in_EN  v.Give_data_quadrangle_interior(id,i+1,j+1,Nxf)

#define quad_in_WS  v.Give_data_quadrangle_interior(id,i-1,j-1,Nxf)
#define quad_in_S   v.Give_data_quadrangle_interior(id,i,  j-1,Nxf)
#define quad_in_ES  v.Give_data_quadrangle_interior(id,i+1,j-1,Nxf)

#define quad_in_W  v.Give_data_quadrangle_interior(id,i-1,j,Nxf)
#define quad_in_M  v.Give_data_quadrangle_interior(id,i,  j,Nxf)
#define quad_in_E  v.Give_data_quadrangle_interior(id,i+1,j,Nxf)







///////////////////////////////////////////////////////////////
// Restriction

class Multigrid_operator;

template <typename TYPE2>

class ExpRes_local_stiffness_matrix {
	public:
		ExpRes_local_stiffness_matrix ( const Local_stiffness_matrix<TYPE2>& lm_,
		                                Multigrid_operator* mgop_ )
				: lm ( lm_ ), mgop ( mgop_ ) {};
				
		const Local_stiffness_matrix<TYPE2>& lm;
		Multigrid_operator* mgop;
};

template <class DTyp>

class ExpressionRestriction {
	public:
		ExpressionRestriction ( const Variable<DTyp>& v_, Multigrid_operator* mgop_ )
				: v ( v_ ), mgop ( mgop_ ) {};
				
		const Variable<DTyp>& v;
		Multigrid_operator* mgop;
};

template <class DTyp>

class ExpressionProlongation {
	public:
		ExpressionProlongation ( const Variable<DTyp>& v_, Multigrid_operator* mgop_ )
				: v ( v_ ), mgop ( mgop_ ) {};
				
		const Variable<DTyp>& v;
		Multigrid_operator* mgop;
};

class Multigrid_operator {
	public:
		Multigrid_operator ( Blockgrid& coarse_grid_, Blockgrid& fine_grid_ );
		~Multigrid_operator();
		
		template <typename TYPE2>
		ExpRes_local_stiffness_matrix<TYPE2> Restrict ( const Local_stiffness_matrix<TYPE2>& lm );
		
		template <class DTyp>
		ExpressionRestriction<DTyp>  Restrict ( const Variable<DTyp>& v );
		
		template <class DTyp>
		ExpressionProlongation<DTyp> Prolongate ( const Variable<DTyp>& v );
		
		coarsening_direction Give_coarse_hex ( int id )    { return coarse_hexahedra[id]; }
		
		coarsening_direction Give_coarse_quad ( int id )   { return coarse_quadrangles[id]; }
		
		bool Give_coarse_quad_exterior ( int id )   { return coarse_quadrangles_exterior[id]; }
		
		bool Give_coarse_quad_interior ( int id )   { return coarse_quadrangles_interior[id]; }
		
		bool    Give_coarse_edge ( int id )         { return coarse_edges[id]; }
		
		double* Give_coarse_weight_edges ( int id ) { return coarse_weight_edges[id]; }
		
		Blockgrid* Give_coarse_grid() { return coarse_grid; }
		
		Blockgrid* Give_fine_grid()   { return fine_grid; }
		
		int* Give_coarse_grid_functions3D() { return coarse_grid_functions3D; }
		
		int* Give_coarse_grid_functions2D() { return coarse_grid_functions2D; }
		
	private:
		bool *coarse;    // coarse[i] means coarsening in direction i
		
		coarsening_direction* coarse_hexahedra;        // num_hexahedra
		
		coarsening_direction* coarse_quadrangles;      // num_quadrangles
		bool* coarse_quadrangles_interior;             // num_quadrangles
		bool* coarse_quadrangles_exterior;             // num_quadrangles
		
		bool*     coarse_edges;                        // num_edges
		double**  coarse_weight_edges;                 // num_edges
		
		Blockgrid* coarse_grid;
		Blockgrid* fine_grid;
		// for coarsening: coarse grid test functions
		int *coarse_grid_functions3D;
		int *coarse_grid_functions2D;
		
		// for parallel
		int my_rank;
		MPI_Comm comm;
};


template <typename TYPE2>
ExpRes_local_stiffness_matrix<TYPE2> Multigrid_operator::Restrict (
    const Local_stiffness_matrix<TYPE2>& lm ) {
	if ( lm.Give_blockgrid() != fine_grid )
		std::cout << " Error in Restriction lm, wrong fine grid! " << std::endl;
		
	return ExpRes_local_stiffness_matrix<TYPE2> ( lm, this );
}

template <class DTyp>
ExpressionRestriction<DTyp> Multigrid_operator::Restrict ( const Variable<DTyp>& v ) {
	if ( v.Give_blockgrid() != fine_grid )
		std::cout << " Error in Restriction v, wrong fine grid! " << std::endl;
		
	return ExpressionRestriction<DTyp> ( v, this );
}

template <class DTyp>
ExpressionProlongation<DTyp> Multigrid_operator::Prolongate ( const Variable<DTyp>& v ) {
	if ( v.Give_blockgrid() != coarse_grid )
		std::cout << " Error in Prolongation v, wrong coarse grid! " << std::endl;
		
	return ExpressionProlongation<DTyp> ( v, this );
}



///////////////////////////////////////////////////////////////
// Restriction stiffness matrix


template <typename TYPE2>
void Local_stiffness_matrix<TYPE2>::operator= ( const
        ExpRes_local_stiffness_matrix<TYPE2>& Expr_res ) {
	int id;
	int Nx, Ny, Nz, N_total;
	int Nx_fine, Ny_fine;
	
	Unstructured_grid* ug = blockgrid->Give_unstructured_grid();
	
	Multigrid_operator* mgop;
	coarsening_direction coarse_hexahedron;
	
	int *coarse_grid_functions3D;
	int *coarse_grid_functions2D;
	
	TYPE2** loc_fine;
	
	loc_fine = Expr_res.lm.loc_m_hexahedra;
	
	mgop = Expr_res.mgop;
	
	coarse_grid_functions3D = mgop->Give_coarse_grid_functions3D();
	coarse_grid_functions2D = mgop->Give_coarse_grid_functions2D();
	
	for ( id = 0;id < blockgrid->Give_unstructured_grid()->Give_number_hexahedra();++id ) {
	
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) ) {
		
			Nx = blockgrid->Give_Nx_hexahedron ( id );
			Ny = blockgrid->Give_Ny_hexahedron ( id );
			Nz = blockgrid->Give_Nz_hexahedron ( id );
			N_total = Nx * Ny * Nz * 64;
			
			coarse_hexahedron = mgop->Give_coarse_hex ( id );
			
			if ( coarse_hexahedron == coarse_xyz ) {
				Nx_fine = Nx * 2;
				Ny_fine = Ny * 2;
				
                                #pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)				
				for(int k = 0;k < Nz;++k ) {
				    TYPE2* Stencil_coarse;
	                            TYPE2* Stencils_fine[8];
				    
				    for(int j = 0;j < Ny;++j )
					for(int i = 0;i < Nx;++i ) {	
					    Stencil_coarse = &loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64];
							
					    Stencils_fine[WSDdir3D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i,  2*j,  2*k  ) *64];
					    Stencils_fine[ESDdir3D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i+1,2*j,  2*k  ) *64];
					    Stencils_fine[WNDdir3D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i,  2*j+1,2*k  ) *64];
					    Stencils_fine[ENDdir3D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i+1,2*j+1,2*k  ) *64];
					    Stencils_fine[WSTdir3D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i,  2*j,  2*k+1 ) *64];
					    Stencils_fine[ESTdir3D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i+1,2*j,  2*k+1 ) *64];
					    Stencils_fine[WNTdir3D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i,  2*j+1,2*k+1 ) *64];
					    Stencils_fine[ENTdir3D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i+1,2*j+1,2*k+1 ) *64];
							    
							    
					    Restrict_local_stiffness_matrix_xyz(Stencil_coarse,
							    Stencils_fine, coarse_grid_functions3D );
					}
				}
			}			
			
			if ( coarse_hexahedron == coarse_yz ) {
				Nx_fine = Nx;
				Ny_fine = Ny * 2;
				
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
				for(int k = 0;k < Nz;++k ) {
		                    TYPE2* Stencil_coarse;
	                            TYPE2* Stencils_fine[8];					  
				    for(int j = 0;j < Ny;++j )
						for(int i = 0;i < Nx;++i ) {
							Stencil_coarse = &loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64];
							
							Stencils_fine[SWdir2D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( i,2*j,  2*k ) *64];
							Stencils_fine[SEdir2D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( i,2*j+1,2*k ) *64];
							Stencils_fine[NWdir2D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( i,2*j,  2*k+1 ) *64];
							Stencils_fine[NEdir2D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( i,2*j+1,2*k+1 ) *64];
							    
							    
							Restrict_local_stiffness_matrix_2D<TYPE2, Index_calc_x > ( Stencil_coarse,
							        Stencils_fine,
							        coarse_grid_functions2D );
						}
				}
			}
			
			
			if ( coarse_hexahedron == coarse_xz ) {
				Nx_fine = Nx * 2;
				Ny_fine = Ny;
				
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
				for(int k = 0;k < Nz;++k ) {
		                    TYPE2* Stencil_coarse;
	                            TYPE2* Stencils_fine[8];
				    for(int j = 0;j < Ny;++j )
						for(int i = 0;i < Nx;++i ) {
							Stencil_coarse = &loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64];
							
							Stencils_fine[SWdir2D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i,  j, 2*k ) *64];
							Stencils_fine[SEdir2D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i+1,j, 2*k ) *64];
							Stencils_fine[NWdir2D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i,  j, 2*k+1 ) *64];
							Stencils_fine[NEdir2D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i+1,j, 2*k+1 ) *64];
							    
							    
							Restrict_local_stiffness_matrix_2D<TYPE2, Index_calc_y > ( Stencil_coarse,
							        Stencils_fine,
							        coarse_grid_functions2D );
						}
				}
			}
			
			if ( coarse_hexahedron == coarse_xy ) {
				Nx_fine = Nx * 2;
				Ny_fine = Ny * 2;
				
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
				for(int k = 0;k < Nz;++k ) {
		                    TYPE2* Stencil_coarse;
	                            TYPE2* Stencils_fine[8];
				    for(int j = 0;j < Ny;++j )
						for(int i = 0;i < Nx;++i ) {
							Stencil_coarse = &loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64];
							
							Stencils_fine[SWdir2D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i,  2*j,  k ) *64];
							Stencils_fine[SEdir2D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i+1,2*j,  k ) *64];
							Stencils_fine[NWdir2D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i,  2*j+1,k ) *64];
							Stencils_fine[NEdir2D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i+1,2*j+1,k ) *64];
							    
							    
							Restrict_local_stiffness_matrix_2D<TYPE2, Index_calc_z > ( Stencil_coarse,
							        Stencils_fine,
							        coarse_grid_functions2D );
						}
				}
			}
			
			
			
			if ( coarse_hexahedron == coarse_x ) {
				Nx_fine = Nx * 2;
				Ny_fine = Ny;
				
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
				for(int k = 0;k < Nz;++k ) {
		                    TYPE2* Stencil_coarse;
	                            TYPE2* Stencils_fine[8];
				    for(int j = 0;j < Ny;++j )
						for(int i = 0;i < Nx;++i ) {
							Stencil_coarse = &loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64];
							
							Stencils_fine[Ldir1D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i,  j,k ) *64];
							Stencils_fine[Rdir1D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( 2*i+1,j,k ) *64];
							    
							    
							Restrict_local_stiffness_matrix_1D<TYPE2, Index_calc_x > ( Stencil_coarse,
							        Stencils_fine );
						}
				}
			}
			
			if ( coarse_hexahedron == coarse_y ) {
				Nx_fine = Nx;
				Ny_fine = Ny * 2;
				
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
				for(int k = 0;k < Nz;++k ) {
		                    TYPE2* Stencil_coarse;
	                            TYPE2* Stencils_fine[8];
				    for(int j = 0;j < Ny;++j )
						for(int i = 0;i < Nx;++i ) {
							Stencil_coarse = &loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64];
							
							Stencils_fine[Ldir1D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( i,2*j,  k ) *64];
							Stencils_fine[Rdir1D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( i,2*j+1,k ) *64];
							    
							    
							Restrict_local_stiffness_matrix_1D<TYPE2, Index_calc_y > ( Stencil_coarse,
							        Stencils_fine );
						}
				}
			}
			
			if ( coarse_hexahedron == coarse_z ) {
				Nx_fine = Nx;
				Ny_fine = Ny;
				
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
				for(int k = 0;k < Nz;++k ) {
		                    TYPE2* Stencil_coarse;
	                            TYPE2* Stencils_fine[8];
				    for(int j = 0;j < Ny;++j )
						for(int i = 0;i < Nx;++i ) {
							Stencil_coarse = &loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64];
							
							Stencils_fine[Ldir1D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( i,j,2*k  ) *64];
							Stencils_fine[Rdir1D] =
							    &loc_fine[id][Ind_loc_matrix_hexahedra_fine ( i,j,2*k+1 ) *64];
							    
							    
							Restrict_local_stiffness_matrix_1D<TYPE2, Index_calc_z > ( Stencil_coarse,
							        Stencils_fine );
						}
				}
			}
			
			if ( coarse_hexahedron == no_coarse ) {
				Nx_fine = Nx;
				Ny_fine = Ny;
				
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
				for(int k = 0;k < Nz;++k ) {
		                    TYPE2* Stencil_coarse;
	                            TYPE2* Stencils_fine[8];
				    for(int j = 0;j < Ny;++j )
						for(int i = 0;i < Nx;++i )
							for(int l = 0;l < 64;++l ) {
								loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64+l]
								= loc_fine[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64+l];
							}
				}
			}
		}
	}
	
	Update();
}


///////////////////////////////////////////////////////////////
// Restriction


template <class DTyp>
void Variable<DTyp>::operator= ( const ExpressionRestriction<DTyp>& Expr_res ) {
	int Nx, Ny, Nz, Nxf, Nyf, Nzf;
	Multigrid_operator* mgop;
	coarsening_direction coarse_hexahedron;
	coarsening_direction coarse_quad;
	bool coarse_quad_exterior;
	bool coarse_quad_interior;
	
	const Variable<DTyp>& v ( Expr_res.v );
	
	mgop = Expr_res.mgop;
	
	if ( Give_blockgrid() != Expr_res.mgop->Give_coarse_grid() )
		std::cout << " Error in Restriction, wrong coares grid! " << std::endl;
		
	// 1. hexahedra
	for ( int id = 0;id < ug->Give_number_hexahedra();++id ) {
	
		v.Update<hexahedronEl> ( id );
		
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) ) {
		
			Nx = blockgrid->Give_Nx_hexahedron ( id );
			Ny = blockgrid->Give_Ny_hexahedron ( id );
			Nz = blockgrid->Give_Nz_hexahedron ( id );
			
			coarse_hexahedron = mgop->Give_coarse_hex ( id );
			
			//    std::cout << " co" << (int)coarse_hexahedron << std::endl;
			
			if ( coarse_hexahedron == coarse_xyz ) {
				Nxf = Nx * 2;       Nyf = Ny * 2;       Nzf = Nz * 2;
				
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)				
				for ( int k = 2;k < Nzf;k = k + 2 ) {
					for ( int j = 2;j < Nyf;j = j + 2 )
						for ( int i = 2;i < Nxf;i = i + 2 )
							data_hexahedra[id][i/2+ ( Nx+1 ) * ( j/2+ ( Ny+1 ) *k/2 ) ] =
							    0.125 * hexESD + 0.25 * hexSD + 0.125 * hexWSD
							    + 0.25  * hexED  + 0.5  * hexD  + 0.25  * hexWD
							    + 0.125 * hexEND + 0.25 * hexND + 0.125 * hexWND
							    
							    + 0.25  * hexES  + 0.5  * hexS  + 0.25  * hexWS
							    + 0.5   * hexE   +        hexM  + 0.5   * hexW
							    + 0.25  * hexEN  + 0.5  * hexN  + 0.25  * hexWN
							    
							    + 0.125 * hexEST + 0.25 * hexST + 0.125 * hexWST
							    + 0.25  * hexET  + 0.5  * hexT  + 0.25  * hexWT
							    + 0.125 * hexENT + 0.25 * hexNT + 0.125 * hexWNT;
				}
			}
			
			if ( coarse_hexahedron == coarse_xy ) {
				Nxf = Nx * 2;       Nyf = Ny * 2;       Nzf = Nz;
				
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)				
				for ( int k = 1;k < Nz;k = k + 1 ) {
					for ( int j = 2;j < Nyf;j = j + 2 )
						for ( int i = 2;i < Nxf;i = i + 2 )
							data_hexahedra[id][i/2+ ( Nx+1 ) * ( j/2+ ( Ny+1 ) *k ) ] =
							    0.25  * hexES  + 0.5  * hexS  + 0.25  * hexWS
							    + 0.5   * hexE   +        hexM  + 0.5   * hexW
							    + 0.25  * hexEN  + 0.5  * hexN  + 0.25  * hexWN;
				}
			}
			
			if ( coarse_hexahedron == coarse_xz ) {
				Nxf = Nx * 2;       Nyf = Ny;           Nzf = Nz * 2;
				
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
				for ( int k = 2;k < Nzf;k = k + 2 ) {
					for ( int j = 1;j < Nyf;j = j + 1 )
						for ( int i = 2;i < Nxf;i = i + 2 )
							data_hexahedra[id][i/2+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k/2 ) ] =
							    0.25  * hexED  + 0.5  * hexD  + 0.25  * hexWD
							    + 0.5   * hexE   +        hexM  + 0.5   * hexW
							    + 0.25  * hexET  + 0.5  * hexT  + 0.25  * hexWT;
				}
			}
			
			if ( coarse_hexahedron == coarse_yz ) {
				Nxf = Nx;           Nyf = Ny * 2;       Nzf = Nz * 2;
				
				
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)				
				for ( int k = 2;k < Nzf;k = k + 2 ) {
					for ( int j = 2;j < Nyf;j = j + 2 )
						for ( int i = 1;i < Nxf;i = i + 1 )
							data_hexahedra[id][i+ ( Nx+1 ) * ( j/2+ ( Ny+1 ) *k/2 ) ] =
							    0.25  * hexSD  + 0.5  * hexD  + 0.25  * hexND
							    + 0.5   * hexS   +        hexM  + 0.5   * hexN
							    + 0.25  * hexST  + 0.5  * hexT  + 0.25  * hexNT;
				}
			}
			
			if ( coarse_hexahedron == coarse_x ) {
				Nxf = Nx * 2;       Nyf = Ny;       Nzf = Nz;
				
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)				
				for ( int k = 1;k < Nz;k = k + 1 ) {
					for ( int j = 1;j < Nyf;j = j + 1 )
						for ( int i = 2;i < Nxf;i = i + 2 )
							data_hexahedra[id][i/2+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] =
							    0.5   * hexE   +        hexM  + 0.5   * hexW;
				}
			}
			
			if ( coarse_hexahedron == coarse_y ) {
				Nxf = Nx;           Nyf = Ny * 2;       Nzf = Nz;
				
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)				
				for ( int k = 2;k < Nzf;k = k + 2 ) {
					for ( int j = 2;j < Nyf;j = j + 2 )
						for ( int i = 1;i < Nxf;i = i + 1 )
							data_hexahedra[id][i+ ( Nx+1 ) * ( j/2+ ( Ny+1 ) *k ) ] =
							    0.5   * hexS   +        hexM  + 0.5   * hexN;
				}
			}
			
			if ( coarse_hexahedron == coarse_z ) {
				Nxf = Nx;           Nyf = Ny;       Nzf = Nz * 2;
				
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
				for ( int k = 2;k < Nzf;k = k + 2 ) {
					for ( int j = 1;j < Nyf;j = j + 1 )
						for ( int i = 1;i < Nxf;i = i + 1 )
							data_hexahedra[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k/2 ) ] =
							    0.5   * hexD   +        hexM  + 0.5   * hexT;
				}
			}
			
			if ( coarse_hexahedron == no_coarse ) {
				Nxf = Nx;           Nyf = Ny;       Nzf = Nz;
				
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)				
				for ( int k = 1;k < Nz;++k ) {
					for ( int j = 1;j < Ny;++j )
						for ( int i = 1;i < Nx;++i )
							data_hexahedra[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] = hexM;
				}
			}
		}
	}
	
	// 2. quadrangles
	for ( int id = 0;id < ug->Give_number_quadrangles();++id ) {
	
		v.Update<quadrangleEl> ( id );
		
		if ( ug->Give_quadrangle ( id )->my_object ( my_rank ) ) {
		
			// Update quadrangle
			Nx = blockgrid->Give_Nx_quadrangle ( id );
			Ny = blockgrid->Give_Ny_quadrangle ( id );
			
			coarse_quad = mgop->Give_coarse_quad ( id );
			coarse_quad = mgop->Give_coarse_quad ( id );
			
			coarse_quad_exterior = mgop->Give_coarse_quad_exterior ( id );
			coarse_quad_interior = mgop->Give_coarse_quad_interior ( id );
			
			if ( coarse_quad == coarse_xy ) {
				Nxf = Nx * 2;       Nyf = Ny * 2;
				
				for ( int j = 2;j < Nyf;j = j + 2 )
					for ( int i = 2;i < Nxf;i = i + 2 ) {
						data_quadrangles[id][i/2+ ( Nx+1 ) *j/2] =
						    0.25  *   quadES  + 0.5  * quadS  + 0.25  * quadWS
						    + 0.5   * quadE   +        quadM  + 0.5   * quadW
						    + 0.25  * quadEN  + 0.5  * quadN  + 0.25  * quadWN;
						    
						if ( coarse_quad_exterior )
							data_quadrangles[id][i/2+ ( Nx+1 ) *j/2] =
							    data_quadrangles[id][i/2+ ( Nx+1 ) *j/2] +
							    0.5 * (
							        0.25  *   quad_ex_ES  + 0.5  * quad_ex_S  + 0.25  * quad_ex_WS
							        + 0.5   * quad_ex_E   +        quad_ex_M  + 0.5   * quad_ex_W
							        + 0.25  * quad_ex_EN  + 0.5  * quad_ex_N  + 0.25  * quad_ex_WN );
							        
						if ( coarse_quad_interior )
							data_quadrangles[id][i/2+ ( Nx+1 ) *j/2] =
							    data_quadrangles[id][i/2+ ( Nx+1 ) *j/2] +
							    0.5 * (
							        0.25  *   quad_in_ES  + 0.5  * quad_in_S  + 0.25  * quad_in_WS
							        + 0.5   * quad_in_E   +        quad_in_M  + 0.5   * quad_in_W
							        + 0.25  * quad_in_EN  + 0.5  * quad_in_N  + 0.25  * quad_in_WN );
					}
			}
			
			if ( coarse_quad == coarse_x ) {
				Nxf = Nx * 2;       Nyf = Ny;
				
				for ( int j = 1;j < Nyf;j = j + 1 )
					for ( int i = 2;i < Nxf;i = i + 2 ) {
						data_quadrangles[id][i/2+ ( Nx+1 ) *j] =
						    0.5 * quadE + quadM + 0.5 * quadW;
						    
						if ( coarse_quad_exterior )
							data_quadrangles[id][i/2+ ( Nx+1 ) *j] =
							    data_quadrangles[id][i/2+ ( Nx+1 ) *j] +
							    0.5 * ( 0.5 * quad_ex_E + quad_ex_M + 0.5 * quad_ex_W );
							    
						if ( coarse_quad_interior )
							data_quadrangles[id][i/2+ ( Nx+1 ) *j] =
							    data_quadrangles[id][i/2+ ( Nx+1 ) *j] +
							    0.5 * ( 0.5 * quad_in_E + quad_in_M + 0.5 * quad_in_W );
					}
			}
			
			if ( coarse_quad == coarse_y ) {
				Nxf = Nx;       Nyf = Ny * 2;
				
				for ( int j = 2;j < Nyf;j = j + 2 )
					for ( int i = 1;i < Nxf;i = i + 1 ) {
						data_quadrangles[id][i+ ( Nx+1 ) *j/2] =
						    0.5 * quadS + quadM + 0.5 * quadN;
						    
						if ( coarse_quad_exterior )
							data_quadrangles[id][i+ ( Nx+1 ) *j/2] =
							    data_quadrangles[id][i+ ( Nx+1 ) *j/2] +
							    0.5 * ( 0.5 * quad_ex_S + quad_ex_M + 0.5 * quad_ex_N );
							    
						if ( coarse_quad_interior )
							data_quadrangles[id][i+ ( Nx+1 ) *j/2] =
							    data_quadrangles[id][i+ ( Nx+1 ) *j/2] +
							    0.5 * ( 0.5 * quad_in_S + quad_in_M + 0.5 * quad_in_N );
					}
			}
			
			if ( coarse_quad == no_coarse ) {
				Nxf = Nx;       Nyf = Ny;
				
				for ( int j = 1;j < Nyf;j = j + 1 )
					for ( int i = 1;i < Nxf;i = i + 1 ) {
						data_quadrangles[id][i+ ( Nx+1 ) *j] = quadM;
					}
			}
		}
	}
	
	// 3. edges
	int num_pos;
	
	double weight;
	
	for ( int id = 0;id < ug->Give_number_edges();++id ) {
	
		v.Update<edgeEl> ( id );
		
		if ( ug->Give_edge ( id )->my_object ( my_rank ) ) {
		
			// Update edge
			Nx      = blockgrid->Give_Nx_edge ( id );
			num_pos = ug->Give_edge ( id )->Give_number_neighbors();
			
			if ( mgop->Give_coarse_edge ( id ) ) { // coarse edge
				Nxf = Nx * 2;
				
				for ( int i = 1;i < Nx;i = i + 1 )
					data_edges[id][i] = 0.0;
					
				for ( int pos = 0;pos < num_pos;++pos ) {
					weight = mgop->Give_coarse_weight_edges ( id ) [pos];
					
					for ( int i = 2;i < Nxf;i = i + 2 )
						data_edges[id][i/2] = data_edges[id][i/2] +
						                      weight * (
						                          0.5 * v.Give_data_neighbor_edge ( id, pos, i - 1 ) +
						                          0.5 * v.Give_data_neighbor_edge ( id, pos, i + 1 ) +
						                          v.Give_data_neighbor_edge ( id, pos, i )
						                      );
				}
			} else { // no coarse edge
				for ( int i = 1;i < Nx;i = i + 1 )
					data_edges[id][i] = 0.0;
					
				for ( int pos = 0;pos < num_pos;++pos ) {
					weight = mgop->Give_coarse_weight_edges ( id ) [pos];
					
					for ( int i = 1;i < Nx;i = i + 1 )
						data_edges[id][i] = data_edges[id][i] +
						                    v.Give_data_neighbor_edge ( id, pos, i ) * weight;
				}
			}
		}
	}
	
	// 3. points
	for ( int id = 0;id < ug->Give_number_points();++id ) {
		if ( ug->Give_point ( id )->my_object ( my_rank ) )
			data_points[id][0] = v.Give_data<pointEl> ( id, 0  , 0  , 0, 0, 0 );
	}
}



///////////////////////////////////////////////////////////////
// Prolongation



template <class DTyp>
void Variable<DTyp>::operator= ( const ExpressionProlongation<DTyp>& Expr_prol ) {
	int Nx, Ny, Nz;
	Blockgrid* coarse_blockgrid;
	Multigrid_operator* mgop;
	coarsening_direction coarse_hexahedron;
	coarsening_direction coarse_quad;
	int fac_x = 0;
	int fac_y = 0;
	int fac_z = 0;
	
	mgop = Expr_prol.mgop;
	
	coarse_blockgrid = mgop->Give_coarse_grid();
	
	const Variable<DTyp>& v ( Expr_prol.v );
	
	if ( Give_blockgrid() != mgop->Give_fine_grid() )
		std::cout << " Error in Prolongation, wrong fine grid! " << std::endl;
		
	// hexahedra
	for ( int id = 0;id < ug->Give_number_hexahedra();++id ) {
	
		v.Update<hexahedronEl> ( id );
		
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) ) {
		
			Nx = coarse_blockgrid->Give_Nx_hexahedron ( id );
			Ny = coarse_blockgrid->Give_Ny_hexahedron ( id );
			Nz = coarse_blockgrid->Give_Nz_hexahedron ( id );
			
			coarse_hexahedron = mgop->Give_coarse_hex ( id );
			
			//    std::cout << " co" << (int)coarse_hexahedron << std::endl;
			
			if ( coarse_hexahedron == coarse_xyz ) {
				fac_x = 2;    fac_y = 2;    fac_z = 2;
			}
			
			if ( coarse_hexahedron == coarse_yz ) {
				fac_x = 1;    fac_y = 2;    fac_z = 2;
			}
			
			if ( coarse_hexahedron == coarse_xz ) {
				fac_x = 2;    fac_y = 1;    fac_z = 2;
			}
			
			if ( coarse_hexahedron == coarse_xy ) {
				fac_x = 2;    fac_y = 2;    fac_z = 1;
			}
			
			if ( coarse_hexahedron == coarse_x ) {
				fac_x = 2;    fac_y = 1;    fac_z = 1;
			}
			
			if ( coarse_hexahedron == coarse_y ) {
				fac_x = 1;    fac_y = 2;    fac_z = 1;
			}
			
			if ( coarse_hexahedron == coarse_z ) {
				fac_x = 1;    fac_y = 1;    fac_z = 2;
			}
			
			if ( coarse_hexahedron == no_coarse ) {
				fac_x = 1;    fac_y = 1;    fac_z = 1;
			}
			
			if ( coarse_hexahedron == coarse_xyz ) {
				// first volumne
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)			  
				for(int k = 1;k <= Nz;++k ) {
				    for(int j = 1;j <= Ny;++j )
					for(int i = 1;i <= Nx;++i )
					    Ind_data_hexahedra ( 2*i - 1, 2*j - 1, 2*k - 1, 2*Nx, 2*Ny )
							= 0.125 *
							  ( v.Give_data<hexahedronEl> ( id, i  , j  , k  , Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i - 1, j  , k  , Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i  , j - 1, k  , Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i - 1, j - 1, k  , Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i  , j  , k - 1, Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i - 1, j  , k - 1, Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i  , j - 1, k - 1, Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i - 1, j - 1, k - 1, Nx, Ny ) );
				}
			}
			
			if ( coarse_hexahedron == coarse_xyz || coarse_hexahedron == coarse_yz ) {
				// x face
				#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)			  
				for(int k = 1;k <= Nz;++k ) {
				    for(int j = 1;j <= Ny;++j )
					for(int i = 1;i < Nx;++i )
					    Ind_data_hexahedra ( fac_x*i, 2*j - 1, 2*k - 1, fac_x*Nx, fac_y*Ny )
							= 0.25 *
							  ( v.Give_data<hexahedronEl> ( id, i, j  , k  , Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i, j - 1, k  , Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i, j  , k - 1, Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i, j - 1, k - 1, Nx, Ny )  );
				}
			}
			
			if ( coarse_hexahedron == coarse_xyz || coarse_hexahedron == coarse_xz ) {
				// y face
			  	#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP) 
				for(int k = 1;k <= Nz;++k ) {
				    for(int j = 1;j < Ny;++j )
					for(int i = 1;i <= Nx;++i ) {
					    Ind_data_hexahedra ( 2*i - 1, fac_y*j, 2*k - 1, fac_x*Nx, fac_y*Ny )
							= 0.25 *
							  ( v.Give_data<hexahedronEl> ( id, i  , j  , k  , Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i - 1, j  , k  , Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i  , j  , k - 1, Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i - 1, j  , k - 1, Nx, Ny )  );
					}
				}
			}
			
			if ( coarse_hexahedron == coarse_xyz || coarse_hexahedron == coarse_xy ) {
				// z face
			  	#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP) 			  
				for(int k = 1;k < Nz;++k ) {
				    for(int j = 1;j <= Ny;++j )
					for(int i = 1;i <= Nx;++i )
					    Ind_data_hexahedra ( 2*i - 1, 2*j - 1, fac_z*k, fac_x*Nx, fac_y*Ny )
							= 0.25 *
							  ( v.Give_data<hexahedronEl> ( id, i  , j  , k  , Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i - 1, j  , k  , Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i  , j - 1, k  , Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i - 1, j - 1, k  , Nx, Ny )  );
				}
			}
			
			if ( coarse_hexahedron == coarse_xyz ||
			        coarse_hexahedron == coarse_xz  || coarse_hexahedron == coarse_yz ||
			        coarse_hexahedron == coarse_z ) {
				// z edge
			  	#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP) 			  
				for(int k = 1;k <= Nz;++k ) {
				    for(int j = 1;j < Ny;++j )
					for(int i = 1;i < Nx;++i )
					    Ind_data_hexahedra ( fac_x*i, fac_y*j, 2*k - 1, fac_x*Nx, fac_y*Ny )
							= 0.5 *
							  ( v.Give_data<hexahedronEl> ( id, i, j  , k  , Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i, j  , k - 1, Nx, Ny ) );
				}
			}
			
			if ( coarse_hexahedron == coarse_xyz ||
			        coarse_hexahedron == coarse_xy  || coarse_hexahedron == coarse_yz ||
			        coarse_hexahedron == coarse_y ) {
				// y edge
			  	#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP) 			  			  
				for(int k = 1;k < Nz;++k ) {
				    for(int j = 1;j <= Ny;++j )
					for(int i = 1;i < Nx;++i )
					    Ind_data_hexahedra ( fac_x*i, 2*j - 1, fac_z*k, fac_x*Nx, fac_y*Ny )
							= 0.5 *
							  ( v.Give_data<hexahedronEl> ( id, i  , j  , k  , Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i  , j - 1, k  , Nx, Ny ) );
				}
			}
			
			if ( coarse_hexahedron == coarse_xyz ||
			        coarse_hexahedron == coarse_xy  || coarse_hexahedron == coarse_xz ||
			        coarse_hexahedron == coarse_x ) {
				// x edge
			  	#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP) 			  
				for(int k = 1;k < Nz;++k ) {
				    for(int j = 1;j < Ny;++j )
					for(int i = 1;i <= Nx;++i )
					    Ind_data_hexahedra ( 2*i - 1, fac_y*j, fac_z*k, fac_x*Nx, fac_y*Ny )
							= 0.5 *
							  ( v.Give_data<hexahedronEl> ( id, i  , j  , k  , Nx, Ny ) +
							    v.Give_data<hexahedronEl> ( id, i - 1, j  , k  , Nx, Ny ) );
				}
			}
			
			// point
			#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP) 			  			  
			for(int k = 1;k < Nz;++k ) {
			    for(int j = 1;j < Ny;++j )
				for(int i = 1;i < Nx;++i )
				    Ind_data_hexahedra ( fac_x*i, fac_y*j, fac_z*k, fac_x*Nx, fac_y*Ny )
						= v.Give_data<hexahedronEl> ( id, i  , j  , k  , Nx, Ny );
			}
		}
	}
	
	// quadrangle
	for ( int id = 0;id < ug->Give_number_quadrangles();++id ) {
	
		v.Update<quadrangleEl> ( id );
		
		if ( ug->Give_quadrangle ( id )->my_object ( my_rank ) ) {
		
			Nx = coarse_blockgrid->Give_Nx_quadrangle ( id );
			Ny = coarse_blockgrid->Give_Ny_quadrangle ( id );
			
			coarse_quad = mgop->Give_coarse_quad ( id );
			
			//    std::cout << " co" << (int)coarse_quad << std::endl;
			
			if ( coarse_quad == coarse_xy ) {
				fac_x = 2;    fac_y = 2;
			}
			
			if ( coarse_quad == coarse_x ) {
				fac_x = 2;    fac_y = 1;
			}
			
			if ( coarse_quad == coarse_y ) {
				fac_x = 1;    fac_y = 2;
			}
			
			if ( coarse_quad == no_coarse ) {
				fac_x = 1;    fac_y = 1;    fac_z = 1;
			}
			
			if ( coarse_quad == coarse_xy ) {
				for ( int j = 1;j <= Ny;++j )
					for ( int i = 1;i <= Nx;++i )
						Ind_data_quadrangle ( 2*i - 1, 2*j - 1, fac_x*Nx )
						= 0.25 *
						  ( v.Give_data<quadrangleEl> ( id, i  , j  , 0, Nx, 0 ) +
						    v.Give_data<quadrangleEl> ( id, i - 1, j  , 0, Nx, 0 ) +
						    v.Give_data<quadrangleEl> ( id, i  , j - 1, 0, Nx, 0 ) +
						    v.Give_data<quadrangleEl> ( id, i - 1, j - 1, 0, Nx, 0 )  );
			}
			
			if ( coarse_quad == coarse_xy  || coarse_quad == coarse_y ) {
				// y edge
				for ( int j = 1;j <= Ny;++j )
					for ( int i = 1;i < Nx;++i )
						Ind_data_quadrangle ( fac_x*i, 2*j - 1, fac_x*Nx )
						= 0.5 *
						  ( v.Give_data<quadrangleEl> ( id, i  , j  , 0, Nx, 0 ) +
						    v.Give_data<quadrangleEl> ( id, i  , j - 1, 0, Nx, 0 ) );
			}
			
			if ( coarse_quad == coarse_xy  || coarse_quad == coarse_x ) {
				// x edge
				for ( int j = 1;j < Ny;++j )
					for ( int i = 1;i <= Nx;++i )
						Ind_data_quadrangle ( 2*i - 1, fac_y*j, fac_x*Nx )
						= 0.5 *
						  ( v.Give_data<quadrangleEl> ( id, i  , j  , 0, Nx, 0 ) +
						    v.Give_data<quadrangleEl> ( id, i - 1, j  , 0, Nx, 0 ) );
			}
			
			// point
			for ( int j = 1;j < Ny;++j )
				for ( int i = 1;i < Nx;++i )
					Ind_data_quadrangle ( fac_x*i, fac_y*j, fac_x*Nx )
					= v.Give_data<quadrangleEl> ( id, i  , j  , 0, Nx, 0 );
		}
	}
	
	// edge
	for ( int id = 0;id < ug->Give_number_edges();++id ) {
	
		v.Update<edgeEl> ( id );
		
		if ( ug->Give_edge ( id )->my_object ( my_rank ) ) {
		
			Nx = coarse_blockgrid->Give_Nx_edge ( id );
			
			if ( mgop->Give_coarse_edge ( id ) ) { // coarse edge
				// x edge
				for ( int i = 1;i <= Nx;++i )
					data_edges[id][2*i-1]
					= 0.5 *
					  ( v.Give_data<edgeEl> ( id, i  , 0  , 0, 0, 0 ) +
					    v.Give_data<edgeEl> ( id, i - 1, 0  , 0, 0, 0 ) );
					    
				// point
				for ( int i = 1;i < Nx;++i )
					data_edges[id][2*i]
					= v.Give_data<edgeEl> ( id, i  , 0, 0, 0, 0 );
					
			} else { // no coarse edge
				// x edge
				for ( int i = 1;i <= Nx;++i )
					data_edges[id][i]
					= v.Give_data<edgeEl> ( id, i  , 0  , 0, 0, 0 );
			}
		}
	}
	
	// point
	for ( int id = 0;id < ug->Give_number_points();++id ) {
		if ( ug->Give_point ( id )->my_object ( my_rank ) )
			data_points[id][0] = v.Give_data<pointEl> ( id, 0  , 0  , 0, 0, 0 );
	}
}



#endif // MGOP_H
