/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#ifndef PETROV_H
#define PETROV_H





template <typename TYPE2>
void Local_stiffness_matrix<TYPE2>::makePetrovGalerkinZ() {
	int Nx, Ny, Nz, N_total;
	
	Unstructured_grid* ug = blockgrid->Give_unstructured_grid();
	
	for ( int id = 0;id < blockgrid->Give_unstructured_grid()->Give_number_hexahedra();++id ) {
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) ) {
			Nx = blockgrid->Give_Nx_hexahedron ( id );
			Ny = blockgrid->Give_Ny_hexahedron ( id );
			Nz = blockgrid->Give_Nz_hexahedron ( id );
			N_total = Nx * Ny * Nz * 64;
			
			// only for k = 0, i.e. at left end of hexaedron
			
			for ( int k = 0;k < Nz;++k ) {
				for ( int j = 0;j < Ny;++j ) {
					for ( int i = 0;i < Nx;++i ) {
						for ( int u = 0;u < 8;++u ) {
							for ( int vxy = 0;vxy < 4;++vxy ) {
								int vL = vxy + 4 * Ldir1D;
								int vR = vxy + 4 * Rdir1D;
								
								loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64+8*vR+u] +=
								    loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64+8*vL+u];
								    
								loc_m_hexahedra[id][Ind_loc_matrix_hexahedra ( i,j,k ) *64+8*vL+u] = 0.0;
							}
						}
					}
				}
			}
		}
	}
	
	Update();
}


#endif // PETROV_H
