/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/


template <class DTyp>
void Restrict_local_stiffness_matrix_xyz(DTyp*  Stencil_coarse,
					 DTyp** Stencils_fine,
					 int* coarse_grid_functions) {
  int* v_function;
  int* u_function;
  int u_f, v_f;

  // 1.Part: Set zero
  for(int i=0;i<8;++i) for(int j=0;j<8;++j) Stencil_coarse[j+8*i] = 0.0;

  // 2.Part: Iterate
  for(int i=0;i<8;++i) { // iteration for U
    u_function = &(coarse_grid_functions[i*27]);
    for(int j=0;j<8;++j) { // iteration for V
      v_function = &(coarse_grid_functions[j*27]);
      for(int k=0;k<8;++k) { // iteration over cells
	for(int l=0;l<8;++l) { // iteration over corners of cell for U
	  u_f = u_function[trans_cellP_27((dir3D_sons)k,(dir3D_sons)l)];
	  if(u_f != 0) for(int m=0;m<8;++m) { // iteration over corners of cell for V
	    v_f = v_function[trans_cellP_27((dir3D_sons)k,(dir3D_sons)m)];
	    if(v_f!=0)
	      Stencil_coarse[j+8*i] = Stencil_coarse[j+8*i] +
		(double)u_f * (double)v_f * 0.015625 *
		Stencils_fine[k][m+8*l];
	  }
	}
      }
    }
  }
}

class Index_calc_x {
 public:
  static int index(int i, int j, int k) { return i + 2*j + 4*k; }
};

class Index_calc_y {
 public:
  static int index(int j, int i, int k) { return i + 2*j + 4*k; }
};

class Index_calc_z {
 public:
  static int index(int k, int i, int j) { return i + 2*j + 4*k; }
};


template <class DTyp, class Tc>
void Restrict_local_stiffness_matrix_2D(DTyp*  Stencil_coarse,
					DTyp** Stencils_fine,
					int* coarse_grid_functions) {
  int i,j,ii,jj,k,l,m;
  int* v_function;
  int* u_function;
  int u_f, v_f;

  // 1.Part: Set zero
  for(i=0;i<8;++i) for(j=0;j<8;++j) Stencil_coarse[j+8*i] = 0.0;

  // 2.Part: Iterate
  for(i=0;i<4;++i) { // iteration for U
    u_function = &(coarse_grid_functions[i*9]);
    for(j=0;j<4;++j) { // iteration for V
      v_function = &(coarse_grid_functions[j*9]);
      for(k=0;k<4;++k) { // iteration over cells
	for(l=0;l<4;++l) { // iteration over corners of cell for U
	  u_f = u_function[trans_cellP_9((dir2D_sons)k,(dir2D_sons)l)];
	  if(u_f != 0) for(m=0;m<4;++m) { // iteration over corners of cell for V
	    v_f = v_function[trans_cellP_9((dir2D_sons)k,(dir2D_sons)m)];
	    if(v_f!=0)
	      for(ii=0;ii<2;++ii) 
		for(jj=0;jj<2;++jj)
		  Stencil_coarse[Tc::index(jj,j&1,j>>1)+8*Tc::index(ii,i&1,i>>1)] = 
		    Stencil_coarse[Tc::index(jj,j&1,j>>1)+8*Tc::index(ii,i&1,i>>1)] +
		(double)u_f * (double)v_f * 0.0625 *
		    Stencils_fine[k][Tc::index(jj,m&1,m>>1)+8*Tc::index(ii,l&1,l>>1)];
	  }
	}
      }
    }
  }
}


template <class DTyp, class Tc>
void Restrict_local_stiffness_matrix_1D(DTyp*  Stencil_coarse,
					DTyp** Stencils_fine) {
  int i,j,ii,jj,iii,jjj,k,l,m;
  int* v_function;
  int* u_function;
  int u_f, v_f;

  int coarse_grid_functions[2][3];

  coarse_grid_functions[0][0] = 2;
  coarse_grid_functions[0][1] = 1;
  coarse_grid_functions[0][2] = 0;

  coarse_grid_functions[1][2] = 2;
  coarse_grid_functions[1][1] = 1;
  coarse_grid_functions[1][0] = 0;

  // 1.Part: Set zero
  for(i=0;i<8;++i) for(j=0;j<8;++j) Stencil_coarse[j+8*i] = 0.0;

  // 2.Part: Iterate
  for(i=0;i<2;++i) { // iteration for U
    u_function = coarse_grid_functions[i];
    for(j=0;j<2;++j) { // iteration for V
      v_function = coarse_grid_functions[j];
      for(k=0;k<2;++k) { // iteration over cells
	for(l=0;l<2;++l) { // iteration over corners of cell for U
	  u_f = u_function[trans_cellP_3((dir1D)k,(dir1D)l)];
	  if(u_f != 0) for(m=0;m<2;++m) { // iteration over corners of cell for V
	    v_f = v_function[trans_cellP_3((dir1D)k,(dir1D)m)];
	    if(v_f!=0)
	      for(ii=0;ii<2;++ii) for(iii=0;iii<2;++iii) 
		for(jj=0;jj<2;++jj) for(jjj=0;jjj<2;++jjj)
	      Stencil_coarse[Tc::index(j,jj,jjj)+8*Tc::index(i,ii,iii)] = 
		Stencil_coarse[Tc::index(j,jj,jjj)+8*Tc::index(i,ii,iii)] +
		(double)u_f * (double)v_f * 0.25 *
		Stencils_fine[k][Tc::index(m,jj,jjj)+8*Tc::index(l,ii,iii)];
	  }
	}
      }
    }
  }
}
