/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#include <string>
#include <stdexcept>
#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "../grid/elements.h"
#include "../grid/parti.h"
#include "../grid/ug.h"
#include "../grid/blockgrid.h"
#include "../grid/marker.h"
#include "../extemp/parallel.h"
#include "../extemp/extemp.h"
#include "../extemp/variable.h"
#include "../extemp/variable_cc.h"
#include "shiftop.h"


Shift_Operator_D::Shift_Operator_D ( Blockgrid& bg )
{
  Unstructured_grid* ug = bg.Give_unstructured_grid();

// quadrangle
  interior_quadrangle_is_down = new shift_quad_type[ug->Give_number_quadrangles() ];
  for ( int id=0;id<ug->Give_number_quadrangles();++id )
  {
    Quadrangle_el* quad = ug->Give_quadrangle ( id );
    if ( quad->my_object ( ug->Give_my_rank() ) )
    {
      D3vector coordInteriorHex =
        ug->Give_hexahedron ( quad->Give_id_interior_vol() )->Give_middle_coord();

      D3vector coordQuad = quad->Give_middle_coord();

      D3vector diffVector = coordInteriorHex - coordQuad;
      if ( ABS ( diffVector.x ) > ABS ( diffVector.y ) && ABS ( diffVector.x ) > ABS ( diffVector.z ) )
      {
        // x-direction
        interior_quadrangle_is_down[id] = no_quad_shift ;
        /// TODO richtig machen für Aussenflächen
      }
      if ( ABS ( diffVector.y ) > ABS ( diffVector.x ) && ABS ( diffVector.y ) > ABS ( diffVector.z ) )
      {
        // y-direction
        interior_quadrangle_is_down[id] = no_quad_shift ;
        /// TODO richtig machen für Aussenflächen
      }
      if ( ABS ( diffVector.z ) > ABS ( diffVector.y ) && ABS ( diffVector.z ) > ABS ( diffVector.x ) )
      {
        // z-direction
        if ( quad->Give_exists_exterior() )
        {
          D3vector coordExteriorHex =
            ug->Give_hexahedron ( quad->Give_id_exterior_vol() )->Give_middle_coord();
          D3vector diffVectorTwo = coordQuad-coordExteriorHex;
          // den richtigen punkt neben bei periodischen randbedingungen !
          if ( ABS ( diffVectorTwo.z ) < ABS ( diffVector.z )) { 
            diffVector = diffVectorTwo;
          }

          if ( diffVector.z < 0.0 )
          {
            interior_quadrangle_is_down[id] = interior_quad_shift;
          }
          else
          {
            interior_quadrangle_is_down[id] = exterior_quad_shift;
          }
        }
        else
        {
          interior_quadrangle_is_down[id] = interior_quad_shift;
        }
      }
    }
  }

  // edges
  id_of_down_neighbor_edge = new int[ug->Give_number_edges() ];
  for ( int id=0;id<ug->Give_number_edges();++id )
  {
    int num_neighbor = 0;  ///TODO richtig machen ausrechnen
    // convention:   num_neighbor == -1   down index on edge
    // convention:   num_neighbor == -2   up   index on edge
    // convention:   num_neighbor >= 0    neighbour edge
    id_of_down_neighbor_edge[id] =  num_neighbor;
  }

// int*  id_of_down_point;                    // num_po
};

