/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

// ------------------------------------------------------------
//
// shiftop.h
//
// ------------------------------------------------------------

#ifndef SHIFTOP_H
#define SHIFTOP_H

enum shift_quad_type { interior_quad_shift, exterior_quad_shift,
                       minus_i_quad_shift, plus_i_quad_shift,
                       minus_j_quad_shift, plus_j_quad_shift,
                       no_quad_shift
                     };
                     
                     
template <typename TYPE>

class Expr_Shift_D :
			public Expr<Expr_Shift_D<TYPE> > {
			
	public:
		Expr_Shift_D ( const Variable<TYPE>& v,
		               shift_quad_type*  interior_quadrangle_is_down_ ) :
				v_ ( v ), interior_quadrangle_is_down ( interior_quadrangle_is_down_ ) {}
				
		Blockgrid* Give_blockgrid() const { return v_.Give_blockgrid(); }
		
		stencil_typ Give_stencil_typ() const { return yes_stencil; }
		control_typ Give_control_typ() const { return thread_save; }
		
		
		template <elementTyp TYP_EL>
		inline TYPE Give_data ( params_in ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			v_.Update<TYP_EL> ( id );
		}
		
	private:
		const Variable<TYPE>& v_;
		shift_quad_type*  interior_quadrangle_is_down;       // num_quadrangles
		
};


// Spezialisierung fuer double und complex double benoetigt !!

template <>
template <>
inline double Expr_Shift_D<double>::Give_data<hexahedronEl> ( params_in ) const {
//   return v_.Give_data<hexahedronEl>(id,i,j,k,Nx,Ny);
	return v_.Give_data<hexahedronEl> ( id, i, j, k - 1, Nx, Ny );
}

template <>
template <>
inline std::complex<double> Expr_Shift_D<std::complex<double> >::Give_data<hexahedronEl> ( params_in ) const {
//   return v_.Give_data<hexahedronEl>(id,i,j,k,Nx,Ny);
	return v_.Give_data<hexahedronEl> ( id, i, j, k - 1, Nx, Ny );
}

template <>
template <>
inline double Expr_Shift_D<double>::Give_data<quadrangleEl> ( params_in ) const {
//   return v_.Give_data_quadrangle_middle(id,i,j,Nx);
	switch ( interior_quadrangle_is_down[id] ) {
	
		case interior_quad_shift :
			return v_.Give_data_quadrangle_interior ( id, i, j, Nx );
			break;
			
		case exterior_quad_shift :
			return v_.Give_data_quadrangle_exterior ( id, i, j, Nx );
			break;
			
		case minus_i_quad_shift  :
			return v_.Give_data_quadrangle_middle ( id, i - 1, j, Nx );
			break;
			
		case plus_i_quad_shift   :
			return v_.Give_data_quadrangle_middle ( id, i + 1, j, Nx );
			break;
			
		case minus_j_quad_shift  :
			return v_.Give_data_quadrangle_middle ( id, i, j - 1, Nx );
			break;
			
		case plus_j_quad_shift   :
			return v_.Give_data_quadrangle_middle ( id, i, j + 1, Nx );
			break;
			
		case no_quad_shift       :
			return v_.Give_data_quadrangle_middle ( id, i, j, Nx );
			break;
			
		default:
			;
	}
	
	throw ( std::runtime_error ( "Error in Expr_Shift_D: No valid shift_type found!" ) );
}

template <>
template <>
inline std::complex<double> Expr_Shift_D<std::complex<double> >::Give_data<quadrangleEl> ( params_in ) const {
//   return v_.Give_data_quadrangle_middle(id,i,j,Nx);
	switch ( interior_quadrangle_is_down[id] ) {
	
		case interior_quad_shift :
			return v_.Give_data_quadrangle_interior ( id, i, j, Nx );
			break;
			
		case exterior_quad_shift :
			return v_.Give_data_quadrangle_exterior ( id, i, j, Nx );
			break;
			
		case minus_i_quad_shift  :
			return v_.Give_data_quadrangle_middle ( id, i - 1, j, Nx );
			break;
			
		case plus_i_quad_shift   :
			return v_.Give_data_quadrangle_middle ( id, i + 1, j, Nx );
			break;
			
		case minus_j_quad_shift  :
			return v_.Give_data_quadrangle_middle ( id, i, j - 1, Nx );
			break;
			
		case plus_j_quad_shift   :
			return v_.Give_data_quadrangle_middle ( id, i, j + 1, Nx );
			break;
			
		case no_quad_shift       :
			return v_.Give_data_quadrangle_middle ( id, i, j, Nx );
			break;
			
		default:
			;
	}
	
	throw ( std::runtime_error ( "Error in Expr_Shift_D: No valid shift_type found!" ) );
}


template <>
template <>
inline double Expr_Shift_D<double>::Give_data<edgeEl> ( params_in ) const {
	///TODO richtig machen
	return v_.Give_data_edge ( id, i );
}

template <>
template <>
inline std::complex<double> Expr_Shift_D<std::complex<double> >::Give_data<edgeEl> ( params_in ) const {
	///TODO richtig machen
	return v_.Give_data_edge ( id, i );
}




template <>
template <>
inline double Expr_Shift_D<double>::Give_data<pointEl> ( params_in ) const {
	///TODO richtig machen
	return v_.Give_data_point ( id );
}

template <>
template <>
inline std::complex<double> Expr_Shift_D<std::complex<double> >::Give_data<pointEl> ( params_in ) const {
	///TODO richtig machen
	return v_.Give_data_point ( id );
}


class Shift_Operator_D {

	public:
		Shift_Operator_D ( Blockgrid& bg_ );
		
		template <typename TYPE>
		Expr_Shift_D<TYPE> operator() ( const Variable<TYPE>& v ) const {
			return Expr_Shift_D<TYPE> ( v, interior_quadrangle_is_down );
		}
		
	private:
		shift_quad_type*  interior_quadrangle_is_down;       // num_quadrangles
		
		///TDO noch nicht richtig
		int* id_of_down_neighbor_edge;                 // num_edges
		int*  id_of_down_point;                    // num_points
};




#endif // SHIFTOP_H
