/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

// ------------------------------------------------------------
//
// stenop.h
//
// ------------------------------------------------------------

#ifndef STENOP_H
#define STENOP_H


///////////////////////////
// 1. stenop type classes
// 2.1. Expression Stencil applied to Variable
// 2.2. Expression Stencil transpose applied to Variable
// 3. class Stencil
// 4. Berechnung der Stencils
///////////////////////////

///////////////////////////
// 1. stenop type classes
///////////////////////////


class Stencil_variable {

	public:
		Stencil_variable() {};
		
		int stenIndex ( int i, int j, int k, int Nx, int Ny, int id ) const {
			return ( i -1 + ( Nx - 1 ) * ( j - 1 + ( Ny - 1 ) * ( k - 1 ) ) ) *27;
		}
		
		bool Not_const_x ( int id ) const { return true; };
		
		bool Not_const_y ( int id ) const { return true; };
		
		bool Not_const_z ( int id ) const { return true; };
};

class Stencil_constant_z {

	public:
		Stencil_constant_z() {};
		
		int stenIndex ( int i, int j, int k, int Nx, int Ny, int id ) const {
			return ( i -1 + ( Nx - 1 ) * ( j - 1 ) ) *27;
		}
		
		bool Not_const_x ( int id ) const { return true; };
		
		bool Not_const_y ( int id ) const { return true; };
		
		bool Not_const_z ( int id ) const { return false; };
};

class Stencil_constant {

	public:
		Stencil_constant() {};
		
		int stenIndex ( int i, int j, int k, int Nx, int Ny, int id ) const {
			return 0;
		}
		
		bool Not_const_x ( int id ) const { return false; };
		
		bool Not_const_y ( int id ) const { return false; };
		
		bool Not_const_z ( int id ) const { return false; };
};

// see ug.h definition of:
// class Not_constant_direction_marker


//////////////////////////////////////////////
// 2.1. Expression Stencil applied to Variable
//////////////////////////////////////////////


template <typename TYPE1, typename TYPE2, class STENCIL_TYP>

class Expr_Sten_matrix :
			public Expr<Expr_Sten_matrix<TYPE1, TYPE2, STENCIL_TYP> > {
			
	public:
		Expr_Sten_matrix ( const Variable<TYPE1>& v, TYPE2**  lm_hex,
		                   TYPE2**  lm_quad_in,
		                   TYPE2**  lm_quad_ex,
		                   TYPE2*** lm_neigh_ed,
		                   TYPE2**  lm_neigh_po,
		                   STENCIL_TYP st ) :
				v_ ( v ), sten_m_hexahedra ( lm_hex ),
				loc_m_interior_quadrangles ( lm_quad_in ),
				loc_m_exterior_quadrangles ( lm_quad_ex ),
				loc_m_neighbor_edges ( lm_neigh_ed ),
				loc_m_neighbor_points ( lm_neigh_po ),
				sten_typ ( st ) {}
				
		Blockgrid* Give_blockgrid() const { return v_.Give_blockgrid(); };
		
		stencil_typ Give_stencil_typ() const { return yes_stencil; }
		
		control_typ Give_control_typ() const { return thread_save; }
		
		typedef _TypeOf0_ ( TYPE1, TYPE2 ) Result;
		
		//template <elementTyp TYP_EL>
		//inline Result Give_data ( params_in ) const;
		template <elementTyp TYP_EL>
		ALWAYS_INLINE Result Give_data ( params_in ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
            v_.template Update<TYP_EL> ( id );
		}
		
	private:
		const Variable<TYPE1>& v_;
		TYPE2**  sten_m_hexahedra;                  // num_hexahedra
		//  TYPE2**  loc_m_hexahedra;                  // num_hexahedra
		TYPE2**  loc_m_interior_quadrangles;       // num_quadrangles
		TYPE2**  loc_m_exterior_quadrangles;       // num_quadrangles
		TYPE2*** loc_m_neighbor_edges;             // num_edges
		TYPE2**  loc_m_neighbor_points;            // num_points
		
		STENCIL_TYP sten_typ;
};

#define Give_functions_Expr_Sten(R,A,B,C)	                               		         \
	template <>                                                                                      \
	template <>                                                                                      \
/*	inline R      Expr_Sten_matrix<A,B,C>::Give_data<hexahedronEl>(params_in) const { */              \
	ALWAYS_INLINE  R      Expr_Sten_matrix<A,B,C>::Give_data<hexahedronEl>(params_in) const {               \
		int ig = sten_typ.stenIndex(i,j,k,Nx,Ny,id);	                              			 \
		return                                                                                         \
		        v_.Give_data<hexahedronEl>(id,i-1,j-1,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,LOrt,LOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j-1,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,LOrt,LOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j-1,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,LOrt,LOrt)] +  \
		        \
		        v_.Give_data<hexahedronEl>(id,i-1,j  ,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,MOrt,LOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j  ,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,MOrt,LOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j  ,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,MOrt,LOrt)] +  \
		        \
		        v_.Give_data<hexahedronEl>(id,i-1,j+1,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,ROrt,LOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j+1,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,ROrt,LOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j+1,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,ROrt,LOrt)] +  \
		        \
		        v_.Give_data<hexahedronEl>(id,i-1,j-1,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,LOrt,MOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j-1,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,LOrt,MOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j-1,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,LOrt,MOrt)] +  \
		        \
		        v_.Give_data<hexahedronEl>(id,i-1,j  ,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,MOrt,MOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j  ,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,MOrt,MOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j  ,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,MOrt,MOrt)] +  \
		        \
		        v_.Give_data<hexahedronEl>(id,i-1,j+1,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,ROrt,MOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j+1,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,ROrt,MOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j+1,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,ROrt,MOrt)] +  \
		        \
		        v_.Give_data<hexahedronEl>(id,i-1,j-1,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,LOrt,ROrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j-1,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,LOrt,ROrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j-1,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,LOrt,ROrt)] +  \
		        \
		        v_.Give_data<hexahedronEl>(id,i-1,j  ,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,MOrt,ROrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j  ,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,MOrt,ROrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j  ,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,MOrt,ROrt)] +  \
		        \
		        v_.Give_data<hexahedronEl>(id,i-1,j+1,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,ROrt,ROrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j+1,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,ROrt,ROrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j+1,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,ROrt,ROrt)];   \
	}                                                                                                           \
	\
	template <>                                                                                                 \
	template <>                                                                                                 \
	inline R Expr_Sten_matrix<A,B,C>::Give_data<quadrangleEl>(params_in) const {                               \
		int u,v;                                                                                                  \
		dir3D_sons ce;                                                                                              \
		R sum;                                                                                                    \
		sum = 0.0;                                                                                                \
		for(v=0;v<4;++v) {                                                                                        \
			ce = opposite3D((dir3D_sons)v);                                                                           \
			for(u=0;u<4;++u)                                                                                        \
				sum = sum +                                                                                           \
				      loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles                                       \
				                                     (x1DCoord(ce)-1+i,y1DCoord(ce)-1+j)*32+8*v+u] *                  \
				      v_.Give_data_quadrangle_middle(id,x1DCoord(ce)+x1DCoord((dir3D_sons)u)-1+i,                       \
				                                     y1DCoord(ce)+y1DCoord((dir3D_sons)u)-1+j,Nx) +                  \
				      loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles                                       \
				                                     (x1DCoord(ce)-1+i,y1DCoord(ce)-1+j)*32+8*v+(u+WSTdir3D)] *           \
				      v_.Give_data_quadrangle_interior(id,x1DCoord(ce)+x1DCoord((dir3D_sons)u)-1+i,                     \
				                                       y1DCoord(ce)+y1DCoord((dir3D_sons)u)-1+j,Nx);                 \
		}                                                                                                         \
		if(Give_blockgrid()->Give_unstructured_grid()->Give_quadrangle(id)->Give_exists_exterior()==false)        \
			return sum;                                                                                           \
		for(v=0;v<4;++v) {                                                                                        \
			ce = opposite3D((dir3D_sons)v);                                                                           \
			for(u=0;u<4;++u)                                                                                        \
				sum = sum +                                                                                           \
				      loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles                                       \
				                                     (x1DCoord(ce)-1+i,y1DCoord(ce)-1+j)*32+8*v+u] *                   \
				      v_.Give_data_quadrangle_middle(id,x1DCoord(ce)+x1DCoord((dir3D_sons)u)-1+i,                       \
				                                     y1DCoord(ce)+y1DCoord((dir3D_sons)u)-1+j,Nx) +                  \
				      loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles                                       \
				                                     (x1DCoord(ce)-1+i,y1DCoord(ce)-1+j)*32+8*v+(u+WSTdir3D)] *           \
				      v_.Give_data_quadrangle_exterior(id,x1DCoord(ce)+x1DCoord((dir3D_sons)u)-1+i,                     \
				                                       y1DCoord(ce)+y1DCoord((dir3D_sons)u)-1+j,Nx);                 \
		}                                                                                                         \
		return sum; }                                                                                             \
	\
	template <>                                                                                                 \
	template <>                                                                                                 \
	inline R Expr_Sten_matrix<A,B,C>::Give_data<edgeEl>(params_in) const {                                     \
		int u,v;                                                                                                  \
		int num;                                                                                                  \
		int id_ND, id_ST, id_NT;                                                                                  \
		dir3D_sons ce;                                                                                              \
		R sum;                                                                                                    \
		sum = 0.0;                                                                                                \
		int number_neighbor_hex =                                                                                 \
		        Give_blockgrid()->Give_unstructured_grid()->Give_edge(id)->Give_number_neighbors_hex();             \
		for(num=0;num<number_neighbor_hex;++num){                                                                 \
			id_ND = Give_blockgrid()->Give_unstructured_grid()->Give_edge(id)->Give_loc_id_hex (num,ND);            \
			id_ST = Give_blockgrid()->Give_unstructured_grid()->Give_edge(id)->Give_loc_id_hex (num,ST);            \
			id_NT = Give_blockgrid()->Give_unstructured_grid()->Give_edge(id)->Give_loc_id_hex (num,NT);            \
			for(v=0;v<2;++v) {                                                                                      \
				ce = opposite3D((dir3D_sons)v);                                                                         \
				for(u=0;u<2;++u)                                                                                      \
					sum = sum +                                                                                          \
					      loc_m_neighbor_edges[id][num][(x1DCoord(ce)-1+i)*16+8*v+u] *                                   \
					      v_.Give_data_edge(id,x1DCoord(ce)+x1DCoord((dir3D_sons)u)-1+i) +                                 \
					      \
					      loc_m_neighbor_edges[id][num][(x1DCoord(ce)-1+i)*16+8*v+(u+WNDdir3D)] *                            \
					      v_.Give_data_neighbor_edge(id,id_ND,x1DCoord(ce)+x1DCoord((dir3D_sons)u)-1+i) +                  \
					      \
					      loc_m_neighbor_edges[id][num][(x1DCoord(ce)-1+i)*16+8*v+(u+WSTdir3D)] *                            \
					      v_.Give_data_neighbor_edge(id,id_ST,x1DCoord(ce)+x1DCoord((dir3D_sons)u)-1+i) +                  \
					      \
					      loc_m_neighbor_edges[id][num][(x1DCoord(ce)-1+i)*16+8*v+(u+WNTdir3D)] *                            \
					      v_.Give_data_neighbor_edge(id,id_NT,x1DCoord(ce)+x1DCoord((dir3D_sons)u)-1+i);                   \
			}                                                                                                       \
		}                                                                                                         \
		return sum; }                                                                                             \
	\
	template <>                                                                                                 \
	template <>                                                                                                 \
	inline R Expr_Sten_matrix<A,B,C>::Give_data<pointEl>(params_in) const {	                            \
		int num;                                                                                                  \
		int id_SED, id_NWD, id_NED, id_SWT, id_SET, id_NWT, id_NET;                                               \
		dir3D_sons ce;                                                                                              \
		R sum;                                                                                                    \
		sum = 0.0;                                                                                                \
		int number_neighbor_hex =                                                                                 \
		        Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->Give_number_neighbors_hex();            \
		for(num=0;num<number_neighbor_hex;++num){                                                                 \
			id_SED = Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->Give_loc_id_hex (num,SED);         \
			id_NWD = Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->Give_loc_id_hex (num,NWD);         \
			id_NED = Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->Give_loc_id_hex (num,NED);         \
			id_SWT = Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->Give_loc_id_hex (num,SWT);         \
			id_SET = Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->Give_loc_id_hex (num,SET);         \
			id_NWT = Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->Give_loc_id_hex (num,NWT);         \
			id_NET = Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->Give_loc_id_hex (num,NET);         \
			ce = ENTdir3D;                                                                                              \
			sum = sum +                                                                                             \
			      loc_m_neighbor_points[id][num*8+WSDdir3D] * v_.Give_data_point(id) +                               \
			      loc_m_neighbor_points[id][num*8+ESDdir3D] * v_.Give_data_neighbor_point(id,id_SED) +               \
			      loc_m_neighbor_points[id][num*8+WNDdir3D] * v_.Give_data_neighbor_point(id,id_NWD) +               \
			      loc_m_neighbor_points[id][num*8+ENDdir3D] * v_.Give_data_neighbor_point(id,id_NED) +               \
			      loc_m_neighbor_points[id][num*8+WSTdir3D] * v_.Give_data_neighbor_point(id,id_SWT) +               \
			      loc_m_neighbor_points[id][num*8+ESTdir3D] * v_.Give_data_neighbor_point(id,id_SET) +               \
			      loc_m_neighbor_points[id][num*8+WNTdir3D] * v_.Give_data_neighbor_point(id,id_NWT) +               \
			      loc_m_neighbor_points[id][num*8+ENTdir3D] * v_.Give_data_neighbor_point(id,id_NET);                \
		}                                                                                                         \
		return sum; }


Give_functions_Expr_Sten ( double, double, double, Stencil_variable )
Give_functions_Expr_Sten ( std::complex<double>, std::complex<double>, double, Stencil_variable )
Give_functions_Expr_Sten ( std::complex<double>, std::complex<double>, std::complex<double>, Stencil_variable )

Give_functions_Expr_Sten ( double, double, double, Stencil_constant_z )
Give_functions_Expr_Sten ( std::complex<double>, std::complex<double>, double, Stencil_constant_z )
Give_functions_Expr_Sten ( std::complex<double>, std::complex<double>, std::complex<double>, Stencil_constant_z )

Give_functions_Expr_Sten ( double, double, double, Stencil_constant )
Give_functions_Expr_Sten ( std::complex<double>, std::complex<double>, double, Stencil_constant )
Give_functions_Expr_Sten ( std::complex<double>, std::complex<double>, std::complex<double>, Stencil_constant )

Give_functions_Expr_Sten ( double, double, double, Not_constant_direction_marker )
Give_functions_Expr_Sten ( std::complex<double>, std::complex<double>, double, Not_constant_direction_marker )
Give_functions_Expr_Sten ( std::complex<double>, std::complex<double>, std::complex<double>, Not_constant_direction_marker )



template <typename TYPE2, class STENCIL_TYP>

class Expr_Diag_Sten_matrix :
			public Expr<Expr_Diag_Sten_matrix<TYPE2, STENCIL_TYP> > {
			
	public:
		Expr_Diag_Sten_matrix ( TYPE2** lm_hex, TYPE2** lm_quad_in, TYPE2** lm_quad_ex,
		                        TYPE2*** lm_neigh_ed, TYPE2** lm_neigh_po, Blockgrid* blockgrid_,
		                        STENCIL_TYP st ) :
				sten_m_hexahedra ( lm_hex ),
				loc_m_interior_quadrangles ( lm_quad_in ),
				loc_m_exterior_quadrangles ( lm_quad_ex ),
				loc_m_neighbor_edges ( lm_neigh_ed ),
				loc_m_neighbor_points ( lm_neigh_po ),
				blockgrid ( blockgrid_ ),
				sten_typ ( st ) {}
				
		Blockgrid* Give_blockgrid() const {
			return blockgrid;
		};
		
		stencil_typ Give_stencil_typ() const { return yes_stencil; }
		control_typ Give_control_typ() const { return thread_save; }
		
		typedef TYPE2 Result;
		
		template <elementTyp TYP_EL>
		inline TYPE2 Give_data ( params_in ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {}
		
	private:
		TYPE2**  sten_m_hexahedra;                  // num_hexahedra
		
		TYPE2**  loc_m_interior_quadrangles;       // num_quadrangles
		TYPE2**  loc_m_exterior_quadrangles;       // num_quadrangles
		TYPE2*** loc_m_neighbor_edges;             // num_edges
		TYPE2**  loc_m_neighbor_points;            // num_points
		
		Blockgrid* blockgrid;
		
		STENCIL_TYP sten_typ;
};

#define Give_functions_Diag_Expr_Sten(A,B)	                               		         \
	template <>                                                                                      \
	template <>                                                                                      \
	inline A Expr_Diag_Sten_matrix<A,B>::Give_data<hexahedronEl>(params_in) const {               \
		return sten_m_hexahedra[id][sten_typ.stenIndex(i,j,k,Nx,Ny,id)  + trans27(MOrt,MOrt,MOrt)];          \
	}                                                                                                \
	\
	template <>                                                                                      \
	template <>                                                                                      \
	inline A Expr_Diag_Sten_matrix<A,B>::Give_data<quadrangleEl>(params_in) const {                 \
		if(Give_blockgrid()->Give_unstructured_grid()->Give_quadrangle(id)->Give_exists_exterior())    \
			return loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles(i  ,j  )*32   ]             \
			       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles(i-1,j  )*32+ 9]           \
			       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles(i  ,j-1)*32+18]           \
			       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles(i-1,j-1)*32+27]           \
			       + loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles(i  ,j  )*32   ]           \
			       + loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles(i-1,j  )*32+ 9]           \
			       + loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles(i  ,j-1)*32+18]           \
			       + loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles(i-1,j-1)*32+27];          \
		return loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles(i  ,j  )*32   ]             \
		       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles(i-1,j  )*32+ 9]           \
		       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles(i  ,j-1)*32+18]           \
		       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles(i-1,j-1)*32+27];          \
	}                                                                                                \
	\
	template <>                                                                                      \
	template <>                                                                                      \
	inline A Expr_Diag_Sten_matrix<A,B>::Give_data<edgeEl>(params_in) const {                       \
		int number_neighbor_hex =                                                                      \
		        blockgrid->Give_unstructured_grid()->Give_edge(id)->Give_number_neighbors_hex();         \
		A sum = 0.0;                                                                                   \
		for(int num=0;num<number_neighbor_hex;++num)                                                   \
			sum += loc_m_neighbor_edges[id][num][ i   *16  ]                                             \
			       + loc_m_neighbor_edges[id][num][(i-1)*16+9];                                            \
		return sum;                                                                                    \
	}                                                                                                \
	\
	template <>                                                                                      \
	template <>                                                                                      \
	inline A Expr_Diag_Sten_matrix<A,B>::Give_data<pointEl>(params_in) const {                      \
		int number_neighbor_hex =                                                                      \
		        blockgrid->Give_unstructured_grid()->Give_point(id)->Give_number_neighbors_hex();        \
		A sum = 0.0;                                                                                   \
		for(int num=0;num<number_neighbor_hex;++num)                                                   \
			sum += loc_m_neighbor_points[id][num*8];                                                   \
		return sum; }





Give_functions_Diag_Expr_Sten ( double, Stencil_variable )
Give_functions_Diag_Expr_Sten ( std::complex<double>, Stencil_variable )

Give_functions_Diag_Expr_Sten ( double, Stencil_constant_z )
Give_functions_Diag_Expr_Sten ( std::complex<double>, Stencil_constant_z )

Give_functions_Diag_Expr_Sten ( double, Stencil_constant )
Give_functions_Diag_Expr_Sten ( std::complex<double>, Stencil_constant )

Give_functions_Diag_Expr_Sten ( double, Not_constant_direction_marker )
Give_functions_Diag_Expr_Sten ( std::complex<double>, Not_constant_direction_marker )



//////////////////////////////////////////////
// 2.2. Expression Stencil applied to Variable
//////////////////////////////////////////////


template <typename TYPE1, typename TYPE2, class STENCIL_TYP>

class Expr_Sten_matrix_transpose :
			public Expr<Expr_Sten_matrix_transpose<TYPE1, TYPE2, STENCIL_TYP> > {
			
	public:
		Expr_Sten_matrix_transpose ( const Variable<TYPE1>& v, TYPE2**  lm_hex,
		                   TYPE2**  lm_quad_in,
		                   TYPE2**  lm_quad_ex,
		                   TYPE2*** lm_neigh_ed,
		                   TYPE2**  lm_neigh_po,
		                   STENCIL_TYP st ) :
				v_ ( v ), sten_m_hexahedra ( lm_hex ),
				loc_m_interior_quadrangles ( lm_quad_in ),
				loc_m_exterior_quadrangles ( lm_quad_ex ),
				loc_m_neighbor_edges ( lm_neigh_ed ),
				loc_m_neighbor_points ( lm_neigh_po ),
				sten_typ ( st ) {}
				
		Blockgrid* Give_blockgrid() const { return v_.Give_blockgrid(); };
		
		stencil_typ Give_stencil_typ() const { return yes_stencil; }
		
		control_typ Give_control_typ() const { return thread_save; }
		
		typedef _TypeOf0_ ( TYPE1, TYPE2 ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params_in ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
            v_.template Update<TYP_EL> ( id );
		}
		
	private:
		const Variable<TYPE1>& v_;
		TYPE2**  sten_m_hexahedra;                  // num_hexahedra
		//  TYPE2**  loc_m_hexahedra;                  // num_hexahedra
		TYPE2**  loc_m_interior_quadrangles;       // num_quadrangles
		TYPE2**  loc_m_exterior_quadrangles;       // num_quadrangles
		TYPE2*** loc_m_neighbor_edges;             // num_edges
		TYPE2**  loc_m_neighbor_points;            // num_points
		
		STENCIL_TYP sten_typ;
};

#define Give_functions_Expr_Sten_transpose(R,A,B,C)	                               		         \
	template <>                                                                                      \
	template <>                                                                                      \
	inline R      Expr_Sten_matrix_transpose<A,B,C>::Give_data<hexahedronEl>(params_in) const {               \
		int ig = sten_typ.stenIndex(i,j,k,Nx,Ny,id);	                              			 \
		return                                                                                         \
		        v_.Give_data<hexahedronEl>(id,i-1,j-1,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,LOrt,LOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j-1,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,LOrt,LOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j-1,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,LOrt,LOrt)] +  \
		        \
		        v_.Give_data<hexahedronEl>(id,i-1,j  ,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,MOrt,LOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j  ,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,MOrt,LOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j  ,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,MOrt,LOrt)] +  \
		        \
		        v_.Give_data<hexahedronEl>(id,i-1,j+1,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,ROrt,LOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j+1,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,ROrt,LOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j+1,k-1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,ROrt,LOrt)] +  \
		        \
		        v_.Give_data<hexahedronEl>(id,i-1,j-1,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,LOrt,MOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j-1,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,LOrt,MOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j-1,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,LOrt,MOrt)] +  \
		        \
		        v_.Give_data<hexahedronEl>(id,i-1,j  ,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,MOrt,MOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j  ,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,MOrt,MOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j  ,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,MOrt,MOrt)] +  \
		        \
		        v_.Give_data<hexahedronEl>(id,i-1,j+1,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,ROrt,MOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j+1,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,ROrt,MOrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j+1,k  ,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,ROrt,MOrt)] +  \
		        \
		        v_.Give_data<hexahedronEl>(id,i-1,j-1,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,LOrt,ROrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j-1,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,LOrt,ROrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j-1,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,LOrt,ROrt)] +  \
		        \
		        v_.Give_data<hexahedronEl>(id,i-1,j  ,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,MOrt,ROrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j  ,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,MOrt,ROrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j  ,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,MOrt,ROrt)] +  \
		        \
		        v_.Give_data<hexahedronEl>(id,i-1,j+1,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(LOrt,ROrt,ROrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i,  j+1,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(MOrt,ROrt,ROrt)] +  \
		        v_.Give_data<hexahedronEl>(id,i+1,j+1,k+1,Nx,Ny) * sten_m_hexahedra[id][ig+trans27(ROrt,ROrt,ROrt)];   \
	}                                                                                                           \
	\
	template <>                                                                                                 \
	template <>                                                                                                 \
	inline R Expr_Sten_matrix_transpose<A,B,C>::Give_data<quadrangleEl>(params_in) const {                               \
		int u,v;                                                                                                  \
		dir3D_sons ce;                                                                                              \
		R sum;                                                                                                    \
		sum = 0.0;                                                                                                \
		for(v=0;v<4;++v) {                                                                                        \
			ce = opposite3D((dir3D_sons)v);                                                                           \
			for(u=0;u<4;++u)                                                                                        \
				sum = sum +                                                                                           \
				      loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles                                       \
				                                                                /* transpose:  vertauscht u,v */        \
				                                     (x1DCoord(ce)-1+i,y1DCoord(ce)-1+j)*32+8*u+v] *                 \
				      v_.Give_data_quadrangle_middle(id,x1DCoord(ce)+x1DCoord((dir3D_sons)u)-1+i,                       \
				                                     y1DCoord(ce)+y1DCoord((dir3D_sons)u)-1+j,Nx) +                  \
				      loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles                                       \
				                                                               /* transpose:  vertauscht u,v */        \
				                                     (x1DCoord(ce)-1+i,y1DCoord(ce)-1+j)*32+8*u+(v+WSTdir3D)] *           \
				      v_.Give_data_quadrangle_interior(id,x1DCoord(ce)+x1DCoord((dir3D_sons)u)-1+i,                     \
				                                       y1DCoord(ce)+y1DCoord((dir3D_sons)u)-1+j,Nx);                 \
		}                                                                                                         \
		if(Give_blockgrid()->Give_unstructured_grid()->Give_quadrangle(id)->Give_exists_exterior()==false)        \
			return sum;                                                                                           \
		for(v=0;v<4;++v) {                                                                                        \
			ce = opposite3D((dir3D_sons)v);                                                                           \
			for(u=0;u<4;++u)                                                                                        \
				sum = sum +                                                                                           \
				      loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles                                       \
				                                                               /* transpose:  vertauscht u,v */        \
				                                     (x1DCoord(ce)-1+i,y1DCoord(ce)-1+j)*32+8*u+v] *                   \
				      v_.Give_data_quadrangle_middle(id,x1DCoord(ce)+x1DCoord((dir3D_sons)u)-1+i,                       \
				                                     y1DCoord(ce)+y1DCoord((dir3D_sons)u)-1+j,Nx) +                  \
				      loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles                                       \
				                                                              /* transpose:  vertauscht u,v */        \
				                                     (x1DCoord(ce)-1+i,y1DCoord(ce)-1+j)*32+8*u+(v+WSTdir3D)] *           \
				      v_.Give_data_quadrangle_exterior(id,x1DCoord(ce)+x1DCoord((dir3D_sons)u)-1+i,                     \
				                                       y1DCoord(ce)+y1DCoord((dir3D_sons)u)-1+j,Nx);                 \
		}                                                                                                         \
		return sum; }                                                                                             \
	\
	template <>                                                                                                 \
	template <>                                                                                                 \
	inline R Expr_Sten_matrix_transpose<A,B,C>::Give_data<edgeEl>(params_in) const {                                     \
		int u,v;                                                                                                  \
		int num;                                                                                                  \
		int id_ND, id_ST, id_NT;                                                                                  \
		dir3D_sons ce;                                                                                              \
		R sum;                                                                                                    \
		sum = 0.0;                                                                                                \
		int number_neighbor_hex =                                                                                 \
		        Give_blockgrid()->Give_unstructured_grid()->Give_edge(id)->Give_number_neighbors_hex();             \
		for(num=0;num<number_neighbor_hex;++num){                                                                 \
			id_ND = Give_blockgrid()->Give_unstructured_grid()->Give_edge(id)->Give_loc_id_hex (num,ND);            \
			id_ST = Give_blockgrid()->Give_unstructured_grid()->Give_edge(id)->Give_loc_id_hex (num,ST);            \
			id_NT = Give_blockgrid()->Give_unstructured_grid()->Give_edge(id)->Give_loc_id_hex (num,NT);            \
			for(v=0;v<2;++v) {                                                                                      \
				ce = opposite3D((dir3D_sons)v);                                                                         \
				for(u=0;u<2;++u)                                                                                      \
					sum = sum +                                                                                          \
					                                              /* transpose:  vertauscht u,v */        \
					      loc_m_neighbor_edges[id][num][(x1DCoord(ce)-1+i)*16+8*u+v] *                                   \
					      v_.Give_data_edge(id,x1DCoord(ce)+x1DCoord((dir3D_sons)u)-1+i) +                                 \
					                                               /* transpose:  vertauscht u,v */        \
					      loc_m_neighbor_edges[id][num][(x1DCoord(ce)-1+i)*16+8*u+(v+WNDdir3D)] *                            \
					      v_.Give_data_neighbor_edge(id,id_ND,x1DCoord(ce)+x1DCoord((dir3D_sons)u)-1+i) +                  \
					                                               /* transpose:  vertauscht u,v */        \
					      loc_m_neighbor_edges[id][num][(x1DCoord(ce)-1+i)*16+8*u+(v+WSTdir3D)] *                            \
					      v_.Give_data_neighbor_edge(id,id_ST,x1DCoord(ce)+x1DCoord((dir3D_sons)u)-1+i) +                  \
					                                               /* transpose:  vertauscht u,v */        \
					      loc_m_neighbor_edges[id][num][(x1DCoord(ce)-1+i)*16+8*u+(v+WNTdir3D)] *                            \
					      v_.Give_data_neighbor_edge(id,id_NT,x1DCoord(ce)+x1DCoord((dir3D_sons)u)-1+i);                   \
			}                                                                                                       \
		}                                                                                                         \
		return sum; }                                                                                             \
	\
	template <>                                                                                                 \
	template <>                                                                                                 \
	inline R Expr_Sten_matrix_transpose<A,B,C>::Give_data<pointEl>(params_in) const {	                            \
		int num;                                                                                                  \
		int id_SED, id_NWD, id_NED, id_SWT, id_SET, id_NWT, id_NET;                                               \
		dir3D_sons ce;                                                                                              \
		R sum;                                                                                                    \
		sum = 0.0;                                                                                                \
		int number_neighbor_hex =                                                                                 \
		        Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->Give_number_neighbors_hex();            \
		for(num=0;num<number_neighbor_hex;++num){                                                                 \
			id_SED = Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->Give_loc_id_hex (num,SED);         \
			id_NWD = Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->Give_loc_id_hex (num,NWD);         \
			id_NED = Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->Give_loc_id_hex (num,NED);         \
			id_SWT = Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->Give_loc_id_hex (num,SWT);         \
			id_SET = Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->Give_loc_id_hex (num,SET);         \
			id_NWT = Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->Give_loc_id_hex (num,NWT);         \
			id_NET = Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->Give_loc_id_hex (num,NET);         \
			ce = ENTdir3D;                                                                                              \
			sum = sum +                                                                                             \
			      loc_m_neighbor_points[id][num*8+WSDdir3D] * v_.Give_data_point(id) +                               \
			      loc_m_neighbor_points[id][num*8+ESDdir3D] * v_.Give_data_neighbor_point(id,id_SED) +               \
			      loc_m_neighbor_points[id][num*8+WNDdir3D] * v_.Give_data_neighbor_point(id,id_NWD) +               \
			      loc_m_neighbor_points[id][num*8+ENDdir3D] * v_.Give_data_neighbor_point(id,id_NED) +               \
			      loc_m_neighbor_points[id][num*8+WSTdir3D] * v_.Give_data_neighbor_point(id,id_SWT) +               \
			      loc_m_neighbor_points[id][num*8+ESTdir3D] * v_.Give_data_neighbor_point(id,id_SET) +               \
			      loc_m_neighbor_points[id][num*8+WNTdir3D] * v_.Give_data_neighbor_point(id,id_NWT) +               \
			      loc_m_neighbor_points[id][num*8+ENTdir3D] * v_.Give_data_neighbor_point(id,id_NET);                \
		}                                                                                                         \
		return sum; }


Give_functions_Expr_Sten_transpose ( double, double, double, Stencil_variable )
Give_functions_Expr_Sten_transpose ( std::complex<double>, std::complex<double>, double, Stencil_variable )
Give_functions_Expr_Sten_transpose ( std::complex<double>, std::complex<double>, std::complex<double>, Stencil_variable )

Give_functions_Expr_Sten_transpose ( double, double, double, Stencil_constant_z )
Give_functions_Expr_Sten_transpose ( std::complex<double>, std::complex<double>, double, Stencil_constant_z )
Give_functions_Expr_Sten_transpose ( std::complex<double>, std::complex<double>, std::complex<double>, Stencil_constant_z )

Give_functions_Expr_Sten_transpose ( double, double, double, Stencil_constant )
Give_functions_Expr_Sten_transpose ( std::complex<double>, std::complex<double>, double, Stencil_constant )
Give_functions_Expr_Sten_transpose ( std::complex<double>, std::complex<double>, std::complex<double>, Stencil_constant )

Give_functions_Expr_Sten_transpose ( double, double, double, Not_constant_direction_marker )
Give_functions_Expr_Sten_transpose ( std::complex<double>, std::complex<double>, double, Not_constant_direction_marker )
Give_functions_Expr_Sten_transpose ( std::complex<double>, std::complex<double>, std::complex<double>, Not_constant_direction_marker )



/*
template <typename TYPE2, class STENCIL_TYP>

class Expr_Diag_Sten_matrix_transpose :
			public Expr<Expr_Diag_Sten_matrix<TYPE2, STENCIL_TYP> > {
			
	public:
		Expr_Diag_Sten_matrix_transpose ( TYPE2** lm_hex, TYPE2** lm_quad_in, TYPE2** lm_quad_ex,
		                        TYPE2*** lm_neigh_ed, TYPE2** lm_neigh_po, Blockgrid* blockgrid_,
		                        STENCIL_TYP st ) :
				sten_m_hexahedra ( lm_hex ),
				loc_m_interior_quadrangles ( lm_quad_in ),
				loc_m_exterior_quadrangles ( lm_quad_ex ),
				loc_m_neighbor_edges ( lm_neigh_ed ),
				loc_m_neighbor_points ( lm_neigh_po ),
				blockgrid ( blockgrid_ ),
				sten_typ ( st ) {}
				
		Blockgrid* Give_blockgrid() const {
			return blockgrid;
		};
		
		stencil_typ Give_stencil_typ() const { return yes_stencil; }
		control_typ Give_control_typ() const { return thread_save; }
		
		typedef TYPE2 Result;
		
		template <elementTyp TYP_EL>
		inline TYPE2 Give_data ( params_in ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {}
		
	private:
		TYPE2**  sten_m_hexahedra;                  // num_hexahedra
		
		TYPE2**  loc_m_interior_quadrangles;       // num_quadrangles
		TYPE2**  loc_m_exterior_quadrangles;       // num_quadrangles
		TYPE2*** loc_m_neighbor_edges;             // num_edges
		TYPE2**  loc_m_neighbor_points;            // num_points
		
		Blockgrid* blockgrid;
		
		STENCIL_TYP sten_typ;
};

#define Give_functions_Diag_Expr_Sten_transpose(A,B)	                               		         \
	template <>                                                                                      \
	template <>                                                                                      \
	inline A Expr_Diag_Sten_matrix_transpose<A,B>::Give_data<hexahedronEl>(params_in) const {               \
		return sten_m_hexahedra[id][sten_typ.stenIndex(i,j,k,Nx,Ny,id)  + trans27(MOrt,MOrt,MOrt)];          \
	}                                                                                                \
	\
	template <>                                                                                      \
	template <>                                                                                      \
	inline A Expr_Diag_Sten_matrix_transpose<A,B>::Give_data<quadrangleEl>(params_in) const {                 \
		if(Give_blockgrid()->Give_unstructured_grid()->Give_quadrangle(id)->Give_exists_exterior())    \
			return loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles(i  ,j  )*32   ]             \
			       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles(i-1,j  )*32+ 9]           \
			       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles(i  ,j-1)*32+18]           \
			       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles(i-1,j-1)*32+27]           \
			       + loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles(i  ,j  )*32   ]           \
			       + loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles(i-1,j  )*32+ 9]           \
			       + loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles(i  ,j-1)*32+18]           \
			       + loc_m_exterior_quadrangles[id][Ind_loc_matrix_quadrangles(i-1,j-1)*32+27];          \
		return loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles(i  ,j  )*32   ]             \
		       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles(i-1,j  )*32+ 9]           \
		       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles(i  ,j-1)*32+18]           \
		       + loc_m_interior_quadrangles[id][Ind_loc_matrix_quadrangles(i-1,j-1)*32+27];          \
	}                                                                                                \
	\
	template <>                                                                                      \
	template <>                                                                                      \
	inline A Expr_Diag_Sten_matrix_transpose<A,B>::Give_data<edgeEl>(params_in) const {                       \
		int number_neighbor_hex =                                                                      \
		        blockgrid->Give_unstructured_grid()->Give_edge(id)->Give_number_neighbors_hex();         \
		A sum = 0.0;                                                                                   \
		for(int num=0;num<number_neighbor_hex;++num)                                                   \
			sum += loc_m_neighbor_edges[id][num][ i   *16  ]                                             \
			       + loc_m_neighbor_edges[id][num][(i-1)*16+9];                                            \
		return sum;                                                                                    \
	}                                                                                                \
	\
	template <>                                                                                      \
	template <>                                                                                      \
	inline A Expr_Diag_Sten_matrix_transpose<A,B>::Give_data<pointEl>(params_in) const {                      \
		int number_neighbor_hex =                                                                      \
		        blockgrid->Give_unstructured_grid()->Give_point(id)->Give_number_neighbors_hex();        \
		A sum = 0.0;                                                                                   \
		for(int num=0;num<number_neighbor_hex;++num)                                                   \
			sum += loc_m_neighbor_points[id][num*8];                                                   \
		return sum; }





Give_functions_Diag_Expr_Sten_transpose ( double, Stencil_variable )
Give_functions_Diag_Expr_Sten_transpose ( std::complex<double>, Stencil_variable )

Give_functions_Diag_Expr_Sten_transpose ( double, Stencil_constant_z )
Give_functions_Diag_Expr_Sten_transpose ( std::complex<double>, Stencil_constant_z )

Give_functions_Diag_Expr_Sten_transpose ( double, Stencil_constant )
Give_functions_Diag_Expr_Sten_transpose ( std::complex<double>, Stencil_constant )

Give_functions_Diag_Expr_Sten_transpose ( double, Not_constant_direction_marker )
Give_functions_Diag_Expr_Sten_transpose ( std::complex<double>, Not_constant_direction_marker )
*/

//////////////////////////////////////////////
// 3. class Stencil
//////////////////////////////////////////////


template <typename TYPE2, class STENCIL_TYP>

class Sten_matrix : public Expr<Sten_matrix<TYPE2, STENCIL_TYP> > {

	public:
	
		typedef TYPE2 Result;
		
		Sten_matrix ( Blockgrid& bg_ );
		Sten_matrix ( Blockgrid& bg_, STENCIL_TYP& not_constant_directions );
		~Sten_matrix();
		
		void Print();
		
		Expr_Diag_Sten_matrix<TYPE2, STENCIL_TYP> diag () const;
		
		template <typename TYPE1>
		Expr_Sten_matrix<TYPE1, TYPE2, STENCIL_TYP> operator() ( const Variable<TYPE1>& v ) const;
		
		template <typename TYPE1>
		Expr_Sten_matrix_transpose<TYPE1, TYPE2, STENCIL_TYP> transposeBoundary( const Variable<TYPE1>& v ) const;
		
		template <sten_part_typ TYP_EL>
		inline TYPE2 Give_data_sten ( params_in_sten ) const;
		
		
		template <class A>
		void operator= ( const Expr<A>& a );
		
		void operator= ( const Sten_matrix& a );
		
		void operator= ( const Local_stiffness_matrix<TYPE2>& local_stiffness_matrix );
		
		
		Blockgrid* Give_blockgrid() const { return blockgrid; };
		

		
		void PrintStencilsToFile() const;
		
		TYPE2** Give_pointer_sten_hex() { return  sten_m_hexahedra; }
		
	private:
		void Allocate_data();
		
		TYPE2**  sten_m_hexahedra;                 // num_hexahedra
		TYPE2**  loc_m_interior_quadrangles;       // num_quadrangles
		TYPE2**  loc_m_exterior_quadrangles;       // num_quadrangles
		TYPE2*** loc_m_neighbor_edges;             // num_edges
		TYPE2**  loc_m_neighbor_points;            // num_points
		
		Blockgrid* blockgrid;
		
		STENCIL_TYP sten_typ;
		
		// for parallel
		int my_rank;
		MPI_Comm comm;
};


#define Give_functions(A,B)  \
	template <>                                                                                 \
	template <>                                                                                 \
	inline A Sten_matrix<A, B>::Give_data_sten<point_sten_part>(params_in_sten) const {         \
		return loc_m_neighbor_points[id][i]; }                                                    \
	template <>                                                                                 \
	template <>                                                                                 \
	inline A Sten_matrix<A, B>::Give_data_sten<edge_sten_part>(params_in_sten) const {          \
		return loc_m_neighbor_edges[id][num][i]; }                                                \
	template <>                                                                                 \
	template <>                                                                                 \
	inline A Sten_matrix<A, B>::Give_data_sten<quadrangle_in_sten_part>(params_in_sten) const { \
		return loc_m_interior_quadrangles[id][i]; }                                               \
	template <>                                                                                 \
	template <>                                                                                 \
	inline A Sten_matrix<A, B>::Give_data_sten<quadrangle_ex_sten_part>(params_in_sten) const { \
		return loc_m_exterior_quadrangles[id][i]; }                                               \
	template <>                                                                                 \
	template <>                                                                                 \
	inline A Sten_matrix<A, B>::Give_data_sten<hexahedron_sten_part>(params_in_sten) const {    \
		return sten_m_hexahedra[id][i]; }


Give_functions ( double,              Stencil_variable )
Give_functions ( std::complex<double>, Stencil_variable )

Give_functions ( double,              Stencil_constant_z )
Give_functions ( std::complex<double>, Stencil_constant_z )

Give_functions ( double,              Stencil_constant )
Give_functions ( std::complex<double>, Stencil_constant )

Give_functions ( double,              Not_constant_direction_marker )
Give_functions ( std::complex<double>, Not_constant_direction_marker )


//////////////////
// Implementierung von Memberfunktionen
//////////////////


template <typename TYPE2, class STENCIL_TYP>
template <typename TYPE1>
Expr_Sten_matrix<TYPE1, TYPE2, STENCIL_TYP> Sten_matrix<TYPE2, STENCIL_TYP>::operator()
( const Variable<TYPE1>& v ) const {
	if ( Give_blockgrid() != v.Give_blockgrid() )
		std::cout << " wrong grids in Sten_matrix( Variable v)! "
		<< std::endl;
		
	return Expr_Sten_matrix<TYPE1, TYPE2, STENCIL_TYP> ( v, sten_m_hexahedra,
	        loc_m_interior_quadrangles,
	        loc_m_exterior_quadrangles,
	        loc_m_neighbor_edges, loc_m_neighbor_points,
	        sten_typ );
}


template <typename TYPE2, class STENCIL_TYP>
template <typename TYPE1>
Expr_Sten_matrix_transpose<TYPE1, TYPE2, STENCIL_TYP> Sten_matrix<TYPE2, STENCIL_TYP>::transposeBoundary( const Variable<TYPE1>& v ) const {
	if ( Give_blockgrid() != v.Give_blockgrid() )
		cout << " wrong grids in Sten_matrix( Variable v)! "
		<< endl;
		
	return Expr_Sten_matrix_transpose<TYPE1, TYPE2, STENCIL_TYP> ( v, sten_m_hexahedra,
	        loc_m_interior_quadrangles,
	        loc_m_exterior_quadrangles,
	        loc_m_neighbor_edges, loc_m_neighbor_points,
	        sten_typ );
}



//////////////////////////////////////////////
// 4. Berechnung der Stencils
//////////////////////////////////////////////


template <typename TYPE2, class STENCIL_TYP>
void Sten_matrix<TYPE2, STENCIL_TYP>::operator= ( const Local_stiffness_matrix<TYPE2>& local_stiffness_matrix ) {
	int id;
	int Nx, Ny, Nz, N_total;
	
	
	Unstructured_grid* ug = blockgrid->Give_unstructured_grid();
	
	///////////////////////////////////////////
	// Calculate stencils in hexaeder
	
	for ( id = 0;id < ug->Give_number_hexahedra();++id ) {
	
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) ) {
		
			Nx = blockgrid->Give_Nx_hexahedron ( id );
			Ny = blockgrid->Give_Ny_hexahedron ( id );
			Nz = blockgrid->Give_Nz_hexahedron ( id );
			
			if ( sten_typ.Not_const_x ( id ) == false ) Nx = 2;
			
			if ( sten_typ.Not_const_y ( id ) == false ) Ny = 2;
			
			if ( sten_typ.Not_const_z ( id ) == false ) Nz = 2;

			//neu eingebaut!!!
			#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake)  if(UGBlocks::useOpenMP)
			for(int k = 0;k < Nz - 1;++k )
				for(int j = 0;j < Ny - 1;++j )
					for(int i = 0;i < Nx - 1;++i ) {
						int gi = ( i + ( Nx - 1 ) * ( j + ( Ny - 1 ) * k ) ) * 27;
						// put zero
						
						for(int s = 0;s < 27;++s )
							sten_m_hexahedra[id][gi+s] = 0.0;
							
							
						// add stencils
						int u, v;
						
						dir3D_sons ce;
						
						for ( v = 0;v < 8;++v ) { // test function v and cell opposite3D(v)
							ce = opposite3D ( ( dir3D_sons ) v );
							
							for ( u = 0;u < 8;++u )
								sten_m_hexahedra[id][gi+trans27 ( Ort1D ( x1DCoord ( ce ) +x1DCoord ( ( dir3D_sons ) u ) ),
								                                  Ort1D ( y1DCoord ( ce ) +y1DCoord ( ( dir3D_sons ) u ) ),
								                                  Ort1D ( z1DCoord ( ce ) +z1DCoord ( ( dir3D_sons ) u ) ) ) ] +=
								                                      local_stiffness_matrix.Give_loc_m_hexahedra() [id][Ind_loc_matrix_hexahedra ( x1DCoord ( ce ) +i,
								                                              y1DCoord ( ce ) +j,
								                                              z1DCoord ( ce ) +k ) *64+8*v+u];
						}
					}
		}
	}
	
	
	// copy local stiffness matrices quadrangle
	for( id = 0; id < ug->Give_number_quadrangles(); ++id ) {
	
		if ( ug->Give_quadrangle ( id )->my_object ( my_rank ) ) {
		
			Nx = blockgrid->Give_Nx_quadrangle ( id );
			Ny = blockgrid->Give_Ny_quadrangle ( id );
			N_total = Nx * Ny * 32;
			
			//neu eingebaut!!!
			//#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake)  if(UGBlocks::useOpenMP)
			for(int i = 0;i < N_total;++i )
				loc_m_interior_quadrangles[id][i] = local_stiffness_matrix.Give_loc_m_interior_quadrangles ( id ) [i];
				
			if ( ug->Give_quadrangle ( id )->Give_exists_exterior() ) {
                                //neu eingebaut!!!
			        //#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake)  if(UGBlocks::useOpenMP)			  
				for(int i = 0;i < N_total;++i )
					loc_m_exterior_quadrangles[id][i] = local_stiffness_matrix.Give_loc_m_exterior_quadrangles ( id ) [i];
			}
		}
	}
	
	// copy local stiffness matrices edges
	for(id = 0;id < ug->Give_number_edges();++id ) {
	
		if(ug->Give_edge ( id )->my_object ( my_rank ) ) {
		
			for ( int num = 0;num < ug->Give_edge ( id )->Give_number_neighbors_hex();++num ) {
				Nx = blockgrid->Give_Nx_edge ( id );
				N_total = Nx  * 16;
				
				for (int i = 0;i < N_total;++i )
					loc_m_neighbor_edges[id][num][i] = local_stiffness_matrix.Give_loc_m_neighbor_edges ( id ) [num][i];
			}
		}
	}
	
	//  copy local stiffness matrices points
	for(id = 0;id < ug->Give_number_points();++id ) {
	
		if ( ug->Give_point ( id )->my_object ( my_rank ) ) {
		
			N_total = ug->Give_point ( id )->Give_number_neighbors_hex() * 8;
                        //loc_m_neighbor_points[id] = new TYPE2[N_total];
			
			for(int i = 0;i < N_total;++i )
				loc_m_neighbor_points[id][i] = local_stiffness_matrix.Give_loc_m_neighbor_points ( id ) [i];
		}
	}
}


template <typename DTyp, class STENCIL_TYP>
template <class A>
void Sten_matrix<DTyp, STENCIL_TYP>::operator= ( const Expr<A>& a ) {
	int num, id, N_total, Nx, Ny, Nz;
	
	Unstructured_grid* ug = blockgrid->Give_unstructured_grid();
	
	const A& ao ( a );
	
	// hexahedra
	
	for ( id = 0;id < ug->Give_number_hexahedra();++id ) {
	
		if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) ) {
		
			Nx = blockgrid->Give_Nx_hexahedron ( id );
			Ny = blockgrid->Give_Ny_hexahedron ( id );
			Nz = blockgrid->Give_Nz_hexahedron ( id );
			
			if ( sten_typ.Not_const_x ( id ) == false ) Nx = 2;
			
			if ( sten_typ.Not_const_y ( id ) == false ) Ny = 2;
			
			if ( sten_typ.Not_const_z ( id ) == false ) Nz = 2;
			
			N_total = ( Nx - 1 ) * ( Ny - 1 ) * ( Nz - 1 ) * 27;
			
                        //#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
            for(int i = 0;i < N_total;++i ) sten_m_hexahedra[id][i] = ao.template Give_data_sten<hexahedron_sten_part> ( id, 0, i );
		}
	}
	
	// quadrangles
	for ( id = 0;id < ug->Give_number_quadrangles();++id ) {
	
		if ( ug->Give_quadrangle ( id )->my_object ( my_rank ) ) {
		
			Nx = blockgrid->Give_Nx_quadrangle ( id );
			Ny = blockgrid->Give_Ny_quadrangle ( id );
			N_total = Nx * Ny * 32;
			

                        //neu eingebaut!!!
			//#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake)  if(UGBlocks::useOpenMP)			
			for(int i = 0;i < N_total;++i )
                loc_m_interior_quadrangles[id][i] = ao.template Give_data_sten<quadrangle_in_sten_part> ( id, 0, i );
				
			if ( ug->Give_quadrangle ( id )->Give_exists_exterior() ) {
                                //neu eingebaut!!!
			        //#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake)  if(UGBlocks::useOpenMP)			  
				for(int i = 0;i < N_total;++i )
                    loc_m_exterior_quadrangles[id][i] = ao.template Give_data_sten<quadrangle_ex_sten_part> ( id, 0, i );
			}
		}
	}
	
	// edges
	for ( id = 0;id < ug->Give_number_edges();++id ) {
	
		if ( ug->Give_edge ( id )->my_object ( my_rank ) ) {
		
			for ( num = 0;num < ug->Give_edge ( id )->Give_number_neighbors_hex();++num ) {
				Nx = blockgrid->Give_Nx_edge ( id );
				N_total = Nx  * 16;
//////#pragma omp parallel for private(i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)

				for(int i = 0;i < N_total;++i )
                    loc_m_neighbor_edges[id][num][i] = ao.template Give_data_sten<edge_sten_part> ( id, num, i );
			}
		}
	}
	
	// points
	for ( id = 0;id < ug->Give_number_points();++id ) {
	
		if ( ug->Give_point ( id )->my_object ( my_rank ) ) {
		
			N_total = ug->Give_point ( id )->Give_number_neighbors_hex() * 8;
//////#pragma omp parallel for private(i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)

			for(int i = 0;i < N_total;++i )
                loc_m_neighbor_points[id][i] = ao.template Give_data_sten<point_sten_part> ( id, 0, i );
		}
	}
}

template <typename TYPE2, class STENCIL_TYP>
void Sten_matrix<TYPE2, STENCIL_TYP>::operator= ( const Sten_matrix & a ) {
	if ( this == &a )
		return;
		
	const Expr<Sten_matrix>& b ( a );
	
	operator= ( b );
}

#endif // STENOP_H
