/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#ifndef OUTPUT_H
#define OUTPUT_H
#ifdef _VTK
#include <iostream>
#include <fstream>
#include <string>
#include <vector>

#include <QtCore>
#include <QApplication>

#ifndef _WIN32
#include <pthread.h>
#else
#include <windows.h>
#include <strsafe.h>
#endif // _WIN32

#include <vtkActor.h>
#include <vtkCubeAxesActor2D.h>
#include <vtkContourFilter.h>
#include <vtkDataSet.h>
#include <vtkFloatArray.h>
#include <vtkImageData.h>
#include <vtkGaussianSplatter.h>
#include <vtkPointData.h>
#include <vtkPoints.h>
#include <vtkPolyDataMapper.h>
#include <vtkDataSetMapper.h>
#include <vtkProperty.h>
#include <vtkRenderWindow.h>
#include <vtkRenderWindowInteractor.h>
#include <vtkRenderer.h>
#include <vtkTubeFilter.h>
#include <vtkUnstructuredGrid.h>
#include <vtkLookupTable.h>
#include <vtkOutlineFilter.h>
#include <vtkImplicitPlaneWidget.h>
#include <vtk3DWidget.h>
#include <vtkLODActor.h>
#include <vtkInteractorEventRecorder.h>
#include <vtkDataSetToPolyDataFilter.h>
#include <vtkCommand.h>
#include <vtkClipDataSet.h>
#include <vtkPlane.h>
#include <vtkCallbackCommand.h>
#include <vtkScalarBarActor.h>
#include <vtkImplicitDataSet.h>
#include <vtkUnstructuredGridReader.h>
#include <vtkScalarBarWidget.h>
#include <vtkCornerAnnotation.h>
#include <vtkTextProperty.h>
#include <vtkCamera.h>


/***
@TODO Output Klasse und viele Gui-Klassen weg
**/

class Output
{
public:
    std::string getName() {return name;}
    virtual void display() = 0;
    virtual ~Output();
protected:
    Output(std::string name_) : name(name_) {};
    Output() {};
    std::string name;
};

class OutputVTK : public Output
{
public:
    OutputVTK(std::vector<std::vector<double> > Points,
        std::vector<std::vector<double> > Cells,
        std::vector<double> Values,
        int Celltype,
        std::string Name);
    OutputVTK(const char* vtkFilepath, std::string Name);
    virtual ~OutputVTK();
    double getValue(double x,double y, double z);
    void display();
    std::string getPath();

private:
    void Init();
    void* t_display();
#ifndef _WIN32
    static void* t_displayHelper(void* This);
    // FIXME: need some thread handle for cleanup (doh!)
#else
    static DWORD WINAPI t_displayHelper( LPVOID lpParam );
    HANDLE m_threadHandle;
#endif // _WIN32
    std::vector<std::vector<double> > v_Points;
    std::vector<std::vector<double> > v_Cells;
    std::vector<double> v_Values;
    int celltype;
    std::string path;
    double min,max;
    float xyz[3];
    double maxPoint[3];
    double color[3];
    int i, npts;
    bool fromFile;


    vtkImplicitDataSet *implicitDataSet;

    ////dataset
    vtkUnstructuredGrid *dataSet;
    vtkPoints *newPts;
    vtkFloatArray *newScalars;
    vtkLookupTable *lut;
    ////normal plot
    vtkDataSetMapper *mapper;
    vtkActor *actor;
    ////scalar bar
    vtkScalarBarActor *scalarbar;
    vtkScalarBarWidget * scalarbarWidget;
    /////legend
    vtkCornerAnnotation *legend;
    /////axes
    vtkCubeAxesActor2D *axes;
    //////outline
    vtkOutlineFilter *outline;
    vtkPolyDataMapper *mapperOutline;
    vtkActor *actorOutline;
    //////plane widget
    vtkImplicitPlaneWidget *planeWidget;
    /////cut to plane
    vtkClipDataSet *plane;
    vtkPlane *implicitFunction;
    vtkLODActor *selectActor;
    vtkDataSetMapper *selectMapper;
    vtkUnstructuredGridReader* reader;

    /////window
    vtkRenderer *renderer;
    vtkRenderWindow *renWin;
    vtkRenderWindowInteractor *iren;

    QMutex m_mutex;
};

#endif // _VTK
#endif /* OUTPUT_H */
