/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#ifndef VTKREADER_H
#define VTKREADER_H
#include <vector>

#include "../mympi.h"
#include "../abbrevi.h"
#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../basics/basic.h"
#include "../grid/elements.h"
#include "../grid/parti.h"
#include "../grid/ug.h"
#include "../grid/blockgrid.h"
#include "../grid/marker.h"
#include "parallel.h"
#include "extemp.h"
#include "variable.h"
#include "variable_cc.h"
#include "print_var_vtk_cc.h"
#include "phaseshift_cc.h"
#include "QFileInfo"


class VTK_Reader {
public:  
  /**
   * liest file mit DateiName und erzeugt dazu blockgrid, ug, variable
   * @param PfadUndDateiName  QString der Pfad und dann Dateiname enthält
   * @param constructionShiftOfZ verschiebt das VTK file in z-Richtung, sodass mehere nebeneinander gerechnet werden können
   **/
  VTK_Reader(QString PfadUndDateiName, double constructionShiftOfZ = 0.0);

  /**
   * liest zweites file 
   * @param PfadUndDateiName  QString der Pfad und dann Dateiname enthält
   * @return NULL falls zweites file anderes ug oder blockgrid, sonst Variable mit eingelesen Daten 
   **/
  Variable<double>* readNext(QString PfadUndDateiName);
  
  
  Blockgrid* give_blockgrid()        const { return blockgrid; }
  Unstructured_grid* give_ug()       const { return ug; }
  Variable<double>*  give_first_variable() const { return variable; }
  
  TypeOfUG findUGType(QStringList splitString, int stringElement);
  std::vector<int> readNumberPoints(QStringList splitString);
  std::vector<double> readConstructionParameters(QStringList splitString);
  
private:

  Blockgrid*       blockgrid;
  UG_WithZdirection*      ug;
  TypeOfUG typGrid;
  int degreesRead, numberConstructionParametersRead;
  Variable<double>* variable;
};



#endif // VTKREADER_H
