//    GUI for solid state laser simulation:
//    Copyright (C) 2016  Christoph Pflaum
//    This program is not a free software;
//    do not redistribute it and/or modify it
//    To use this software you need
//    a permission by Christoph Pflaum
/////////////////////////////////////////////////


#ifndef GAUSSSEIDEL_H
#define GAUSSSEIDEL_H

#include "../source/extemp/variable2D.h"
#include "../source/pde_op/diffop2D.h"

/**
 Gauss Seidel iterativer Löser für \f$ A x = b \f$
 **/
template<class DTyp>
class  GaussSeidelSolver : public IterativeSolverAbstract<DTyp>  {
   public:
         /** 
	  * @param blockgrid Gitter auf dem die Gleichung gelösst werden soll
	  **/
         GaussSeidelSolver(Blockgrid2D& blockgrid_) : IterativeSolverAbstract<DTyp>(blockgrid_) {}	
	
	/**
	 * solves the linear equation system \f$ A x = b \f$
	 **/    
	void solve(Local_stiffness_matrix2D<DTyp>& A, Variable2D<DTyp>& b, Variable2D<DTyp>& x);
	
	bool isResiduumCalculated() { return false; }
};

///////////////////////////////
// Implementation
///////////////////////////////

template<class DTyp>
void GaussSeidelSolver<DTyp>::solve(Local_stiffness_matrix2D<DTyp>& A, Variable2D<DTyp>& b, Variable2D<DTyp>& x) {
  this->numberIterationsPerformed = this->max_iter;
  if(this->stopByEps==false) {
    if(this->inner_points==NULL) {
       for(int i=0;i< this->max_iter;++i) {  
           x = x - (A(x) -b) / A.diag();
       }
    }
    else {
      assert(false);
    }
  }
  else assert(false);
}



#endif //ITERATIVESOLVER_H
