/**********************************************************************************
 * Copyright 2016 Christoph Pflaum and Rainer Hartmann
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

//#define _USE_MATH_DEFINES
#include <cmath>
#include <complex>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
using std::vector;


#ifndef GMRES_ALLG_H
#define GMRES_ALLG_H

/**
 GMRES Löser für \f$ A x = b \f$
 **/
template<class DTyp>
  class GmRes : public IterativeSolverAbstract<DTyp>  {

  public:
    /** 
     * @param blockgrid Gitter auf dem die Gleichung gelösst werden soll
     **/
    GmRes(Blockgrid2D& blockgrid);
    //~IterativeSolverAbstract() { };
    ~GmRes();

    /** 
     * set maximum size of Kryloc-space
     * @param numberRestart_ important GMRES parameter, which is default 100, changing this parameter changes memory allocation
     **/
    void setRestartNumber(int numberRestart_);
    
    /**
     * solves the linear equation system \f$ A x = b \f$
     **/    
    void solve(Local_stiffness_matrix2D<DTyp>& A, Variable2D<DTyp>& b, Variable2D<DTyp>& x);

    bool isResiduumCalculated() { return true; }    
  private:
    Variable2D<DTyp> r;
    std::vector< Variable2D<DTyp>* > Q;

    int Restart;

    DTyp** H;     // size [Restart + 1][Restart]
    DTyp* C;      // size [Restart]
    DTyp* S;      // size [Restart]
    DTyp* y;      // size [Restart]    
    DTyp* gamma;  // size [Restart+1]

    void GivensRotation(const DTyp& x1, DTyp& x2, DTyp& c, DTyp& s);    
};




/*
 * für System
template<class DTyp>
  class GmRes2
  {
  private:
    std::vector< Variable2D< DTyp > > Q_x;
    std::vector< Variable2D< DTyp > > Q_y;
    Blockgrid2D& blockgrid;

    const double eps;
    const int max_iter;

    DTyp H[RESTART + 1][RESTART];
    DTyp C[RESTART];
    DTyp S[RESTART];
    DTyp gamma[RESTART+1];

    double L2Norm(Variable2D<DTyp>& x);
    double L2InnerNorm(Variable2D<DTyp>& x, Unstructured2DGrid_Marker2D& inner_points);

    void GivensRotation(const DTyp& x1, DTyp& x2, DTyp& c, DTyp& s);

  public:
    GmRes2(Blockgrid2D& blockgrid);
    GmRes2(Blockgrid2D& blockgrid, double eps, int max_iter = 500);

    double solve(Local_stiffness_matrix2D<DTyp>& A_xx, Local_stiffness_matrix2D<DTyp>& A_xy, Local_stiffness_matrix2D<DTyp>& A_yx,
        Local_stiffness_matrix2D<DTyp>& A_yy, Variable2D<DTyp>& b_x, Variable2D<DTyp>& b_y, Variable2D<DTyp>& x_x, Variable2D<DTyp>& x_y,
        Unstructured2DGrid_Marker2D& inner_points, int& iter);
  };
*/  
  
#endif
