#include "../source/ugblock2D.h"

//#define L2NORM(vector) sqrt(product(vector,vector).real())
//#define L2INORM(vector,iproduct) sqrt(product(vector,vector,iproduct).real())
/*
#define L2NORM2(vector1,vector2) sqrt(product(vector1,vector1).real()+product(vector2,vector2).real())
#define L2INORM2(vector1,vector2,iproduct) sqrt(product(vector1,vector1,iproduct).real()+product(vector2,vector2,iproduct).real())
*/


template<class DTyp>
GmRes<DTyp>::GmRes(Blockgrid2D& blockgrid_) : IterativeSolverAbstract<DTyp>(blockgrid_), r(blockgrid_) {
   Restart = 100;
     
   Q.resize(Restart+1);
   
   H = new DTyp*[Restart+1];
   C = new DTyp[Restart];
   S = new DTyp[Restart];
   y = new DTyp[Restart];
   gamma = new DTyp[Restart+1];
   
   for(int i=0;i<Restart+1;++i) {
       H[i] = new DTyp[Restart];
       Q.at(i) = new Variable2D<DTyp>(*(this->blockgrid));
   }
}
template<class DTyp>
GmRes<DTyp>::~GmRes() {
   for(int i=0;i<Restart+1;++i) {
       delete[] H[i];
       delete Q.at(i);
   }
   delete H;
   delete C;
   delete S;
   delete y;
   delete gamma;
} 


template<class DTyp>
void GmRes<DTyp>::setRestartNumber(int numberRestart_) {
  int oldRestart = Restart;
  Restart = numberRestart_;
  if(oldRestart==Restart) return;

  for(int i=0;i<oldRestart+1;++i) delete H[i];
  delete H;
  delete C;
  delete S;
  delete y;
  delete gamma;
    
   H = new DTyp*[Restart+1];
   C = new DTyp[Restart];
   S = new DTyp[Restart];
   y = new DTyp[Restart];
   gamma = new DTyp[Restart+1];
   
   for(int i=0;i<Restart+1;++i) {
       H[i] = new DTyp[Restart];
   }
  
  
  if(oldRestart<Restart) {
     Q.resize(Restart+1);
     for(int i=oldRestart+1;i<Restart+1;++i) {
         Q.at(i) = new Variable2D<DTyp>(*(this->blockgrid));
     }
  }
  else {
     for(int i=Restart+1;i<oldRestart+1;++i) {
         delete Q.at(i);
     }
     Q.resize(Restart+1);
  }
}




template<>
inline void GmRes<complex<double> >::GivensRotation(const complex<double>& x1, complex<double>& x2, 
						    complex<double>& c, complex<double>& s)  {
    complex<double> rho = x1 / abs(x1) * sqrt(x1.real() * x1.real() + x1.imag() * x1.imag() + x2.real() * x2.real() + x2.imag() * x2.imag());
    c = x1 / rho;
    s = std::conj(x2 / rho);
  }

template<>
inline void GmRes<double>::GivensRotation(const double& x1, double& x2, double& c, double& s)  {
    double rho = x1 / fabs(x1) * sqrt(x1 * x1 + x2 * x2);
    c = x1 / rho;
    s = x2 / rho;
  }

  



template<class DTyp>
void GmRes<DTyp>::solve(Local_stiffness_matrix2D<DTyp>& A, Variable2D<DTyp>& b, Variable2D<DTyp>& x) {
    this->numberIterationsPerformed = 0;

    Marker2D* activePoints;
    if(this->inner_points==NULL) activePoints = &b.Give_blockgrid()->Give_unstructured_grid()->Give_all_points();
    else                         activePoints = this->inner_points;

      
//test
//cout << " max_iter: " << this->max_iter << endl;    
    
    double r_abs;
    do {
      r = 0.0;
      if(this->inner_points==NULL)  r = (b - A(x));
      else                          r = (b - A(x))   | *this->inner_points;

     
      r_abs = L2Norm(r,*activePoints);
      if(r_abs < 1e-12) break;

      gamma[0] = r_abs;
      *Q[0] = r / gamma[0];


      
      for(int j = 1; j <= Restart; ++j) {
        this->numberIterationsPerformed++;
        *Q[j] = 0.0;
        if(this->inner_points==NULL) *Q[j] = (A(*Q[j - 1]));
        else                         *Q[j] = (A(*Q[j - 1])) | *this->inner_points;

        for (int i = 0; i < j; ++i) {
          H[i][j - 1] = product(*Q[j], *Q[i]);
          *Q[j] = *Q[j] - H[i][j - 1] * *Q[i];
        }

        //H[j][j - 1] = L2NORM(*Q[j]);  // old
        H[j][j - 1] = L2Norm(*Q[j],*activePoints);    
        *Q[j] = *Q[j] / H[j][j - 1];

        for (int i = 1; i < j; ++i) {
          const DTyp h_tmp = H[i - 1][j - 1];
          H[i - 1][j - 1] = (C[i - 1]) * h_tmp + S[i - 1] * H[i][j - 1];
          H[i][j - 1] = -std::conj(S[i - 1]) * h_tmp + C[i - 1] * H[i][j - 1];
        }
      
        GivensRotation(H[j - 1][j - 1], H[j][j - 1], C[j - 1], S[j - 1]);

        H[j - 1][j - 1] = (C[j - 1]) * H[j - 1][j - 1] + S[j - 1] * H[j][j - 1];

        gamma[j] = -std::conj(S[j - 1]) * gamma[j - 1];
        gamma[j - 1] = (C[j - 1]) * gamma[j - 1];
        r_abs = abs(gamma[j]);

        if (j == Restart || r_abs < this->eps || this->numberIterationsPerformed >= this->max_iter) {
          for (int i = j - 1; i >= 0; i--) {
            DTyp tmp_sum = gamma[i];
            for (int l = i + 1; l < j; ++l) {
              tmp_sum -= H[i][l] * y[l];

            }
            y[i] = tmp_sum / H[i][i];

            if(this->inner_points==NULL)  x = x + y[i] * *Q[i];
	    else                          x = x + y[i] * *Q[i] | *this->inner_points;
          }
          if (r_abs < this->eps || this->numberIterationsPerformed >= this->max_iter) {
	    this->residuum = r_abs;
	    return;
	  }
        }
      }
      
      
      
      
      
    } while (r_abs > this->eps && this->numberIterationsPerformed < this->max_iter);
    this->residuum = r_abs;
    return;
}


  
  
  
  
  
  /*
template<class DTyp>
GmRes2<DTyp>::GmRes2(Blockgrid2D& blockgrid)
      : blockgrid(blockgrid), Q_x(RESTART + 1, Variable2D<DTyp>(blockgrid)), Q_y(RESTART + 1, Variable2D<DTyp>(blockgrid)), 
      eps(1e-10),
      max_iter(5000) {
  }

template<class DTyp>
  GmRes2<DTyp>::GmRes2(Blockgrid2D& blockgrid, double eps, int max_iter)
      : blockgrid(blockgrid), Q_x(RESTART + 1, Variable2D<DTyp>(blockgrid))
        , Q_y(RESTART + 1, Variable2D<DTyp>(blockgrid)), eps(eps), max_iter(max_iter)
  {
  }

template<>
  inline void
  GmRes2<complex<double> >::GivensRotation(const complex<double>& x1, complex<double>& x2, complex<double>& c, complex<double>& s)
  {
    complex<double> rho = x1 / abs(x1) * sqrt(x1.real() * x1.real() + x1.imag() * x1.imag() + x2.real() * x2.real() + x2.imag() * x2.imag());
    c = x1 / rho;
    s = std::conj(x2 / rho);
  }

template<>
  inline void
  GmRes2<double>::GivensRotation(const double& x1, double& x2, double& c, double& s)
  {
    double rho = x1 / fabs(x1) * sqrt(x1 * x1 + x2 * x2);
    c = x1 / rho;
    s = x2 / rho;
  }

template<>
  inline double
  GmRes2<double>::L2Norm(Variable2D<double>& x)
  {
    return sqrt(product(x, x));
  }
template<>
  inline double
  GmRes2<std::complex<double> >::L2Norm(Variable2D<std::complex<double> >& x)
  {
    return sqrt(product(x, x).real());
  }

template<>
  inline double
  GmRes2<double>::L2InnerNorm(Variable2D<double>& x, Unstructured2DGrid_Marker2D& inner_points)
  {
    return sqrt(product(x, x, inner_points));
  }
template<>
  inline double
  GmRes2<std::complex<double> >::L2InnerNorm(Variable2D<std::complex<double> >& x, Unstructured2DGrid_Marker2D& inner_points)
  {
    return sqrt(product(x, x, inner_points).real());
  }

  
template<class DTyp>
  double
  GmRes2<DTyp>::solve(Local_stiffness_matrix2D<DTyp>& A_xx, Local_stiffness_matrix2D<DTyp>& A_xy, Local_stiffness_matrix2D<DTyp>& A_yx,
      Local_stiffness_matrix2D<DTyp>& A_yy, Variable2D<DTyp>& b_x, Variable2D<DTyp>& b_y, Variable2D<DTyp>& x_x, Variable2D<DTyp>& x_y,
      Unstructured2DGrid_Marker2D& inner_points, int& iter)
  {
  Variable2D<DTyp> r_x(blockgrid);
  Variable2D<DTyp> r_y(blockgrid);
  iter = 0;
  double r_abs;
  do {
    r_x = 0.0;
    r_x = (b_x - A_xx(x_x) - A_xy(x_y)) | inner_points;
    r_y = 0.0;
    r_y = (b_y - A_yx(x_x) - A_yy(x_y)) | inner_points;

    r_abs = L2INORM2(r_x, r_y, inner_points);
    if(r_abs < 1e-12) break;

    gamma[0] = r_abs;
    Q_x[0] = r_x / gamma[0];
    Q_y[0] = r_y / gamma[0];

    for (int j = 1; j <= RESTART; ++j) {
      iter++;
      Q_x[j] = 0.0;
      Q_y[j] = 0.0;
      Q_x[j] = (A_xx(Q_x[j - 1])+A_xy(Q_y[j - 1])) | inner_points;
      Q_y[j] = (A_yx(Q_x[j - 1])+A_yy(Q_y[j - 1])) | inner_points;

      for (int i = 0; i < j; ++i) {
        H[i][j - 1] = product(Q_x[j], Q_x[i])+product(Q_y[j], Q_y[i]);
        Q_x[j] = Q_x[j] - H[i][j - 1] * Q_x[i];
        Q_y[j] = Q_y[j] - H[i][j - 1] * Q_y[i];
      }

      H[j][j - 1] = L2NORM2(Q_x[j],Q_y[j]);
      Q_x[j] = Q_x[j] / H[j][j - 1];
      Q_y[j] = Q_y[j] / H[j][j - 1];

      for (int i = 1; i < j; ++i) {
        const DTyp h_tmp = H[i - 1][j - 1];
        H[i - 1][j - 1] = (C[i - 1]) * h_tmp + S[i - 1] * H[i][j - 1];
        H[i][j - 1] = -std::conj(S[i - 1]) * h_tmp + C[i - 1] * H[i][j - 1];
      }

      GivensRotation(H[j - 1][j - 1], H[j][j - 1], C[j - 1], S[j - 1]);

      H[j - 1][j - 1] = (C[j - 1]) * H[j - 1][j - 1] + S[j - 1] * H[j][j - 1];

      gamma[j] = -std::conj(S[j - 1]) * gamma[j - 1];
      gamma[j - 1] = (C[j - 1]) * gamma[j - 1];
      r_abs = abs(gamma[j]);

      if (j == RESTART || r_abs < eps) {
        DTyp y[RESTART];

        for (int i = j - 1; i >= 0; i--) {
          DTyp tmp_sum = gamma[i];
          for (int l = i + 1; l < j; ++l) {
            tmp_sum -= H[i][l] * y[l];

          }
          y[i] = tmp_sum / H[i][i];

          x_x = x_x + y[i] * Q_x[i] | inner_points;
          x_y = x_y + y[i] * Q_y[i] | inner_points;
        }
        if (r_abs < eps || iter >= max_iter) return r_abs;
      }
    }
  } while (r_abs > eps && iter < max_iter);
  return r_abs;
    return 0.0;
  }


  */
  