/**********************************************************************************
 * Copyright 2013 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/


//////////////////////////////////////////////////////////
//                  abbbreviations
//////////////////////////////////////////////////////////

#ifndef ABBREVI2D_H_
#define ABBREVI2D_H_

#include "../../common_source/abbrevi_common.h"

//#include "grid/elements.h"
//#include "grid/ug.h"

// ersetzen:
// hexahedra  -> rectanglEs
// hexahedron -> rectangle
// id_hex     -> id_rec
// quadrangle ->  alle weg!!!
// _el        -> _el2D 
// element_typ -> elementTyp
// 3Dvector    -> D2vector
// local_id_of_hex3D -> local_id_of_rec2D
// Edge_el  -> EdgeEl
// Point_el -> PointEl
// Geometric_Object -> Geometric2DObject
// Object_based_on_ug -> wegggggg!!!!!!!
// Unstructured_grid  -> Unstructured2DGrid
// data_hexahedra -> data_rectangles
// Expr ->  Expr2D
// Add  ->  Add2D
// ....
// Local_stiffness_matrix -> Local_stiffness_matrix2D

#define params2D_in  int id, int i, int j, int Nx
#define params2D_out id, i, j, Nx


#define params2D_in_loc_matrix   int id_rec, int ind_loc_matrix, int ind_cell, int i, int j, int Nx
#define params2D_out_loc_matrix      id_rec,     ind_loc_matrix,     ind_cell,     i,     j,     Nx

#define params2D_in_cell   int id_rec, int ind_cell, int i, int j, int Nx
#define params2D_out_cell      id_rec,     ind_cell,     i,     j,     Nx


// params for stencil
#define params2D_in_sten   int id, int num, int i
#define params2D_out_sten  id,     num,     i


#define Ind_data_rectangle(i,j,Nnx) data_rectangle[id][(i)+(Nnx+1)*(j)] 

#define Ind_loc_matrix_rectangle(i,j) ((j)*(Nx)+(i))
#define Ind_loc_matrix_rectangle_fine(i,j) ((j)*(Nx_fine)+(i))

#define Ind_loc_variable_rectangle(i,j) ((j)*(Nx+1)+(i))


// 1D
//-------------
enum dir1d { Wd,  Ed };
enum Ort1d { Wort, Mort, Eort };

#ifndef DIR2D_DEFINES
#define DIR2D_DEFINES


// integer-value of a point ds in a neighbour cell dz
inline int trans_cellP_9(dir2D_sons dz, dir2D_sons ds) {
  return ((ds&1) + (dz&1)) +
    3*(((ds>>1)&1) + ((dz>>1)&1));
}

inline void Print(dir2D_sons d) {
  if(d==SWdir2D)  cout << "SWdir2D";
  if(d==SEdir2D)  cout << "SEdir2D";
  if(d==NWdir2D)  cout << "NWdir2D";
  if(d==NEdir2D)  cout << "NEdir2D";
}

inline void Print(dir2D d) {
  if(d==Wdir2D) cout << "Wdir2D";
  if(d==Edir2D) cout << "Edir2D";
  if(d==Sdir2D) cout << "Sdir2D";
  if(d==Ndir2D) cout << "Ndir2D";
}


#endif

// Stern in 2D
//enum stencil_2D { SWsten, Ssten, SEsten,
//                  Wsten, Msten,  Esten,
//                  NWsten, Nsten, NEsten  };
		  
// local directions of rectangle with fixed SW-direction
enum local_dir_of_rec2D { SErec, NWrec, NErec };


// main direction in 2D
enum mainDirection2d { xDirection2d, yDirection2d };


/////////////////////////////////////////////////////////
//  funktionen for transformation
/////////////////////////////////////////////////////////

inline mainDirection2d mainDirection(dir2D_sons from, dir2D_sons to) {
   if((from&1) == (to&1)) return yDirection2d;
   return xDirection2d;
}

inline mainDirection2d orthogonalDirection(mainDirection2d dd) {
  if(dd==xDirection2d) return yDirection2d;
   return xDirection2d;
}

inline dir2D_sons nextInDirection(dir2D_sons from, mainDirection2d dd) {
  if(from==SWdir2D) {
    if(dd==xDirection2d) return SEdir2D;
    return NWdir2D;
  }
  if(from==SEdir2D) {
    if(dd==xDirection2d) return SWdir2D;
    return NEdir2D;
  }
  if(from==NWdir2D) {
    if(dd==xDirection2d) return NEdir2D;
    return SWdir2D;
  }
  if(from==NEdir2D) {
    if(dd==xDirection2d) return NWdir2D;
    return SEdir2D;
  }
}

inline dir2D edgeOfCorners(dir2D_sons A, dir2D_sons B) {
  if((A&1) == (B&1)) {
     if(A&1 == 1) return Edir2D;
     return Wdir2D;
  }
  if(((A>>1)&1) == 1) return Ndir2D;
  return Sdir2D;
}

// direction of the edge of a cell
//---------------------------------
inline dir2D oppositeEdge(dir2D ed) {
  if(ed==Wdir2D) return Edir2D;
  if(ed==Edir2D) return Wdir2D;
  if(ed==Sdir2D) return Ndir2D;
  return Sdir2D;
}

// opposite directions in 2D
//-----------------------------------
inline dir2D_sons opposite2D(dir2D_sons dir) {
  if(dir==SWdir2D) return NEdir2D;
  if(dir==NWdir2D) return SEdir2D;
  if(dir==SEdir2D) return NWdir2D;
  return SWdir2D;
}

// dir1D coordinate of direction "dir3D_sons"
//---------------------------------------------
inline dir1d x1DCoord(dir2D_sons d) {
  return (dir1d)(d&1);
}
inline dir1d y1DCoord(dir2D_sons d) {
  return (dir1d)((d>>1)&1);
}
 
// Berechnet lokales Koordinatensystem der Ecken
//
//       WN_corner
//
//         ^
//         |
//         c   ->   ES_corner
//
//----------------------------------------------
inline void Calculate_corresponding_corners(dir2D_sons corner,dir2D_sons& ES_corner,dir2D_sons& WN_corner) {
  //     3    4        NW  NE 
  //     c    2        SW  SE
  if(corner == SWdir2D){
     ES_corner = SEdir2D;
     WN_corner = NWdir2D;
  }
  //    4    2        NW  NE
  //    3    c        SW  SE
  else if(corner == SEdir2D){
     ES_corner = NEdir2D;
     WN_corner = SWdir2D;
  }
  //    c    3        NW  NE
  //    2    4        SW  SE
  else if(corner == NWdir2D){
     ES_corner = SWdir2D;
     WN_corner = NEdir2D;
  }
  //    2    c        NW  NE
  //    4    3        SW  SE
  else if(corner == NEdir2D){
     ES_corner = NWdir2D;
     WN_corner = SEdir2D;
  }
}
 
inline void calculate_map(dir2D_sons corner, bool clockOrientation, dir2D_sons* map_sons) {
  //     3    4        NW  NE 
  //     c    2        SW  SE
  if(corner == SWdir2D){
    map_sons[SWdir2D] = SWdir2D;
    map_sons[SEdir2D] = SEdir2D;
    map_sons[NWdir2D] = NWdir2D;
    map_sons[NEdir2D] = NEdir2D;
  }
  //    4    2        WN  EN
  //    3    c        WS  ES
  else if(corner == SEdir2D){
    map_sons[SWdir2D] = SEdir2D;
    map_sons[SEdir2D] = NEdir2D;
    map_sons[NWdir2D] = SWdir2D;
    map_sons[NEdir2D] = NWdir2D;
  }
  //    c    3        WN  EN
  //    2    4        WS  ES
  else if(corner == NWdir2D){
    map_sons[SWdir2D] = NWdir2D;
    map_sons[SEdir2D] = SWdir2D;
    map_sons[NWdir2D] = NEdir2D;
    map_sons[NEdir2D] = SEdir2D;
  }
  //    2    c        WN  EN
  //    4    3        WS  ES
  else if(corner == NEdir2D){
    map_sons[SWdir2D] = NEdir2D;
    map_sons[SEdir2D] = NWdir2D;
    map_sons[NWdir2D] = SEdir2D;
    map_sons[NEdir2D] = SWdir2D;
  }
  
  if(clockOrientation==false) {
    dir2D_sons cSp = map_sons[SEdir2D];
    map_sons[SEdir2D] = map_sons[NWdir2D];
    map_sons[NWdir2D] = cSp;
  }
}
 
// direction from one corner to another
//-------------------------------------
inline dir2D direction_from_to(dir2D_sons from, dir2D_sons to) {
  if((from&1) != (to&1)) {
    if((from&1) < (to&1)) return Edir2D;
    else return Wdir2D;
  }
  if((from&2) != (to&2)) {
    if((from&2) < (to&2)) return Ndir2D;
    else return Sdir2D;
  }
  assert(false);
}
 



class Abbreviations {   
public:
  static void Print(dir2D_sons d) {
    if(d==SWdir2D)  cout << "SWdir2D";
    if(d==SEdir2D)  cout << "SEdir2D";
    if(d==NWdir2D)  cout << "NWdir2D";
    if(d==NEdir2D)  cout << "NEdir2D";
  }

  static void Print(dir2D d) {
    if(d==Wdir2D)  cout << "Wdir2D";
    if(d==Edir2D)  cout << "Edir2D";
    if(d==Ndir2D)  cout << "Ndir2D";
    if(d==Sdir2D)  cout << "Sdir2D";
  } 
  static void Print(dir1d d) {
    if(d==Wd)  cout << "Wdir1D";
    if(d==Ed)  cout << "EdirID";
  } 
  static void Print(mainDirection2d dd) {
    if(dd==xDirection2d) cout << " x-direction "; 
    else                 cout << " y-direction "; 
  }
};

inline Ort1d trans_cellP_3(dir1d dz, dir1d ds) {
  return (Ort1d)((ds&1) + (dz&1));
}

#endif // ABBREVI2D_H_
