/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// calcArg.h
//
// ------------------------------------------------------------
#include "../grid/ug2D.h"
#include "../grid/blockgrid2D.h"

#ifndef CALCARG2D_H
#define CALCARG2D_H
#ifdef _OPENMP
#include <omp.h>
#include "../ugopenmp.h"
#endif

#include <list>

class PointForPhase {
  public:
    PointForPhase(int i_, int j_, double r_, int iSmall_, int jSmall_);
    double getRadius() { return r; }
    
    int getI() { return i; }
    int getJ() { return j; }
    int getIsmall() { return iSmall; }
    int getJsmall() { return jSmall; }    
  private:
    int i;
    int j;
    double r;
    int iSmall;
    int jSmall;
};

 
class BlockForPhase {
  public:
    BlockForPhase(int id_, double r_) { id = id_;  r = r_; }
    double getRadius() { return r; }
    double getId() { return id; }      
  private:
    int id;
    double r;
};

/** \addtogroup ExpressionTemplates2D **/
/* @{ */   
/**
 * Calculates the continuous Phase of a complex 2D Beam
 **/
class CalcContinuousArg {
  public:  
    CalcContinuousArg(Blockgrid2D& blockgrid_);
    void calcArg(Variable2D<double>& phase, Variable2D<std::complex<double> >& Efield, bool zeroAtOrigin = false);    
    
  private:
    Blockgrid2D*        blockgrid;
    Unstructured2DGrid* ug;
    std::list<PointForPhase>* orderedPoints; 
    std::list<BlockForPhase> orderedBlocks; 
    int numberBlocks;
   
    int idOrigin;
    int iOrigin;
    int jOrigin;
    int NxOrigin;
};
/* @} */


#endif // CALCARG2D_H



