

/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// variable.cc
//
// ------------------------------------------------------------

#include <string>
#include <stdexcept>
#include "../../../program/source/mympi.h"  // von 3D UGBlocks
#include "../abbrevi2D.h"
//#include "../../../linAlgExptemp/source/linAlg.h"
//#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../grid/elements2D.h"
#include "../grid/parti2D.h"
#include "../grid/ug2D.h"
#include "../grid/blockgrid2D.h"
#include "../grid/marker2D.h"
//#include "parallel.h"
#include "extemp2D.h"
#include "variable2D.h"
#include "update2D.h"
#include "cellvar2D.h"

template<>
void Cell_variable2D<double>::operator=(Variable2D<double>& varPoint) {
  int Nx, Ny, N_total;

  varPoint.updateAllRectangles();
  
  // rectangle
  for(int id=0;id<ug->Give_number_rectangle();++id) {

    if(ug->Give_rectangle(id)->my_object(my_rank)) {

      Nx = blockgrid->Give_Nx_rectangle(id);
      Ny = blockgrid->Give_Ny_rectangle(id);

#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
        for(int j=0;j<Ny;++j) {
            for(int i=0;i<Nx;++i){
                data_cell[id][Ind_loc_matrix_rectangle(i,j)] =
                          varPoint.Give_interpolated_cell_rectangle(id, i, j, Nx);
	  }
	}
    }
  }
}

template<>
void Cell_variable2D<std::complex<double> >::operator=(Variable2D<double>& varPoint) {
  int Nx, Ny, N_total;

  varPoint.updateAllRectangles();
  
  // rectangle
  for(int id=0;id<ug->Give_number_rectangle();++id) {

    if(ug->Give_rectangle(id)->my_object(my_rank)) {

      Nx = blockgrid->Give_Nx_rectangle(id);
      Ny = blockgrid->Give_Ny_rectangle(id);

#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
        for(int j=0;j<Ny;++j) {
            for(int i=0;i<Nx;++i){
                data_cell[id][Ind_loc_matrix_rectangle(i,j)] =
                          varPoint.Give_interpolated_cell_rectangle(id, i, j, Nx);
	  }
	}
    }
  }
}
