/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// cellvar2D.h
//
// ------------------------------------------------------------

#ifndef CE_VA2D_H
#define CE_VA2D_H

#ifdef _OPENMP
#include <omp.h>
#endif

//////////////////////////////////////////////////////////////
// 1. Marker2D: Restriction to subgrid
// 2. interpolation: variable -> cell variable
// 3. cell variable
// 3.1. definition
// 3.2. assign operators
// 3.3. product_cell, L_infty_cell
//////////////////////////////////////////////////////////////


///////////////////////////
// 1. Restriction to subgrid
///////////////////////////



// 2. interpolation: variable -> cell variable
//////////////////////////////////////////////////////
inline double dist(D2vector& v1, D2vector& v2)
{
  return sqrt((v1.x-v2.x)*(v1.x-v2.x)+(v1.y-v2.y)*(v1.y-v2.y));
}

template <class A>
class Expr2D_interpolant_point_to_cell : public Expr2D<Expr2D_interpolant_point_to_cell< A > >  {
 private:
  const A& a_;

 public:
  Expr2D_interpolant_point_to_cell(const Expr2D<A>& a) : a_(a) {}
    
  typedef typename A::Result Result;
    
  Result Give_cell_rectangle(params2D_in_cell) const {
//    Blockgrid2D*        blockgrid = a_.Give_blockgrid();
//    //Unstructured2DGrid*        ug = blockgrid->Give_unstructured_grid();
//    D2vector SW = blockgrid->Give_coord_rectangle(id_rec, i, j);
//    D2vector SE = blockgrid->Give_coord_rectangle(id_rec, i+1, j);
//    D2vector NW = blockgrid->Give_coord_rectangle(id_rec, i, j+1);
//    D2vector NE = blockgrid->Give_coord_rectangle(id_rec, i+1, j+1);
//    D2vector center(0.25*(SW.x+SE.x+NW.x+NE.x),0.25*(SW.y+SE.y+NW.y+NE.y));
////    std::cout << SW.x << "  " << SW.y << "  ";
////    std::cout << SE.x << "  " << SE.y << "  ";
////    std::cout << NW.x << "  " << NW.y << "  ";
////    std::cout << NE.x << "  " << NE.y << "  ";
////    std::cout << center.x << "  " << center.y << "\n";
//
//    double distSW = dist(SW,center);
//    double distSE = dist(SE,center);
//    double distNW = dist(NW,center);
//    double distNE = dist(NE,center);
//    double distges = distSW+distSE+distNW+distNE;
//
////    std::cout << distSW << "  " << distSE << "  " <<distNW << "  " << distNE  << "\n";
//
////    D2vector Give_coord_edge(int id, int i);
////    D2vector Give_coord_point(int id);
//
//    return (distSW/distges*a_.template Give_data<rectangleEl>(id_rec,i  ,j  ,Nx) +
//            distSE/distges*a_.template Give_data<rectangleEl>(id_rec,i+1,j  ,Nx) +
//            distNW/distges*a_.template Give_data<rectangleEl>(id_rec,i  ,j+1,Nx) +
//            distNE/distges*a_.template Give_data<rectangleEl>(id_rec,i+1,j+1,Nx));

    return (0.25*(a_.template Give_data<rectangleEl>(id_rec,i  ,j  ,Nx) +
            a_.template Give_data<rectangleEl>(id_rec,i+1,j  ,Nx) +
            a_.template Give_data<rectangleEl>(id_rec,i  ,j+1,Nx) +
            a_.template Give_data<rectangleEl>(id_rec,i+1,j+1,Nx)));
  }
  
  Result Give_matrix_rectangle(params2D_in_loc_matrix) const {
    return 0.25  * (a_.template Give_data<rectangleEl>(id_rec,i  ,j  ,Nx) +
		    a_.template Give_data<rectangleEl>(id_rec,i+1,j  ,Nx) +
		    a_.template Give_data<rectangleEl>(id_rec,i  ,j+1,Nx) +
		    a_.template Give_data<rectangleEl>(id_rec,i+1,j+1,Nx));
  }

  template <elementTyp TYP_EL>
  void Update(int id) const {
    a_.template Update<TYP_EL>(id);
  }
  
};
/** \addtogroup InterpolationOperators **/
/* @{ */ 
/**
 * interpolates from point data to cell data
 **/
template <class A>
inline Expr2D_interpolant_point_to_cell<A> Cell_interpolation(const Expr2D<A>& ao) {
	const A& a ( ao );  
	
        Unstructured2DGrid* ug = a.Give_blockgrid()->Give_unstructured_grid();	
        for(int id = 0;id < ug->Give_number_rectangle();++id ) {
                        ao.template Update<rectangleEl> ( id );	
	}
  return Expr2D_interpolant_point_to_cell<A>(ao);
}

/* @} */


//////////////////////////////////////////////////////
// 3. cell variable
//////////////////////////////////////////////////////
// 3.1. definition
//////////////////////////////////////////////////////////////
/** \addtogroup ExpressionTemplates2D **/
/* @{ */  
template <class DTyp>
class Cell_variable2D : public Expr2D<Cell_variable2D< DTyp > > {

 public:

  typedef DTyp Result;


  Cell_variable2D(Blockgrid2D& blockgrid_);
  Blockgrid2D* getBlockgrid2D() { return blockgrid; }
  
  ~Cell_variable2D();

  template <elementTyp TYP_EL>
  void Update(int id) const {}

  Blockgrid2D* Give_blockgrid() const { return blockgrid; };

  /// Prints absolute value in case of complex variable
  void Print_VTK ( std::ostream& Datei, std::string title = "myData");

  inline DTyp Give_cell_rectangle(params2D_in_cell)         const  { return data_cell[id_rec][ind_cell]; }
  inline DTyp Give_matrix_rectangle(params2D_in_loc_matrix) const  { return data_cell[id_rec][ind_cell]; }

  template <class A>
  void operator=(const Expr2D<A>& a);
  void operator=(DTyp x);

  void operator=(Variable2D<double>& varPoint);    
  
  void operator=(const DTyp_Restriction2D<DTyp>& a);

  template <class A>
  void operator=(const Expr2D_Restriction<A>& a);

  bool containsNaN();
  
  private:

  Blockgrid2D*        blockgrid;
  Unstructured2DGrid*       ug;

  // own data
  DTyp** data_cell;        // num_rectangle

  // for parallel
  int my_rank;
  MPI_Comm comm;
};

template <class A, class B>
_TypeOf2_(A,B) product_cell(Expr2D<A>& a, Expr2D<B>& b, Marker2D& marker);

template <class A, class B>
_TypeOf2_(A,B) product_cell(Expr2D<A>& a, Expr2D<B>& b);


template <class A>
double L_infty_cell(const Expr2D<A>& a, Marker2D& marker);

template <class A>
double L_infty_cell(const Expr2D<A>& a);

template <class A>
double Maximum_cell(const Expr2D<A>& ao);

template <class A>
double Minimum_cell(const Expr2D<A>& ao);
/* @} */


//////////////////////////////////////////
// 3.3. product_cell, L_infty_cell
//////////////////////////////////////////////////////////////


template <class DTyp>
Cell_variable2D<DTyp>::Cell_variable2D(Blockgrid2D& blockgrid_) {
  int Nx, Ny, N_total;

  blockgrid = &blockgrid_;
  ug = blockgrid->Give_unstructured_grid();

  // for parallel
  my_rank = ug->Give_my_rank();
  comm    = ug->Give_MPI_comm();

  // rectangle
  data_cell = new DTyp*[ug->Give_number_rectangle()];
  for(int id=0;id<ug->Give_number_rectangle();++id) {

    if(ug->Give_rectangle(id)->my_object(my_rank)) {

      Nx = blockgrid->Give_Nx_rectangle(id);
      Ny = blockgrid->Give_Ny_rectangle(id);

      N_total = Nx * Ny;

      data_cell[id] = new DTyp[N_total];

#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
      for(int i=0;i<N_total;++i)
        data_cell[id][i] = (DTyp)0;
    }
    else data_cell[id] = NULL;
  }
}

template <class DTyp>
Cell_variable2D<DTyp>::~Cell_variable2D() {
  if(data_cell != NULL){
      for(int i=0; i<ug->Give_number_rectangle(); ++i)
          if(data_cell[i] != NULL)
              delete [] data_cell[i];
      delete [] data_cell;
  }
  data_cell = NULL;
}






template <class DTyp>
void Cell_variable2D<DTyp>::operator=(DTyp x) {
  int Nx, Ny, N_total;

  // rectangle
  for(int id=0;id<ug->Give_number_rectangle();++id) {

    if(ug->Give_rectangle(id)->my_object(my_rank)) {

      Nx = blockgrid->Give_Nx_rectangle(id);
      Ny = blockgrid->Give_Ny_rectangle(id);
      N_total = Nx * Ny;

#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
      for(int i=0;i<N_total;++i)
        data_cell[id][i] = x;
    }
  }
}

template <class DTyp>
bool Cell_variable2D<DTyp>::containsNaN() {
  int Nx, Ny, N_total;

  // rectangle
  for(int id=0;id<ug->Give_number_rectangle();++id) {

    if(ug->Give_rectangle(id)->my_object(my_rank)) {

      Nx = blockgrid->Give_Nx_rectangle(id);
      Ny = blockgrid->Give_Ny_rectangle(id);
      N_total = Nx * Ny;
      for(int i=0;i<N_total;++i)
         if(isNaN(data_cell[id][i])) return true;
    }
  }
  return false;
}
  
  
template <class DTyp>
void Cell_variable2D<DTyp>::operator=(const DTyp_Restriction2D<DTyp>& a) {
  int Nx, Ny, N_total;
  DTyp   x      = a.Give_x();

  // rectangle
  for(int id=0;id<ug->Give_number_rectangle();++id) {

    if(ug->Give_rectangle(id)->my_object(my_rank) &&
       a.template Give_marker<rectangleEl>(id)==yes_mark) {

      Nx = blockgrid->Give_Nx_rectangle(id);
      Ny = blockgrid->Give_Ny_rectangle(id);
      N_total = Nx * Ny;

#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
      for(int i=0;i<N_total;++i)
        data_cell[id][i] = x;
    }
  }
}

template <class DTyp>
template <class A>
void Cell_variable2D<DTyp>::operator=(const Expr2D<A>& a) {
  int Nx, Ny, i, j;

  const A& ao ( a );

  // rectangle
  for(int id=0;id<ug->Give_number_rectangle();++id) {
    ao.template Update<rectangleEl>(id);

    if(ug->Give_rectangle(id)->my_object(my_rank)) {

      Nx = blockgrid->Give_Nx_rectangle(id);
      Ny = blockgrid->Give_Ny_rectangle(id);

// pragma laut gcc4.2 hier nicht erlaubt
#pragma omp parallel for private(i)  num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
        for(j=0;j<Ny;++j) {
          for(i=0;i<Nx;++i){
            data_cell[id][Ind_loc_matrix_rectangle(i,j)] =
              ao.Give_cell_rectangle(id,Ind_loc_matrix_rectangle(i,j),i,j,Nx);
	  }
      }
    }
  }
}

template <class DTyp>
template <class A>
void Cell_variable2D<DTyp>::operator=(const Expr2D_Restriction<A>& a)  {
  int Nx, Ny, i, j;

  //  const A& ao ( a );

  // rectangle
  for(int id=0;id<ug->Give_number_rectangle();++id) {
    a.template Update<rectangleEl>(id);

    if(ug->Give_rectangle(id)->my_object(my_rank) &&
       a.template Give_marker<rectangleEl>(id)==yes_mark) {
      
      Nx = blockgrid->Give_Nx_rectangle(id);
      Ny = blockgrid->Give_Ny_rectangle(id);

// pragma laut gcc4.2 hier nicht erlaubt
#pragma omp parallel for private(i) num_threads(UGBlocks::numThreadsToTake)  if(UGBlocks::useOpenMP)
        for(j=0;j<Ny;++j) {
          for(i=0;i<Nx;++i){
            data_cell[id][Ind_loc_matrix_rectangle(i,j)] =
              a.Give_cell_rectangle(id,Ind_loc_matrix_rectangle(i,j),i,j,Nx);
	  }
      }
    }
  }
}

//////////////////////////////////////////
// 3.3. product_cell, L_infty_cell
//////////////////////////////////////////////////////////////

template <class A, class B>
  _TypeOf2_(A,B) product_cell(Expr2D<A>& a, Expr2D<B>& b) {
  const A& ao ( a );

  return product_cell(a,b,ao.Give_blockgrid()->Give_unstructured_grid()->Give_all_points());
}

template <class A>
double  L_infty_cell(Expr2D<A>& a) {
  const A& ao ( a );

  return L_infty_cell(a,ao.Give_blockgrid()->Give_unstructured_grid()->Give_all_points());
}



template <class A, class B>
_TypeOf2_(A,B) product_cell(Expr2D<A>& ao, Expr2D<B>& bo, Marker2D& marker) {
  int Nx, Ny;

  typedef _TypeOf2_(A,B) DTyp;
  const A& a ( ao );
  const B& b ( bo );

  DTyp sum;
  DTyp sum_total;

  sum = (DTyp)0;

  assert(a.Give_blockgrid() == b.Give_blockgrid());

  Unstructured2DGrid* ug = a.Give_blockgrid()->Give_unstructured_grid();
  int my_rank   = ug->Give_my_rank();
  MPI_Comm comm = ug->Give_MPI_comm();

  // rectangle
  for(int id=0;id<ug->Give_number_rectangle();++id) {
     if(ug->Give_rectangle(id)->my_object(my_rank)   &&
        marker.template Give_marker<rectangleEl>(id)==yes_mark) {
      Nx = a.Give_blockgrid()->Give_Nx_rectangle(id);
      Ny = a.Give_blockgrid()->Give_Ny_rectangle(id);

      /// todo hier sollte noch ein reduction rein
      //#pragma omp parallel for private(i,j) num_threads(UGBlocks::numThreadsToTake)  if(UGBlocks::useOpenMP)
      for(int j=0;j<Ny;++j) {
          for(int i=0;i<Nx;++i)
	      sum+= a.Give_cell_rectangle(id,
					  Ind_loc_matrix_rectangle(i,j),i,j,Nx)
	             * MyConj<DTyp>::Give(b.Give_cell_rectangle(id,
							 Ind_loc_matrix_rectangle(i,j),i,j,Nx));
      }
     }
  }

  sum_total = sum;
//  sum_total = Make_MPI_All_Sum<DTyp>::Do(sum,comm);

  return sum_total;
}




template <class A>
double L_infty_cell(const Expr2D<A>& ao, Marker2D& marker) {
  int Nx, Ny;

  double infty;
  double infty_total;

  const A& a ( ao );

  infty = 0.0;

  Unstructured2DGrid* ug = a.Give_blockgrid()->Give_unstructured_grid();
  int my_rank   = ug->Give_my_rank();
  MPI_Comm comm = ug->Give_MPI_comm();

  for(int id=0; id<ug->Give_number_rectangle(); ++id) {
    if(ug->Give_rectangle(id)->my_object(my_rank) &&
        marker.template Give_marker<rectangleEl>(id)==yes_mark) {
      Nx = a.Give_blockgrid()->Give_Nx_rectangle(id);
      Ny = a.Give_blockgrid()->Give_Ny_rectangle(id);

      /// todo hier sollte noch ein reduction rein 
      //#pragma omp parallel for private(i,j) num_threads(UGBlocks::numThreadsToTake)  if(UGBlocks::useOpenMP)
      for(int j=0;j<Ny;++j) {
          for(int i=0;i<Nx;++i)
	    if(infty < ABS(a.Give_cell_rectangle(id,
						 Ind_loc_matrix_rectangle(i,j),i,j,Nx)))
	      infty = ABS(a.Give_cell_rectangle(id,
						Ind_loc_matrix_rectangle(i,j),i,j,Nx));
      }
    }
  }

  MPI_Allreduce(&infty,&infty_total,1,MPI_DOUBLE,MPI_MAX,comm);

  return infty_total;
}

template <class A>
double Maximum_cell(const Expr2D<A>& ao) {
  int Nx, Ny;

  double maximum;
  double maximum_total;

  maximum = -1.0e50;
  
  const A& a ( ao );


  Unstructured2DGrid* ug = a.Give_blockgrid()->Give_unstructured_grid();
  int my_rank   = ug->Give_my_rank();
  MPI_Comm comm = ug->Give_MPI_comm();

  
  
  for(int id=0; id<ug->Give_number_rectangle(); ++id) {
    if(ug->Give_rectangle(id)->my_object(my_rank)) {
      Nx = a.Give_blockgrid()->Give_Nx_rectangle(id);
      Ny = a.Give_blockgrid()->Give_Ny_rectangle(id);

      /// todo hier sollte noch ein reduction rein 
      //#pragma omp parallel for private(i,j) num_threads(UGBlocks::numThreadsToTake)  if(UGBlocks::useOpenMP)
        for(int j=0;j<Ny;++j) {
          for(int i=0;i<Nx;++i) {
	    if(maximum < ABS(a.Give_cell_rectangle(id,
						 Ind_loc_matrix_rectangle(i,j),i,j,Nx)))
	      maximum = ABS(a.Give_cell_rectangle(id,
						Ind_loc_matrix_rectangle(i,j),i,j,Nx));
      }
    }
   }
  }

  MPI_Allreduce(&maximum,&maximum_total,1,MPI_DOUBLE,MPI_MAX,comm);

  return maximum_total;
}


template <class A>
double Minimum_cell(const Expr2D<A>& ao) {
  int Nx, Ny;

  double minimum;
  double minimum_total;

  minimum = 1.0e50;
  
  const A& a ( ao );


  Unstructured2DGrid* ug = a.Give_blockgrid()->Give_unstructured_grid();
  int my_rank   = ug->Give_my_rank();
  MPI_Comm comm = ug->Give_MPI_comm();

  
  
  for(int id=0; id<ug->Give_number_rectangle(); ++id) {
    if(ug->Give_rectangle(id)->my_object(my_rank)) {
      Nx = a.Give_blockgrid()->Give_Nx_rectangle(id);
      Ny = a.Give_blockgrid()->Give_Ny_rectangle(id);

      /// todo hier sollte noch ein reduction rein 
      //#pragma omp parallel for private(i,j) num_threads(UGBlocks::numThreadsToTake)  if(UGBlocks::useOpenMP)
        for(int j=0;j<Ny;++j) {
          for(int i=0;i<Nx;++i) {
	    if(minimum > ABS(a.Give_cell_rectangle(id,
						 Ind_loc_matrix_rectangle(i,j),i,j,Nx)))
	      minimum = ABS(a.Give_cell_rectangle(id,
						Ind_loc_matrix_rectangle(i,j),i,j,Nx));
      }
    }
   }
  }

  MPI_Allreduce(&minimum,&minimum_total,1,MPI_DOUBLE,MPI_MAX,comm);

  return minimum_total;
}




#endif // CE_VA2D_H
