/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#include "../grid/ug2D.h"
#include "../grid/blockgrid2D.h"
//#include "../../../linAlgExptemp/source/linAlg.h"
//#include "../../../linAlgExptemp/source/extemp/extemp_type.h"
#include "../../../common_source/extemp_type.h"

#ifndef EXTEMP2D_H
#define EXTEMP2D_H



template <class A> struct Expr2D {
	inline operator const A&() const {
		return *static_cast<const A*> ( this );
	}
};

//----------------------------------------------------------------------------
template <class A, class B>
class Add2D : public Expr2D<Add2D<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline Add2D ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid2D* Give_blockgrid() const {
			if ( a_.Give_blockgrid() == NULL )
				return b_.Give_blockgrid();
				
			return a_.Give_blockgrid();
		};
		
		stencil_typ Give_stencil_typ() const {
			if ( a_.Give_stencil_typ() == yes_stencil ) return yes_stencil;
			
			return b_.Give_stencil_typ();
		};
				
		
		typedef _TypeOf2_ ( A, B ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params2D_in ) const {
			return a_.template Give_data<TYP_EL> ( params2D_out ) +
			       b_.template Give_data<TYP_EL> ( params2D_out );
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params2D_in_sten ) const {
			return a_.template Give_data_sten<TYP_EL> ( params2D_out_sten ) +
			       b_.template Give_data_sten<TYP_EL> ( params2D_out_sten );
		}
		
		
		/*
		template <elementTyp TYP_EL>
		  inline Result Give_matrix(params2D_in_loc_matrix) const
		  {
		    return a_.template Give_matrix<TYP_EL>(params2D_out_loc_matrix) +
		  b_.template Give_matrix<TYP_EL>(params2D_out_loc_matrix);
		  }
		*/
		inline Result Give_matrix_rectangle ( params2D_in_loc_matrix ) const {
			return a_.Give_matrix_rectangle ( params2D_out_loc_matrix ) +
			       b_.Give_matrix_rectangle ( params2D_out_loc_matrix );
		}
		
		inline Result Give_cell_rectangle ( params2D_in_cell ) const {
			return a_.Give_cell_rectangle ( params2D_out_cell ) +
			       b_.Give_cell_rectangle ( params2D_out_cell );
		}
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
			b_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class A, class B>
inline Add2D<A, B> operator+ ( const Expr2D<A>& a, const Expr2D<B>& b ) {
	return Add2D<A, B> ( a, b );
}

//----------------------------------------------------------------------------
template <class A, typename TYPE>
class CAdd2D : public Expr2D<CAdd2D<A, TYPE> > {
		const A& a_;
		TYPE b_;
		
	public:
		inline CAdd2D ( const A& a, TYPE b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid2D* Give_blockgrid() const {
			return a_.Give_blockgrid();
		}
		
		stencil_typ Give_stencil_typ() const {
			return a_.Give_stencil_typ();
		}
		
		
		typedef _TypeOf1_ ( A, TYPE ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params2D_in ) const {
			return a_.template Give_data<TYP_EL> ( params2D_out ) + b_;
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params2D_in_sten ) const {
			return a_.template  Give_data_sten<TYP_EL> ( params2D_out_sten ) + b_;
		}
		
		
		inline Result Give_matrix_rectangle ( params2D_in_loc_matrix ) const {
			return a_.Give_matrix_rectangle ( params2D_out_loc_matrix ) + b_;
		}
		
		inline Result Give_cell_rectangle ( params2D_in_cell ) const {
			return a_.Give_cell_rectangle ( params2D_out_cell ) + b_;
		}
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class A>
inline CAdd2D<A, double>
operator+ ( const Expr2D<A>& a, double b ) {
	return CAdd2D<A, double> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAdd2D<A, double>
operator+ ( double b, const Expr2D<A>& a ) {
	return CAdd2D<A, double> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAdd2D<A, double>
operator- ( const Expr2D<A>& a, double b ) {
	return CAdd2D<A, double> ( a, -b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAdd2D<A, std::complex<double> >
operator+ ( const Expr2D<A>& a, std::complex<double>  b ) {
	return CAdd2D<A, std::complex<double> > ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAdd2D<A, std::complex<double> >
operator+ ( std::complex<double>  b, const Expr2D<A>& a ) {
	return CAdd2D<A, std::complex<double> > ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline CAdd2D<A, std::complex<double> >
operator- ( const Expr2D<A>& a, std::complex<double>  b ) {
	return CAdd2D<A, std::complex<double> > ( a, -b );
}

//----------------------------------------------------------------------------
template <class A, class B>

class Sub2D : public Expr2D<Sub2D<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline Sub2D ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid2D* Give_blockgrid() const {
			if ( a_.Give_blockgrid() == NULL )
				return b_.Give_blockgrid();
				
			return a_.Give_blockgrid();
		};
		
		stencil_typ Give_stencil_typ() const {
			if ( a_.Give_stencil_typ() == yes_stencil ) {
				return yes_stencil;
			}
			
			return b_.Give_stencil_typ();
		}
		
	
		typedef _TypeOf2_ ( A, B ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params2D_in ) const {
			return a_.template Give_data<TYP_EL> ( params2D_out ) -
			       b_.template Give_data<TYP_EL> ( params2D_out );
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params2D_in_sten ) const {
			return a_.template Give_data_sten<TYP_EL> ( params2D_out_sten ) -
			       b_.template Give_data_sten<TYP_EL> ( params2D_out_sten );
		}
		
		inline Result Give_matrix_rectangle ( params2D_in_loc_matrix ) const {
			return a_.Give_matrix_rectangle ( params2D_out_loc_matrix ) -
			       b_.Give_matrix_rectangle ( params2D_out_loc_matrix );
		}
		
		inline Result Give_cell_rectangle ( params2D_in_cell ) const {
			return a_.Give_cell_rectangle ( params2D_out_cell ) -
			       b_.Give_cell_rectangle ( params2D_out_cell );
		}
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
			b_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class A, class B>
inline Sub2D<A, B> operator- ( const Expr2D<A>& a, const Expr2D<B>& b ) {
	return Sub2D<A, B> ( a, b );
}

//----------------------------------------------------------------------------
template <typename TYPE, class B>

class CSub2D : public Expr2D<CSub2D<TYPE, B> > {
		TYPE a_;
		const B& b_;
		
	public:
		inline CSub2D ( TYPE a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid2D* Give_blockgrid() const {
			return b_.Give_blockgrid();
		}
		
		stencil_typ Give_stencil_typ() const {
			return b_.Give_stencil_typ();
		}
		
		typedef _TypeOf1_ ( B, TYPE ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params2D_in ) const {
			return a_ - b_.template Give_data<TYP_EL> ( params2D_out );
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params2D_in_sten ) const {
			return a_ - b_.template Give_data_sten<TYP_EL> ( params2D_out_sten );
		}
		
		inline Result Give_matrix_rectangle ( params2D_in_loc_matrix ) const {
			return a_ - b_.Give_matrix_rectangle ( params2D_out_loc_matrix );
		}
		
		inline Result Give_cell_rectangle ( params2D_in_cell ) const {
			return a_ - b_.Give_cell_rectangle ( params2D_out_cell );
		}
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			b_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class B>
inline CSub2D<double, B> operator- ( double a, const Expr2D<B>& b ) {
	return CSub2D<double, B> ( a, b );
}

template <class B>
inline CSub2D<std::complex<double>, B> operator- ( std::complex<double> a, const Expr2D<B>& b ) {
	return CSub2D<std::complex<double>, B> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>

class Minus2D : public Expr2D<Minus2D<A> > {
		const A& a_;
		
	public:
		inline Minus2D ( const A& a ) : a_ ( a ) {}
		
		Blockgrid2D* Give_blockgrid() const {
			return a_.Give_blockgrid();
		}
		
		stencil_typ Give_stencil_typ() const {
			return a_.Give_stencil_typ();
		}
		
		typedef typename A::Result Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params2D_in ) const {
			return - a_.template Give_data<TYP_EL> ( params2D_out );
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params2D_in_sten ) const {
			return - a_.template Give_data_sten<TYP_EL> ( params2D_out_sten );
		}
		
		inline Result Give_matrix_rectangle ( params2D_in_loc_matrix ) const {
			return - a_.Give_matrix_rectangle ( params2D_out_loc_matrix );
		}
		
		inline Result Give_cell_rectangle ( params2D_in_cell ) const {
			return - a_.Give_cell_rectangle ( params2D_out_cell );
		}
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class A>
inline Minus2D<A> operator- ( const Expr2D<A>& a ) {
	return Minus2D<A> ( a );
}

//----------------------------------------------------------------------------
template <class A, typename TYPE>

class Times2D : public Expr2D<Times2D<A, TYPE> > {
		const A& a_;
		TYPE b_;
		
	public:
		inline Times2D ( const A& a, const TYPE& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid2D* Give_blockgrid() const {
			return a_.Give_blockgrid();
		};
		
		stencil_typ Give_stencil_typ() const {
			return a_.Give_stencil_typ();
		}
			
		typedef _TypeOf1_ ( A, TYPE ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params2D_in ) const {
			return a_.template Give_data<TYP_EL> ( params2D_out ) * b_;
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params2D_in_sten ) const {
			return a_.template Give_data_sten<TYP_EL> ( params2D_out_sten ) * b_;
		}
		
		inline Result Give_matrix_rectangle ( params2D_in_loc_matrix ) const {
			return a_.Give_matrix_rectangle ( params2D_out_loc_matrix ) * b_;
		}
		
		inline Result Give_cell_rectangle ( params2D_in_cell ) const {
			return a_.Give_cell_rectangle ( params2D_out_cell ) * b_;
		}
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class A>
inline Times2D<A, double>
operator* ( const Expr2D<A>& a, double b ) {
	return Times2D<A, double> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline Times2D<A, double>
operator* ( double b, const Expr2D<A>& a ) {
	return Times2D<A, double> ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline Times2D<A, std::complex<double> >
operator* ( const Expr2D<A>& a, std::complex<double> b ) {
	return Times2D<A, std::complex<double> > ( a, b );
}

//----------------------------------------------------------------------------
template <class A>
inline Times2D<A, std::complex<double> >
operator* ( std::complex<double> b, const Expr2D<A>& a ) {
	return Times2D<A, std::complex<double> > ( a, b );
}

//----------------------------------------------------------------------------

// spezielle Konstruktion falls ein Operant bool
//----------------------------------------------------------------------------
template <class A>
class TimesBool2D : public Expr2D<TimesBool2D<A> > {
		const A& a_;
		bool b_;
		
	public:
		inline TimesBool2D ( const A& a, const bool& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid2D* Give_blockgrid() const {
			return a_.Give_blockgrid();
		};
		
		stencil_typ Give_stencil_typ() const {
			return a_.Give_stencil_typ();
		}
			
		typedef typename A::Result Result;
				
		template <elementTyp TYP_EL>
		inline Result Give_data ( params2D_in ) const {
			if(b_) return a_.template Give_data<TYP_EL> ( params2D_out );
			return 0.0;
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params2D_in_sten ) const {
			if(b_) return a_.template Give_data_sten<TYP_EL> ( params2D_out_sten );
			return 0.0;
		}
		
		inline Result Give_matrix_rectangle ( params2D_in_loc_matrix ) const {
			if(b_) return a_.Give_matrix_rectangle ( params2D_out_loc_matrix );
			return 0.0;
		}
		
		inline Result Give_cell_rectangle ( params2D_in_cell ) const {
			if(b_) return a_.Give_cell_rectangle ( params2D_out_cell );
			return 0.0;
		}
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
		}
};


template <class A>
inline TimesBool2D<A>
operator*( bool b, const Expr2D<A>& a ) {
	return TimesBool2D<A> ( a, b );
}


template <class A>
inline TimesBool2D<A>
operator*( const Expr2D<A>& a, bool b) {
	return TimesBool2D<A> ( a, b );
}



//----------------------------------------------------------------------------
template <typename TYPE, class B>

class Div2D : public Expr2D<Div2D<TYPE, B> > {
		const TYPE a_;
		const B& b_;
		
	public:
		inline Div2D ( const TYPE& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid2D* Give_blockgrid() const {
			return b_.Give_blockgrid();
		};
		
		stencil_typ Give_stencil_typ() const {
			return b_.Give_stencil_typ();
		}
		
		typedef _TypeOf1_ ( B, TYPE ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params2D_in ) const {
			return a_ / b_.template Give_data<TYP_EL> ( params2D_out );
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params2D_in_sten ) const {
			return a_ / b_.template Give_data_sten<TYP_EL> ( params2D_out_sten );
		}
		
		inline Result Give_matrix_rectangle ( params2D_in_loc_matrix ) const {
			return a_ / b_.Give_matrix_rectangle ( params2D_out_loc_matrix );
		}
		
		inline Result Give_cell_rectangle ( params2D_in_cell ) const {
			return a_ / b_.Give_cell_rectangle ( params2D_out_cell );
		}
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			b_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class A>
inline Times2D<A, double> operator/ ( const Expr2D<A>& a, double b ) {
	return Times2D<A, double> ( a, 1. / b );
}

//----------------------------------------------------------------------------
template <class A>
inline Times2D<A, std::complex<double> > operator/ ( const Expr2D<A>& a, std::complex<double> b ) {
	return Times2D<A, std::complex<double> > ( a, 1. / b );
}

//----------------------------------------------------------------------------
template <class B>
inline Div2D<double, B> operator/ ( double a, const Expr2D<B>& b ) {
	return Div2D<double, B> ( a, b );
}

//----------------------------------------------------------------------------
template <class B>
inline Div2D<std::complex<double>, B> operator/ ( std::complex<double> a, const Expr2D<B>& b ) {
	return Div2D<std::complex<double>, B> ( a, b );
}

//----------------------------------------------------------------------------
template <class A, class B>

class VTimes2D : public Expr2D<VTimes2D<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline VTimes2D ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid2D* Give_blockgrid() const {
			if ( a_.Give_blockgrid() == NULL )
				return b_.Give_blockgrid();
				
			return a_.Give_blockgrid();
		};
		
		stencil_typ Give_stencil_typ() const {
			if ( a_.Give_stencil_typ() == yes_stencil ) {
				return yes_stencil;
			}
			
			return b_.Give_stencil_typ();
		}
		
		typedef _TypeOf2_ ( A, B ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params2D_in ) const {
			return a_.template Give_data<TYP_EL> ( params2D_out ) *
			       b_.template Give_data<TYP_EL> ( params2D_out );
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params2D_in_sten ) const {
			return a_.template Give_data_sten<TYP_EL> ( params2D_out_sten ) *
			       b_.template Give_data_sten<TYP_EL> ( params2D_out_sten );
		}
		
		inline Result Give_matrix_rectangle ( params2D_in_loc_matrix ) const {
			return a_.Give_matrix_rectangle ( params2D_out_loc_matrix ) *
			       b_.Give_matrix_rectangle ( params2D_out_loc_matrix );
		}
		
		inline Result Give_cell_rectangle ( params2D_in_cell ) const {
			return a_.Give_cell_rectangle ( params2D_out_cell ) *
			       b_.Give_cell_rectangle ( params2D_out_cell );
		}
		
		inline Unstructured2DGrid * Give_Ug() const {
		  return a_.Give_Ug();
		}

		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
			b_.template Update<TYP_EL> ( id );
		}
};

//----------------------------------------------------------------------------
template <class A, class B>
inline VTimes2D<A, B> operator* ( const Expr2D<A>& a, const Expr2D<B>& b ) {
	return VTimes2D<A, B> ( a, b );
}

//----------------------------------------------------------------------------
template <class A, class B>

class VDiv2D : public Expr2D<VDiv2D<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline VDiv2D ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid2D* Give_blockgrid() const {
			if ( a_.Give_blockgrid() == NULL )
				return b_.Give_blockgrid();
				
			return a_.Give_blockgrid();
		};
		
		stencil_typ Give_stencil_typ() const {
			if ( a_.Give_stencil_typ() == yes_stencil ) {
				return yes_stencil;
			}
			
			return b_.Give_stencil_typ();
		}
		
		typedef _TypeOf2_ ( A, B ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params2D_in ) const {
			return a_.template Give_data<TYP_EL> ( params2D_out ) /
			       b_.template Give_data<TYP_EL> ( params2D_out );
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params2D_in_sten ) const {
			return a_.template Give_data_sten<TYP_EL> ( params2D_out_sten ) /
			       b_.template Give_data_sten<TYP_EL> ( params2D_out_sten );
		}
		
		inline Result Give_matrix_rectangle ( params2D_in_loc_matrix ) const {
			return a_.Give_matrix_rectangle ( params2D_out_loc_matrix ) /
			       b_.Give_matrix_rectangle ( params2D_out_loc_matrix );
		}
		
		inline Result Give_cell_rectangle ( params2D_in_cell ) const {
			return a_.Give_cell_rectangle ( params2D_out_cell ) /
			       b_.Give_cell_rectangle ( params2D_out_cell );
		}
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
			b_.template Update<TYP_EL> ( id );
		}

                inline Unstructured2DGrid * Give_Ug() const {
                  return a_.Give_Ug();
                }
};

//----------------------------------------------------------------------------
template <class A, class B>
inline VDiv2D<A, B> operator/ ( const Expr2D<A>& a, const Expr2D<B>& b ) {
	return VDiv2D<A, B> ( a, b );
}

/** \addtogroup ExpressionTemplatesTwoD **/
/* @{ */  
/**
 * epsilin Parameter for save division
 **/
struct saveDivision2D {
    static double eps;	
};
/* @} */

//----------------------------------------------------------------------------
template <class A, class B>

class SaveDiv2D : public Expr2D<SaveDiv2D<A, B> > {
		const A& a_;
		const B& b_;
		
	public:
		inline SaveDiv2D ( const A& a, const B& b ) : a_ ( a ), b_ ( b ) {}
		
		Blockgrid2D* Give_blockgrid() const {
			if ( a_.Give_blockgrid() == NULL )
				return b_.Give_blockgrid();
				
			return a_.Give_blockgrid();
		};
		
		stencil_typ Give_stencil_typ() const {
			if ( a_.Give_stencil_typ() == yes_stencil ) {
				return yes_stencil;
			}
			
			return b_.Give_stencil_typ();
		}
		
		typedef _TypeOf2_ ( A, B ) Result;
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params2D_in ) const {
		       Result valueB = b_.template Give_data<TYP_EL> ( params2D_out );
		       if(myABS(valueB) < saveDivision2D::eps) return 0.0;
                       return a_.template Give_data<TYP_EL> ( params2D_out ) / valueB;			  
		}
		
		template <sten_part_typ TYP_EL>
		inline Result Give_data_sten ( params2D_in_sten ) const {
		       Result valueB = b_.template Give_data_sten<TYP_EL> ( params2D_out_sten );
		       if(myABS(valueB) < saveDivision2D::eps) return 0.0;
                       return a_.template Give_data_sten<TYP_EL> ( params2D_out_sten ) / valueB;			  
		}
		
		inline Result Give_matrix_rectangle ( params2D_in_loc_matrix ) const {
		       Result valueB = b_.Give_matrix_rectangle ( params2D_out_loc_matrix );
		       if(myABS(valueB) < saveDivision2D::eps) return 0.0;
                       return a_.Give_matrix_rectangle ( params2D_out_loc_matrix ) / valueB;			  
		}
		
		inline Result Give_cell_rectangle ( params2D_in_cell ) const {
		       Result valueB = b_.Give_cell_rectangle ( params2D_out_cell );
		       if(myABS(valueB) < saveDivision2D::eps) return 0.0;
                       return a_.Give_cell_rectangle ( params2D_out_cell ) / valueB;			  
		}
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			a_.template Update<TYP_EL> ( id );
			b_.template Update<TYP_EL> ( id );
		}

                inline Unstructured2DGrid * Give_Ug() const {
                  return a_.Give_Ug();
                }
                
   private:
     static double myABS(double a) {  return a < 0.0 ? -a : a; }
     static std::complex<double> myABS(std::complex<double> c) {  
         return (c.real() < 0.0 ? -c.real() : c.real()) + (c.imag() < 0.0 ? -c.imag() : c.imag());
     }
};

//----------------------------------------------------------------------------
template <class A, class B>
inline SaveDiv2D<A, B> saveDiv( const Expr2D<A>& a, const Expr2D<B>& b ) {
	return SaveDiv2D<A, B> ( a, b );
}


//----------------------------------------------------------------------------
#endif // EXTEMP_H
