/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// interpolateFrom3D.cc
//
// ------------------------------------------------------------



#include <string>
#include <stdexcept>
#include <complex>
#include "../../../program/source/mympi.h"  // von 3D UGBlocks

// von 3D
#include "../../program/source/abbrevi.h"
#include "../../program/source/parameter.h"
#include "../../program/source/math_lib/math_lib.h"
#include "../../program/source/basics/basic.h"
#include "../../program/source/grid/elements.h"
#include "../../program/source/grid/parti.h"
#include "../../program/source/grid/ug.h"
#include "../../program/source/grid/blockgrid.h"
#include "../../program/source/grid/marker.h"
#include "../../program/source/grid/examples_ug.h"
#include "../../program/source/extemp/parallel.h"
#include "../../program/source/extemp/extemp.h"
#include "../../program/source/extemp/variable.h"


// von 2D
#include "../abbrevi2D.h"
//#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../grid/elements2D.h"
#include "../grid/parti2D.h"
#include "../grid/ug2D.h"
#include "../grid/blockgrid2D.h"
#include "../grid/marker2D.h"
//#include "parallel.h"
#include "extemp2D.h"
#include "variable2D.h"
#include "variable2D_cc.h"

// 2D 3D
#include "interpolateFrom3D.h"

/////////////////////////////////////////////////////////
// 1. z-Iterator
/////////////////////////////////////////////////////////

IteratorZDirection::IteratorZDirection(Blockgrid* blockgrid_) {
   blockgrid = blockgrid_;
   Unstructured_grid* ug3Dp  = blockgrid->Give_unstructured_grid();
   
   typeUg = ug3Dp->getUgTyp();
   assert(typeUg!=noWithZTyp);
   ug_withZ = static_cast<UG_WithZdirection*>(ug3Dp);

   nummernBloecke = ug_withZ->getNummernBloecke();

   //Start-Werte setzen:
   startBlock(0);
   
   
   totalNumberSlices = 0;
   index = 0;
   for(int nummerBlock=0;nummerBlock<nummernBloecke;++nummerBlock) {
       if(typeUg==hexahedronUgTyp) {
          Nz = blockgrid->Give_Nz_hexahedron(nummerBlock);
       }
       else if(typeUg==cylinderUgTyp) {
          Nz = blockgrid->Give_Nz_hexahedron(nummerBlock*5);
       }
       else if(typeUg==cylinderListUgTyp) {
          Nz = blockgrid->Give_Nz_hexahedron(nummerBlock*5);
       }
       else if(typeUg==hexahedronAdaptiveUgTyp) {
         Nz = blockgrid->Give_Nz_hexahedron(nummerBlock*5);
       }
       else {
         assert(false);
       }
       totalNumberSlices = totalNumberSlices + Nz;
   }
   totalNumberSlices = totalNumberSlices + 1;  
}

void IteratorZDirection::startBlock(int numBlock) {
  nummerBlock = numBlock;
  if(numBlock==0)  k = 0;
  else             k = 1;
  if(typeUg==hexahedronUgTyp) {
     Nz = blockgrid->Give_Nz_hexahedron(nummerBlock);
  }
  else if(typeUg==hexahedronListUgTyp) {
    Unstructured_grid* ug3Dp  = blockgrid->Give_unstructured_grid();
    Nz = blockgrid->Give_Nz_hexahedron(nummerBlock*((HexahedronXYList*)ug3Dp)->n_x*((HexahedronXYList*)ug3Dp)->n_y);
  }
  else if(typeUg==hexahedronListTestUgTyp) {
    Unstructured_grid* ug3Dp  = blockgrid->Give_unstructured_grid();
    Nz = blockgrid->Give_Nz_hexahedron(nummerBlock*((HexahedronXYListTest*)ug3Dp)->n_x*((HexahedronXYListTest*)ug3Dp)->n_y);
  }
  else if(typeUg==cylinderUgTyp) {
    Nz = blockgrid->Give_Nz_hexahedron(nummerBlock*5);
  }
  else if(typeUg==cylinderListUgTyp) {
    Nz = blockgrid->Give_Nz_hexahedron(nummerBlock*5);
  }
  else if(typeUg==hexahedronAdaptiveUgTyp) {
    Nz = blockgrid->Give_Nz_hexahedron(nummerBlock*5);
  }
  else {
    assert(false);
  }
  
  hz     = (ug_withZ->zRight(nummerBlock) - ug_withZ->zLeft(nummerBlock)) / Nz;
  zLeft  =  ug_withZ->zLeft(nummerBlock);;

  if(numBlock==0)   z_Wert = zLeft;
  else              z_Wert = zLeft + hz;
}

void IteratorZDirection::endBlock(int numBlock) {
  nummerBlock = numBlock;

  if(typeUg==hexahedronUgTyp) {
     Nz = blockgrid->Give_Nz_hexahedron(nummerBlock);
  }
  else if(typeUg==cylinderUgTyp) {
    Nz = blockgrid->Give_Nz_hexahedron(nummerBlock*5);
  }
  else if(typeUg==cylinderListUgTyp) {
    Nz = blockgrid->Give_Nz_hexahedron(nummerBlock*5);
  }
  else if(typeUg==hexahedronAdaptiveUgTyp) {
    Nz = blockgrid->Give_Nz_hexahedron(nummerBlock*5);
  }
  else {
    assert(false);
  }
  
  if(numBlock==nummernBloecke-1)  k = Nz;
  else             k = Nz-1;
  
  zLeft          =  ug_withZ->zLeft(nummerBlock);;
  double zRight  =  ug_withZ->zRight(nummerBlock);;
  hz     = (zRight - zLeft) / Nz;

  if(numBlock==nummernBloecke-1)   z_Wert = zRight;
  else              z_Wert = zRight - hz;
}

bool IteratorZDirection::nextDirection(bool right) {
     if(right) return next();
     return previous();
}

bool IteratorZDirection::next() {  
   if(k==Nz) {
     if(nummernBloecke == nummerBlock+1) return false;
     startBlock(nummerBlock+1);
     ++index;
     return true;
   }
   ++index;   
   ++k;
   z_Wert = zLeft + hz * k;
   return true;
}

bool IteratorZDirection::has_next() {  
   if(k==Nz) {
     if(nummernBloecke == nummerBlock+1) return false;
     return true;
   }
   return true;
}

bool IteratorZDirection::previous() {
   if(k==0) {
     if(0 == nummerBlock) return false;
     endBlock(nummerBlock-1);
     --index;
     return true;
   }
   --index;   
   --k;
   z_Wert = zLeft + hz * k;
   return true;
}

void IteratorZDirection::gotoStartDirection(bool right) {
   if(right) gotoFront();
   else      gotoEnd();
}

void IteratorZDirection::gotoFront() {
  index = 0;
   //Start-Werte setzen:
   startBlock(0);
}

void IteratorZDirection::gotoEnd() {
   //Start-Werte setzen:
   index = totalNumberSlices-1;
   endBlock(nummernBloecke-1);
}


/////////////////////////////////////////////////////////
// 2. find_k
/////////////////////////////////////////////////////////

     
int find_k(double zLeft, double zRight, double z, int Nz) {
   if(z<=zLeft)  return 0;
   if(z>=zRight) return Nz; 
   for(int k=1;k<=Nz;k++) {
       if(z <= zLeft + (zRight-zLeft)/Nz * k) return k;  
   }
   return Nz;
}
//
//
//template <>
//template <>
//void Variable2D< std::complex<double> >::interpolateSlizeZ(const Expr2D<A>&  U, double z) {
//   Blockgrid* blockgrid3D    = U->Give_blockgrid();
//   Unstructured_grid* ug3Dp  = blockgrid3D->Give_unstructured_grid();
//   TypeOfUG typeUg = ug3Dp->getUgTyp();
//   assert(typeUg!=noWithZTyp);
//   UG_WithZdirection* ug3DZ = static_cast<UG_WithZdirection*>(ug3Dp);
//
//   int numberCorners2D = ug->Give_number_points();
//
//   // Hexaeder -> Rectangle
//   if(typeUg==hexahedronUgTyp) {
//      assert(numberCorners2D==4);
//      int id = ug3DZ->getNummerBlockZ(z);
//
//      int Nx = blockgrid->Give_Nx_rectangle(0);
//      int Ny = blockgrid->Give_Ny_rectangle(0);
//      int Nz = blockgrid3D->Give_Nz_hexahedron(id);
//
//      double zLeft  = ug3DZ->zLeft(id);                           //ug3Dp->Give_point(0)->Give_coordinate().z;
//      double zRight = ug3DZ->zRight(id);                          //ug3Dp->Give_point(4)->Give_coordinate().z;
//      int kf = find_k(zLeft,zRight,z,Nz);
//
//      if(kf==0 || kf==Nz) { //links und rechts
//         int k=kf;
//         U->Update<hexahedronEl>(id);
//         for(int j = 0;j <= Ny;++j)
//             for(int i = 0;i <= Nx;++i) {
//                 data_rectangles[0][i+(Nx+1)*j] = std::complex<double>(U->Give_data<hexahedronEl> ( params_out ), 0.0);
//             }
//         Update_back(0);
//      }
//      else { //mitte
//         int k;
//         double zl = zLeft + (zRight-zLeft)/Nz * (kf-1);
//         double zr = zLeft + (zRight-zLeft)/Nz * (kf  );
//
//         U->Update<hexahedronEl>(id);
//         for(int j = 0;j <= Ny;++j)
//             for(int i = 0;i <= Nx;++i) {
//                 k=kf-1;
//                 double Ul = U->Give_data<hexahedronEl> ( params_out );
//                 k=kf;
//                 double Ur = U->Give_data<hexahedronEl> ( params_out );
//
//                 data_rectangles[0][i+(Nx+1)*j] = std::complex<double>(Ul + (Ur-Ul) * (z-zl)/(zr-zl),0.0);
//             }
//          Update_back(0);
//      }
//   }
//   else if(typeUg==cylinderUgTyp || typeUg==hexahedronAdaptiveUgTyp) { // Cylinder -> disc oder adaptiveHex -> adaptiveRec
//      assert(numberCorners2D==8);
//
//      int id_block = ug3DZ->getNummerBlockZ(z);
//      int Nz = blockgrid3D->Give_Nz_hexahedron(id_block);
//      int num_blocks = ug3DZ->getNummernBloecke();
//
//      double zLeft  = ug3DZ->zLeft(id_block);
//      double zRight = ug3DZ->zRight(id_block);
//
//      for(int id_rec = 0;id_rec<5;++id_rec) {
//          int Nx = blockgrid->Give_Nx_rectangle(id_rec);
//          int Ny = blockgrid->Give_Ny_rectangle(id_rec);
//          int id = id_rec*num_blocks + id_block;
//
//          int kf = find_k(zLeft,zRight,z,Nz);
//
//          if(kf==0 || kf==Nz) { //links und rechts
//             int k=kf;
//             U->Update<hexahedronEl>(id);
//             for(int j = 0;j <= Ny;++j)
//                 for(int i = 0;i <= Nx;++i) {
//                     data_rectangles[id_rec][i+(Nx+1)*j] = std::complex<double>(U->Give_data<hexahedronEl> ( params_out ),0.0);
//                 }
//             Update_back(id_rec);
//          }
//          else { //mitte
//             int k;
//             double zl = zLeft + (zRight-zLeft)/Nz * (kf-1);
//             double zr = zLeft + (zRight-zLeft)/Nz * (kf  );
//
//             U->Update<hexahedronEl>(id);
//             for(int j = 0;j <= Ny;++j)
//                 for(int i = 0;i <= Nx;++i) {
//                     k=kf-1;
//                     double Ul = U->Give_data<hexahedronEl> ( params_out );
//                     k=kf;
//                     double Ur = U->Give_data<hexahedronEl> ( params_out );
//
//                     data_rectangles[id_rec][i+(Nx+1)*j] = std::complex<double>(Ul + (Ur-Ul) * (z-zl)/(zr-zl), 0.0);
//                 }
//              Update_back(id_rec);
//          }
//      }
//   }
//   else {
//     assert(false);
//   }
//}

template <>
template <>
void Variable2D< std::complex<double> >::interpolateSlizeZ(const Variable< double >* U, double z) {
   Blockgrid* blockgrid3D    = U->Give_blockgrid();
   Unstructured_grid* ug3Dp  = blockgrid3D->Give_unstructured_grid();
   TypeOfUG typeUg = ug3Dp->getUgTyp();
   assert(typeUg!=noWithZTyp);
   UG_WithZdirection* ug3DZ = static_cast<UG_WithZdirection*>(ug3Dp);

   int numberCorners2D = ug->Give_number_points();

   // Hexaeder -> Rectangle
   if(typeUg==hexahedronUgTyp) {
      assert(numberCorners2D==4);
      int id = ug3DZ->getNummerBlockZ(z);

      int Nx = blockgrid->Give_Nx_rectangle(0);
      int Ny = blockgrid->Give_Ny_rectangle(0);
      int Nz = blockgrid3D->Give_Nz_hexahedron(id);

      double zLeft  = ug3DZ->zLeft(id);                           //ug3Dp->Give_point(0)->Give_coordinate().z;
      double zRight = ug3DZ->zRight(id);                          //ug3Dp->Give_point(4)->Give_coordinate().z;
      int kf = find_k(zLeft,zRight,z,Nz);

      if(kf==0 || kf==Nz) { //links und rechts
         int k=kf;
         U->Update<hexahedronEl>(id);
         for(int j = 0;j <= Ny;++j)
             for(int i = 0;i <= Nx;++i) {
                 data_rectangles[0][i+(Nx+1)*j] = std::complex<double>(U->Give_data<hexahedronEl> ( params_out ), 0.0);
             }
         Update_back(0);
      }
      else { //mitte
         //int k;
         double zl = zLeft + (zRight-zLeft)/Nz * (kf-1);
         double zr = zLeft + (zRight-zLeft)/Nz * (kf  );

         U->Update<hexahedronEl>(id);

	 #pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
         for(int j = 0;j <= Ny;++j)
             for(int i = 0;i <= Nx;++i) {
                 int k=kf-1;
                 double Ul = U->Give_data<hexahedronEl> ( params_out );
                 k=kf;
                 double Ur = U->Give_data<hexahedronEl> ( params_out );

                 data_rectangles[0][i+(Nx+1)*j] = std::complex<double>(Ul + (Ur-Ul) * (z-zl)/(zr-zl),0.0);
             }
          Update_back(0);
      }
   }
   else if(typeUg==cylinderUgTyp || typeUg==hexahedronAdaptiveUgTyp) { // Cylinder -> disc oder adaptiveHex -> adaptiveRec
      assert(numberCorners2D==8);

      int id_block = ug3DZ->getNummerBlockZ(z);
      int Nz = blockgrid3D->Give_Nz_hexahedron(id_block);
      int num_blocks = ug3DZ->getNummernBloecke();

      double zLeft  = ug3DZ->zLeft(id_block);
      double zRight = ug3DZ->zRight(id_block);

      for(int id_rec = 0;id_rec<5;++id_rec) {
          int Nx = blockgrid->Give_Nx_rectangle(id_rec);
          int Ny = blockgrid->Give_Ny_rectangle(id_rec);
          int id = id_rec*num_blocks + id_block;

          int kf = find_k(zLeft,zRight,z,Nz);

          if(kf==0 || kf==Nz) { //links und rechts
             int k=kf;
             U->Update<hexahedronEl>(id);
	     
	     #pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) {
                     data_rectangles[id_rec][i+(Nx+1)*j] = std::complex<double>(U->Give_data<hexahedronEl> ( params_out ),0.0);
                 }
          }
          else { //mitte
             //int k;
             double zl = zLeft + (zRight-zLeft)/Nz * (kf-1);
             double zr = zLeft + (zRight-zLeft)/Nz * (kf  );

             U->Update<hexahedronEl>(id);

	     #pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
             for(int j = 0;j <= Ny;++j)
                 for(int i = 0;i <= Nx;++i) {
                     int k=kf-1;
                     double Ul = U->Give_data<hexahedronEl> ( params_out );
                     k=kf;
                     double Ur = U->Give_data<hexahedronEl> ( params_out );

                     data_rectangles[id_rec][i+(Nx+1)*j] = std::complex<double>(Ul + (Ur-Ul) * (z-zl)/(zr-zl), 0.0);
                 }
              Update_back(id_rec);
          }
      }
   }
   else {
     assert(false);
   }
}
