/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#ifndef PRINTVAR2DVTKCCH
#define PRINTVAR2DVTKCCH

#include "variable2D.h"
//#include "../../../linAlgExptemp/source/linAlg.h"

//#include "cellvar.h"
#include <QtGui/QtGui>
#include <typeinfo>



template<class DTyp>
void Variable2D<DTyp>::Print_VTK ( std::ostream& Datei, std::string title) {
    D2vector v;
    Datei.precision(20);
    Datei.setf(std::ios::fixed,std::ios::floatfield);
    int num_total  = 0;
    int cell_total = 0;

    // Teil 0: Write information
    Datei << "# vtk DataFile Version 2.0\n"
            << title << std::endl
            << "ASCII\n"
            << "DATASET UNSTRUCTURED_GRID\n";

     for(int id_rec=0;id_rec < ug->Give_number_rectangle();++id_rec) {
         Update<rectangleEl>(id_rec);
         num_total  += blockgrid->Give_N_total_rectangle(id_rec);
         cell_total +=     (blockgrid->Give_Nx_rectangle(id_rec))
                          *(blockgrid->Give_Ny_rectangle(id_rec));
      }

    // Teil 1: Kopfzeile schreiben
    Datei << "POINTS " << num_total << " float\n";

    // Teil 2: Koordinaten der Punkte ausgeben
    for(int id_rec=0;id_rec<ug->Give_number_rectangle();++id_rec) {
      int Ny = blockgrid->Give_Ny_rectangle(id_rec);
      int Nx = blockgrid->Give_Nx_rectangle(id_rec);


        for(int j=0;j<=Ny;++j)
            for(int i=0;i<=Nx;++i) {
              v = blockgrid->Give_coord_rectangle(id_rec,i,j);
              Datei << v.x
                    << " " << v.y
                    << " " << 0.0
                    << "\n";

          }
    }

      // Teil 3: Zellen ausgeben
      Datei << "\nCELLS " << cell_total << " " <<  cell_total*5  << "\n";

      int id_start;
      id_start = 0;
      for(int id_rec=0;id_rec<ug->Give_number_rectangle();++id_rec) {
         int Nx = blockgrid->Give_Nx_rectangle(id_rec);
         int Ny = blockgrid->Give_Ny_rectangle(id_rec);

         for(int j=0;j<Ny;++j) for(int i=0;i<Nx;++i) {
                 Datei << "4 "
                       << id_start + i   + (Nx+1)*j     << " " // WS
                       << id_start + i+1 + (Nx+1)*j     << " " // ES
                       << id_start + i+1 + (Nx+1)*(j+1) << " " // EN
                       << id_start + i   + (Nx+1)*(j+1) << " " // WN
                       << "\n";
             }
        id_start = id_start +  blockgrid->Give_N_total_rectangle(id_rec);
      }

      Datei << "\nCELL_TYPES " << cell_total << "\n";

      for(int id_rec=0;id_rec<ug->Give_number_rectangle();++id_rec) {
        int Nx = blockgrid->Give_Nx_rectangle(id_rec);
        int Ny = blockgrid->Give_Ny_rectangle(id_rec);
        for(int j=0;j<Ny;++j) for(int i=0;i<Nx;++i)
              Datei << "9"
                    << "\n";
      }

    // Teil 4:
      Datei << "POINT_DATA " << num_total << "\n";
      if(typeid(DTyp) == typeid(std::complex<double>) )
      {
          Datei << "SCALARS " << title << " float 2\n";
      }
      else
          Datei << "SCALARS " << title << " float 1\n";

      Datei << "LOOKUP_TABLE default\r\n";

    // Teil 6: Ausgabe der Werte der Variable2Dref
    for(int id_rec=0;id_rec<ug->Give_number_rectangle();++id_rec) {
      int Ny = blockgrid->Give_Ny_rectangle(id_rec);
      int Nx = blockgrid->Give_Nx_rectangle(id_rec);

      for(int j=0;j<=Ny;++j) for(int i=0;i<=Nx;++i) {
          if(typeid(DTyp) == typeid(std::complex<double>) ) {
                  Datei << ((std::complex<double>)(data_rectangles[id_rec][i+(Nx+1)*j])).real() << "\t"
                            << ((std::complex<double>)(data_rectangles[id_rec][i+(Nx+1)*j])).imag() << "\n";
          }
          else
                  Datei << data_rectangles[id_rec][i+(Nx+1)*j] << "\n";
  /*
  //test
  v = blockgrid->Give_coord_rectangle(id_rec,i,j);
  if((v.x*v.x+v.y*v.y) > 3.9) 	     cout << " x: " << v.x
                    << " y: " << v.y
                    << " value: " << data_rectangles[id_rec][i+(Nx+1)*j]
                    << "\n";
  */

          }
    }

    // Teil 7: Ausgabe der Werte der Variable2D mit driter Spalte 0.0 fuer paraview-vector-darstellung
    if(typeid(DTyp) == typeid(std::complex<double>) ) {
      Datei << "VECTORS " << title << "_vectors float\n";
      for(int id_rec=0;id_rec<ug->Give_number_rectangle();++id_rec) {
        int Ny = blockgrid->Give_Ny_rectangle(id_rec);
        int Nx = blockgrid->Give_Nx_rectangle(id_rec);

        for(int j=0;j<=Ny;++j) for(int i=0;i<=Nx;++i) {
          Datei << ((std::complex<double>)(data_rectangles[id_rec][i+(Nx+1)*j])).real() << "\t"
                << ((std::complex<double>)(data_rectangles[id_rec][i+(Nx+1)*j])).imag() << "\t"
                << 0.0 << "\n";
        }
      }
    }
}


#endif //#ifndef PRINTVARVTKCCH
