/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// update.h
//
// ------------------------------------------------------------

#ifndef UPDATE2D_H
#define UPDATE2D_H

#ifdef _OPENMP
#include <omp.h>
//#include <source/abbrevi.h>
#endif




template <class DTyp>
void Update_back_rectangle(int id_quad, Blockgrid2D* blockgrid, DTyp** data_rectangles,
                           DTyp** data_edges, DTyp** data_points,
                           int my_rank, MPI_Comm comm ) {

        Unstructured2DGrid* ug = blockgrid->Give_unstructured_grid();
  
        int Nxedge, NconstEdge, Fedge;  // fuer Indizes auf Kante           i*Fedge+NconstEdge
	int Nconst, Ni;                 // fuer Indizes auf Rechteck
        int id_edge;
	int Nxquad = blockgrid->Give_Nx_rectangle ( id_quad );
	int Nyquad = blockgrid->Give_Ny_rectangle ( id_quad );
        // index:  i + j * (Nxquad+1),  i=0,...,Nxquad, j=0,...,Nyquad
	
	// I. edges Wdir2D, Edir2D, Sdir2D, Ndir2D
        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Edir2D
	//--------
	// a) Kanten Indizes
	id_edge = ug->Give_rectangle(id_quad)->Give_id_edge (Edir2D);
	Nxedge     = blockgrid->Give_Nx_edge (id_edge);
        if(ug->getOrientationEdge(id_quad,Edir2D)) { NconstEdge = Nxedge;  Fedge      = -1; }
	else                                       { NconstEdge = 0;	   Fedge      =  1; }
	
	// b) Rechteck Indizes
	Nconst = Nxquad;
	Ni     = Nxquad+1;

	// c) Iteration
	for(int j=1; j<Nxedge; ++j )
	    // vorwaerts:
            //data_rectangles[id_quad][Nconst + j*Ni] = data_edges[id_edge][j*Fedge+NconstEdge];
            data_edges[id_edge][j*Fedge+NconstEdge] = data_rectangles[id_quad][Nconst + j*Ni];

	// Wdir2D
	//--------
	
	// a) Kanten Indizes
	id_edge = ug->Give_rectangle(id_quad)->Give_id_edge (Wdir2D);
	Nxedge     = blockgrid->Give_Nx_edge(id_edge);
        if(ug->getOrientationEdge(id_quad,Wdir2D)) { NconstEdge = Nxedge;  Fedge      = -1; }
	else                                       { NconstEdge = 0;	   Fedge      =  1; }
	
	// b) Rechteck Indizes
	Nconst = 0;
	Ni     = Nxquad+1;

	// c) Iteration
	for(int j=1; j<Nxedge; ++j )
	    // vorwaerts:
            // data_rectangles[id_quad][Nconst + j*Ni] = data_edges[id_edge][j*Fedge+NconstEdge];
            data_edges[id_edge][j*Fedge+NconstEdge] = data_rectangles[id_quad][Nconst + j*Ni];

	// Sdir2D
	//--------
	// a) Kanten Indizes
	id_edge = ug->Give_rectangle(id_quad)->Give_id_edge (Sdir2D);
	Nxedge     = blockgrid->Give_Nx_edge(id_edge);
        if(ug->getOrientationEdge(id_quad,Sdir2D)) { NconstEdge = Nxedge;  Fedge      = -1; }
	else                                       { NconstEdge = 0;	   Fedge      =  1; }
	
	// b) Rechteck Indizes

	// c) Iteration
	for(int i=1; i<Nxedge; ++i )
	    // vorwaerts:
            // data_rectangles[id_quad][i] = data_edges[id_edge][i*Fedge+NconstEdge];
	    data_edges[id_edge][i*Fedge+NconstEdge] = data_rectangles[id_quad][i];
	
	// Ndir2D
	//--------
	// a) Kanten Indizes
	id_edge = ug->Give_rectangle(id_quad)->Give_id_edge (Ndir2D);
	Nxedge     = blockgrid->Give_Nx_edge(id_edge);
        if(ug->getOrientationEdge(id_quad,Ndir2D)) { NconstEdge = Nxedge;  Fedge      = -1; }
	else                                       { NconstEdge = 0;	   Fedge      =  1; }
	
	// b) Rechteck Indizes
	Nconst = (Nxquad+1)*(Nyquad+1-1);

	// c) Iteration
	for(int i=1; i<Nxedge; ++i )
            //data_rectangles[id_quad][i+Nconst] = data_edges[id_edge][i*Fedge+NconstEdge];
	    data_edges[id_edge][i*Fedge+NconstEdge] = data_rectangles[id_quad][i+Nconst];
	
	// II. corners
        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// index:  i + j * (Nxquad+1),  i=0,...,Nxquad, j=0,...,Nyquad
	// WS corner
	// vorwaerts    
	//data_rectangles[id_quad][0]                       = data_points[ug->Give_rectangle(id_quad)->Give_id_corner(SWdir2D)][0];
	//data_rectangles[id_quad][Nxquad]                  = data_points[ug->Give_rectangle(id_quad)->Give_id_corner(SEdir2D)][0];
        //data_rectangles[id_quad][(Nxquad+1)*Nyquad]       = data_points[ug->Give_rectangle(id_quad)->Give_id_corner(NWdir2D)][0];
    	//data_rectangles[id_quad][(Nxquad+1)*(Nyquad+1)-1] = data_points[ug->Give_rectangle(id_quad)->Give_id_corner(NEdir2D)][0];
	data_points[ug->Give_rectangle(id_quad)->Give_id_corner(SWdir2D)][0] = data_rectangles[id_quad][0];
	data_points[ug->Give_rectangle(id_quad)->Give_id_corner(SEdir2D)][0] = data_rectangles[id_quad][Nxquad];
        data_points[ug->Give_rectangle(id_quad)->Give_id_corner(NWdir2D)][0] = data_rectangles[id_quad][(Nxquad+1)*Nyquad];
    	data_points[ug->Give_rectangle(id_quad)->Give_id_corner(NEdir2D)][0] = data_rectangles[id_quad][(Nxquad+1)*(Nyquad+1)-1];
}


template <class DTyp>
void Update_rectangle(int id_quad, Blockgrid2D* blockgrid, DTyp** data_rectangles,
                      DTyp** data_edges, DTyp** data_points,
                      int my_rank, MPI_Comm comm ) {

        Unstructured2DGrid* ug = blockgrid->Give_unstructured_grid();
  
        int Nxedge, NconstEdge, Fedge;  // fuer Indizes auf Kante           i*Fedge+NconstEdge
	int Nconst, Ni;                 // fuer Indizes auf Rechteck
        int id_edge;

	int Nxquad = blockgrid->Give_Nx_rectangle ( id_quad );
	int Nyquad = blockgrid->Give_Ny_rectangle ( id_quad );
        // index:  i + j * (Nxquad+1),  i=0,...,Nxquad, j=0,...,Nyquad
	
	// I. edges Wdir2D, Edir2D, Sdir2D, Ndir2D
        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Edir2D
	//--------
	// a) Kanten Indizes
	id_edge = ug->Give_rectangle(id_quad)->Give_id_edge (Edir2D);
	Nxedge     = blockgrid->Give_Nx_edge (id_edge);
        if(ug->getOrientationEdge(id_quad,Edir2D)) { NconstEdge = Nxedge;  Fedge      = -1; }
	else                                       { NconstEdge = 0;	   Fedge      =  1; }
	
	// b) Rechteck Indizes
	Nconst = Nxquad;
	Ni     = Nxquad+1;

	// c) Iteration
	for(int j=1; j<Nxedge; ++j )
            data_rectangles[id_quad][Nconst + j*Ni] = data_edges[id_edge][j*Fedge+NconstEdge];

	// Wdir2D
	//--------
	
	// a) Kanten Indizes
	id_edge = ug->Give_rectangle(id_quad)->Give_id_edge (Wdir2D);
	Nxedge     = blockgrid->Give_Nx_edge(id_edge);
        if(ug->getOrientationEdge(id_quad,Wdir2D)) { NconstEdge = Nxedge;  Fedge      = -1; }
	else                                       { NconstEdge = 0;	   Fedge      =  1; }
	
	// b) Rechteck Indizes
	Nconst = 0;
	Ni     = Nxquad+1;

	// c) Iteration
	for(int j=1; j<Nxedge; ++j )
            data_rectangles[id_quad][Nconst + j*Ni] = data_edges[id_edge][j*Fedge+NconstEdge];

	// Sdir2D
	//--------
	// a) Kanten Indizes
	id_edge = ug->Give_rectangle(id_quad)->Give_id_edge (Sdir2D);
	Nxedge     = blockgrid->Give_Nx_edge(id_edge);
        if(ug->getOrientationEdge(id_quad,Sdir2D)) { NconstEdge = Nxedge;  Fedge      = -1; }
	else                                       { NconstEdge = 0;	   Fedge      =  1; }
	
	// b) Rechteck Indizes

	// c) Iteration
	for(int i=1; i<Nxedge; ++i )
            data_rectangles[id_quad][i] = data_edges[id_edge][i*Fedge+NconstEdge];
	
	// Ndir2D
	//--------
	// a) Kanten Indizes
	id_edge = ug->Give_rectangle(id_quad)->Give_id_edge (Ndir2D);
	Nxedge     = blockgrid->Give_Nx_edge(id_edge);
        if(ug->getOrientationEdge(id_quad,Ndir2D)) { NconstEdge = Nxedge;  Fedge      = -1; }
	else                                       { NconstEdge = 0;	   Fedge      =  1; }
	
	// b) Rechteck Indizes
	Nconst = (Nxquad+1)*(Nyquad+1-1);

	// c) Iteration
	for(int i=1; i<Nxedge; ++i )
            data_rectangles[id_quad][i+Nconst] = data_edges[id_edge][i*Fedge+NconstEdge];
	
	// II. corners
        // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// index:  i + j * (Nxquad+1),  i=0,...,Nxquad, j=0,...,Nyquad
	// WS corner
	data_rectangles[id_quad][0]                       = data_points[ug->Give_rectangle(id_quad)->Give_id_corner(SWdir2D)][0];
	data_rectangles[id_quad][Nxquad]                  = data_points[ug->Give_rectangle(id_quad)->Give_id_corner(SEdir2D)][0];
        data_rectangles[id_quad][(Nxquad+1)*Nyquad]       = data_points[ug->Give_rectangle(id_quad)->Give_id_corner(NWdir2D)][0];
    	data_rectangles[id_quad][(Nxquad+1)*(Nyquad+1)-1] = data_points[ug->Give_rectangle(id_quad)->Give_id_corner(NEdir2D)][0];

}


template <class DTyp>
void Update_edge(int id_edge, Blockgrid2D* blockgrid, DTyp** data_rectangles,
                 DTyp** data_edges, DTyp** data_points,
                 DTyp**  data_interior_edge,
                 DTyp**  data_exterior_edge, int my_rank, MPI_Comm comm ) {
        Unstructured2DGrid* ug = blockgrid->Give_unstructured_grid();

        int Nxedge = blockgrid->Give_Nx_edge(id_edge);
        EdgeEl* edge = ug->Give_edge(id_edge);
	
	int id_W_poi = edge->Give_id_corner_W();
	int id_E_poi = edge->Give_id_corner_E();
	
	// Ecken der Kante
	//~~~~~~~~~~~~~~~~~~
	data_edges[id_edge][0]      = data_points[id_W_poi][0];
	data_edges[id_edge][Nxedge] = data_points[id_E_poi][0];

	// rectangle
	//~~~~~~~~~~~~~~~~~~~
	for(int k=0;k < edge->getNumberNextRectangles();++k) {
	   int id_rec = edge->getIdNextRectangle((bool)k);  
	   int Nxrec = blockgrid->Give_Nx_rectangle(id_rec);
           int Nyrec = blockgrid->Give_Ny_rectangle(id_rec);
       
	   // a) Find corresponding points in rectangle
           dir2D_sons c_W;
           dir2D_sons c_E;

   	   for(int c = 0;c < 4;++c) {
	       int idsp = ug->Give_rectangle(id_rec)->Give_id_corner((dir2D_sons)c);
					
               if(idsp == id_W_poi ) c_W = (dir2D_sons) c;
	       if(idsp == id_E_poi ) c_E = (dir2D_sons) c;
	   }
				
           // b) Finde Orientierung und map
           bool clockOrientation = false;
	   if(c_W == SWdir2D && c_E == NWdir2D)  clockOrientation = true;
	   if(c_W == NWdir2D && c_E == NEdir2D)  clockOrientation = true;
	   if(c_W == NEdir2D && c_E == SEdir2D)  clockOrientation = true;
	   if(c_W == SEdir2D && c_E == SWdir2D)  clockOrientation = true;

	   dir2D_sons map_sons[4];
	   calculate_map(c_W,clockOrientation,map_sons);				
				
				
	   // c) Indizes Konstanten			
           int Nconst = Blockgrid2D::Give_Nconst(c_W, Nxrec, Nyrec);
	   int Ni; // beachte: (j*(Nx+1)+i)
	   dir2D direction = direction_from_to(c_W,c_E);
	   if(direction==Wdir2D) Ni=-1;
	   if(direction==Edir2D) Ni= 1;
	   if(direction==Ndir2D) Ni= (Nxrec+1);
	   if(direction==Sdir2D) Ni=-(Nxrec+1);
						       
	   // d) Determine if rec is interior or exterior volume		
	   DTyp* loc_pointer;				       
	   if(k==0) { //interior
	       loc_pointer = data_interior_edge[id_edge];
	   } 
	   else { // exterior
	       loc_pointer = data_exterior_edge[id_edge];
	   }
				
	    // e) kopiere		
	    for(int i = 0;i < Nxedge-1;++i) {
		loc_pointer[i+1] = data_rectangles[id_rec][i*Ni+Nconst];
	    }
	    
	    
	   // edge an Ecken der Kante
	   //~~~~~~~~~~~~~~~~~~~ 
	   mainDirection2d mainDir  = mainDirection(c_W,c_E);
	   mainDirection2d orthoDir = orthogonalDirection(mainDir);

/*	   
assert(id_W_poi < id_E_poi);	   
	   
cout << " copy: " << endl;
Abbreviations::Print(c_W); cout << endl;
Abbreviations::Print(c_E); cout << endl;
cout << " mainDir: "; Abbreviations::Print(mainDir); cout << endl;
cout << " orthoDir: "; Abbreviations::Print(orthoDir); cout << endl;
*/
	   //next West
	   {
	     dir2D_sons nextW = nextInDirection(c_W,orthoDir);
	     dir2D nextEdgeW = edgeOfCorners(c_W,nextW);
	     
//cout << " copy: " << endl;
//cout << " nextEdgeW: "; Abbreviations::Print(nextEdgeW); cout << endl;
	   
	     
	     int id_edgeW = ug->Give_rectangle(id_rec)->Give_id_edge(nextEdgeW);
	     if(ug->Give_edge(id_edgeW)->Give_id_corner_W()==id_W_poi) {
	        loc_pointer[0] = data_edges[id_edgeW][1];
	     }
             else {
	        int Nx_W = blockgrid->Give_Nx_edge(id_edgeW);
	        loc_pointer[0] = data_edges[id_edgeW][Nx_W-1];
	     }
	   }
	   //next East
	   {
	     dir2D_sons nextE = nextInDirection(c_E,orthoDir);
	     dir2D nextEdgeE = edgeOfCorners(c_E,nextE);

	     int id_edgeE = ug->Give_rectangle(id_rec)->Give_id_edge(nextEdgeE);

/*	     
if(id_edgeE==2) {
cout << " copy: " << endl;
cout << " nextEdgeE: "; Abbreviations::Print(nextEdgeE); cout << endl;
cout << " id_edgeE : " <<  id_edgeE << endl;
cout << " id_edge : " <<  id_edge << endl;
}
*/

	     if(ug->Give_edge(id_edgeE)->Give_id_corner_W()==id_E_poi) {
	        loc_pointer[Nxedge] = data_edges[id_edgeE][1];
	     }
             else {
	        int Nx_E = blockgrid->Give_Nx_edge(id_edgeE);
	        loc_pointer[Nxedge] = data_edges[id_edgeE][Nx_E-1];

	     }
	   }
	}
}


template <class DTyp>
void Update_point(int id_point, Blockgrid2D* blockgrid, DTyp** data_rectangles,
                  DTyp** data_edges, DTyp** data_points,
                  DTyp**  data_neighbor_points, int my_rank, MPI_Comm comm ) {

        Unstructured2DGrid* ug = blockgrid->Give_unstructured_grid();
	PointEl* point = ug->Give_point(id_point);
	
	// a) Iteration über Daten an Kanten
	int num_neighbor_edges = point->give_number_neighbors_edges();
	for(int pos=0; pos<num_neighbor_edges;++pos) {
	    int edge_id = point->give_edgeId(pos);
	    int Nxedge = blockgrid->Give_Nx_edge(edge_id);
	    if(ug->Give_edge(edge_id)->Give_id_corner(Wd) == id_point) { // normale Orientierung
	       data_neighbor_points[id_point][pos] = data_edges[edge_id][1];
//               data_neighbor_points[id_point][pos] = data_edges[edge_id][0];
	    }
	    else { // umgekehrte Orientierung
	       data_neighbor_points[id_point][pos] = data_edges[edge_id][Nxedge-1];
//               data_neighbor_points[id_point][pos] = data_edges[edge_id][Nxedge];
	    }

//cout << " update at id: " <<  	id_point << " position: " << pos << endl;
	    
	}
	// b) Iteration über Daten an Rectangle
	for(int pos_rec=0;pos_rec<point->give_number_neighbors_rec();++pos_rec) {
	    int pos = num_neighbor_edges + pos_rec;
	    int rec_id = point->give_id_rec(pos_rec);
	    int Nxrec = blockgrid->Give_Nx_rectangle(rec_id);
	    int Nyrec = blockgrid->Give_Ny_rectangle(rec_id);
	    RectangleEl* rectangle = ug->Give_rectangle(rec_id);
	    dir2D_sons corner;
	    for(int c=0; c<4; ++c ) {
		   int id_corner = rectangle->Give_id_corner((dir2D_sons)c);
		   if(id_corner==id_point) corner = (dir2D_sons)c;
	    }
	    int Nconst = Blockgrid2D::Give_Nconst(corner, Nxrec, Nyrec);

	    assert(Nconst < (Nxrec+1)*(Nyrec+1));
	    data_neighbor_points[id_point][pos] = data_rectangles[rec_id][Nconst];
/*	    
cout << " update at id: " <<  	id_point << " position: " << pos 
                          << " wert: " << data_neighbor_points[id_point][pos]  << endl;
*/	    
	}
}

//----------------------------------------------------------------------------




#endif // UPDATE2D_H
