/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// variable.cc
//
// ------------------------------------------------------------

#include <string>
#include <stdexcept>
#include "../../../program/source/mympi.h"  // von 3D UGBlocks
#include "../abbrevi2D.h"
//#include "../../../linAlgExptemp/source/linAlg.h"
//#include "../parameter.h"
#include "../math_lib/math_lib.h"
#include "../grid/elements2D.h"
#include "../grid/parti2D.h"
#include "../grid/ug2D.h"
#include "../grid/blockgrid2D.h"
#include "../grid/marker2D.h"
//#include "parallel.h"
#include "extemp2D.h"
#include "variable2D.h"
#include "variable2D_cc.h"
#include "update2D.h"
#include "print_var2D_vtk_cc.h"
#include "../grid/examples_ug2D.h"
#include "../../../common_source/mathlib/fouriertransform.h"

template <>
void Variable2D<double>::Update_back(int id_rec) const{
   Update_back_rectangle<double>(id_rec, blockgrid,
                            data_rectangles,
                            data_edges, data_points,
			     my_rank, comm);

}


template <>
void Variable2D<std::complex<double> >::Update_back(int id_rec) const{
   Update_back_rectangle<std::complex<double> >(id_rec, blockgrid,
                                           data_rectangles,
                                           data_edges, data_points,
			                    my_rank, comm);

}



template <>
template <>
void Variable2D<int>::Update<rectangleEl>(int id_rec) const{
   Update_rectangle<int>(id_rec, blockgrid,
                            data_rectangles,
                            data_edges, data_points,
			     my_rank, comm);

}

template <>
template <>
void Variable2D<bool>::Update<rectangleEl>(int id_rec) const{
   Update_rectangle<bool>(id_rec, blockgrid,
                            data_rectangles,
                            data_edges, data_points,
			     my_rank, comm);

}


template <>
template <>
void Variable2D<double>::Update<rectangleEl>(int id_rec) const{
   Update_rectangle<double>(id_rec, blockgrid,
                            data_rectangles,
                            data_edges, data_points,
			     my_rank, comm);

}


template <>
template <>
void Variable2D<std::complex<double> >::Update<rectangleEl>(int id_rec) const{
   Update_rectangle<std::complex<double> >(id_rec, blockgrid,
                                           data_rectangles,
                                           data_edges, data_points,
			                    my_rank, comm);

}




template <>
template <>
void Variable2D<double>::Update<edgeEl>(int id_edge) const{
 Update_edge<double>(id_edge, blockgrid, data_rectangles, data_edges,
                     data_points,
		      data_interior_edges, data_exterior_edges,
		      my_rank, comm);
}


template <>
template <>
void Variable2D<std::complex<double> >::Update<edgeEl>(int id_edge) const{
   Update_edge<std::complex<double> >(id_edge, blockgrid, data_rectangles, data_edges,
                     data_points,
		      data_interior_edges, data_exterior_edges,
		      my_rank, comm);
}



template <>
template <>
void Variable2D<int>::Update<edgeEl>(int id_edge) const{
   Update_edge<int>(id_edge, blockgrid, data_rectangles, data_edges,
                    data_points,
		     data_interior_edges, data_exterior_edges,
		     my_rank, comm);
}


template <>
template <>
void Variable2D<double>::Update<pointEl>(int id_point) const{
 Update_point<double>(id_point, blockgrid, data_rectangles, data_edges,
                      data_points, data_neighbor_points, my_rank, comm);
}

template <>
template <>
void Variable2D<std::complex<double> >::Update<pointEl>(int id_point) const{
 Update_point<std::complex<double> >(id_point, blockgrid, data_rectangles,
                                     data_edges, data_points,
                                     data_neighbor_points, my_rank, comm);
}



template <>
template <>
void Variable2D<int>::Update<pointEl>(int id_point) const{
 Update_point<int>(id_point, blockgrid, data_rectangles,
                   data_edges, data_points,
                   data_neighbor_points, my_rank, comm);
}


Iteration_description2D::Iteration_description2D ( Marker2D& m ) : marker ( &m ) { red_black = false; }



