/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// variable2D.h
//
// ------------------------------------------------------------
#include "../grid/marker2D.h"
#include "extemp2D.h"
#ifndef VARIABLE2D_H
#define VARIABLE2D_H
#ifdef _OPENMP
#include <omp.h>
#include "../ugopenmp.h"
#endif

class Rectangle;

template <typename TYPE>       class ExpressionRestriction2D;

template <typename TYPE>       class ExpressionProlongation2D;

template <class A, class DTyp> class Expr2D_Equation_of_Array;

class IteratorZDirection;

class QString;

class Iteration_description2D {

        public:
                Iteration_description2D ( Marker2D& m );

                Marker2D* marker;
                bool red_black;
};

template <class A>
class Expr2D_Restriction : public Expr2D<Expr2D_Restriction< A > > {
                const A& a_;
                const Marker2D& marker;

        public:
                Expr2D_Restriction ( const Expr2D<A>& a, const Marker2D& mark ) : a_ ( a ), marker ( mark ) {}

                Blockgrid2D* Give_blockgrid() const { return a_.Give_blockgrid(); };

                stencil_typ Give_stencil_typ() const { return a_.Give_stencil_typ(); }

                typedef typename A::Result Result;

                template <elementTyp TYP_EL>
                inline Result Give_data ( params2D_in ) const {
                        return a_.template Give_data<TYP_EL> ( params2D_out );
                }

                inline Result Give_cell_hexahedra ( params2D_in_cell ) const  {
                        return a_.Give_cell_hexahedra ( params2D_out_cell );
                }

                template <elementTyp TYP_EL>
                void Update ( int id ) const {
                        a_.template Update<TYP_EL> ( id );
                }

                template <elementTyp TYP_EL>
                marker_typ Give_marker ( int i ) const{ return marker.Give_marker<TYP_EL> ( i ); }
};

template <typename DTyp>
class DTyp_Restriction2D {
                DTyp x_;
                const Marker2D& marker;

        public:
                DTyp_Restriction2D ( DTyp x, const Marker2D& mark ) : x_ ( x ), marker ( mark ) {}

                ~DTyp_Restriction2D() {}

                template <elementTyp TYP_EL>
                marker_typ Give_marker ( int i ) const{ return marker.Give_marker<TYP_EL> ( i ); }

                inline DTyp Give_x() const { return x_; }
};

inline DTyp_Restriction2D<bool> operator| ( bool x, Marker2D& sg ) {
        return DTyp_Restriction2D<bool> ( x, sg );
}

inline DTyp_Restriction2D<double> operator| ( double x, Marker2D& sg ) {
        return DTyp_Restriction2D<double> ( x, sg );
}

inline DTyp_Restriction2D<std::complex<double> > operator| ( std::complex<double> x, Marker2D& sg ) {
        return DTyp_Restriction2D<std::complex<double> > ( x, sg );
}


template <class A>
class Expr2D_Restriction_Iterdes {
                const A& a_;

        public:
                Expr2D_Restriction_Iterdes ( const Expr2D<A>& a, const Iteration_description2D& ides ) : a_ ( a ), iterdes ( ides ) {}

                const Iteration_description2D& iterdes;

                Blockgrid2D* Give_blockgrid() const { return a_.Give_blockgrid(); };

                stencil_typ Give_stencil_typ() const { return a_.Give_stencil_typ(); }

                typedef typename A::Result Result;

                template <elementTyp TYP_EL>
                inline Result Give_data ( params2D_in ) const {
                        return a_.template Give_data<TYP_EL> ( params2D_out );
                }

                template <elementTyp TYP_EL>
                void Update ( int id ) const {
                        a_.template Update<TYP_EL> ( id );
                }


                template <elementTyp TYP_EL>
                marker_typ Give_marker ( int i ) const{ return iterdes.marker->Give_marker<TYP_EL> ( i ); }
};


inline DTyp_Restriction2D<bool> operator| ( bool x, Marker2D& sg );

inline DTyp_Restriction2D<int> operator| ( int x, Marker2D& sg );

inline DTyp_Restriction2D<double> operator| ( double x, Marker2D& sg );

inline DTyp_Restriction2D<std::complex<double> > operator| ( std::complex<double> x, Marker2D& sg );

///////////////////////////
// 2. Variable2D for points
///////////////////////////

class OutputVTK;

template <class DTyp>                       class Variable;
template <class DTyp>                       class Variable2D;

template <class A, class DTyp, class DTyp1> class Exp_Function1;
template <class A, class B, class C, class D, class DTyp, class DTyp_Result> class Exp_Function4;

class CalcContinuousArg;

template <class DTyp, class Vari> class VariableVector;


/** \addtogroup ExpressionTemplates2D **/
/* @{ */  
template <class DTyp>
class Variable2D : public Expr2D<Variable2D< DTyp > > {
        friend class CalcContinuousArg;
        friend  VariableVector<DTyp, Variable2D<DTyp> >;	
		
        public:
                typedef DTyp Result;

                Variable2D ( Blockgrid2D& blockgrid_ );

                Variable2D ( const Variable2D<DTyp>& other );
                         
                ~Variable2D();          

		        int getGridId() { return idGrid; }  // muss nochmal richtig implementiert werden

                void test();

                template <class DTyp1>
                void interpolateSlizeZ(const Variable<DTyp1>* U, double z);

                template <class DTyp1>
                void interpolateSlizeZ(const Variable<DTyp1>* U, IteratorZDirection* iterZ);		
		
                template <class DTyp1, class A>
                void interpolateSlizeZ(const Exp_Function1<A, DTyp1, DTyp>&  U, double z);

                template <class A, class B, class C, class D, class DTyp1>
                void interpolateSlizeZ(const Exp_Function4<A, B, C, D, DTyp1,DTyp>&  U, double z);
		
                void integrateZ(const Variable<DTyp>* U);

                Blockgrid2D* Give_blockgrid() const { return blockgrid; };

                Unstructured2DGrid * Give_Ug() const { return ug;};

                //! QT-VTK print
                void QPrint_VTK(QString DateiName, QString title = QString("myfile"));
                //! QT-VTK print with convert
                void QPrint_VTK(QString DateiName,
                                double ( *convert ) ( std::complex<double> x ), 
                                QString title = QString("myfile"));
                
                //! Prints in VTK using std, auch vector bei complex ...
                void Print_VTK ( std::ostream& Datei, std::string title = "myData");
                void Print_VTK ( std::ostream& Datei, 
                                 double ( *convert ) ( std::complex<double> x ), 
                                 std::string title = "myData");
                
                // copy of data from hex_a to hex_b
                void Copy_invert_z ( int hex_b, Variable2D<DTyp>& a, int hex_a );

                template <elementTyp TYP_EL>
                inline DTyp Give_data ( params2D_in ) const;
		
		// fuer debug-zwecke mit positionsrueckgabe
                template <elementTyp TYP_EL>
                inline DTyp Give_data ( params2D_in, D2vector &posVector ) const ;

                inline DTyp Give_interpolated_cell_rectangle(int id_rec, int i, int j, int Nx) const; ///> benoetiged updated rectangle daten, daher wird dies nur selten verwendet

                /// todo update nur falls notwendig
                template <elementTyp TYP_EL>
                void Update(int id) const;
                
                void updateAllRectangles() {
                    for(int id_rec=0;id_rec<ug->Give_number_rectangle();++id_rec) Update<rectangleEl>(id_rec);
		}
		
                inline DTyp Give_data_edge_middle( int id, int i ) const;
                inline DTyp Give_data_edge_interior( int id, int i ) const;
                inline DTyp Give_data_edge_exterior( int id, int i ) const;

                inline DTyp Give_data_point ( int id ) const;
                inline DTyp Give_data_neighbor_point ( int id, int num ) const;

                inline DTyp** Give_pointer_data_points() const { return data_points; };
                inline DTyp** Give_pointer_data_edges() const { return data_edges; };
                inline DTyp** Give_pointer_data_rectangles() const { return data_rectangles; };

		inline DTyp Give_data_at_point(const D2vector& point_coord, DTyp defaultForOutside) const;
		
                void operator= ( const DTyp_Restriction2D<DTyp>& a );

                template <class A>
                void operator= ( const Expr2D_Restriction<A>& a );

                template <class A>
                void operator= ( const Expr2D_Restriction_Iterdes<A>& a );

                template <class A>
                void operator= ( const Expr2D<A>& a );

                void operator= ( const Variable2D<DTyp>& v );
                void operator= ( DTyp x );

                void operator= ( const ExpressionRestriction2D<DTyp>& Expr2D_res );
                void operator= ( const ExpressionProlongation2D<DTyp>& Expr2D_prol );
                int ref();

                /**
                 * Verbindung mit algebraischem Vektor
                 */
                template <class AlgV, class Ind>
                void assignToVector(AlgV* algVector,Ind* indexVector, int id_rec);

                /**
                 * Verbindung mit algebraischem Vektor
                 */
                template <class AlgV, class Ind>
                void assignFromVector(AlgV* algVector,Ind* indexVector, int id_rec);
                stencil_typ Give_stencil_typ() const { return no_stencil; }

                // copies data from a in hex_a to data in hex_b
                void Copy_invert_z ( int hex_b, Variable2D<DTyp> a, int hex_a );

                int  getTotalNumberData()   const { return totalNumberData; }	
                 
        protected:
                Blockgrid2D*        blockgrid;
                Unstructured2DGrid*        ug;                 
                 
	        inline DTyp* give_startTotal() const  { assert(dataTotal!=NULL);  return dataTotal; }  	
	        
	        void allocateData();
	        
                //template <elementTyp TYP_EL>
                void Update_back(int id) const;

		void setValueAtRectangularPoint(DTyp val, int id, int i, int j, int Nx) { data_rectangles[id][i+ ( Nx+1 ) *j] = val; }		
		


                // own data
                DTyp** data_rectangles;        // num_rectangles
                DTyp** data_edges;             // num_edges
                DTyp** data_points;            // num_points,
                                               // 0: own data
                                               // 1,...????  data of next rectangle
                                               // ....????
                // rest unstructured

                // neighbor unstructured data
                DTyp** data_interior_edges;             // num_edges
                DTyp** data_exterior_edges;             // num_edges

                DTyp**  data_neighbor_points;           // num_points

                // for parallel
                int my_rank;
                MPI_Comm comm;
		
		//Daten zum Löschen
		int  ug_number_rectangles;
		int  ug_number_edges;
		int  ug_number_points;
		int  idGrid;
		
		// own data
                int totalNumberData;
                DTyp* dataTotal;		
};


/* @} */

template <class DTyp>
DTyp Variable2D<DTyp>::Give_interpolated_cell_rectangle(int id_rec, int i, int j, int Nx) const { 
  return 0.25 * (data_rectangles[id_rec][i+ ( Nx+1 ) *j    ] + data_rectangles[id_rec][(i+1) + ( Nx+1 ) * j    ] +
                 data_rectangles[id_rec][i+ ( Nx+1 ) *(j+1)] + data_rectangles[id_rec][(i+1) + ( Nx+1 ) * (j+1)]);
}




#endif // VARIABLE_H
