/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// variable2D_cc.h
//
// ------------------------------------------------------------
#include "variable2D.h"
//#include "../../../linAlgExptemp/source/linAlg.h"
#ifndef VARIABLE2D_CC_H
#define VARIABLE2D_CC_H

#ifdef _OPENMP
#include <omp.h>
//#include <source/abbrevi.h>
#endif

#include "update2D.h"

/*
template<class DTyp, bool isSystem>
void Variable2D_helper<DTyp, pointEl, isSystem>::Update(int id, const Variable2D<DTyp>* variable ) {
  Update_point(id, variable->blockgrid, variable->data_rectangles,
      variable->data_edges, variable->data_points,
      variable->data_neighbor_points, variable->my_rank, variable->comm);
}


template<class DTyp, bool isSystem>
void Variable2D_helper<DTyp, edgeEl, isSystem>::Update(int id, const Variable2D<DTyp>* variable ) {
  Update_edge(id, variable->blockgrid, variable->data_rectangles,
      variable->data_edges, variable->data_points,
      variable->data_interior_edges, variable->data_exterior_edges, variable->my_rank, variable->comm);
}


template<class DTyp, bool isSystem>
void Variable2D_helper<DTyp, rectangleEl, isSystem>::Update(int id, const Variable2D<DTyp>* variable ) {
  Update_rectangle(id, variable->blockgrid, variable->data_rectangles,
      variable->data_edges, variable->data_points, variable->my_rank, variable->comm);
}

//
//template<class DTyp, bool isSystem>
//void Variable2D_helper<DTyp, rectangleEl, isSystem>::Update_back(int id, const Variable2D<DTyp>* variable ) {
//
//}
//
//template<class DTyp, bool isSystem>
//void Variable2D_helper<DTyp, edgeEl, isSystem>::Update_back(int id, const Variable2D<DTyp>* variable ) {
//}
//
//template<class DTyp, bool isSystem>
//void Variable2D_helper<DTyp, pointEl, isSystem>::Update_back(int id, const Variable2D<DTyp>* variable ) {
//}
template <class DTyp>
inline void Variable2D<DTyp>::Update_back(int id) const {
  //Variable2D_helper<DTyp, rectangleEl, isSystem<DTyp>::value>::Update_back(id,this);
  Update_back_rectangle(id, blockgrid, data_rectangles,
      data_edges, data_points, my_rank, comm);
}
*/

template <class DTyp>
Variable2D<DTyp>::Variable2D ( const Variable2D<DTyp>& other ) {
	blockgrid = other.blockgrid;
	ug = blockgrid->Give_unstructured_grid();

    
	// for parallel
	my_rank = ug->Give_my_rank();
	comm    = ug->Give_MPI_comm();

	// Kopie der UG Daten zum Löschen
	ug_number_rectangles = ug->Give_number_rectangle();
	ug_number_edges      = ug->Give_number_edges();
	ug_number_points     = ug->Give_number_points();
	idGrid = blockgrid->getId();
	
	allocateData();	
	
	this->operator=(other);

/*	
	// rectangles
	data_rectangles = new DTyp*[ug->Give_number_rectangle()];
	for(int id=0; id<ug->Give_number_rectangle(); ++id ) {
		int num_total = blockgrid->Give_N_total_rectangle ( id );

		if(ug->Give_rectangle(id)->my_object(my_rank)) {
			data_rectangles[id] = new DTyp[num_total];
		   {
#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
		      for(int i=0; i<num_total; ++i )
			  data_rectangles[id][i] = ( DTyp ) 0;
		   }
		} else data_rectangles[id] = NULL;
	}

	// edges
	data_edges = new DTyp*[ug->Give_number_edges()];
	data_interior_edges = new DTyp*[ug->Give_number_edges()];
	data_exterior_edges = new DTyp*[ug->Give_number_edges()];

	for(int id=0; id<ug->Give_number_edges(); ++id ) {
		int num_total = blockgrid->Give_Nx_edge(id) +1;
		if ( ug->Give_edge ( id )->my_object ( my_rank ) ) {
			data_edges[id] = new DTyp[num_total];
			data_interior_edges[id] = new DTyp[num_total];
			data_exterior_edges[id] = new DTyp[num_total];

			for(int i=0; i<num_total; ++i ) {
			    data_edges[id][i] = ( DTyp ) 0;
			    data_interior_edges[id][i] = ( DTyp ) 0;
			    data_exterior_edges[id][i] = ( DTyp ) 0;
			}
		}
		else {
		  data_edges[id] = NULL;
		  data_interior_edges[id] = NULL;
		  data_exterior_edges[id] = NULL;
		}
	}


	// points
	data_points = new DTyp*[ug->Give_number_points() ];
	for ( int id=0; id<ug->Give_number_points(); ++id ) {
		int num_total = 1;
		data_points[id] = new DTyp[num_total];

		for(int i=0; i<num_total; ++i)
				data_points[id][i] = ( DTyp ) 0;
	}

	data_neighbor_points = new DTyp*[ug->Give_number_points() ];
	for(int id=0; id<ug->Give_number_points(); ++id ) {
	    int num_neighbor = ug->Give_point(id)->give_number_neighbors_points();

		if (ug->Give_point(id)->my_object(my_rank)) {
			data_neighbor_points[id] = new DTyp[num_neighbor];

			for(int i=0; i<num_neighbor; ++i )
				data_neighbor_points[id][i] = ( DTyp ) 0;
		} else data_neighbor_points[id] = NULL;
	}
	*/
}

template <class DTyp>
Variable2D<DTyp>::Variable2D ( Blockgrid2D& blockgrid_ ) {
	blockgrid = &blockgrid_;
	ug = blockgrid->Give_unstructured_grid();

	// for parallel
	my_rank = ug->Give_my_rank();
	comm    = ug->Give_MPI_comm();

	// Kopie der UG Daten zum Löschen
	ug_number_rectangles = ug->Give_number_rectangle();
	ug_number_edges      = ug->Give_number_edges();
	ug_number_points     = ug->Give_number_points();
	idGrid = blockgrid->getId();

	allocateData();
}
	
template <class DTyp>
void Variable2D<DTyp>::allocateData() {  
  //Part I
  //-------
	totalNumberData = 0;  
	// rectangles
	data_rectangles = new DTyp*[ug->Give_number_rectangle()];
	for(int id=0; id<ug->Give_number_rectangle(); ++id ) {
		int num_total = blockgrid->Give_N_total_rectangle ( id );

		if(ug->Give_rectangle(id)->my_object(my_rank)) {
		   totalNumberData = totalNumberData + num_total;
		} else data_rectangles[id] = NULL;
	}

	// edges
	data_edges = new DTyp*[ug->Give_number_edges()];
	data_interior_edges = new DTyp*[ug->Give_number_edges()];
	data_exterior_edges = new DTyp*[ug->Give_number_edges()];

	for(int id=0; id<ug->Give_number_edges(); ++id ) {
		int num_total = blockgrid->Give_Nx_edge(id) +1;
		if ( ug->Give_edge ( id )->my_object ( my_rank ) ) {
		        totalNumberData = totalNumberData + 3*num_total;		  
		} 
		else {
		  data_edges[id] = NULL;
		  data_interior_edges[id] = NULL;
		  data_exterior_edges[id] = NULL;
		}
	}


	// points
	data_points = new DTyp*[ug->Give_number_points() ];
	for ( int id=0; id<ug->Give_number_points(); ++id ) {
		int num_total = 1; 
		totalNumberData = totalNumberData + num_total;	
	}
	
	data_neighbor_points = new DTyp*[ug->Give_number_points() ];
	for(int id=0; id<ug->Give_number_points(); ++id ) {
	    int num_total = ug->Give_point(id)->give_number_neighbors_points();

		if (ug->Give_point(id)->my_object(my_rank)) {
 		        totalNumberData = totalNumberData + num_total;	
		} else data_neighbor_points[id] = NULL;
	}
	
  //Part II
  //-------
	dataTotal = new DTyp[totalNumberData];
	int indexStart = 0; 
  
  
  //Part III
  //---------
	// rectangles
	data_rectangles = new DTyp*[ug->Give_number_rectangle()];
	for(int id=0; id<ug->Give_number_rectangle(); ++id ) {
		int num_total = blockgrid->Give_N_total_rectangle ( id );

		if(ug->Give_rectangle(id)->my_object(my_rank)) {
			data_rectangles[id] = &(dataTotal[indexStart]);
                        indexStart = indexStart + num_total;			
		   {
#pragma omp parallel for num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
		      for(int i=0; i<num_total; ++i )
			  data_rectangles[id][i] = ( DTyp ) 0;
		   }
		} else data_rectangles[id] = NULL;
	}

	// edges
	data_edges = new DTyp*[ug->Give_number_edges()];
	data_interior_edges = new DTyp*[ug->Give_number_edges()];
	data_exterior_edges = new DTyp*[ug->Give_number_edges()];

	for(int id=0; id<ug->Give_number_edges(); ++id ) {
		int num_total = blockgrid->Give_Nx_edge(id) +1;
		if ( ug->Give_edge ( id )->my_object ( my_rank ) ) {
			data_edges[id] = &(dataTotal[indexStart]);
                        indexStart = indexStart + num_total;	

			data_interior_edges[id] = &(dataTotal[indexStart]);
                        indexStart = indexStart + num_total;	

			data_exterior_edges[id] = &(dataTotal[indexStart]);
                        indexStart = indexStart + num_total;				
			
			for(int i=0; i<num_total; ++i ) {
			    data_edges[id][i] = ( DTyp ) 0;
			    data_interior_edges[id][i] = ( DTyp ) 0;
			    data_exterior_edges[id][i] = ( DTyp ) 0;
			}
		} 
		else {
		  data_edges[id] = NULL;
		  data_interior_edges[id] = NULL;
		  data_exterior_edges[id] = NULL;
		}
	}


	// points
	data_points = new DTyp*[ug->Give_number_points() ];
	for ( int id=0; id<ug->Give_number_points(); ++id ) {
		int num_total = 1; 
		data_points[id] = &(dataTotal[indexStart]);
                indexStart = indexStart + num_total;	
		
		for(int i=0; i<num_total; ++i)
				data_points[id][i] = ( DTyp ) 0;
	}
	
	data_neighbor_points = new DTyp*[ug->Give_number_points() ];
	for(int id=0; id<ug->Give_number_points(); ++id ) {
	    int num_total = ug->Give_point(id)->give_number_neighbors_points();

		if (ug->Give_point(id)->my_object(my_rank)) {
		        data_neighbor_points[id] = &(dataTotal[indexStart]);
                        indexStart = indexStart + num_total;	
			
			for(int i=0; i<num_total; ++i )
				data_neighbor_points[id][i] = ( DTyp ) 0;
		} else data_neighbor_points[id] = NULL;
	}	
}

template <class DTyp>
Variable2D<DTyp>::~Variable2D() {
        delete [] dataTotal;  
  
	if ( data_rectangles != NULL ) {
		delete [] data_rectangles;
	}
	data_rectangles = NULL;

	if(data_edges != NULL ) {
	       delete [] data_edges;
	}
	data_interior_edges = NULL;
	if(data_interior_edges != NULL ) {
	       delete [] data_interior_edges;
	}
	data_exterior_edges = NULL;
	if(data_exterior_edges != NULL ) {
	       delete [] data_exterior_edges;
	}
	data_exterior_edges = NULL;

	if ( data_points != NULL ) {
		delete [] data_points;
	}
	data_points = NULL;

	if(data_neighbor_points != NULL ) {
	       delete [] data_neighbor_points;
	}
	data_neighbor_points = NULL;

	ug=NULL;
}


template <class DTyp>
void Variable2D<DTyp>::operator= ( DTyp x ) {
        int Nx, Ny, i, j;
	
	
	// rectangles
	for ( int id = 0;id < ug->Give_number_rectangle();++id ) {
		if ( ug->Give_rectangle(id)->my_object ( my_rank ) ) {		
			Nx = blockgrid->Give_Nx_rectangle(id);
			Ny = blockgrid->Give_Ny_rectangle(id);

			#pragma omp parallel for private(i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
			for(j = 1;j < Ny;++j )
				for(i = 1;i < Nx;++i ) {
					data_rectangles[id][i+ ( Nx+1 ) *j] = x;
				}
		}
	}
	
	// edges
	for ( int id = 0;id < ug->Give_number_edges();++id ) {
		if ( ug->Give_edge ( id )->my_object ( my_rank ) ) {
			Nx = blockgrid->Give_Nx_edge ( id );
			for ( i = 1;i < Nx;++i )  {
				data_edges[id][i] = x;
			}
		}
	}
	
	// points
	for ( int id = 0;id < ug->Give_number_points();++id ) {
		if ( ug->Give_point ( id )->my_object ( my_rank ) ) {
			data_points[id][0] = x;
		}
	}	
}


template <class DTyp>
void Variable2D<DTyp>::operator=(const DTyp_Restriction2D<DTyp>& a ) {
        int Nx, Ny, i, j;
	
	DTyp   x      = a.Give_x();
	// rectangles
	for ( int id = 0;id < ug->Give_number_rectangle();++id ) {
		if ( ug->Give_rectangle(id)->my_object ( my_rank ) &&
		     a.template Give_marker<rectangleEl> ( id ) ==yes_mark) {		
			Nx = blockgrid->Give_Nx_rectangle(id);
			Ny = blockgrid->Give_Ny_rectangle(id);

			#pragma omp parallel for private(i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
			for(j = 1;j < Ny;++j )
				for(i = 1;i < Nx;++i ) {
					data_rectangles[id][i+ ( Nx+1 ) *j] = x;
				}
		}
	}
	
	// edges
	for ( int id = 0;id < ug->Give_number_edges();++id ) {
		if ( ug->Give_edge ( id )->my_object(my_rank) &&
		     a.template Give_marker<edgeEl>(id) == yes_mark) {
			Nx = blockgrid->Give_Nx_edge(id);
			for ( i = 1;i < Nx;++i )  {
				data_edges[id][i] = x;
			}
		}
	}
	
	// points
	for ( int id = 0;id < ug->Give_number_points();++id ) {
		if ( ug->Give_point ( id )->my_object(my_rank) &&
		     a.template Give_marker<pointEl>(id) ==yes_mark ) {
			data_points[id][0] = x;
		}
	}	
}



template <class DTyp>
void Variable2D<DTyp>::operator= ( const Variable2D<DTyp>& v ) {
    int Nx, Ny, i, j;
    assert(blockgrid->hasSameStructure(*v.blockgrid));
	// rectangles
	for ( int id = 0;id < ug->Give_number_rectangle();++id ) {
		if ( ug->Give_rectangle(id)->my_object ( my_rank )) {
			Nx = blockgrid->Give_Nx_rectangle(id);
			Ny = blockgrid->Give_Ny_rectangle(id);

			#pragma omp parallel for private(i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)
			for(j = 1;j < Ny;++j )
				for(i = 1;i < Nx;++i ) {
					data_rectangles[id][i+ ( Nx+1 ) *j] = v.data_rectangles[id][i+ ( Nx+1 ) *j];
				}
		}
	}

	// edges
	for ( int id = 0;id < ug->Give_number_edges();++id ) {
		if ( ug->Give_edge ( id )->my_object(my_rank)) {
			Nx = blockgrid->Give_Nx_edge(id);
			for ( i = 1;i < Nx;++i )  {
				data_edges[id][i] = v.data_edges[id][i];
			}
		}
	}

	// points
	for ( int id = 0;id < ug->Give_number_points();++id ) {
		if ( ug->Give_point ( id )->my_object(my_rank)) {
			data_points[id][0] = v.data_points[id][0];
		}
	}
}

///////////////////////////
// 1.a Restriction to subgrid
///////////////////////////

template <class A>
inline Expr2D_Restriction<A> operator| ( const Expr2D<A>& a, const Marker2D& sg ) {
        return Expr2D_Restriction<A> ( a, sg );
}




///////////////////////////
// 1.b Restriction to subgrid and iteration description
///////////////////////////

template <class A>
inline Expr2D_Restriction_Iterdes<A> operator| ( const Expr2D<A>& a, const Iteration_description2D& iterdes ) {
        return Expr2D_Restriction_Iterdes<A> ( a, iterdes );
}


/////////////////////////////////////////
// Verbindung mit algebraischem Vektor
////////////////////////////////////////
template <class DTyp>
template <class AlgV, class Ind>
void Variable2D<DTyp>::assignFromVector(AlgV* algVector,Ind* indexVector, int id_rec) {
   assert(algVector->getSize() == indexVector->getSize());

   int size = indexVector->getSize();
   int Nx = blockgrid->Give_Nx_rectangle(id_rec);

   for(int s=0;s<size;++s) {
       int i = indexVector->getI(s);
       int j = indexVector->getJ(s);
       data_rectangles[id_rec][i+(Nx+1)*j] =  algVector->getValueS(s);
   }

   Update_back(id_rec);
}




template <class DTyp>
template <class AlgV, class Ind>
void Variable2D<DTyp>::assignToVector(AlgV* algVector,Ind* indexVector, int id_rec) {
   assert(algVector->getSize() == indexVector->getSize());

   Update<rectangleEl>(id_rec);

   int size = indexVector->getSize();
   int Nx = blockgrid->Give_Nx_rectangle(id_rec);

   for(int s=0;s<size;++s) {
       int i = indexVector->getI(s);
       int j = indexVector->getJ(s);
       algVector->setValueS(s,data_rectangles[id_rec][i+(Nx+1)*j]);
   }
}

/////////////////////////////////////////
// Give_Data
////////////////////////////////////////
	
template <>
template <>
inline double Variable2D<double>::Give_data<pointEl> ( params2D_in ) const {
	return data_points[id][0];
}

template <>
template <>
inline std::complex<double> Variable2D<std::complex<double> >::Give_data<pointEl> ( params2D_in ) const {
	return data_points[id][0];
}

template <>
template <>
inline int Variable2D<int>::Give_data<pointEl> ( params2D_in ) const {
	return data_points[id][0];
}


template <>
template <>
inline double Variable2D<double>::Give_data<edgeEl> ( params2D_in ) const {
	return data_edges[id][i];
}

template <>
template <>
inline std::complex<double> Variable2D<std::complex<double> >::Give_data<edgeEl>(params2D_in) const {
	return data_edges[id][i];
}

template <>
template <>
inline int Variable2D<int>::Give_data<edgeEl>(params2D_in) const {
	return data_edges[id][i];
}


template <>
template <>
inline double Variable2D<double>::Give_data<rectangleEl> ( params2D_in ) const {
	return data_rectangles[id][i+ ( Nx+1 ) *j];
}

template <>
template <>
inline std::complex<double> Variable2D<std::complex<double> >::Give_data<rectangleEl> ( params2D_in ) const {
	return data_rectangles[id][i+ ( Nx+1 ) *j];
}

template <>
template <>
inline int Variable2D<int>::Give_data<rectangleEl> ( params2D_in ) const {
	return data_rectangles[id][i+ ( Nx+1 ) *j];
}

template <>
template <>
inline bool Variable2D<bool>::Give_data<rectangleEl> ( params2D_in ) const {
	return data_rectangles[id][i+ ( Nx+1 ) *j];
}





///////////////////////////
// 3. Other functions, max, ...
///////////////////////////

/** \addtogroup ExpressionTemplates2D **/
/* @{ */ 
/*
template <class A, class B>
_TypeOf2_ ( A, B ) product ( Expr2D<A>& a, Expr2D<B>& b, Marker2D& marker );
*/
template <class A, class B>
_TypeOf2_ ( A, B ) product (const Expr2D<A>& a,const Expr2D<B>& b );




template <class A>
double L_infty ( const Expr2D<A>& a, Marker2D& marker );

template <class A>
double L_infty ( const Expr2D<A>& a );


template <class A>
double Maximum ( const Expr2D<A>& a, Marker2D& marker );

template <class A>
double Maximum ( const Expr2D<A>& a );


template <class A>
double Minimum ( const Expr2D<A>& a, Marker2D& marker );

template <class A>
double Minimum ( const Expr2D<A>& a );

template <class DTyp>
double Minimum ( const Variable2D<DTyp>& v);

template <class DTyp>
double Maximum ( const Variable2D<DTyp>& v);

template <class DTyp>
double L2Norm ( const Variable2D<DTyp>& v, Marker2D& marker );

template <class DTyp>
double L2Norm ( const Variable2D<DTyp>& v);
/* @} */

//////////////////////////////////////////
// and it's implementation

template <class A, class B>
_TypeOf2_ ( A, B ) product (const Expr2D<A>& a,const Expr2D<B>& b ) {
        const A& ao ( a );

        return product ( a, b, ao.Give_blockgrid()->Give_unstructured_grid()->Give_all_points() );
}
template <class A, class B>
_TypeOf2_ ( A, B ) product (Expr2D<A>& a,Expr2D<B>& b ) {
        const A& ao ( a );

        return product ( a, b, ao.Give_blockgrid()->Give_unstructured_grid()->Give_all_points() );
}
template <class A>
double  L_infty ( const Expr2D<A>& a ) {
        const A& ao ( a );

        return L_infty ( a, ao.Give_blockgrid()->Give_unstructured_grid()->Give_all_points() );
}


template <class A>
double  Minimum ( const Expr2D<A>& a ) {
        const A& ao ( a );

        return Minimum ( a, ao.Give_blockgrid()->Give_unstructured_grid()->Give_all_points() );
}

template <class DTyp>
double Minimum ( const Variable2D<DTyp>& v ) {
        return Minimum ( v, v.Give_blockgrid()->Give_unstructured_grid()->Give_all_points() );
}

template <class DTyp>
double Maximum ( const Variable2D<DTyp>& v ) {
        return Maximum ( v, v.Give_blockgrid()->Give_unstructured_grid()->Give_all_points() );
}

template <class A>
double  Maximum ( const Expr2D<A>& a ) {
        const A& ao ( a );

        return Maximum ( a, ao.Give_blockgrid()->Give_unstructured_grid()->Give_all_points() );
}



template <class A, class B>
_TypeOf2_ ( A, B ) product (Expr2D<A>& ao,Expr2D<B>& bo, Marker2D& marker ) {
        int i,j, Nx, Ny;
        typedef _TypeOf2_ ( A, B ) DTyp;
        const A& a ( ao );
        const B& b ( bo );

        DTyp sum;
        DTyp sum_total;

        sum = ( DTyp ) 0;

        assert(a.Give_blockgrid() == b.Give_blockgrid());

        Unstructured2DGrid* ug = a.Give_blockgrid()->Give_unstructured_grid();

        int my_rank   = ug->Give_my_rank();
        MPI_Comm comm = ug->Give_MPI_comm();

        for(int id = 0;id < ug->Give_number_rectangle();++id) {
            if(marker.template Give_marker<rectangleEl> ( id ) == yes_mark) {
               Nx = a.Give_blockgrid()->Give_Nx_rectangle(id);
               Ny = a.Give_blockgrid()->Give_Ny_rectangle(id);

/// @todo: problem mit kombination aus reduction (+:sum) und MyConj
//#pragma omp parallel for private(k,j,i) reduction(+ : sum) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)

               for(j = 1;j < Ny;++j )
                   for(i = 1;i < Nx;++i )
                       sum += a.template Give_data<rectangleEl> ( params2D_out )
                              * MyConj<DTyp>::Give(b.template Give_data<rectangleEl>(params2D_out));
            }
        }


        for(int id = 0;id < ug->Give_number_edges();++id ) {
            if(marker.template Give_marker<edgeEl>(id) == yes_mark) {
               Nx = a.Give_blockgrid()->Give_Nx_edge ( id );

               for(i = 1;i < Nx;++i )
                   sum += a.template Give_data<edgeEl> ( params2D_out )
                          * MyConj<DTyp>::Give(b.template Give_data<edgeEl>(params2D_out));
            }
        }

        for(int id = 0;id < ug->Give_number_points();++id ) {
            if(marker.template Give_marker<pointEl> ( id ) == yes_mark ) {
               sum += a.template Give_data<pointEl> ( params2D_out )
                      * MyConj<DTyp>::Give ( b.template Give_data<pointEl> ( params2D_out ) );
            }
        }

        sum_total = sum;
        //sum_total = Make_MPI_All_Sum<DTyp>::Do ( sum, comm );

        return sum_total;
}



template <class A, class B>
_TypeOf2_ ( A, B ) product (const Expr2D<A>& ao,const Expr2D<B>& bo, Marker2D& marker ) {
        int i,j, Nx, Ny;
        typedef _TypeOf2_ ( A, B ) DTyp;
        const A& a ( ao );
        const B& b ( bo );

        DTyp sum;
        DTyp sum_total;

        sum = ( DTyp ) 0;

        assert(a.Give_blockgrid() == b.Give_blockgrid());

        Unstructured2DGrid* ug = a.Give_blockgrid()->Give_unstructured_grid();

        int my_rank   = ug->Give_my_rank();
        MPI_Comm comm = ug->Give_MPI_comm();

        for(int id = 0;id < ug->Give_number_rectangle();++id) {
            if(marker.template Give_marker<rectangleEl> ( id ) == yes_mark) {
               Nx = a.Give_blockgrid()->Give_Nx_rectangle(id);
               Ny = a.Give_blockgrid()->Give_Ny_rectangle(id);

/// @todo: problem mit kombination aus reduction (+:sum) und MyConj
//#pragma omp parallel for private(k,j,i) reduction(+ : sum) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)

               for(j = 1;j < Ny;++j )
                   for(i = 1;i < Nx;++i )
                       sum += a.template Give_data<rectangleEl> ( params2D_out )
                              * MyConj<DTyp>::Give(b.template Give_data<rectangleEl>(params2D_out));
            }
        }


        for(int id = 0;id < ug->Give_number_edges();++id ) {
            if(marker.template Give_marker<edgeEl>(id) == yes_mark) {
               Nx = a.Give_blockgrid()->Give_Nx_edge ( id );

               for(i = 1;i < Nx;++i )
                   sum += a.template Give_data<edgeEl> ( params2D_out )
                          * MyConj<DTyp>::Give(b.template Give_data<edgeEl>(params2D_out));
            }
        }

        for(int id = 0;id < ug->Give_number_points();++id ) {
            if(marker.template Give_marker<pointEl> ( id ) == yes_mark ) {
               sum += a.template Give_data<pointEl> ( params2D_out )
                      * MyConj<DTyp>::Give ( b.template Give_data<pointEl> ( params2D_out ) );
            }
        }

        sum_total = sum;
        //sum_total = Make_MPI_All_Sum<DTyp>::Do ( sum, comm );

        return sum_total;
}


template <class A>
double  L2Norm(const Variable2D<A>& a, Marker2D& marker ) {
        return sqrt(product ( a, a, marker ));
}

inline double  L2Norm(const Variable2D<std::complex<double> >& a, Marker2D& marker ) {
        return sqrt(product ( a, a, marker ).real());
}

inline double  L2Norm(const Variable2D<std::complex<double> >& a) {
        return sqrt(product ( a, a).real());
}



template <class A>
double L_infty ( const Expr2D<A>& ao, Marker2D& marker ) {
        int i, j, k, id;
        int Nx, Ny, Nz;

        double infty;
        double infty_total;

        const A& a ( ao );

        infty = 0.0;

        Unstructured2DGrid* ug = a.Give_blockgrid()->Give_unstructured_grid();
        int my_rank   = ug->Give_my_rank();
        MPI_Comm comm = ug->Give_MPI_comm();

        for ( id = 0; id < ug->Give_number_rectangle(); ++id ) {
                if(ug->Give_rectangle( id )->my_object ( my_rank ) &&
                   marker.template Give_marker<rectangleEl> ( id ) == yes_mark ) {
                   Nx = a.Give_blockgrid()->Give_Nx_rectangle(id);
                   Ny = a.Give_blockgrid()->Give_Ny_rectangle(id);

                   for(j = 1;j < Ny;++j )
                       for(i = 1;i < Nx;++i )
                           if(infty < ABS ( a.template Give_data<rectangleEl> ( params2D_out ) ) )
                              infty = ABS ( a.template Give_data<rectangleEl> ( params2D_out ) );
                }
        }


        for ( id = 0; id < ug->Give_number_edges(); ++id ) {
                if ( ug->Give_edge ( id )->my_object ( my_rank ) &&
                        marker.template Give_marker<edgeEl> ( id ) == yes_mark ) {
                        Nx = a.Give_blockgrid()->Give_Nx_edge ( id );

                        for ( i = 1;i < Nx;++i )
                                if ( infty < ABS ( a.template Give_data<edgeEl> ( params2D_out ) ) )
                                        infty = ABS ( a.template Give_data<edgeEl> ( params2D_out ) );
                }
        }

        for ( id = 0; id < ug->Give_number_points(); ++id ) {
                if ( ug->Give_point ( id )->my_object ( my_rank ) &&
                        marker.template Give_marker<pointEl> ( id ) == yes_mark ) {
                        if ( infty < ABS ( a.template Give_data<pointEl> ( params2D_out ) ) )
                                infty = ABS ( a.template Give_data<pointEl> ( params2D_out ) );
                }
        }

        MPI_Allreduce ( &infty, &infty_total, 1, MPI_DOUBLE, MPI_MAX, comm );

        return infty_total;
}



template <class A>
double Maximum ( const Expr2D<A>& ao, Marker2D& marker ) {
        int i, j, k, id;
        int Nx, Ny, Nz;

        double maximum = 0.0;
        double maximum_total;
        bool first_set = false;

        const A& a ( ao );


        Unstructured2DGrid* ug = a.Give_blockgrid()->Give_unstructured_grid();
        int my_rank   = ug->Give_my_rank();
        MPI_Comm comm = ug->Give_MPI_comm();

        for ( id = 0; id < ug->Give_number_rectangle(); ++id ) {
                if(ug->Give_rectangle( id )->my_object ( my_rank ) &&
                   marker.template Give_marker<rectangleEl> ( id ) == yes_mark ) {
                   Nx = a.Give_blockgrid()->Give_Nx_rectangle(id);
                   Ny = a.Give_blockgrid()->Give_Ny_rectangle(id);

                   for(j = 1;j < Ny;++j )
                       for(i = 1;i < Nx;++i ) {
                           double value = a.template Give_data<rectangleEl>(params2D_out);
                           if(first_set==false) { maximum = value; first_set=true; }
                           if(maximum < value)  maximum = value;
                       }
                }
        }


        for ( id = 0; id < ug->Give_number_edges(); ++id ) {
                if ( ug->Give_edge ( id )->my_object ( my_rank ) &&
                        marker.template Give_marker<edgeEl> ( id ) == yes_mark ) {
                        Nx = a.Give_blockgrid()->Give_Nx_edge ( id );

                        for(i = 1;i < Nx;++i ) {
                           double value = a.template Give_data<edgeEl>(params2D_out);
                           if(first_set==false) { maximum = value; first_set=true; }
                           if(maximum < value)  maximum = value;
                       }
                }
        }

        for ( id = 0; id < ug->Give_number_points(); ++id ) {
                if ( ug->Give_point ( id )->my_object ( my_rank ) &&
                        marker.template Give_marker<pointEl> ( id ) == yes_mark ) {

                        double value = a.template Give_data<pointEl>(params2D_out);
                           if(first_set==false) { maximum = value; first_set=true; }
                           if(maximum < value)  maximum = value;
                }
        }

        MPI_Allreduce ( &maximum, &maximum_total, 1, MPI_DOUBLE, MPI_MAX, comm );

        return maximum_total;
}



template <class A>
double Minimum(const Expr2D<A>& ao, Marker2D& marker ) {
        int i, j, k, id;
        int Nx, Ny, Nz;

        double minimum = 0.0;
        double minimm_total;
        bool first_set = false;

        const A& a ( ao );


        Unstructured2DGrid* ug = a.Give_blockgrid()->Give_unstructured_grid();
        int my_rank   = ug->Give_my_rank();
        MPI_Comm comm = ug->Give_MPI_comm();

        for ( id = 0; id < ug->Give_number_rectangle(); ++id ) {
                if(ug->Give_rectangle( id )->my_object ( my_rank ) &&
                   marker.template Give_marker<rectangleEl> ( id ) == yes_mark ) {
                   Nx = a.Give_blockgrid()->Give_Nx_rectangle(id);
                   Ny = a.Give_blockgrid()->Give_Ny_rectangle(id);

                   for(j = 1;j < Ny;++j )
                       for(i = 1;i < Nx;++i ) {
                           double value = a.template Give_data<rectangleEl>(params2D_out);			   
                           if(first_set==false) { minimum = value; first_set=true; }
                           if(minimum > value)  minimum = value;
                       }
                }
        }


        for ( id = 0; id < ug->Give_number_edges(); ++id ) {
                if ( ug->Give_edge ( id )->my_object ( my_rank ) &&
                        marker.template Give_marker<edgeEl> ( id ) == yes_mark ) {
                        Nx = a.Give_blockgrid()->Give_Nx_edge ( id );

                        for(i = 1;i < Nx;++i ) {
                           double value = a.template Give_data<edgeEl>(params2D_out);
                           if(first_set==false) { minimum = value; first_set=true; }
                           if(minimum > value)  minimum = value;
                       }
                }
        }

        for ( id = 0; id < ug->Give_number_points(); ++id ) {
                if ( ug->Give_point ( id )->my_object ( my_rank ) &&
                        marker.template Give_marker<pointEl> ( id ) == yes_mark ) {

                        double value = a.template Give_data<pointEl>(params2D_out);
                           if(first_set==false) { minimum = value; first_set=true; }
                           if(minimum > value)  minimum = value;
                }
        }

        MPI_Allreduce ( &minimum, &minimum, 1, MPI_DOUBLE, MPI_MAX, comm );
	
        return minimum;
}





/*
template <class A>
double Minimum ( const Expr2D<A>& ao, Marker2D& marker ) {
        int i, j, k, id;
        int Nx, Ny, Nz;


        double minimum = 1e60;
        double minimum_total;

        const A& a ( ao );

        Unstructured2DGrid* ug = a.Give_blockgrid()->Give_unstructured_grid();
        int my_rank   = ug->Give_my_rank();
        MPI_Comm comm = ug->Give_MPI_comm();

        int set       = false;
        int set_total = false;

        for ( id = 0; id < ug->Give_number_rectangle(); ++id ) {
                if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) &&
                        marker.template Give_marker<rectangleEl> ( id ) == yes_mark ) {
                        Nx = a.Give_blockgrid()->Give_Nx_rectangle(id);
                        Ny = a.Give_blockgrid()->Give_Ny_rectangle(id);
                        Nz = a.Give_blockgrid()->Give_Nz_hexahedron ( id );

                        for ( k = 1;k < Nz;++k )
                                for ( j = 1;j < Ny;++j )
                                        for ( i = 1;i < Nx;++i ) {
                                                if ( minimum > myREAL ( a.template Give_data<rectangleEl> ( params2D_out ) ) ||
                                                        set == false ) {
                                                        minimum = myREAL ( a.template Give_data<rectangleEl> ( params2D_out ) );
                                                        set = true;
//       cout << "New Minimum = " << minimum << "in Hexahedron " << id <<" at point with indices (" << i << "," << j << "," << k << "): " << a.template Give_data<rectangleEl>(params2D_out)<< endl;
                                                }

                                                // cout << a.template Give_data<rectangleEl>(params2D_out) << endl;
                                        }
                }
        }

        for ( id = 0; id < ug->Give_number_quadrangles(); ++id ) {
                if ( ug->Give_quadrangle ( id )->my_object ( my_rank ) &&
                        marker.template Give_marker<quadrangleEl> ( id ) == yes_mark ) {
                        Nx = a.Give_blockgrid()->Give_Nx_quadrangle ( id );
                        Ny = a.Give_blockgrid()->Give_Ny_quadrangle ( id );

                        for ( j = 1;j < Ny;++j )
                                for ( i = 1;i < Nx;++i ) {
                                        if ( minimum > myREAL ( a.template Give_data<quadrangleEl> ( params2D_out ) ) ||
                                                set == false ) {
                                                minimum = myREAL ( a.template Give_data<quadrangleEl> ( params2D_out ) );
                                                set = true;
//     cout << "New Minimum = " << minimum << "in Quadrangle " << id <<" at point with indices (" << i << "," << j <<  "): " << a.template Give_data<quadrangleEl>(params2D_out) << endl;
                                        }
                                }
                }
        }

        for ( id = 0; id < ug->Give_number_edges(); ++id ) {
                if ( ug->Give_edge ( id )->my_object ( my_rank ) &&
                        marker.template Give_marker<edgeEl> ( id ) == yes_mark ) {
                        Nx = a.Give_blockgrid()->Give_Nx_edge ( id );

                        for ( i = 1;i < Nx;++i ) {
                                if ( minimum > myREAL ( a.template Give_data<edgeEl> ( params2D_out ) ) ||
                                        set == false ) {
                                        minimum = myREAL ( a.template Give_data<edgeEl> ( params2D_out ) );
                                        set = true;
//     cout << "New Minimum = " << minimum << "in Edge " << id <<" at point with index (" << i << "): " << a.template Give_data<edgeEl>(params2D_out) << endl;
                                }

                                //      cout << a.template Give_data<edgeEl>(params2D_out) << endl;

                        }
                }
        }

        for ( id = 0; id < ug->Give_number_points(); ++id ) {
                if ( ug->Give_point ( id )->my_object ( my_rank ) &&
                        marker.template Give_marker<pointEl> ( id ) == yes_mark ) {
                        if ( minimum > myREAL ( a.template Give_data<pointEl> ( params2D_out ) ) ||
                                set == false ) {
                                minimum = myREAL ( a.template Give_data<pointEl> ( params2D_out ) );
//  cout << "New Minimum = " << minimum << "at point with (" << id << "): " << a.template Give_data<pointEl>(params2D_out) << endl;
                                a.Give_blockgrid()->Give_coord_point ( id ).Print();


                                set = true;
                        }
                }
        }

//   cout << "Overall minimum bevor MPI: " << minimum << endl;
        MPI_Allreduce ( &minimum, &minimum_total, 1, MPI_DOUBLE, MPI_MIN, comm );

        int p;

        MPI_Comm_size ( comm, &p );

        MPI_Allreduce ( &set, &set_total, 1, MPI_INT, MPI_SUM, comm );

        if ( set_total != p && minimum_total > 1e59 ) cout << "error in Minimum" << endl;

//   cout << "Overall minimum after MPI: " << minimum_total << endl;
        return minimum_total;
}

template <class A>
double Maximum ( const Expr2D<A>& ao, Marker2D& marker ) {
        int i, j, k, id;
        int Nx, Ny, Nz;

        double maximum = -1e60;
        double maximum_total;

        const A& a ( ao );

        Unstructured2DGrid* ug = a.Give_blockgrid()->Give_unstructured_grid();
        int my_rank   = ug->Give_my_rank();
        MPI_Comm comm = ug->Give_MPI_comm();

        int set = false;
        int set_total = false;

        for ( id = 0; id < ug->Give_number_rectangle(); ++id ) {
                if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) &&
                        marker.template Give_marker<rectangleEl> ( id ) == yes_mark ) {
                        Nx = a.Give_blockgrid()->Give_Nx_rectangle(id);
                        Ny = a.Give_blockgrid()->Give_Ny_rectangle(id);
                        Nz = a.Give_blockgrid()->Give_Nz_hexahedron ( id );

                        for ( k = 1;k < Nz;++k )
                                for ( j = 1;j < Ny;++j )
                                        for ( i = 1;i < Nx;++i )
                                                if ( maximum < a.template Give_data<rectangleEl> ( params2D_out ) ||
                                                        set == false ) {
                                                        maximum = myREAL ( a.template Give_data<rectangleEl> ( params2D_out ) );
                                                        set = true;
                                                }
                }
        }

        for ( id = 0; id < ug->Give_number_quadrangles(); ++id ) {
                if ( ug->Give_quadrangle ( id )->my_object ( my_rank )  &&
                        marker.template Give_marker<quadrangleEl> ( id ) == yes_mark ) {
                        Nx = a.Give_blockgrid()->Give_Nx_quadrangle ( id );
                        Ny = a.Give_blockgrid()->Give_Ny_quadrangle ( id );

                        for ( j = 1;j < Ny;++j )
                                for ( i = 1;i < Nx;++i )
                                        if ( maximum < a.template Give_data<quadrangleEl> ( params2D_out ) ||
                                                set == false ) {
                                                maximum = myREAL ( a.template Give_data<quadrangleEl> ( params2D_out ) );
                                                set = true;
                                        }
                }
        }

        for ( id = 0; id < ug->Give_number_edges(); ++id ) {
                if ( ug->Give_edge ( id )->my_object ( my_rank )  &&
                        marker.template Give_marker<edgeEl> ( id ) == yes_mark ) {
                        Nx = a.Give_blockgrid()->Give_Nx_edge ( id );

                        for ( i = 1;i < Nx;++i )
                                if ( maximum < a.template Give_data<edgeEl> ( params2D_out ) ||
                                        set == false ) {
                                        maximum = myREAL ( a.template Give_data<edgeEl> ( params2D_out ) );
                                        set = true;
                                }
                }
        }

        for ( id = 0; id < ug->Give_number_points(); ++id ) {
                if ( ug->Give_point ( id )->my_object ( my_rank )  &&
                        marker.template Give_marker<pointEl> ( id ) == yes_mark ) {
                        if ( maximum < a.template Give_data<pointEl> ( params2D_out ) ||
                                set == false ) {
                                maximum = myREAL ( a.template Give_data<pointEl> ( params2D_out ) );
                                set = true;
                        }
                }
        }

        MPI_Allreduce ( &maximum, &maximum_total, 1, MPI_DOUBLE, MPI_MAX, comm );

        int p;
        MPI_Comm_size ( comm, &p );

        MPI_Allreduce ( &set, &set_total, 1, MPI_INT, MPI_SUM, comm );

        if ( set_total != p && maximum_total < -1e59 ) cout << "error in Maximum! " << endl;

        //max_total = " << maximum_total << ".\t set_total = " << set_total << ".\t p = " << p  << endl;


        return maximum_total;
}
*/

//////////////////////////////////////////
// 4. Implementation of inline functions
//////////////////////////////////////////


template <class DTyp>
inline DTyp Variable2D<DTyp>::Give_data_edge_middle( int id, int i ) const {
        return data_edges[id][i];
}

template <class DTyp>
inline DTyp Variable2D<DTyp>::Give_data_edge_interior( int id, int i ) const {
        return data_interior_edges[id][i];
}

template <class DTyp>
inline DTyp Variable2D<DTyp>::Give_data_edge_exterior( int id, int i ) const {
        return data_exterior_edges[id][i];
}
		
		
template <class DTyp>
inline DTyp Variable2D<DTyp>::Give_data_at_point(const D2vector& point_coord, DTyp defaultForOutside) const {
   int id;
   double I;
   double J;
   if(ug->getCoordinate(point_coord, id, I, J, *blockgrid)==false) return defaultForOutside;
     
   
   int i = round(I);
   int j = round(J);
   int Nx = blockgrid->Give_Nx_rectangle(id);
   int Ny = blockgrid->Give_Ny_rectangle(id);

   assert(i>=0);  assert(j>=0);
   assert(i<=Nx); assert(j<=Ny);
   
   if(i<Nx && j<Ny) {
      DTyp uSW = data_rectangles[id][i+  (Nx+1)*j];
      DTyp uSE = data_rectangles[id][i+1+(Nx+1)*j];
      DTyp uNW = data_rectangles[id][i+  (Nx+1)*(j+1)];
      DTyp uNE = data_rectangles[id][i+1+(Nx+1)*(j+1)];
      double xi  = I-i; 
      double eta = J-j;
      return uSW*(1.0-xi)*(1.0-eta) + uSE*xi*(1.0-eta) + uNW*(1.0-xi)*eta + uNE*xi*eta;
   }
   else if(i==Nx && j==Ny) {
      return data_rectangles[id][i+  (Nx+1)*j];
   }
   else if(i<Nx) {
      DTyp uW = data_rectangles[id][i+  (Nx+1)*j];
      DTyp uE = data_rectangles[id][i+1+(Nx+1)*j];
      double xi  = I-i; 
      return uW*(1.0-xi) + uE*xi;
   }
   else { // (j<Ny) {
      DTyp uS = data_rectangles[id][i+  (Nx+1)*j];
      DTyp uN = data_rectangles[id][i+  (Nx+1)*(j+1)];
      double eta = J-j;
      return uS*(1.0-eta) + uN*eta;
   }
}
		

template <class DTyp>
inline DTyp Variable2D<DTyp>::Give_data_point ( int id ) const {
   return data_points[id][0];
}

template <class DTyp>
inline DTyp Variable2D<DTyp>::Give_data_neighbor_point ( int id, int num ) const {
   return data_neighbor_points[id][num];
}



template <class DTyp>
template <class A>
void Variable2D<DTyp>::operator= ( const Expr2D_Restriction<A>& a ) {
        int Nx, Ny, i, j;

        //const A& ao ( a );

        stencil_typ sten_typ = a.Give_stencil_typ();

        // rectangles
        for(int id = 0;id < ug->Give_number_rectangle();++id ) {
                //if(sten_typ == yes_stencil )
                        a.template Update<rectangleEl> ( id );

                if(a.template Give_marker<rectangleEl> ( id ) == yes_mark ) {
                   Nx = blockgrid->Give_Nx_rectangle(id);
                   Ny = blockgrid->Give_Ny_rectangle(id);

		   #pragma omp parallel for private(i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)		   
                   for(j = 1;j < Ny;++j )
                       for(i = 1;i < Nx;++i ) {
                           data_rectangles[id][i+(Nx+1)*j] =  a.template Give_data<rectangleEl> ( id, i, j, Nx);
                       }
                }
        }

        // edges
        for(int id = 0;id < ug->Give_number_edges();++id ) {
                //if ( sten_typ == yes_stencil )
                        a.template Update<edgeEl> ( id );

                if(a.template Give_marker<edgeEl> ( id ) == yes_mark ) {
                   Nx = blockgrid->Give_Nx_edge ( id );

                   for(i = 1;i < Nx;++i )  {
                       data_edges[id][i] = a.template Give_data<edgeEl> ( id, i, 0, 0);
                   }
                }
        }

        // points
        for(int id = 0;id < ug->Give_number_points();++id ) {
                //if ( sten_typ == yes_stencil )
                        a.template Update<pointEl> ( id );

                if(a.template Give_marker<pointEl> ( id ) == yes_mark ) {
                   data_points[id][0] = a.template Give_data<pointEl> ( id, 0, 0, 0);
                }
        }
}




/*
template <class DTyp>
template <class A>
void Variable2D<DTyp>::operator= ( const Expr2D_Restriction_Iterdes<A>& a ) {
        int Nx, Ny, Nz, i, j, k, kk, jj, ii;

        stencil_typ sten_typ;

        sten_typ = a.Give_stencil_typ();

        // hexahedra

        for ( int id = 0;id < ug->Give_number_rectangle();++id ) {
                if ( sten_typ == yes_stencil ) {
                        a.template Update<rectangleEl> ( id );
                }

                if ( ug->Give_hexahedron ( id )->my_object ( my_rank ) &&
                        a.template Give_marker<rectangleEl> ( id ) == yes_mark ) {
                        Nx = blockgrid->Give_Nx_rectangle(id);
                        Ny = blockgrid->Give_Ny_rectangle(id);
                        Nz = blockgrid->Give_Nz_hexahedron ( id );

                        if ( a.iterdes.red_black ) {
                                for ( kk = 0;kk < 2;++kk ) {
#pragma omp parallel for private(jj,j,i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)

                                        for ( k = 1 + kk;k < Nz;k = k + 2 ) {
                                                for ( jj = 0;jj < 2;++jj )
                                                        for ( j = 1 + jj;j < Ny;j = j + 2 ) {
                                                                data_rectangles[id][1+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] =
                                                                    a.template Give_data<rectangleEl> ( id, 1, j, k, Nx, Ny );

                                                                for ( i = 2;i < Nx;i = i + 2 ) {
                                                                        //for(i=2;i<Nx;i=i+1) {  // so kann man etwas line_blocking machen
                                                                        data_rectangles[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] =
                                                                            a.template Give_data<rectangleEl> ( id, i, j, k, Nx, Ny );
                                                                        data_rectangles[id][i-1+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] =
                                                                            a.template Give_data<rectangleEl> ( id, i - 1, j, k, Nx, Ny );
                                                                }

                                                                if ( Nx >= 2 )
                                                                        data_rectangles[id][Nx-1+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] =
                                                                            a.template Give_data<rectangleEl> ( id, Nx - 1, j, k, Nx, Ny );
                                                        }
                                        }
                                }
                        } else {
#pragma omp parallel for private(j,i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)

                                for ( k = 1;k < Nz;++k ) {
                                        for ( j = 1;j < Ny;++j ) {
                                                for ( i = 1;i < Nx;i = i + 1 ) {
                                                        data_rectangles[id][i+ ( Nx+1 ) * ( j+ ( Ny+1 ) *k ) ] =
                                                            a.template Give_data<rectangleEl> ( id, i, j, k, Nx, Ny );
                                                }
                                        }
                                }
                        }
                }
        }


        // quadrangles
        for ( int id = 0;id < ug->Give_number_quadrangles();++id ) {
                if ( sten_typ == yes_stencil ) {
                        a.template Update<quadrangleEl> ( id );
                }

                if ( ug->Give_quadrangle ( id )->my_object ( my_rank ) &&
                        a.template Give_marker<quadrangleEl> ( id ) == yes_mark ) {

                        Nx = blockgrid->Give_Nx_quadrangle ( id );
                        Ny = blockgrid->Give_Ny_quadrangle ( id );

                        if ( a.iterdes.red_black ) {
                                for ( jj = 0;jj < 2;++jj ) {
                                        for ( j = 1 + jj;j < Ny;j = j + 2 ) {
                                                data_quadrangles[id][1+ ( Nx+1 ) *j] =
                                                    a.template Give_data<quadrangleEl> ( id, 1, j, 0, Nx, 0 );

                                                for ( i = 2;i < Nx;i = i + 2 ) {
                                                        data_quadrangles[id][i+ ( Nx+1 ) *j] =
                                                            a.template Give_data<quadrangleEl> ( id, i, j, 0, Nx, 0 );
                                                        data_quadrangles[id][i-1+ ( Nx+1 ) *j] =
                                                            a.template Give_data<quadrangleEl> ( id, i - 1, j, 0, Nx, 0 );
                                                }

                                                if ( Nx >= 2 )
                                                        data_quadrangles[id][Nx-1+ ( Nx+1 ) *j] =
                                                            a.template Give_data<quadrangleEl> ( id, Nx - 1, j, 0, Nx, 0 );
                                        }
                                }
                        } else {
                                for ( j = 1;j < Ny;++j )
                                        for ( i = 1;i < Nx;++i ) {
                                                data_quadrangles[id][i+ ( Nx+1 ) *j] =
                                                    a.template Give_data<quadrangleEl> ( id, i, j, 0, Nx, 0 );
                                        }
                        }
                }
        }


        // edges
        for ( int id = 0;id < ug->Give_number_edges();++id ) {
                if ( sten_typ == yes_stencil ) {
                        a.template Update<edgeEl> ( id );
                }

                if ( ug->Give_edge ( id )->my_object ( my_rank ) &&
                        a.template Give_marker<edgeEl> ( id ) == yes_mark ) {
                        Nx = blockgrid->Give_Nx_edge ( id );

                        if ( a.iterdes.red_black ) {
                                for ( ii = 0;ii < 2;++ii ) {
                                        for ( i = 1 + ii;i < Nx;i = i + 2 )
                                                data_edges[id][i] =
                                                    a.template Give_data<edgeEl> ( id, i, 0, 0, 0, 0 );
                                }
                        } else {
                                for ( i = 1;i < Nx;++i )
                                        data_edges[id][i] =
                                            a.template Give_data<edgeEl> ( id, i, 0, 0, 0, 0 );
                        }
                }
        }

        // points
        for ( int id = 0;id < ug->Give_number_points();++id ) {
                if ( sten_typ == yes_stencil ) {
                        a.template Update<pointEl> ( id );
                }

                if ( ug->Give_point ( id )->my_object ( my_rank ) &&
                        a.template Give_marker<pointEl> ( id ) == yes_mark ) {
                        data_points[id][0] =
                            a.template Give_data<pointEl> ( id, 0, 0, 0, 0, 0 );
                }
        }
}

*/


template <class DTyp>
void Variable2D<DTyp>::test()
  {
    int Nx, Ny, i, j;
    // rectangles
    for (int id = 0; id < ug->Give_number_rectangle(); ++id) {
      Nx = blockgrid->Give_Nx_rectangle(id);
      Ny = blockgrid->Give_Ny_rectangle(id);

      for (j = 1; j < Ny; ++j)
        for (i = 1; i < Nx; ++i) {
          data_rectangles[id][i + (Nx + 1) * j] = 1.0;
        }
    }

    for ( int id = 0;id < ug->Give_number_edges();++id )
    {
      Update<edgeEl> ( id );
    }
}

template <class DTyp>
template <class A>
void Variable2D<DTyp>::operator=(const Expr2D<A>& a) {
        int Nx, Ny, i, j;

        const A& ao ( a );

        stencil_typ sten_typ = ao.Give_stencil_typ();
        // rectangles
        for ( int id = 0;id < ug->Give_number_rectangle();++id ) {
                //if(sten_typ == yes_stencil )
                        ao.template Update<rectangleEl> ( id );
                Nx = blockgrid->Give_Nx_rectangle(id);
                Ny = blockgrid->Give_Ny_rectangle(id);
		
		#pragma omp parallel for private(i) num_threads(UGBlocks::numThreadsToTake) if(UGBlocks::useOpenMP)		
                for(j = 1;j < Ny;++j )
                    for(i = 1;i < Nx;++i ) {
                      data_rectangles[id][i+(Nx+1)*j] =  ao.template Give_data<rectangleEl> ( id, i, j, Nx);
                    }
        }

        // edges
        for ( int id = 0;id < ug->Give_number_edges();++id ) {
                //if ( sten_typ == yes_stencil )
                        ao.template Update<edgeEl> ( id );

                Nx = blockgrid->Give_Nx_edge ( id );

                for(i = 1;i < Nx;++i )  {
                    data_edges[id][i] = ao.template Give_data<edgeEl> ( id, i, 0, 0);
                }
        }

        // points
        for (int id = 0;id < ug->Give_number_points();++id ) {
                //if ( sten_typ == yes_stencil )
                        ao.template Update<pointEl> ( id );

                data_points[id][0] = ao.template Give_data<pointEl> ( id, 0, 0, 0);
        }

        for ( int id = 0;id < ug->Give_number_rectangle();++id )
          Update<rectangleEl> ( id );
        for ( int id = 0;id < ug->Give_number_edges();++id )
          Update<edgeEl> ( id );
        for (int id = 0;id < ug->Give_number_points();++id )
          Update<pointEl> ( id );
}

//----------------------------------------------------------------------


#endif // VARIABLE2D_CC_H
