/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
// ------------------------------------------------------------
//
// variableFFT.h
//
// ------------------------------------------------------------
#include "../grid/marker2D.h"
#include "extemp2D.h"
#ifndef VARIABLEFFT_H
#define VARIABLEFFT_H
#ifdef _OPENMP
#include <omp.h>
#include "../ugopenmp.h"
#endif

class Rechteck;
class Interpolate_on_structured_2Dgrid;

/** \addtogroup ExpressionTemplates2D **/
/* @{ */  

/**
 * 2D Variable which allows Fast Fourier Transkormation (FFT)
 **/
class VariableFFT : public Variable2D<std::complex<double> >  {
public:
    /**
     * @param tx depth of grid in x-direction, Nx=2^tx
     * @param ty depth of grid in y-direction, Nx=2^ty 
     * @param ugRectangle  description of physical domain
     **/
    VariableFFT(int tx, int ty, Rechteck& ugRectangle);
    VariableFFT(VariableFFT* other);    
    VariableFFT(const VariableFFT& other);      
    ~VariableFFT();
    void setDomainStrech(double strechX, double strechY); 

    void FFT();
    void inversFFT();     
    
    template <class A>
    void operator=(const Expr2D<A>& a){
           (static_cast<Variable2D<std::complex<double> >*>(this))->operator=(a);
    }

    void operator=(VariableFFT& v);
    
    void operator=(std::complex<double> x);       
    
    void interpolate(Variable2D<std::complex<double> >& v, 
                     std::complex<double> defaultInterpolation);

    
    double getHx() const;
    double getHy() const;
    double getNx();
    double getNy();
    double getSizeX();
    double getSizeY();
    
    
    /*
    double getHx() const { return  getSizeX() / Nx; }  
    double getHy() const { return  getSizeY() / Ny; }
    double getNx() const { return Nx; }
    double getNy() const { return Ny; }
    double getSizeX() const { return assignmentBlockgrid->MaximumStrechedX() - assignmentBlockgrid->MinimumStrechedX(); }
    double getSizeY() const { return assignmentBlockgrid->MaximumStrechedY() - assignmentBlockgrid->MinimumStrechedY(); }
    */
    
//     
private:
    void init();
     bool ownBlockGrid;  
     std::complex<double>* startValueY;

    int Nx;
    int Ny;        

    double Hx;
    double Hy;

     int powerTwo(int t);
     
     Interpolate_on_structured_2Dgrid* assignmentInterpolator;
     Blockgrid2D* assignmentBlockgrid;
};

/* @} */


#endif // VARIABLEFFT_H
