/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#include "../abbrevi2D.h"
#include "../math_lib/math_lib.h"
#include "elements2D.h"

NoTransform2D NoTransform2D::noTransform2D;

// PointEl

PointEl::PointEl() {
//   rel_neigh_locstiff    = NULL;
}

/*
void PointEl::Set_number_neighbors(int n) { 
  assert(number_neighbors!=-1);
  assert(n>1);
  number_neighbors = n; 
}; 
*/

PointEl::~PointEl(){
//   if(rel_neigh_locstiff != NULL)    delete [] rel_neigh_locstiff;
//   rel_neigh_locstiff = NULL;
}

void PointEl::Print() {
  cout<< "   Ort: "; coordinate.Print(); cout << endl;
  for(int pos=0;pos<edgeInformation.size();++pos) {
    cout << "      edgeInfo[" << pos << "] ";
    edgeInformation[pos].print();
  }
  for(int pos=0;pos<rectangleInformation.size();++pos) {
    cout << "      rectangleInfo[" << pos << "] ";
    rectangleInformation[pos].print();
  }
}

/*
 *  gibt id der Kante zurück falls diese existiert, ansonsten Rückgabe -1
 *  id_right:  id des anderen Punktes an der Kante 
 */
int PointEl::getEdge(int id_right) {
   for(int i=0;i<edgeInformation.size();++i) {
     if(edgeInformation.at(i).id_next_point == id_right)
       return edgeInformation.at(i).id_edge;
   }
   return -1;
}

/*
  fügt Kante mit Id id_edge mit Eckpunkt id_right hinzu
*/
void PointEl::addEdge(int id_right, int id_edge) {
  edgeInformation.push_back(EdgeInfo(id_edge,id_right));
}

void PointEl::add_rectangle(Id_of_rec2D neighbourCorners, int id_rectangle) {
   rectangleInformation.push_back(RectangleInfo(id_rectangle,neighbourCorners));
}
    
int PointEl::getLocalIdPoint(int globalId, int pos_rec) {
    int pos = 0;
    for(pos=0;pos<edgeInformation.size();++pos) {
       if(edgeInformation[pos].id_next_point==globalId)
	 return pos;       
    }
    assert(pos_rec<rectangleInformation.size());
    return pos+pos_rec;
}

int PointEl::getLocalIdRectangle(int globalId) {
    for(int pos=0;pos<rectangleInformation.size();++pos) {
	if(rectangleInformation[pos].id_rectangle==globalId)
	  return pos;
    }
    assert(false);
}

	  
void PointEl::addLocalPointInfosToRectangles() {
  for(int pos=0;pos<rectangleInformation.size();++pos) {
     rectangleInformation[pos].id_local_points_ofRectangle.id_SE =
         getLocalIdPoint(rectangleInformation[pos].id_points_ofRectangle.id_SE,
	                 pos);
     rectangleInformation[pos].id_local_points_ofRectangle.id_NE =
         getLocalIdPoint(rectangleInformation[pos].id_points_ofRectangle.id_NE,
	          	 pos);
     rectangleInformation[pos].id_local_points_ofRectangle.id_NW =
         getLocalIdPoint(rectangleInformation[pos].id_points_ofRectangle.id_NW,
	                 pos);
  }
}



// EdgeEl

EdgeEl::EdgeEl() : transformer(NULL) {
  color_of_edge = -1; // means: no color
  transformer     = &(NoTransform2D::noTransform2D);
  is_at_boundary = true;
}

EdgeEl::~EdgeEl(){
}

void EdgeEl::Print() { 
  cout<< "   Ort: " << endl;; //Give_middle_coord(); cout << endl;
  if(isAtBoundary()) cout << "      ist am Rand! " << endl;
  else               cout << "      ist Innen! " << endl;
  cout << "      Farbe: " << color_of_edge << endl;  
  cout << "      klein: " << id_corner[0] << ", gross: " << id_corner[1] << endl;
  cout << "      rec first: " << id_rectangle[0] << ", next: " << id_rectangle[1] << endl;
}

void EdgeEl::Set_id_corners(int id_L, int id_R) { 
  if(id_L < id_R) {
     id_corner[Wd] = id_L; 
     id_corner[Ed] = id_R; 
  }
  else {
     id_corner[Wd] = id_R; 
     id_corner[Ed] = id_L; 
  }
}


// rectangle
RectangleEl::RectangleEl() {
  for(int i=0;i<4;++i){
     transformer[i]  = NULL;
     otherOrientation[i] = false;
  }
  transform_rec = NULL;
}

void RectangleEl::Print() { 
  cout << "   Ort: "; Give_middle_coord().Print();; cout << endl;
  cout << " Ecken: ~~~~~~~~~~~~~~~" << endl;
  for(int c=0;c<4;++c) {
    cout << "      Ecke: "; Abbreviations::Print((dir2D_sons)c);
    cout << " Id: " << id_corner[c] << endl;
  }
  cout << " Kanten: ~~~~~~~~~~~~~~~" << endl;
  for(int c=0;c<4;++c) {
    cout << "      Kante: "; Abbreviations::Print((dir2D)c);
    cout << " Id:" << id_edge[c];
    cout << " andere Orientierung:" << otherOrientation[c] << endl;
  }
}


D2vector RectangleEl::Give_middle_coord() {
  D2vector middle = D2vector(0.0,0.0);
  for(int i=0;i<4;++i) middle = middle + Give_coord((dir2D_sons)i);
  return (middle / 4.0);
}
