/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#ifndef ELEMENTS2D_H
#define ELEMENTS2D_H
#include "../abbrevi2D.h"
#include "../math_lib/math_lib.h"

#include <vector>

/**
 *  Wenn ein Punkt zu einem rectangle gehört, dann sind dies die ids der anderen Ecken des rectangle
 *  Der Punkt selber ist am Ort SD
 */
class Id_of_rec2D {
public:
  Id_of_rec2D(int iSE, int iNW, int iNE) : id_SE(iSE), id_NW(iNW), id_NE(iNE) {} 
  //Id_of_rec2D(Id_of_rec2D& X) : id_SE(X.id_SE), id_NW(X.id_NW), id_NE(X.id_NE) {} 
  int id_SE, id_NW, id_NE;
  void print() { std::cout << " id_SE: " << id_SE << " id_NW: " << id_NW<< " id_NE: " << id_NE; }
};

/**
 * Datenspeicher von Kanten an einem Punkt 
 */
class EdgeInfo {
public:
  EdgeInfo(int idE, int idP) : id_edge(idE), id_next_point(idP) {}
   int id_edge;
   int id_next_point;
   void print() { std::cout << "id_edge: " << id_edge << " id_next_point: " << id_next_point << std::endl; }
};

/**
 * Datenspeicher von Rectangle an einem Punkt 
 * a) id des rectangle
 * b) Wenn ein Punkt zu einem rectangle gehört, dann sind dies die globalen Ids der anderen Ecken des rectangle
 *    Der Punkt selber ist am Ort WS
 * c) Das selbe wie b) aber locale Id bezüglich des Punktes
 */
class RectangleInfo {
public:
  RectangleInfo(int id_rec, Id_of_rec2D id_poiRec) : id_rectangle(id_rec), 
                                                     id_points_ofRectangle(id_poiRec),
                                                     id_local_points_ofRectangle(0,0,0) {}
  int id_rectangle;  ///> zu a)
  Id_of_rec2D id_points_ofRectangle;  ///> zu b) 
  Id_of_rec2D id_local_points_ofRectangle;  ///> zu c) 
  void print() { std::cout << "rec: " << id_rectangle;
                 std::cout << std::endl << "                  id_points_ofRectangle:       "; id_points_ofRectangle.print();
		 std::cout << std::endl << "                  id_local_points_ofRectangle: "; id_local_points_ofRectangle.print();
		 std::cout << std::endl;
  }
};

class Geometric2DObject{
  public:
  Geometric2DObject() { rank_owner = 0; };
    virtual ~Geometric2DObject(){};

    int  Give_rank()            { return rank_owner; }
    void set_rank(int rank)     { rank_owner = rank; }
    bool my_object(int my_rank) { return my_rank==rank_owner; }
      //      cout << " my_rank: " << my_rank << " rank_owner: " << rank_owner << endl;
    
 private:
    int rank_owner;
};

/**
 * Datenstruktur der Punkte:
 * Es gibt: "edgeInformation.size()+rectangleInformation.size()=number_neighbors" Nachbarpunkte
 * Deren lokale Indizes (pos) ist
 *        0, ..., edgeInformation.size() -1:               pos der Punkte an Kanten
 *        edgeInformation.size(), ...,number_neighbors -1: pos der gegenüberliegenden Punkte in Rechtecken
 * Im Uhrzeigersinn sind die Nachbarpunkte und dann die NachbarRectangle
 */
class PointEl : public Geometric2DObject {
  public:
    PointEl();
    ~PointEl();

    void Print();
    
    void Set_coordinate ( D2vector coordinate_ ) { coordinate = coordinate_; }
    D2vector Give_coordinate() { return coordinate; }

    void add_rectangle(Id_of_rec2D neighbourCorners, int id_rectangle); ///> füge neuen Rectangle hinzu
  
    /**
     * gib lokale PunktId der Ecke x im NachbarRectangle pos
     * Wert ist zwischen 0 und edgeInformation.size-1
     */
    inline int give_local_pointId_rec(int pos, local_dir_of_rec2D x);

    /**
    * gib lokale PunktId der Ecke x im NachbarRectangle pos
    * Wert ist zwischen 0 und points.size-1
    */
    inline int give_pointId_rec(int pos, local_dir_of_rec2D x);
  
    /**
     * @param pos lokale Position: 0,..., edgeInformation.size()-1
     * @return gib Id der Kante mit Position pos
     * Wert ist zwischen 0 und edge.size-1
    */
    int give_edgeId(int pos) { return edgeInformation[pos].id_edge; }
  
    
  
    inline int give_id_rec(int pos); ///> gib Id rectangle des NachbarRectangle pos

    int give_number_neighbors_rec()    { return rectangleInformation.size(); } ///> Anzahl der benachbarten Rechtecke
    /**
     * gib Anzahl der Nachbar-Kanten
     */
    int give_number_neighbors_edges() { return edgeInformation.size(); }
     /**
     * gib Anzahl der Nachbar-Punkte, welche die Anzahl der Nachbar-Kanten plus Nachbar-Rechtecke ist
     */
    int give_number_neighbors_points() { return edgeInformation.size()+rectangleInformation.size(); }
  
    /**
     * @param globalId: globale Punkt-Id
     * @param pos_rec:   lokale Position des Rechtecks in dem der Punkt ist
     *                  lokale Position zwischen 0 und rectangleInformation.size-1
     * @return lokale position zwischen 0 und 
     *         (edgeInformation.size + rectangleInformation.size) -1 
    **/
    int getLocalIdPoint(int globalId, int pos_rec);

    /**
     * gibt lokale id zwischen 0 und rectangleInformation.size-1 
     * der rectangle globalId zwischen 0 und rectangle.size-1
    **/
    int getLocalIdRectangle(int globalId);
  
    /**
     * @return gibt id der Kante zurück falls diese existiert, ansonsten Rückgabe -1
     * @param  id_right:  id des anderen Punktes an der Kante 
     **/
    int getEdge(int id_right);
      /**
       * fügt Kante mit Id id_edge mit Eckpunkt id_right hinzu
       **/
    void addEdge(int id_right, int id_edge);
  
  
//  void Set_size_rel_neigh_locstiff (int size); ///> ??

//  void Set_rel_neigh_locstiff  (int neigh, int pos){ rel_neigh_locstiff[neigh] = pos;    }; ///> ???
//  int  Give_rel_neigh_locstiff (int neigh)         { return rel_neigh_locstiff[neigh]; };   ///> ???
 
    void addLocalPointInfosToRectangles();
 
  private:
    D2vector coordinate;

    std::vector<EdgeInfo>                edgeInformation;  ///>  Informationen zu den Nachbar Kanten
    std::vector<RectangleInfo>      rectangleInformation;  ///>  Informationen zu den Nachbar Rectangles
  
    bool boundaryPoint;
};


class EdgeTransformator2D {
 public:
  virtual D2vector transform(double) = 0;
};

class NoTransform2D: public EdgeTransformator2D{
 public:
  D2vector transform(double) {
    return D2vector(0.0,0.0);
  }

  static NoTransform2D noTransform2D;
};

/**
 *  Kante mit folgender Nummerierung:
 *                   id_rectangle[1];
 * id_corner[0];    ----------------------  id_corner[1]
 *                   id_rectangle[0];
 */
class EdgeEl : public Geometric2DObject {
 public:
  EdgeEl();
  ~EdgeEl();

  void Print();
  
  int Give_id_corner_W() { return id_corner[Wd]; }
  int Give_id_corner_E() { return id_corner[Ed]; }

  int Give_id_corner(dir1d EWdd) { return id_corner[EWdd]; }

  /**
   * setzt id_corner[Wd] immer den kleineren Index von id_L, id_R
   * setzt id_corner[Ed] immer den größeren  Index von id_L, id_R
   */
  void Set_id_corners(int id_L, int id_R);

  void Set_color_of_edge(int c) { color_of_edge = c; }; ///> setze Farbe des Freiheitsgrades
  int  Give_color_of_edge()     { return color_of_edge;};  ///> gib Farbe des Freiheitsgrades
  
  // Nachbar Rechtecke
  void setFirst_Id_Rectangle(int id_rec) { id_rectangle[0] = id_rec; }
  void setSecond_Id_Rectangle(int id_rec) { id_rectangle[1] = id_rec; is_at_boundary= false; }
  int getNumberNextRectangles()  { return (2-is_at_boundary); }
  int getIdNextRectangle(bool i) { return id_rectangle[i]; }
  
  // information about element mapping [0,1] ->  R^3 
  //D2vector (*transform)(double, double*); // transformation of edge
  EdgeTransformator2D* transformer;
  
  D2vector Give_coord(dir1d c)        { return coordinates[c]; }
  void Set_coord(dir1d c, D2vector v) { coordinates[c] = v; }
    
   // exterior
  bool isAtBoundary() { return is_at_boundary; }
  
  D2vector Give_middle_coord(); ///> geometric middle of coordinates of corner points
  
 private:
  int id_corner[2];
  int id_rectangle[2];

  bool is_at_boundary;
  
  int color_of_edge;
  D2vector coordinates[2]; ///> for  local coordinates in edge
};



class RectangleEl : public Geometric2DObject {
 public:
  RectangleEl();
  ~RectangleEl(){};

  void Print();
    
  // structured part:
  int Give_id_corner(dir2D_sons i)         { return id_corner[i]; }
  void Set_id_corner(dir2D_sons i, int id) { id_corner[i] = id; }

  int Give_id_edge(dir2D i)         { return id_edge[i]; }
  void Set_id_edge(dir2D i, int id) { id_edge[i] = id; }


  D2vector Give_coord(dir2D_sons c)        { return coordinates[c]; }
  void Set_coord(dir2D_sons c, D2vector v) { coordinates[c] = v; }

  // information edges 4 times [0,1] -> R^2
  EdgeTransformator2D* transformer[4];
	
  // information hex [0,1]^3 -> R^3
  D2vector (*transform_rec)(int, double,double,double); // transformation of volumne

  
  bool getOrientationEdge(dir2D i) { return otherOrientation[i]; }  ///> false heisst dass rectangle innen/zuerst
  void edgeIsOtherOrientation(dir2D i)  { otherOrientation[i] = true; }
  void edgeIsNormalOrientation(dir2D i) { otherOrientation[i] = false; }
  
  // geometric middle of coordinates of corner points
  D2vector Give_middle_coord();
  
 private:
  int id_corner[4];
  int id_edge[4];

  bool otherOrientation[4];   // Dieser Speicher wird zweimal verwendet
                              // a) Bei der Konstruktion der Kanten, false:  Kante ist an nur einem Rechteck
                              //                                     true:   Kante ist an zwei Rechtecken           
                              // b) Die Informationen von b) werden überschrieben:
                              //                                    false: Kante hat Orientierung von W->E, S->N 
                              //                                    true:  Kante hat umgekehrte Orientierung
  // for local coordinates in hexahedron
  D2vector coordinates[4];
};

//////////////////////////////////////////
// Implementation of inline functions
//////////////////////////////////////////



/*

inline void PointEl::Set_loc_id_rec(int pos, local_dir_of_rec2D x, int loc_id){
  assert(pos>=0 && pos<number_neighbors);
  if     (x==SErec) loc_id_rec[pos].id_SE = loc_id;
  else if(x==NWrec) loc_id_rec[pos].id_NW = loc_id;
  else {
     assert(x != NErec);
     loc_id_rec[pos].id_NE = loc_id;
  }
}
*/
inline int  PointEl::give_pointId_rec(int pos, local_dir_of_rec2D x){
  assert(pos>=0 && pos<rectangleInformation.size());
  if(x==SErec) return rectangleInformation[pos].id_points_ofRectangle.id_SE;
  if(x==NWrec) return rectangleInformation[pos].id_points_ofRectangle.id_NW;
  assert(x == NErec); 
  return rectangleInformation[pos].id_points_ofRectangle.id_NE;
}


inline int  PointEl::give_local_pointId_rec(int pos, local_dir_of_rec2D x){
  assert(pos>=0 && pos<rectangleInformation.size());
  if(x==SErec) return rectangleInformation[pos].id_local_points_ofRectangle.id_SE;
  if(x==NWrec) return rectangleInformation[pos].id_local_points_ofRectangle.id_NW;
  assert(x == NErec); 
  return rectangleInformation[pos].id_local_points_ofRectangle.id_NE;
}








inline int PointEl::give_id_rec(int pos) {
  return rectangleInformation[pos].id_rectangle;
}



#endif // ELEMENTS2D_H


