


/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#ifndef GRIDFROM3D_H
#define GRIDFROM3D_H

//#include "../mympi.h"
#include "../../../program/source/abbrevi.h"
#include "../../../program/source/parameter.h"
#include "../../../program/source/math_lib/math_lib.h"
#include "../../../program/source/basics/basic.h"
#include "../../../program/source/grid/elements.h"
#include "../../../program/source/grid/parti.h"
#include "../../../program/source/grid/ug.h"
#include "../../../program/source/grid/blockgrid.h"
#include "../extemp/functor2D.h"
#include "ug2D.h"
#include "blockgrid2D.h"
#include "marker2D.h"
#include "examples_ug2D.h"

#ifdef CPP11
template <class DTyp_Result, class DTyp, typename... Arguments>
class Function2d2;
#else
template <class DTyp_Result, class DTyp>
class Function2d2;
#endif

class Blockgrid2DFrom3D;
class X_coordinate2d;
class Y_coordinate2d;
class PMLGrid_Marker2D_Inner;
class PMLGrid_Marker2D_PML;
/**
 * Rectangle or Disc from Hexahedron or Cylinder
 * 
 * **/
class UGFrom3DSlice : public Unstructured2DGrid {
 public:
  UGFrom3DSlice(Unstructured_grid* ug3D, double PMLwidth = -1.0);
  ~UGFrom3DSlice(){};
  bool inDomain(const D2vector& point_coord);  
  Unstructured_grid* Give_unstructured_grid() { return ug3Dp; }
  
  void getCoordinate(const D2vector& point_coord, int& id, int& i, int& j, const Blockgrid2D& blockgrid) const ;

  bool getCoordinate(const D2vector& point_coord, int& id, double& i, double& j, const Blockgrid2D& blockgrid) const ;

  Variable2D< std::complex<double> > getPMLlayer(Blockgrid2D& blockgrid, const double lambda);

  PMLGrid_Marker2D_Inner getInnerMarker();
  PMLGrid_Marker2D_PML getPMLMarker();

  std::complex<double> evaluate(const double x, const double y);
  
  void Test() const;

 private:
  Unstructured_grid* ug3Dp;
      
  //static std::complex<double> set_pml(const double x, const double y, const double pml_start, const double pml_end, const double lambda);

  //static Function2d2< std::complex<double>,double,double, double, double> Set_pml;

  Functor2D2<std::complex<double>,double,UGFrom3DSlice> PMLfunctor;
  double r_inner;
  double r_outer;
  double lambda;

   enum UG3Dtype {
     CylinderUG,
     HexaederUG
   } ug3Dtype;

   void constructRectangle();
   void constructSectionedRectangle();

   void constructDisc();
   void constructPMLDisc(double PMLwidth);

   void getDiscCoordinate(const D2vector& point_coord, int& id, int& i, int& j, const Blockgrid2D& blockgrid) const ;
   bool getDiscCoordinate(const D2vector& point_coord, int& id, double& i, double& j, const Blockgrid2D& blockgrid) const ;

   void getRectangleCoordinate(const D2vector& point_coord, int& id, int& i, int& j, const Blockgrid2D& blockgrid) const ;
   bool getRectangleCoordinate(const D2vector& point_coord, int& id, double& i, double& j, const Blockgrid2D& blockgrid) const ;

   void inverseTransformation(const D2vector& point_coord, double& i, double& j) const;

   void constructAdaptiveRectangle();
   void constructTestRectangle();

   //Disc data
   NETransform2D transform_disc_NE;
   NETransform2D transform_disc_NE_inner;
   NETransform2D transform_disc_NE_PML;
   NWTransform2D transform_disc_NW;
   NWTransform2D transform_disc_NW_inner;
   NWTransform2D transform_disc_NW_PML;
   SETransform2D transform_disc_SE;
   SETransform2D transform_disc_SE_inner;
   SETransform2D transform_disc_SE_PML;
   SWTransform2D transform_disc_SW;
   SWTransform2D transform_disc_SW_inner;
   SWTransform2D transform_disc_SW_PML;

   double PMLwidth;
};


class Blockgrid2DFrom3D : public Blockgrid2D {
 public:
    Blockgrid2DFrom3D(UGFrom3DSlice* ug, Blockgrid* bg3D);

    Blockgrid2DFrom3D(UGFrom3DSlice* ug, const std::vector<int>& N_fine);
};

class  PMLGrid_Marker2D_PML: public Unstructured2DGrid_Marker2D {
public:
  PMLGrid_Marker2D_PML ( UGFrom3DSlice& ug_slize ) : Unstructured2DGrid_Marker2D ( (Unstructured2DGrid*)&ug_slize ) {
    for ( int id=0; id<ug_slize.Give_number_rectangle(); ++id ) {
      if(id >= 5)
            Set_marker<pointEl> ( id, no_mark );
      else
        Set_marker<pointEl> ( id, yes_mark );
    }
    for ( int id=0; id<ug_slize.Give_number_edges(); ++id ) {
      if(id >= 12)
            Set_marker<pointEl> ( id, no_mark );
      else
        Set_marker<pointEl> ( id, yes_mark );
    }
    for ( int id=0; ug_slize.Give_number_points(); ++id ) {
      if(id >= 8)
            Set_marker<pointEl> ( id, no_mark );
      else
        Set_marker<pointEl> ( id, yes_mark );
    }
  }
//  PMLGrid_Marker2D_PML(const PMLGrid_Marker2D_PML& copy) : Unstructured2DGrid_Marker2D((Unstructured2DGrid*)(copy.ug))
//  {
//    for ( int id=0; id<ug->Give_number_rectangle(); ++id ) {
//      if(id >= 5)
//            Set_marker<pointEl> ( id, no_mark );
//      else
//        Set_marker<pointEl> ( id, yes_mark );
//    }
//    for ( int id=0; id<ug->Give_number_edges(); ++id ) {
//      if(id >= 12)
//            Set_marker<pointEl> ( id, no_mark );
//      else
//        Set_marker<pointEl> ( id, yes_mark );
//    }
//    for ( int id=0; ug->Give_number_points(); ++id ) {
//      if(id >= 8)
//            Set_marker<pointEl> ( id, no_mark );
//      else
//        Set_marker<pointEl> ( id, yes_mark );
//    }
//  }
};

class PMLGrid_Marker2D_Inner: public Unstructured2DGrid_Marker2D {
 public:
  PMLGrid_Marker2D_Inner ( UGFrom3DSlice& ug_slize ) : Unstructured2DGrid_Marker2D ( (Unstructured2DGrid*)&ug_slize ) {
    for ( int id=0; id<ug_slize.Give_number_rectangle(); ++id ) {
      if(id >= 5)
            Set_marker<pointEl> ( id, yes_mark );
      else
        Set_marker<pointEl> ( id, no_mark );
    }
    for ( int id=0; id<ug_slize.Give_number_edges(); ++id ) {
      if(id >= 12)
            Set_marker<pointEl> ( id, yes_mark );
      else
        Set_marker<pointEl> ( id, no_mark );
    }
    for ( int id=0; ug_slize.Give_number_points(); ++id ) {
      if(id >= 8)
            Set_marker<pointEl> ( id, yes_mark );
      else
        Set_marker<pointEl> ( id, no_mark );
    }
  }
//
//  PMLGrid_Marker2D_Inner (const PMLGrid_Marker2D_Inner& copy) : Unstructured2DGrid_Marker2D ( (Unstructured2DGrid*)copy.ug ) {
//    for ( int id=0; id<ug->Give_number_rectangle(); ++id ) {
//      if(id >= 5)
//            Set_marker<pointEl> ( id, yes_mark );
//      else
//        Set_marker<pointEl> ( id, no_mark );
//    }
//    for ( int id=0; id<ug->Give_number_edges(); ++id ) {
//      if(id >= 12)
//            Set_marker<pointEl> ( id, yes_mark );
//      else
//        Set_marker<pointEl> ( id, no_mark );
//    }
//    for ( int id=0; ug->Give_number_points(); ++id ) {
//      if(id >= 8)
//            Set_marker<pointEl> ( id, yes_mark );
//      else
//        Set_marker<pointEl> ( id, no_mark );
//    }
//  }
};


#endif // EXAMPLES_UG_2D3D_H






