/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

// #include <assert.h>
#include "../../../program/source/mympi.h"  // von 3D UGBlocks
#include "../abbrevi2D.h"
#include "../math_lib/math_lib.h"
#include "elements2D.h"
#include "parti2D.h"
#include "ug2D.h"
//#include "blockgrid.h"
#include "marker2D.h"
#include <stdexcept>
#include <vector>


Marker2D::Marker2D ( Unstructured2DGrid* ug_ ) {
	ug = ug_;
	marker_rectangles.resize(ug->Give_number_rectangle());
	marker_edges.resize(ug->Give_number_edges());       
	marker_points.resize(ug->Give_number_points());

	on();
}

Marker2D::~Marker2D() {
}



void Marker2D::on() {
	for ( int id=0; id<marker_rectangles.size(); ++id )
		Set_marker<rectangleEl> ( id, yes_mark );

	for ( int id=0; id<marker_edges.size(); ++id )
		Set_marker<edgeEl> ( id, yes_mark );

	for ( int id=0; id<marker_points.size(); ++id )
		Set_marker<pointEl> ( id, yes_mark );
}


void Marker2D::off() {
	for ( int id=0; id<marker_rectangles.size(); ++id )
		Set_marker<rectangleEl> ( id, no_mark );

	for ( int id=0; id<marker_edges.size(); ++id )
		Set_marker<edgeEl> ( id, no_mark );

	for ( int id=0; id<marker_points.size(); ++id )
		Set_marker<pointEl> ( id, no_mark );
}
/*
void Marker2D::Mark_hexahedra_only( void ) {
	for ( int id=0; id<ug->Give_number_rectangle(); ++id )
		Set_marker<rectangleEl> ( id, yes_mark );

	for ( int id=0; id<ug->Give_number_edges(); ++id )
		Set_marker<edgeEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_points(); ++id )
		Set_marker<pointEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_quadrangles(); ++id )
		Set_marker<quadrangleEl> ( id, no_mark );
}

void Marker2D::Mark_quadrangles_only( void ) {
	for ( int id=0; id<ug->Give_number_rectangle(); ++id )
		Set_marker<rectangleEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_edges(); ++id )
		Set_marker<edgeEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_points(); ++id )
		Set_marker<pointEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_quadrangles(); ++id )
		Set_marker<quadrangleEl> ( id, yes_mark );
}

void Marker2D::Mark_edges_only( void ) {
	for ( int id=0; id<ug->Give_number_rectangle(); ++id )
		Set_marker<rectangleEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_edges(); ++id )
		Set_marker<edgeEl> ( id, yes_mark );

	for ( int id=0; id<ug->Give_number_points(); ++id )
		Set_marker<pointEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_quadrangles(); ++id )
		Set_marker<quadrangleEl> ( id, no_mark );
}

void Marker2D::Mark_points_only( void ) {
	for ( int id=0; id<ug->Give_number_rectangle(); ++id )
		Set_marker<rectangleEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_edges(); ++id )
		Set_marker<edgeEl> ( id, no_mark );

	for ( int id=0; id<ug->Give_number_points(); ++id )
		Set_marker<pointEl> ( id, yes_mark );

	for ( int id=0; id<ug->Give_number_quadrangles(); ++id )
		Set_marker<quadrangleEl> ( id, no_mark );
}

void Marker2D::Unmark_hexahedra( void ) {
	for ( int id=0; id<ug->Give_number_rectangle(); ++id )
		Set_marker<rectangleEl> ( id, no_mark );
}

void Marker2D::Unmark_quadrangles( void ) {
	for ( int id=0; id<ug->Give_number_quadrangles(); ++id )
		Set_marker<quadrangleEl> ( id, no_mark );

}
void Marker2D::Unmark_edges( void ) {
	for ( int id=0; id<ug->Give_number_edges(); ++id )
		Set_marker<edgeEl> ( id, no_mark );
}
void Marker2D::Unmark_points( void ) {
	for ( int id=0; id<ug->Give_number_points(); ++id )
		Set_marker<pointEl> ( id, no_mark );
}


// Marks hexaeder and adjacent elements.
void Marker2D::selectHexBlock ( unsigned int chosenHexBlock ) {
	if ( ( int ) chosenHexBlock >= ug->Give_number_rectangle() )
		throw ( std::runtime_error ( std::string ( "Marker2D::selectHexBlock(): Selected block does not exist" ) ) );
//   cout << "select Block " << chosenHexBlock << endl;
	Set_marker<rectangleEl> ( chosenHexBlock, yes_mark );

	Hexahedron_el* pHex = ug->Give_hexahedron ( chosenHexBlock );

	for ( int i=0; i<12; ++i ) {
		Set_marker<edgeEl> ( pHex->Give_id_edge ( ( Edges_cell ) i ), yes_mark );
	}

	for ( int i=0; i<6; ++i ) {
		Set_marker<quadrangleEl> ( pHex->Give_id_quadrangle ( ( dir3D ) i ), yes_mark );
	}
	for ( int i=0; i<3; ++i ) {
		Set_marker<pointEl> ( pHex->Give_id_corner ( ( dir3D_sons ) i ), yes_mark );
	}
}

// Marks hexaeder and adjacent elements.
void Marker2D::selectSingleHexBlock ( unsigned int chosenHexBlock ) {
	this->off();
	selectHexBlock(chosenHexBlock);
}


void Marker2D::selectHexBlocks ( unsigned int firstBlock, unsigned int numBlocks) {
	this->off();
	for ( unsigned int i = 0; i < numBlocks; ++i ) {
		selectHexBlock(firstBlock+i);
	}
}

void Marker2D::selectHexBlocks ( std::vector<unsigned int> & chosenHexBlocks ) {
	this->off();
	for ( unsigned int count = 0;
	        count <chosenHexBlocks.size();
	        ++count ) {
		unsigned chosenBlock = chosenHexBlocks.at ( count );
		selectHexBlock(chosenBlock);
	}
}
*/


Boundary_Marker2D::Boundary_Marker2D ( Unstructured2DGrid* ug_ ) : Marker2D ( ug_ )    {
	EdgeEl* edge;
	for ( int id=0; id<ug->Give_number_rectangle(); ++id )     {
		Set_marker<rectangleEl> ( id, no_mark );
	}
	for ( int id=0; id<ug->Give_number_points(); ++id )    {
		Set_marker<pointEl> ( id, no_mark );
	}
	for ( int id=0; id<ug->Give_number_edges(); ++id ) {
	        edge = ug->Give_edge(id);
		if(edge->isAtBoundary() == false)
		   Set_marker<edgeEl>(id,no_mark);
		else  {
			Set_marker<edgeEl> ( id, yes_mark );
			for ( int i=0; i<2; ++i )
				Set_marker<pointEl>(edge->Give_id_corner((dir1d)i), yes_mark);
		}
	}

}

Boundary_Marker2D::~Boundary_Marker2D() {}


Unstructured2DGrid_Marker2D::Unstructured2DGrid_Marker2D ( Unstructured2DGrid* ug_ ) : Marker2D ( ug_ ) {
	for ( int id=0; id<ug->Give_number_rectangle(); ++id ) {
		Set_marker<rectangleEl> ( id, yes_mark );
	}
	for ( int id=0; id<ug->Give_number_edges(); ++id ) {
		Set_marker<edgeEl> ( id, yes_mark );
	}
	for ( int id=0; id<ug->Give_number_points(); ++id ) {
		Set_marker<pointEl> ( id, yes_mark );
	}
}

Unstructured2DGrid_Marker2D::~Unstructured2DGrid_Marker2D() {}

Blockgrid_Marker2D::Blockgrid_Marker2D ( Unstructured2DGrid* ug_ ) : Marker2D ( ug_ ) {
}

Blockgrid_Marker2D::~Blockgrid_Marker2D() {}

void Unstructured2DGrid_Marker2D::complement ( Marker2D* marker ) {
	for ( int id=0; id<ug->Give_number_rectangle(); ++id ) {
		if ( marker->Give_marker<rectangleEl> ( id ) == no_mark )
			Set_marker<rectangleEl> ( id, yes_mark );
		else Set_marker<rectangleEl> ( id, no_mark );
	}
	for ( int id=0; id<ug->Give_number_edges(); ++id ) {
		if ( marker->Give_marker<edgeEl> ( id ) == no_mark )
			Set_marker<edgeEl> ( id, yes_mark );
		else Set_marker<edgeEl> ( id, no_mark );
	}
	for ( int id=0; id<ug->Give_number_points(); ++id ) {
		if ( marker->Give_marker<pointEl> ( id ) == no_mark )
			Set_marker<pointEl> ( id, yes_mark );
		else Set_marker<pointEl> ( id, no_mark );
	}
}

/*
void Unstructured2DGrid_Marker2D::interior_boundary ( Marker2D* marker ) {
  // kopieren
  for ( int id=0; id<ug->Give_number_rectangle(); ++id ) {
     if ( marker->Give_marker<rectangleEl> ( id ) == no_mark )
	Set_marker<rectangleEl> ( id, no_mark );
     else Set_marker<rectangleEl> ( id, yes_mark );
  }
  	for ( int id=0; id<ug->Give_number_quadrangles(); ++id ) {
		if ( marker->Give_marker<quadrangleEl> ( id ) == no_mark )
			Set_marker<quadrangleEl> ( id, no_mark );
		else Set_marker<quadrangleEl> ( id, yes_mark );
	}
	for ( int id=0; id<ug->Give_number_edges(); ++id ) {
		if ( marker->Give_marker<edgeEl> ( id ) == no_mark )
			Set_marker<edgeEl> ( id, no_mark );
		else Set_marker<edgeEl> ( id, yes_mark );
	}
	for ( int id=0; id<ug->Give_number_points(); ++id ) {
		if ( marker->Give_marker<pointEl> ( id ) == no_mark )
			Set_marker<pointEl> ( id, no_mark );
		else Set_marker<pointEl> ( id, yes_mark );
	}
	
	
  //aendern
  for ( int id_quad=0; id_quad<ug->Give_number_quadrangles(); ++id_quad ) {
     if ( marker->Give_marker<quadrangleEl> ( id_quad ) == no_mark ) {
       
       for(int e=0;e < 4;++e) {
          Set_marker<edgeEl> (ug->Give_quadrangle ( id_quad )->Give_id_edge ( (dir2D)e ),
			       no_mark );
       }
       for(int c=0;c < 4;++c) {
          Set_marker<pointEl> (ug->Give_quadrangle ( id_quad )->Give_id_corner ( (dir2D_sons)c ), 
				no_mark );
       } 
     }
  }
}



void Unstructured2DGrid_Marker2D::operator= ( bool ( *Marker2D_formula ) ( double x, double y, double z ) ) {
	D3vector middle;

	Hexahedron_el* hex;
	Quadrangle_el* quad;
	Edge_el*       edge;

	for ( int id=0; id<ug->Give_number_rectangle(); ++id ) {
		hex = ug->Give_hexahedron ( id );


		middle = hex->Give_middle_coord();

		if ( Marker2D_formula ( middle.x,middle.y,middle.z ) )
			Set_marker<rectangleEl> ( id, yes_mark );
		else Set_marker<rectangleEl> ( id, no_mark );
	}
	for ( int id=0; id<ug->Give_number_quadrangles(); ++id ) {
		quad = ug->Give_quadrangle ( id );


		middle = quad->Give_middle_coord();

		if ( Marker2D_formula ( middle.x,middle.y,middle.z ) )
			Set_marker<quadrangleEl> ( id, yes_mark );
		else Set_marker<quadrangleEl> ( id, no_mark );
	}
	for ( int id=0; id<ug->Give_number_edges(); ++id ) {
		edge = ug->Give_edge ( id );

		middle = ( ug->Give_point ( edge->Give_id_corner_W() )->Give_coordinate() +
		           ug->Give_point ( edge->Give_id_corner_E() )->Give_coordinate() ) / 2.0;

		if ( Marker2D_formula ( middle.x,middle.y,middle.z ) )
			Set_marker<edgeEl> ( id, yes_mark );
		else Set_marker<edgeEl> ( id, no_mark );
	}
	for ( int id=0; id<ug->Give_number_points(); ++id ) {
		middle = ug->Give_point ( id )->Give_coordinate();

		if ( Marker2D_formula ( middle.x,middle.y,middle.z ) )
			Set_marker<pointEl> ( id, yes_mark );
		else Set_marker<pointEl> ( id, no_mark );
	}
}




void Unstructured2DGrid_Marker2D::setBoundary( bool ( *Marker2D_formula ) ( double x, double y, double z ) )  {
	D3vector middle;


	Quadrangle_el* quad;
	Edge_el*       edge;

	for ( int id=0; id<ug->Give_number_quadrangles(); ++id ) {
		quad = ug->Give_quadrangle ( id );
		if ( quad->Give_exists_exterior() == false) {

			// mark quadrangles
			middle = quad->Give_middle_coord();
			if ( Marker2D_formula ( middle.x,middle.y,middle.z ) )
				Set_marker<quadrangleEl> ( id, yes_mark );
			else Set_marker<quadrangleEl> ( id, no_mark );

			// mark edges
			for ( int i=0; i<4; ++i ) {
				int idquad = quad->Give_id_edge ( ( dir2D ) i );
				edge = ug->Give_edge ( idquad );

				middle = ( ug->Give_point ( edge->Give_id_corner_W() )->Give_coordinate() +
				           ug->Give_point ( edge->Give_id_corner_E() )->Give_coordinate() ) / 2.0;

				if ( Marker2D_formula ( middle.x,middle.y,middle.z ) )
					Set_marker<edgeEl> ( idquad, yes_mark );
				else Set_marker<edgeEl> ( idquad, no_mark );
			}

			// mark points
			for ( int i=0; i<4; ++i ) {
				int idcorner =  quad->Give_id_corner ( ( dir2D_sons ) i );
				middle = ug->Give_point ( idcorner )->Give_coordinate();

				if ( Marker2D_formula ( middle.x,middle.y,middle.z ) )
					Set_marker<pointEl> ( idcorner, yes_mark );
				else Set_marker<pointEl> ( idcorner, no_mark );
			}
		}
	}
}
*/
