/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/
 
#ifndef MARKER2D_H
#define MARKER2D_H
#include "../abbrevi2D.h"
#include "ug2D.h"
#include <vector>

class Marker2D {
public:
	Marker2D( Unstructured2DGrid* ug_ );
	virtual ~Marker2D();

	// select everything
	void on();

	//select nothing
	void off();


	template <elementTyp TYP_EL>
	inline marker_typ Give_marker(int i) const;

	template <elementTyp TYP_EL>
	inline void Set_marker(int i, marker_typ mt);

	void operator=(bool setValue) {
		if(setValue) on();
		else off();
	}

	void Mark_rectangles_only( void ); 		///< Marks hexahedra and unmarks everything else
	void Mark_edges_only( void );			///< Marks edges and unmarks everything else
	void Mark_points_only( void );			///< Marks points and unmarks everything else

	void Unmark_rectangles( void ); 		///< Unmarks hexahedra, keeps everything else
	void Unmark_edges( void );			///< Unmarks edges, keeps everything else
	void Unmark_points( void );			///< Unarks points, keeps everything else
		
protected:
	Unstructured2DGrid*       ug;

private:
	// own data
	std::vector<marker_typ> marker_rectangles;      
	std::vector<marker_typ> marker_edges;         
	std::vector<marker_typ> marker_points;    
};

class Boundary_Marker2D : public Marker2D {
public:
	Boundary_Marker2D(Unstructured2DGrid* ug_);
	~Boundary_Marker2D();

};

class Unstructured2DGrid_Marker2D :  public Marker2D {
public:
	Unstructured2DGrid_Marker2D(Unstructured2DGrid* ug_);
	~Unstructured2DGrid_Marker2D();
	void complement (Marker2D* marker);
	void interior_boundary (Marker2D* marker);
	void operator= (bool  (*Marker2D_formula)(double x, double y));
	void operator=(bool setValue) {
		if(setValue) on();
		else off();
	}
	void setBoundary(bool  (*Marker2D_formula)(double x, double y));
};

class Blockgrid_Marker2D :  public Marker2D {
public:
	Blockgrid_Marker2D(Unstructured2DGrid* ug_);
	~Blockgrid_Marker2D();

};


template <>
inline marker_typ Marker2D::Give_marker<rectangleEl>(int i) const {
	return marker_rectangles[i];
}

template <>
inline marker_typ Marker2D::Give_marker<edgeEl>      (int i) const {
	return marker_edges[i];
}

template <>
inline marker_typ Marker2D::Give_marker<pointEl>     (int i) const {
	return marker_points[i];
}


template <>
inline void Marker2D::Set_marker<rectangleEl>(int i, marker_typ mt) {
	marker_rectangles[i] = mt;
}

template <>
inline void Marker2D::Set_marker<edgeEl>      (int i, marker_typ mt) {
	marker_edges[i] = mt;
}

template <>
inline void Marker2D::Set_marker<pointEl>     (int i, marker_typ mt) {
	marker_points[i] = mt;
}

#endif // MARKER2D_H
