/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/


#include "../../../program/source/mympi.h"  // von 3D UGBlocks
#include "../abbrevi2D.h"
#include "../math_lib/math_lib.h"
#include "elements2D.h"
#include "parti2D.h"
#include "ug2D.h"

void Unstructured2DGrid::PrintVtk(ofstream *Datei) {
  D2vector v;
  Datei->precision(3);
  Datei->setf(std::ios::fixed,std::ios::floatfield);

  // Teil 0: Write information
  *Datei << "# vtk DataFile Version 2.0\n"
          << "ug2D" << endl
          << "ASCII\n"
          << "DATASET UNSTRUCTURED_GRID\n";

  // Teil 1: Kopfzeile schreiben
  *Datei << "POINTS " << points.size()*1 << " float\n";
    
  // Teil 2: Koordinaten der Punkte ausgeben
  for(int id=0;id < points.size();++id) {
    v = points.at(id).Give_coordinate();
    *Datei << v.x << "  "
	   << v.y << "  "
	   << 0.0 << "\n";
  }

  // Teil 3: Zellen ausgeben
  int cell_total = rectanglEs.size();
  *Datei << "\nCELLS " << cell_total << " " << cell_total*5  << "\n";
  for(int id=0;id < rectanglEs.size();++id) {
      *Datei <<  "4 "
      		  << rectanglEs.at(id).Give_id_corner(SWdir2D) << " " // WS
                  << rectanglEs.at(id).Give_id_corner(SEdir2D) << " " // ES
                  << rectanglEs.at(id).Give_id_corner(NEdir2D) << " " // EN
                  << rectanglEs.at(id).Give_id_corner(NWdir2D) << " " // WN
                  << "\n";
  }

  // Teil 4:
  *Datei << "\nCELL_TYPES " << cell_total << "\n";
  for(int id=0;id < rectanglEs.size();++id)
      *Datei <<  "9 " << endl;
  
  
/*
  // Teil 5: Ausgabe der id der Points
  *Datei << "\nPOINT_DATA " << points.size()*1 << "\n";
  *Datei << "SCALARS " << " ug2D " << " float 1\n";
  *Datei << "LOOKUP_TABLE default\n";

  for(int id=0;id < points.size()*1;++id) {
//    v = points.at(id).Give_coordinate();
    *Datei << " " << id << "\n";
  }
*/
  // Teil 5: Ausgabe der id der Points
  *Datei << "\nCELL_DATA " << rectanglEs.size()*1 << "\n";
  *Datei << "SCALARS " << " ug2D " << " float 1\n";
  *Datei << "LOOKUP_TABLE default\n";

  for(int id=0;id < rectanglEs.size()*1;++id) {
    *Datei << " " << id << "\n";
  }


}
