/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

//////////////////////////////////////////////////////////
//   dummympi.h  (include file for library)
/////////////////////////////////////////////////////////

// for CLOCKS_PER_SEC macro:
#ifdef _WIN32
#include <time.h>
#endif

#ifdef _MPI_PARALLEL
#include <mpi.h>
#endif 

#include <iostream>

#ifndef _MPI_PARALLEL


#ifndef DUMMYMPI_H_
#define DUMMYMPI_H_



#define MPI_MAX_PROCESSOR_NAME 6
inline int MPI_Get_processor_name(char s[MPI_MAX_PROCESSOR_NAME], int* i) {
  return 0;
}

#define MPI_Comm         int
#define MPI_COMM_WORLD   0
#define MPI_Request      int
#define MPI_Status       int
#define MPI_Datatype     int

#define MPI_INT        0
#define MPI_DOUBLE     0
#define MPI_LOR        0
#define MPI_SUM        0
#define MPI_MIN        0
#define MPI_BOOL       0
#define MPI_OR         0
#define MPI_DOUBLE_COMPLEX        0
#define MPI_BYTE        0
#define MPI_MAX        0

inline void Error_Dummy_mpi() {
  std::cout << " Error_Dummy_mpi! " << std::endl;
}


inline void MPI_Comm_rank(int, int* my_rank) { *my_rank = 0; }
inline void MPI_Comm_size(int, int* p)       { *p       = 1; }

template <class A>
inline void MPI_Irecv(A*,int,int,int,int,int,int*) {}

template <class A>
inline void MPI_Isend(A*,int,int,int,int,int,int*) {}

template <class A>
inline void MPI_Recv(A*,int,int,int,int,int) {}
template <class A>
inline void MPI_Send(A*,int,int,int,int,int) {}

template <class A>
inline void MPI_Recv(A*,int,int,int,int,int,int*) {}
template <class A>
inline void MPI_Send(A*,int,int,int,int,int,int*) {}

inline void MPI_Waitall(int,int*,int*) {}
inline void MPI_Barrier(int) {}
inline void MPI_Gather(int*,int,int,int*,int,int,int,int) {}
inline void MPI_Type_commit(int*) {}
inline void MPI_Type_free(int*) {}
inline void MPI_Type_vector(int,int,int,int,int*) {}


inline void MPI_Init(int*,char***) {}

inline double MPI_Wtime() {
  return clock() / CLOCKS_PER_SEC; 
}

inline void MPI_Allreduce(int* input,int* output,int,int,int,int) {
  *output = *input;
}
inline void MPI_Allreduce(double* input,double* output,int,int,int,int) {
  *output = *input;
}
inline void MPI_Allreduce(bool* input,bool* output,int,int,int,int) {
  *output = *input;
}

inline void MPI_Finalize() {
  std::cout << " This is the serial version of EXPDE! " << std::endl;
}




#endif

#endif
