/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

// ------------------------------------------------------------
//
// diffop.h
// local stiffness matrix
//
// ------------------------------------------------------------


#ifndef DIFFOP2D_H
#define DIFFOP2D_H
using namespace ::_COLSAMM_;


#define Ind_loc_matrix_rectangle_BL(i,j) ((j)*(Nx+1)+(i))
#define Ind_loc_matrix_rectangle_BR(i,j) ((j)*(Nx+1)+(i+1))
#define Ind_loc_matrix_rectangle_TL(i,j) ((j+1)*(Nx+1)+(i))
#define Ind_loc_matrix_rectangle_TR(i,j) ((j+1)*(Nx+1)+(i+1))
//-----------------------------------------------------------------------------

///////////////////////////
// 0. Predefinition of class for MG
///////////////////////////
#include "../abbrevi2D.h"
using std::complex;

template <typename TYPE2>
class ExpRes2D_local_stiffness_matrix;

template <typename TYPE1, typename TYPE2>
class Expr2D_Local_stiffness_matrix2D;


/////////////////////////////
//// 1.
/////////////////////////////
//
//// Gitter:
//
//
///////////////////////////////////////
////
////  Orientierung:   lokales Gebiet fuer Jochens code mit
////                  _COLSAMM_::ELEMENTS::_Quadrangle_<_COLSAMM_::Gauss2>
////
////
////                    3--2
////                   /  /
////                  0--1
////
/////////////////////////////////////
template <typename TYPE1, typename TYPE2>
class Expr2D_Local_stiffness_matrix2D :
			public Expr2D<Expr2D_Local_stiffness_matrix2D<TYPE1, TYPE2> > {

	public:
		Expr2D_Local_stiffness_matrix2D ( const Variable2D<TYPE1>& v, TYPE2**  lm_hex,
		                              TYPE2**  lm_edge_in,
		                              TYPE2**  lm_edge_ex,
		                              TYPE2**  lm_neigh_po ) :
				v_ ( v ), loc_m_rectangles ( lm_hex ),
				loc_m_interior_edges ( lm_edge_in ),
				loc_m_exterior_edges ( lm_edge_ex ),
				loc_m_neighbor_points ( lm_neigh_po ) {}

		Blockgrid2D* Give_blockgrid() const { return v_.Give_blockgrid(); };
		stencil_typ Give_stencil_typ() const { return yes_stencil; }

		typedef _TypeOf0_ ( TYPE1, TYPE2 ) Result;		
		
		template <elementTyp TYP_EL>
		inline Result Give_data ( params2D_in ) const;

		template <elementTyp TYP_EL>
		void Update ( int id ) const {
			v_.template Update<TYP_EL> ( id );
		}

	private:
		const Variable2D<TYPE1>& v_;
		TYPE2**  loc_m_rectangles;                  // num_rectangles
		TYPE2**  loc_m_interior_edges;       // num_edges
		TYPE2**  loc_m_exterior_edges;       // num_edges
		TYPE2**  loc_m_neighbor_points;            // num_points
};


template <typename TYPE2>
class Expr2D_Diag_Local_stiffness_matrix2D :
			public Expr2D<Expr2D_Diag_Local_stiffness_matrix2D<TYPE2> > {
			
	public:
		Expr2D_Diag_Local_stiffness_matrix2D(TYPE2** lm_hex, TYPE2** lm_edges_in, TYPE2** lm_edges_ex,
		                                     TYPE2** lm_neigh_po, Blockgrid2D* blockgrid_ ) :
				loc_m_rectangles ( lm_hex ),
				loc_m_interior_edges ( lm_edges_in ),
				loc_m_exterior_edges ( lm_edges_ex ),
				loc_m_neighbor_points ( lm_neigh_po ),
				blockgrid ( blockgrid_ ) {}
				
		Blockgrid2D* Give_blockgrid() const {
			return blockgrid;
		};
		
		stencil_typ Give_stencil_typ() const { return yes_stencil; }
		
//		typedef typename TYPE2::Vector Result;  // Rainer: warum hier vector????
		typedef _TypeOf0_ ( TYPE2, TYPE2 ) Result;
		
		template <elementTyp TYP_EL>
		inline TYPE2 Give_data ( params2D_in ) const;
		
		template <elementTyp TYP_EL>
		void Update ( int id ) const {}
		
	private:
		TYPE2**  loc_m_rectangles;                 // num_rectangles*8
		TYPE2**  loc_m_interior_edges;             // num_edges
		TYPE2**  loc_m_exterior_edges;             // num_edges
		TYPE2**  loc_m_neighbor_points;            // num_points
		
		Blockgrid2D* blockgrid;
};





template <>
template <>
inline double Expr2D_Diag_Local_stiffness_matrix2D<double>::Give_data<rectangleEl> ( params2D_in ) const {
          return loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i  ,j  ) *16   ]
               + loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i-1,j  ) *16+ 5]
               + loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i  ,j-1) *16+10]
               + loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i-1,j-1) *16+15];
}


template <>
template <>
inline std::complex<double> Expr2D_Diag_Local_stiffness_matrix2D<std::complex<double> >::Give_data<rectangleEl> ( params2D_in ) const {
          return loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i  ,j  ) *16   ]
               + loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i-1,j  ) *16+ 5]
               + loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i  ,j-1) *16+10]
               + loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i-1,j-1) *16+15];
}


template <>
template <>
inline double Expr2D_Diag_Local_stiffness_matrix2D<double>::Give_data<edgeEl> ( params2D_in ) const {
        if(Give_blockgrid()->Give_unstructured_grid()->Give_edge(id)->isAtBoundary()==false)
                return   loc_m_interior_edges[id][( i   ) *8   ]
                       + loc_m_interior_edges[id][( i-1 ) *8+ 5]
                       + loc_m_exterior_edges[id][( i   ) *8   ]
                       + loc_m_exterior_edges[id][( i-1 ) *8+ 5];

        return   loc_m_interior_edges[id][( i   ) *8   ]
               + loc_m_interior_edges[id][( i-1 ) *8+ 5];
}

template <>
template <>
inline std::complex<double> Expr2D_Diag_Local_stiffness_matrix2D<std::complex<double> >::Give_data<edgeEl> ( params2D_in ) const {
        if(Give_blockgrid()->Give_unstructured_grid()->Give_edge(id)->isAtBoundary()==false)
                return   loc_m_interior_edges[id][( i   ) *8   ]
                       + loc_m_interior_edges[id][( i-1 ) *8+ 5]
                       + loc_m_exterior_edges[id][( i   ) *8   ]
                       + loc_m_exterior_edges[id][( i-1 ) *8+ 5];

        return   loc_m_interior_edges[id][( i   ) *8   ]
               + loc_m_interior_edges[id][( i-1 ) *8+ 5];
}


template <>
template <>
inline double Expr2D_Diag_Local_stiffness_matrix2D<double>::Give_data<pointEl> ( params2D_in ) const {
        int number_neighbor_hex =
            blockgrid->Give_unstructured_grid()->Give_point(id)->give_number_neighbors_rec();
        double sum = 0.0;

        for ( int num = 0;num < number_neighbor_hex;++num )
                sum += loc_m_neighbor_points[id][num*4];

        return sum;
}


template <>
template <>
inline std::complex<double> Expr2D_Diag_Local_stiffness_matrix2D<std::complex<double> >::Give_data<pointEl> ( params2D_in ) const {
        int number_neighbor_hex =
            blockgrid->Give_unstructured_grid()->Give_point(id)->give_number_neighbors_rec();
        std::complex<double> sum = 0.0;

        for ( int num = 0;num < number_neighbor_hex;++num )
                sum += loc_m_neighbor_points[id][num*4];

        return sum;
}




/**
Class for local stiffness matrix
**/

template <typename TYPE2>
class Local_stiffness_matrix2D : public Expr2D<Local_stiffness_matrix2D<TYPE2> > {
	public:	
		Local_stiffness_matrix2D ( Blockgrid2D& bg_ );
		~Local_stiffness_matrix2D();
		
		typedef TYPE2 Result;
		
		template <class B>
		void Calculate ( const B& bilinear_form );
		
                template <class B>
                void CalculateSystem ( int i, int k, const B& bilinear_form );

		void Calculate_HelmBoundary();		
		void Multiply_Boundary_Part(double factor); //!< Multipliziert Stencils die den Randterm liefern mit Faktor factor			
		
		void Print();
		void Update();
		
		Expr2D_Diag_Local_stiffness_matrix2D<TYPE2> diag () const;
		
		template <typename TYPE1>
		Expr2D_Local_stiffness_matrix2D<TYPE1, TYPE2> operator() ( const Variable2D<TYPE1>& v ) const;
		
		template <class A>
		void operator= ( const Expr2D<A>& a );
		
		void operator= ( const ExpRes2D_local_stiffness_matrix<TYPE2>& Expr2D_res );
		
		void operator= ( const Local_stiffness_matrix2D& a );

                void Print_MAT ( std::ostream& Datei, std::string title = "myData") const;
		
		void multiplyTestfunction_v_with(const TYPE2& coeff );
                void multiplyTestfunction_v_with(const Variable2D<TYPE2>& coeff );
                void multiplyFunction_u_with(const TYPE2& coeff );
		void multiplyFunction_u_with(const Variable2D<TYPE2>& coeff );
		
		inline TYPE2 Give_matrix_rectangle(params2D_in_loc_matrix) const;
		
		Blockgrid2D* Give_blockgrid() const { return blockgrid; };
			
		TYPE2** Give_loc_m_rectangles() const { return loc_m_rectangles; };
		TYPE2*  Give_loc_m_interior_edges (int id)    const { return loc_m_interior_edges[id]; }
		TYPE2*  Give_loc_m_exterior_edges (int id)    const { return loc_m_exterior_edges[id]; }
		TYPE2*  Give_loc_m_neighbor_points(int id)    const { return loc_m_neighbor_points[id]; }
	private:
		TYPE2**  loc_m_rectangles;                  // num_rectangles
		TYPE2**  loc_m_interior_edges;              // num_edges
		TYPE2**  loc_m_exterior_edges;              // num_edges
		TYPE2**  loc_m_neighbor_points;            // num_points
		
		Blockgrid2D* blockgrid;
		
		static int transform(int d) { ///> transform orientation for COLSAM
		  if(d==0) return 0;
		  if(d==1) return 1;
		  if(d==2) return 3;
		  if(d==3) return 2;
		}		
		// for parallel
		int my_rank;
		MPI_Comm comm;
};

//////////////////
// Implementierung von Memberfunktionen
//////////////////


template <typename TYPE2>
inline TYPE2 Local_stiffness_matrix2D<TYPE2>::Give_matrix_rectangle(params2D_in_loc_matrix ) const {
   return loc_m_rectangles[id_rec][ind_loc_matrix];
}



template <typename TYPE2>
template <typename TYPE1>
Expr2D_Local_stiffness_matrix2D<TYPE1, TYPE2> Local_stiffness_matrix2D<TYPE2>::operator()
                                                                  (const Variable2D<TYPE1>& v) const {
	assert(Give_blockgrid()->hasSameStructure(*v.Give_blockgrid()));

	return Expr2D_Local_stiffness_matrix2D<TYPE1, TYPE2>( v, loc_m_rectangles, loc_m_interior_edges,
	                                                         loc_m_exterior_edges,
	                                                         loc_m_neighbor_points);
}



template <typename TYPE2>
void Local_stiffness_matrix2D<TYPE2>::Multiply_Boundary_Part(double factor) {
        Unstructured2DGrid* ug = blockgrid->Give_unstructured_grid();
	
	for(int id = 0;id < blockgrid->Give_unstructured_grid()->Give_number_rectangle(); ++id ) {
	    int Nx = blockgrid->Give_Nx_rectangle(id);
	    int Ny = blockgrid->Give_Ny_rectangle(id);            
	    
	    RectangleEl* rectangle = blockgrid->Give_unstructured_grid()->Give_rectangle(id);  
	    	    
	    bool Wex = ug->Give_edge(ug->Give_rectangle(id)->Give_id_edge(Wdir2D))->isAtBoundary();
	    bool Eex = ug->Give_edge(ug->Give_rectangle(id)->Give_id_edge(Edir2D))->isAtBoundary();
	    bool Sex = ug->Give_edge(ug->Give_rectangle(id)->Give_id_edge(Sdir2D))->isAtBoundary();
	    bool Nex = ug->Give_edge(ug->Give_rectangle(id)->Give_id_edge(Ndir2D))->isAtBoundary();

	    // Kanten 
	    if(Wex) {		    
	        int i = 0;
		for(int j = 1;j < Ny-1;++j ) {    
		    int ux = 0;
		    int vx = 0;		    
		    
		    for(int vy = 0;vy < 2;++vy )
	                for(int uy = 0;uy < 2;++uy) {
			    loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i,j)*16+4*(2*vy+vx)+(2*uy+ux)]
								*= factor;
			}
		}
	    }
	    if(Eex) {
	        int i = Nx-1;
		for(int j = 1;j < Ny-1;++j ) {		    
		    int ux = 1;
		    int vx = 1;		    

		    for(int vy = 0;vy < 2;++vy )
	                for(int uy = 0;uy < 2;++uy) {
			    loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i,j)*16+4*(2*vy+vx)+(2*uy+ux)]
								*= factor;
			}
		}
	    }
	    if(Sex) {
	        int j = 0;
		for(int i = 1;i < Nx-1;++i ) {		    
		    int uy = 0;
		    int vy = 0;		    

		    for(int vx = 0;vx < 2;++vx )
	                for(int ux = 0;ux < 2;++ux) {
			    loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i,j)*16+4*(2*vy+vx)+(2*uy+ux)]
								*= factor;
			}
		}
	    }	    
	    if(Nex) {
	        int j = Ny-1;
		for(int i = 1;i < Nx-1;++i ) {
		    int uy = 1;
		    int vy = 1;		    

		    for(int vx = 0;vx < 2;++vx )
	                for(int ux = 0;ux < 2;++ux) {
			    loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i,j)*16+4*(2*vy+vx)+(2*uy+ux)]
								+= factor;
			}
		}
	    }
	    // Ecken
	    if(Wex && Sex) {		    
	        int i = 0;
		int j = 0;
		{
		    int ux = 0;
		    int vx = 0;		    
		    
		    for(int vy = 0;vy < 2;++vy )
	                for(int uy = 0;uy < 2;++uy) {
			    loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i,j)*16+4*(2*vy+vx)+(2*uy+ux)]
								*= factor;
			}
		}
	    }    
	    if(Wex && Nex) {		    
	        int i = 0;
		int j = Ny-1;
		{
		    int ux = 0;
		    int vx = 0;		    
		    
		    for(int vy = 0;vy < 2;++vy )
	                for(int uy = 0;uy < 2;++uy) {
			    loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i,j)*16+4*(2*vy+vx)+(2*uy+ux)]
								*= factor;
			}
		}
	    }
	    if(Eex && Sex) {	
	        int i = Nx-1;	      
		int j = 0;
		{
		    int ux = 0;
		    int vx = 0;		    
		    
		    for(int vy = 0;vy < 2;++vy )
	                for(int uy = 0;uy < 2;++uy) {
			    loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i,j)*16+4*(2*vy+vx)+(2*uy+ux)]
								*= factor;
			}
		}
	    }    
	    if(Eex && Nex) {	
	        int i = Nx-1;	      
		int j = Ny-1;
		{
		    int ux = 0;
		    int vx = 0;		    
		    
		    for(int vy = 0;vy < 2;++vy )
	                for(int uy = 0;uy < 2;++uy) {
			    loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i,j)*16+4*(2*vy+vx)+(2*uy+ux)]
								*= factor;
			}
		}
	    }	    
	}
	
	Update();
}


template <typename TYPE2>
void Local_stiffness_matrix2D<TYPE2>::Calculate_HelmBoundary() {
        Unstructured2DGrid* ug = blockgrid->Give_unstructured_grid();
	
	double loc_stiff[2][2];
	loc_stiff[0][0] = 1.0/3.0;
	loc_stiff[1][0] = 1.0/6.0;
	loc_stiff[0][1] = 1.0/6.0;	
	loc_stiff[1][1] = 1.0/3.0;
	
	for(int id = 0;id < blockgrid->Give_unstructured_grid()->Give_number_rectangle(); ++id ) {
	    int Nx = blockgrid->Give_Nx_rectangle(id);
	    int Ny = blockgrid->Give_Ny_rectangle(id);            
	    
	    for(int j=0;j<Ny;++j) for(int i=0;i<Nx;++i)
		for(int uv = 0;uv < 16;++uv)
		    loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i,j)*16+uv] = 0.0;
	    
	    RectangleEl* rectangle = blockgrid->Give_unstructured_grid()->Give_rectangle(id);  
	    	    
	    bool Wex = ug->Give_edge(ug->Give_rectangle(id)->Give_id_edge(Wdir2D))->isAtBoundary();
	    bool Eex = ug->Give_edge(ug->Give_rectangle(id)->Give_id_edge(Edir2D))->isAtBoundary();
	    bool Sex = ug->Give_edge(ug->Give_rectangle(id)->Give_id_edge(Sdir2D))->isAtBoundary();
	    bool Nex = ug->Give_edge(ug->Give_rectangle(id)->Give_id_edge(Ndir2D))->isAtBoundary();

	      
	    if(Wex) {		    
	        int i = 0;
		for(int j = 0;j < Ny;++j ) {
                    D2vector pS = blockgrid->Give_coord_rectangle( id, i, j);
                    D2vector pN = blockgrid->Give_coord_rectangle( id, i, j+1);
		
		    double h = ABS(pS-pN);
		    		    
		    int ux = 0;
		    int vx = 0;		    
		    
		    for(int vy = 0;vy < 2;++vy )
	                for(int uy = 0;uy < 2;++uy) {
			    loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i,j)*16+4*(2*vy+vx)+(2*uy+ux)]
								+= h * loc_stiff[vy][vx];
			}
		}
	    }
	    if(Eex) {
	        int i = Nx-1;
		for(int j = 0;j < Ny;++j ) {
                    D2vector pS = blockgrid->Give_coord_rectangle( id, i+1, j);
                    D2vector pN = blockgrid->Give_coord_rectangle( id, i+1, j+1);
		
		    double h = ABS(pS-pN);
		    
		    int ux = 1;
		    int vx = 1;		    

		    for(int vy = 0;vy < 2;++vy )
	                for(int uy = 0;uy < 2;++uy) {
			    loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i,j)*16+4*(2*vy+vx)+(2*uy+ux)]
								+= h * loc_stiff[vy][vx];
			}
		}
	    }
	    if(Sex) {
	        int j = 0;
		for(int i = 0;i < Nx;++i ) {
                    D2vector pW = blockgrid->Give_coord_rectangle( id, i,   j);
                    D2vector pE = blockgrid->Give_coord_rectangle( id, i+1, j);
		
		    double h = ABS(pE-pW);
		    
		    int uy = 0;
		    int vy = 0;		    

		    for(int vx = 0;vx < 2;++vx )
	                for(int ux = 0;ux < 2;++ux) {
			    loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i,j)*16+4*(2*vy+vx)+(2*uy+ux)]
								+= h * loc_stiff[vy][vx];
			}
		}
	    }	    
	    if(Nex) {
	        int j = Ny-1;
		for(int i = 0;i < Nx;++i ) {
                    D2vector pW = blockgrid->Give_coord_rectangle( id, i,   j+1);
                    D2vector pE = blockgrid->Give_coord_rectangle( id, i+1, j+1);
		
		    double h = ABS(pE-pW);
		    
		    int uy = 1;
		    int vy = 1;		    

		    for(int vx = 0;vx < 2;++vx )
	                for(int ux = 0;ux < 2;++ux) {
			    loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i,j)*16+4*(2*vy+vx)+(2*uy+ux)]
								+= h * loc_stiff[vy][vx];
			}
		}
	    }	
	}
	
	Update();
}
  
  

template <typename TYPE2>
template <class B>
void Local_stiffness_matrix2D<TYPE2>::Calculate ( const B& bilinear_form ) {
	Unstructured2DGrid* ug = blockgrid->Give_unstructured_grid();

//test
//bool set = false;	
	
	std::vector<std::vector<TYPE2> > loc_stiff;
	std::vector<double> points ( 8, 0. ); //Länge 4*2=8

//	_COLSAMM_::ELEMENTS::_Quadrangle_2<_COLSAMM_::Gauss2, TYPE2> G;
	_COLSAMM_::ELEMENTS::_Quadrangle_<_COLSAMM_::Gauss2, TYPE2> G;
		
	//std::vector< std::vector<std::complex<double> > > loc_stiff;
	
	for(int id = 0;id < blockgrid->Give_unstructured_grid()->Give_number_rectangle(); ++id ) {
		int Nx = blockgrid->Give_Nx_rectangle ( id );
		int Ny = blockgrid->Give_Ny_rectangle ( id );

		int N_total = Nx * Ny * 8;
		for(int j = 0;j < Ny;++j )
	            for(int i = 0;i < Nx;++i ) {
				assert(Ind_loc_matrix_rectangle(i,j) *8 < N_total);
				
				points[0] = blockgrid->Give_coord_rectangle ( id, i, j).x;		
				points[1] = blockgrid->Give_coord_rectangle ( id, i, j).y;						
				points[2] = blockgrid->Give_coord_rectangle ( id, i + 1, j).x;		
				points[3] = blockgrid->Give_coord_rectangle ( id, i + 1, j).y;
				points[4] = blockgrid->Give_coord_rectangle ( id, i + 1, j + 1).x;
				points[5] = blockgrid->Give_coord_rectangle ( id, i + 1, j + 1).y;
				points[6] = blockgrid->Give_coord_rectangle ( id, i    , j + 1).x;
				points[7] = blockgrid->Give_coord_rectangle ( id, i    , j + 1).y;
				
				G ( points );
				loc_stiff = G.integrate ( bilinear_form );

				for(int v = 0;v < 4;++v )
			            for(int u = 0;u < 4;++u )
			            {
			              loc_m_rectangles[id][Ind_loc_matrix_rectangle ( i,j)*16+4*v+u]
								= loc_stiff[transform(u)][transform(v)];
			            }
			    }
	}
	
	Update();
}



template <typename TYPE2>
void Local_stiffness_matrix2D<TYPE2>::Update() {
   Unstructured2DGrid* ug = blockgrid->Give_unstructured_grid();

   // Iteriere über alle Rechtecke und kopiere lokale Steifigkeitsmatrizen zu Kanten und zu Punkten
   // ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   for(int id_rec = 0;id_rec < ug->Give_number_rectangle();++id_rec) {
       int Nxrec = blockgrid->Give_Nx_rectangle(id_rec);
       int Nyrec = blockgrid->Give_Ny_rectangle(id_rec);
		
		
	//Update local stiffness matrix for edges
        //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	for(int fc = 0;fc < 4;++fc) {  // Iteration über 4 Kanten
	    int id_edge  = ug->Give_rectangle(id_rec)->Give_id_edge((dir2D) fc);
			
	    int Nxedge = blockgrid->Give_Nx_edge(id_edge);
	    int N_total = Nxedge * 8;
				
	    // a) coordinate points of edge
	    int id_W = ug->Give_edge(id_edge)->Give_id_corner(Wd);
	    int id_E = ug->Give_edge(id_edge)->Give_id_corner(Ed);
			
	    dir2D_sons c_W;
	    dir2D_sons c_E;
	    
	    // b) Find corresponding points in rectangle
	    for(int c = 0;c < 4;++c) {
		int idsp = ug->Give_rectangle(id_rec)->Give_id_corner((dir2D_sons)c);
					
		if(idsp == id_W ) c_W = (dir2D_sons) c;
		if(idsp == id_E ) c_E = (dir2D_sons) c;
	    }
				
            // c) Finde Orientierung und map
            bool clockOrientation = true;
	    
	    if(c_W == SWdir2D && c_E == NWdir2D)  clockOrientation = false;
	    if(c_W == NWdir2D && c_E == NEdir2D)  clockOrientation = false;
	    if(c_W == NEdir2D && c_E == SEdir2D)  clockOrientation = false;
	    if(c_W == SEdir2D && c_E == SWdir2D)  clockOrientation = false;

	    /*  das hilft nicht!!!::::
	    if(ug->Give_edge(id_edge)->getIdNextRectangle(false) != id_rec) { // exterior
	       if(clockOrientation) clockOrientation=false;
	       else                 clockOrientation=true;
	    }
*/	    
	    dir2D_sons map_sons[4];
//test	    
	    calculate_map(c_W,clockOrientation,map_sons);				
		 				
//if(clockOrientation) calculate_map(c_W,true,map_sons);	 	    
//else                 calculate_map(c_E,true,map_sons);	 	    
	    
	    // d) Indizes Konstanten			
            int Nconst = Blockgrid2D::Give_Nconst_cell(c_W, Nxrec, Nyrec);
	    int Ni; // beachte: (j*Nx+i)
	    dir2D direction = direction_from_to(c_W,c_E);

	    if(direction==Wdir2D) Ni=-1;
	    if(direction==Edir2D) Ni= 1;
	    if(direction==Ndir2D) Ni= Nxrec;
	    if(direction==Sdir2D) Ni=-Nxrec;
						       
	    // d) Determine if rec is interior or exterior volume		
	    TYPE2* loc_pointer;				       
	    if(ug->Give_edge(id_edge)->getIdNextRectangle(false) == id_rec) { //interior
	       loc_pointer = loc_m_interior_edges[id_edge];
	    } 
	    else { // exterior
	       loc_pointer = loc_m_exterior_edges[id_edge];
	    }
				
	    // e) kopiere		
	    for(int i = 0;i < Nxedge;++i) {
	        for(int v = 0;v < 2;++v)
		    for(int u = 0;u < 4;++u)
		        loc_pointer[i*8+4*v+u]
			    = loc_m_rectangles[id_rec][(i*Ni+Nconst)*16 + 4*map_sons[v] + map_sons[u]];
	    }
	}
      //Update local stiffness matrix for points
      //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      dir2D_sons map_sons[4];
      for(int c = 0;c < 4;++c ) {
	  // a) Punkt finden
	  int id_point = ug->Give_rectangle(id_rec)->Give_id_corner((dir2D_sons)c);
	  PointEl* point = ug->Give_point(id_point);
	  
	  // b)lokaler Nachbar rectangle Nummer in diesem Punkt, simpler Suchalgorithmus
	  int loc_num_rec = point->getLocalIdRectangle(id_rec);
          
          // c) Berechne Richtungen des lokalen Koordinatensystems
	  /* old
	  dir2D_sons ES_corner, WN_corner;
	  Calculate_corresponding_corners((dir2D_sons)c, ES_corner, WN_corner);  // Kann man besser machen!!! gleich c_x, c_y berechnen
	  dir2D c_x = direction_from_to ((dir2D_sons) c, ES_corner );
	  dir2D c_y = direction_from_to ((dir2D_sons) c, WN_corner );
	  calculate_map((dir2D_sons)c,map_sons);
	
	  // d) Berechne Start-Wert des Index fuer rectangle
	    // x- direction
	    int Nconst;
	    Nconst = Blockgrid2D::Give_Nconst_cell(c_x, Nxrec, Nyrec);				
	    // y- direction
	    Nconst = Blockgrid2D::Give_Nconst_cell(c_y, Nxrec, Nyrec) + Nconst;
		*/
	  
          int Nconst = Blockgrid2D::Give_Nconst_cell((dir2D_sons)c, Nxrec, Nyrec);
	  calculate_map((dir2D_sons)c,true,map_sons);
	  
	  // e) Kopiere 
          for(int u=0;u<4;++u)
	       loc_m_neighbor_points[id_point][loc_num_rec*4+u]
	           = loc_m_rectangles[id_rec][Nconst*16+4*map_sons[0] + map_sons[u]];
									
      }
   }
}


template <typename TYPE2>
template <class A>
void Local_stiffness_matrix2D<TYPE2>::operator= ( const Expr2D<A>& a ) {
   Unstructured2DGrid* ug = blockgrid->Give_unstructured_grid();
	
   const A& ao ( a );
	
   for(int id = 0;id < blockgrid->Give_unstructured_grid()->Give_number_rectangle();++id ) {
    //   if(ug->Give_rectangle(id)->my_object(my_rank)) {
	  int Nx = blockgrid->Give_Nx_rectangle(id);
	  int Ny = blockgrid->Give_Ny_rectangle(id);
	  int N_total = Nx * Ny * 16;
			
	  for(int j=0;j<Ny;++j)
	      for(int i=0;i<Nx;++i) {
                  assert(Ind_loc_matrix_rectangle(i,j) *16 < N_total);
		  for(int uv = 0;uv < 16;++uv)
		      loc_m_rectangles[id][Ind_loc_matrix_rectangle(i,j) * 16+uv]
			   = ao.Give_matrix_rectangle(id,Ind_loc_matrix_rectangle(i,j) * 16 + uv,
							 Ind_loc_matrix_rectangle(i,j),
							 i, j, Nx);
	      }
   //    }
   }
   Update();
}


template <typename TYPE2>
void Local_stiffness_matrix2D<TYPE2>::operator= ( const Local_stiffness_matrix2D & a ) {
	if ( this == &a )
		return;
		
	const Expr2D<Local_stiffness_matrix2D>& b ( a );
	
	operator= ( b );
}

template <typename TYPE2>
void Local_stiffness_matrix2D<TYPE2>::multiplyTestfunction_v_with ( const Variable2D<TYPE2>& coeffV ) {
  Unstructured2DGrid* ug = blockgrid->Give_unstructured_grid();
	
  for(int id = 0;id < blockgrid->Give_unstructured_grid()->Give_number_rectangle();++id ) {
    //  if(ug->Give_rectangle(id)->my_object(my_rank)) {
	 coeffV.template Update<rectangleEl> ( id );		
	 int Nx = blockgrid->Give_Nx_rectangle ( id );
	 int Ny = blockgrid->Give_Ny_rectangle ( id );
	 int N_total = Nx * Ny * 16;
			
	 for(int j = 0;j < Ny;++j)
	     for(int i = 0;i < Nx;++i) {
		 assert(Ind_loc_matrix_rectangle(i,j) * 16 < N_total);
								
		 for(int v = 0;v<4;++v) { // test function v
		     for(int u = 0;u <4;++u) {
			 loc_m_rectangles[id][Ind_loc_matrix_rectangle(i,j) * 16 + 4*v+u] =
			     loc_m_rectangles[id][Ind_loc_matrix_rectangle(i,j) * 16 + 4*v+u] *
				(coeffV.template Give_data<rectangleEl>(id,
								        x1DCoord((dir2D_sons)v) + i,
								        y1DCoord((dir2D_sons)v) + j,Nx));
		     }
		 }
	     }
      //}
  }
  Update();
}

template <typename TYPE2>
void Local_stiffness_matrix2D<TYPE2>::multiplyTestfunction_v_with ( const TYPE2& coeffV ) {
  Unstructured2DGrid* ug = blockgrid->Give_unstructured_grid();

  for(int id = 0;id < blockgrid->Give_unstructured_grid()->Give_number_rectangle();++id ) {
    //  if(ug->Give_rectangle(id)->my_object(my_rank)) {
         int Nx = blockgrid->Give_Nx_rectangle ( id );
         int Ny = blockgrid->Give_Ny_rectangle ( id );
         int N_total = Nx * Ny * 16;

         for(int j = 0;j < Ny;++j)
             for(int i = 0;i < Nx;++i) {
                 assert(Ind_loc_matrix_rectangle(i,j) * 16 < N_total);

                 for(int v = 0;v<4;++v) { // test function v
                     for(int u = 0;u <4;++u) {
                         loc_m_rectangles[id][Ind_loc_matrix_rectangle(i,j) * 16 + 4*v+u] =
                             loc_m_rectangles[id][Ind_loc_matrix_rectangle(i,j) * 16 + 4*v+u] * coeffV;
                     }
                 }
             }
      //}
  }
  Update();
}

template <typename TYPE2>
void Local_stiffness_matrix2D<TYPE2>::multiplyFunction_u_with ( const Variable2D<TYPE2>& coeffV ) {
  Unstructured2DGrid* ug = blockgrid->Give_unstructured_grid();
	
  for(int id = 0;id < blockgrid->Give_unstructured_grid()->Give_number_rectangle();++id ) {
    //  if(ug->Give_rectangle(id)->my_object(my_rank)) {
	 coeffV.template Update<rectangleEl> ( id );
	 int Nx = blockgrid->Give_Nx_rectangle ( id );
	 int Ny = blockgrid->Give_Ny_rectangle ( id );
	 int N_total = Nx * Ny * 16;

	 for(int j = 0;j < Ny;++j) {
	     for(int i = 0;i < Nx;++i) {
		 assert(Ind_loc_matrix_rectangle(i,j) * 16 < N_total);
								
		 for(int v = 0;v < 4;++v) { // test function v
		     for(int u = 0;u < 4;++u) {
		       //std::cout << loc_m_rectangles[id][Ind_loc_matrix_rectangle(i,j) * 16 + 4*v+u] << "  ";
			 loc_m_rectangles[id][Ind_loc_matrix_rectangle(i,j) * 16 + 4*v+u] =
			     loc_m_rectangles[id][Ind_loc_matrix_rectangle(i,j) * 16 + 4*v+u] *
				(coeffV.template Give_data<rectangleEl>(id,
								        x1DCoord((dir2D_sons)u) + i,
								        y1DCoord((dir2D_sons)u) + j,Nx));
	                      // std::cout << loc_m_rectangles[id][Ind_loc_matrix_rectangle(i,j) * 16 + 4*v+u] << std::endl;
		     }
		 }
	     }
	 }
  }
  Update();
}

template <typename TYPE2>
void Local_stiffness_matrix2D<TYPE2>::multiplyFunction_u_with ( const TYPE2& coeffV ) {
  Unstructured2DGrid* ug = blockgrid->Give_unstructured_grid();

  for(int id = 0;id < blockgrid->Give_unstructured_grid()->Give_number_rectangle();++id ) {
    //  if(ug->Give_rectangle(id)->my_object(my_rank)) {
         int Nx = blockgrid->Give_Nx_rectangle ( id );
         int Ny = blockgrid->Give_Ny_rectangle ( id );
         int N_total = Nx * Ny * 16;

         for(int j = 0;j < Ny;++j) {
             for(int i = 0;i < Nx;++i) {
                 assert(Ind_loc_matrix_rectangle(i,j) * 16 < N_total);

                 for(int v = 0;v < 4;++v) { // test function v
                     for(int u = 0;u < 4;++u) {
                       //std::cout << loc_m_rectangles[id][Ind_loc_matrix_rectangle(i,j) * 16 + 4*v+u] << "  ";
                         loc_m_rectangles[id][Ind_loc_matrix_rectangle(i,j) * 16 + 4*v+u] =
                             loc_m_rectangles[id][Ind_loc_matrix_rectangle(i,j) * 16 + 4*v+u] * coeffV;
                              // std::cout << loc_m_rectangles[id][Ind_loc_matrix_rectangle(i,j) * 16 + 4*v+u] << std::endl;
                     }
                 }
             }
         }
  }
  Update();
}


#endif // DIFFOP_H
