/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#ifndef DIFFOP2D_CC_H
#define DIFFOP2D_CC_H
using namespace ::_COLSAMM_;



template <>
template <>
inline double Expr2D_Local_stiffness_matrix2D<double, double>::Give_data<rectangleEl>(params2D_in) const {
 int u, v;
 dir2D_sons ce;
 double sum;
 sum = 0.0;

 for(v = 0;v < 4;++v ) { // test function v and cell opposite3D(v)
     ce = opposite2D((dir2D_sons)v);
     for(u = 0;u < 4;++u ) {
	  sum = sum +
		loc_m_rectangles[id][Ind_loc_matrix_rectangle(x1DCoord ( ce )-1+i,
			                                      y1DCoord ( ce )-1+j)
		                                                  * 16+4*v+u]
		   *  v_.Give_data<rectangleEl>(id,x1DCoord(ce)+x1DCoord((dir2D_sons)u)-1+i,
			                          y1DCoord(ce)+y1DCoord((dir2D_sons)u)-1+j,
			                          Nx);
     }
 }

 return sum;
}


template <>
template <>
inline double Expr2D_Local_stiffness_matrix2D<double, double>::Give_data<edgeEl> ( params2D_in ) const {
	dir2D_sons ce;
	double sum = 0.0;


	// Zelle EN
	//----------
	//          Zelle
	//    N     |- - - |
	//          |      |
	//          |      |
	//    S     v- - - u         v=WS ^ Wd    u=ES      v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_interior_edges[id][(i-1+Ed)*8 +  Wd*4 + SEdir2D] *
	                                     v_.Give_data_edge_middle(id,i-1 + Eort);

	//----------
	//          Zelle
	//    N     |- - - |
	//          |      |
	//          |      |
	//    S     vu - - |         v=u=WS ^ Wd            v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_interior_edges[id][(i-1+Ed)*8 +  Wd*4 + SWdir2D] *
	                                     v_.Give_data_edge_middle(id,i-1 + Mort);

	//----------
	//          Zelle
	//    N     |- - - u                    u=EN
	//          |      |
	//          |      |
	//    S     v- - - |         v=WS ^ Wd              v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_interior_edges[id][(i-1+Ed)*8 +  Wd*4 + NEdir2D] *
	                                     v_.Give_data_edge_interior(id,i-1 + Eort);

	//----------
	//          Zelle
	//    N     u- - - |         v=WN
	//          |      |
	//          |      |
	//    S     v- - - |         v=WS ^ Wd              v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_interior_edges[id][(i-1+Ed)*8 +  Wd*4 + NWdir2D] *
	                                     v_.Give_data_edge_interior(id,i-1 + Mort);




	// Zelle WN -----------------------------------------------------------------------------
	//----------
	//          Zelle
	//    N     |- - - |
	//          |      |
	//          |      |
	//    S     |- - -vu               v=u=ES ^ Ed      v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_interior_edges[id][(i-1+Wd)*8 +  Ed*4 + SEdir2D] *
	                                     v_.Give_data_edge_middle(id,i-1 + Mort);

	//----------
	//          Zelle
	//    N     |- - - |
	//          |      |
	//          |      |
	//    S     u- - - v       u=WS      v=ES^ Ed       v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_interior_edges[id][(i-1+Wd)*8 +  Ed*4 + SWdir2D] *
	                                     v_.Give_data_edge_middle(id,i-1 + Wort);


	//----------
	//          Zelle
	//    N     |- - - u               u=EN
	//          |      |
	//          |      |
	//    S     |- - - v               v=ES ^ Ed        v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_interior_edges[id][(i-1+Wd)*8 +  Ed*4 + NEdir2D] *
	                                     v_.Give_data_edge_interior(id,i-1 + Mort);

	//----------
	//          Zelle
	//    N     u- - - |       u=WN
	//          |      |
	//          |      |
	//    S      - - - v                 v=ES^ Ed       v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_interior_edges[id][(i-1+Wd)*8 +  Ed*4 + NWdir2D] *
	                                     v_.Give_data_edge_interior(id,i-1 + Wort);


	//---------------------------------------------------------------------------------

	if(Give_blockgrid()->Give_unstructured_grid()->Give_edge(id)->isAtBoundary())
		return sum;



	// Zelle ES
	//----------
	//          Zelle
	//    S     v- - - |
	//          |      |
	//          |      |
	//    N     |- - - u         v=WN ^ Wd    u=EN      v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_exterior_edges[id][(i-1+Ed)*8 +  Wd*4 + NEdir2D] *
	                                     v_.Give_data_edge_exterior(id,i-1 + Eort);

	//----------
	//          Zelle
	//    S     v- - - |         v=WS ^ Wd
	//          |      |
	//          |      |
	//    N     u- - - |         u=WN                   v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_exterior_edges[id][(i-1+Ed)*8 +  Wd*4 + NWdir2D] *
	                                     v_.Give_data_edge_exterior(id,i-1 + Mort);



	//----------
	//          Zelle
	//    S     v- - - u         v=WS ^ Wd     u=ES
	//          |      |
	//          |      |
	//    N     |- - - |                                v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_exterior_edges[id][(i-1+Ed)*8 +  Wd*4 + SEdir2D] *
	                                     v_.Give_data_edge_middle(id,i-1 + Eort);

	//----------
	//          Zelle
	//    S     vu - - |         v=u=WS ^ Wd
	//          |      |
	//          |      |
	//    N     |- - - |                                v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_exterior_edges[id][(i-1+Ed)*8 +  Wd*4 + SWdir2D] *
	                                     v_.Give_data_edge_middle(id,i-1 + Mort);




	// Zelle WS -----------------------------------------------------------------------------
	//----------
	//          Zelle
	//    S     |- - - v               v=ES ^ Ed
	//          |      |
	//          |      |
	//    N     |- - - u               u=EN             v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_exterior_edges[id][(i-1+Wd)*8 +  Ed*4 + NEdir2D] *
	                                     v_.Give_data_edge_exterior(id,i-1 + Mort);

	//----------
	//          Zelle
	//    S     |- - - v                 v=ES ^ Ed
	//          |      |
	//          |      |
	//    N     u- - - |       u=WN                     v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_exterior_edges[id][(i-1+Wd)*8 +  Ed*4 + NWdir2D] *
	                                     v_.Give_data_edge_exterior(id,i-1 + Wort);



	//----------
	//          Zelle
	//    S     |- - -uv               v=u=ES ^ Ed
	//          |      |
	//          |      |
	//    N     |- - - |                                v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_exterior_edges[id][(i-1+Wd)*8 +  Ed*4 + SEdir2D] *
	                                     v_.Give_data_edge_middle(id,i-1 + Mort);

	//----------
	//          Zelle
	//    S     u- - - v       u=WS     v=ES ^ Ed
	//          |      |
	//          |      |
	//    N      - - - |                                v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_exterior_edges[id][(i-1+Wd)*8 +  Ed*4 + SWdir2D] *
	                                     v_.Give_data_edge_middle(id,i-1 + Wort);

	return sum;

}


template <>
template <>
inline double Expr2D_Local_stiffness_matrix2D<double, double>::Give_data<pointEl> ( params2D_in ) const {
	int num;
	int pos_SE, pos_NW, pos_NE;
	dir2D_sons ce;
	double sum;
	sum = 0.0;


// test Diff
//sum = v_.Give_data_point(id);
//sum = 0.0;
//return sum;




	int number_neighbor_rec =
	    Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->give_number_neighbors_rec();

	for ( num = 0;num < number_neighbor_rec;++num ) {
	        PointEl* poi = Give_blockgrid()->Give_unstructured_grid()->Give_point(id);
		pos_SE = poi->give_local_pointId_rec(num,SErec);
		pos_NW = poi->give_local_pointId_rec(num,NWrec);
		pos_NE = poi->give_local_pointId_rec(num,NErec);
		ce = NEdir2D; //cell
		sum = sum +
		      loc_m_neighbor_points[id][num*4+SWdir2D] *  // SW
		      v_.Give_data_point(id ) +

		      loc_m_neighbor_points[id][num*4+SEdir2D] *  // SE
		      v_.Give_data_neighbor_point(id, pos_SE ) +

		      loc_m_neighbor_points[id][num*4+NWdir2D] *  // NW
		      v_.Give_data_neighbor_point(id, pos_NW ) +

		      loc_m_neighbor_points[id][num*4+NEdir2D] *  // NE
		      v_.Give_data_neighbor_point(id, pos_NE );
	}

	return sum;
}

template <>
template <>
inline complex<double> Expr2D_Local_stiffness_matrix2D< complex<double>,  complex<double> >::Give_data<rectangleEl>(params2D_in) const {
 int u, v;
 dir2D_sons ce;
 complex<double> sum;
 sum = 0.0;


 for(v = 0;v < 4;++v ) { // test function v and cell opposite3D(v)
     ce = opposite2D((dir2D_sons)v);
     for(u = 0;u < 4;++u ) {
	  sum = sum +
		loc_m_rectangles[id][Ind_loc_matrix_rectangle(x1DCoord ( ce )-1+i,
			                                      y1DCoord ( ce )-1+j)
		                                                  * 16+4*v+u]
		   *  v_.Give_data<rectangleEl>(id,x1DCoord(ce)+x1DCoord((dir2D_sons)u)-1+i,
			                          y1DCoord(ce)+y1DCoord((dir2D_sons)u)-1+j,
			                          Nx);
     }
 }

 return sum;
}


template <>
template <>
inline complex<double> Expr2D_Local_stiffness_matrix2D< complex<double>,  complex<double> >::Give_data<edgeEl> ( params2D_in ) const {
	dir2D_sons ce;
	complex<double> sum = 0.0;

	// Zelle EN
	//----------
	//          Zelle
	//    N     |- - - |
	//          |      |
	//          |      |
	//    S     v- - - u         v=WS ^ Wd    u=ES      v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_interior_edges[id][(i-1+Ed)*8 +  Wd*4 + SEdir2D] *
	                                     v_.Give_data_edge_middle(id,i-1 + Eort);

	//----------
	//          Zelle
	//    N     |- - - |
	//          |      |
	//          |      |
	//    S     vu - - |         v=u=WS ^ Wd            v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_interior_edges[id][(i-1+Ed)*8 +  Wd*4 + SWdir2D] *
	                                     v_.Give_data_edge_middle(id,i-1 + Mort);


	//----------
	//          Zelle
	//    N     |- - - u                    u=EN
	//          |      |
	//          |      |
	//    S     v- - - |         v=WS ^ Wd              v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_interior_edges[id][(i-1+Ed)*8 +  Wd*4 + NEdir2D] *
	                                     v_.Give_data_edge_interior(id,i-1 + Eort);

	//----------
	//          Zelle
	//    N     u- - - |         v=WN
	//          |      |
	//          |      |
	//    S     v- - - |         v=WS ^ Wd              v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_interior_edges[id][(i-1+Ed)*8 +  Wd*4 + NWdir2D] *
	                                     v_.Give_data_edge_interior(id,i-1 + Mort);




	// Zelle WN -----------------------------------------------------------------------------
	//----------
	//          Zelle
	//    N     |- - - |
	//          |      |
	//          |      |
	//    S     |- - -vu               v=u=ES ^ Ed      v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_interior_edges[id][(i-1+Wd)*8 +  Ed*4 + SEdir2D] *
	                                     v_.Give_data_edge_middle(id,i-1 + Mort);

	//----------
	//          Zelle
	//    N     |- - - |
	//          |      |
	//          |      |
	//    S     u- - - v       u=WS      v=ES^ Ed       v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_interior_edges[id][(i-1+Wd)*8 +  Ed*4 + SWdir2D] *
	                                     v_.Give_data_edge_middle(id,i-1 + Wort);



	//----------
	//          Zelle
	//    N     |- - - u               u=EN
	//          |      |
	//          |      |
	//    S     |- - - v               v=ES ^ Ed        v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_interior_edges[id][(i-1+Wd)*8 +  Ed*4 + NEdir2D] *
	                                     v_.Give_data_edge_interior(id,i-1 + Mort);

	//----------
	//          Zelle
	//    N     u- - - |       u=WN
	//          |      |
	//          |      |
	//    S      - - - v                 v=ES^ Ed       v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_interior_edges[id][(i-1+Wd)*8 +  Ed*4 + NWdir2D] *
	                                     v_.Give_data_edge_interior(id,i-1 + Wort);


	//---------------------------------------------------------------------------------

	if(Give_blockgrid()->Give_unstructured_grid()->Give_edge(id)->isAtBoundary())
		return sum;



	// Zelle ES
	//----------
	//          Zelle
	//    S     v- - - |
	//          |      |
	//          |      |
	//    N     |- - - u         v=WN ^ Wd    u=EN      v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_exterior_edges[id][(i-1+Ed)*8 +  Wd*4 + NEdir2D] *
	                                     v_.Give_data_edge_exterior(id,i-1 + Eort);

	//----------
	//          Zelle
	//    S     v- - - |         v=WS ^ Wd
	//          |      |
	//          |      |
	//    N     u- - - |         u=WN                   v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_exterior_edges[id][(i-1+Ed)*8 +  Wd*4 + NWdir2D] *
	                                     v_.Give_data_edge_exterior(id,i-1 + Mort);



	//----------
	//          Zelle
	//    S     v- - - u         v=WS ^ Wd     u=ES
	//          |      |
	//          |      |
	//    N     |- - - |                                v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_exterior_edges[id][(i-1+Ed)*8 +  Wd*4 + SEdir2D] *
	                                     v_.Give_data_edge_middle(id,i-1 + Eort);

	//----------
	//          Zelle
	//    S     vu - - |         v=u=WS ^ Wd
	//          |      |
	//          |      |
	//    N     |- - - |                                v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_exterior_edges[id][(i-1+Ed)*8 +  Wd*4 + SWdir2D] *
	                                     v_.Give_data_edge_middle(id,i-1 + Mort);




	// Zelle WS -----------------------------------------------------------------------------
	//----------
	//          Zelle
	//    S     |- - - v               v=ES ^ Ed
	//          |      |
	//          |      |
	//    N     |- - - u               u=EN             v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_exterior_edges[id][(i-1+Wd)*8 +  Ed*4 + NEdir2D] *
	                                     v_.Give_data_edge_exterior(id,i-1 + Mort);

	//----------
	//          Zelle
	//    S     |- - - v                 v=ES ^ Ed
	//          |      |
	//          |      |
	//    N     u- - - |       u=WN                     v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_exterior_edges[id][(i-1+Wd)*8 +  Ed*4 + NWdir2D] *
	                                     v_.Give_data_edge_exterior(id,i-1 + Wort);



	//----------
	//          Zelle
	//    S     |- - -uv               v=u=ES ^ Ed
	//          |      |
	//          |      |
	//    N     |- - - |                                v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_exterior_edges[id][(i-1+Wd)*8 +  Ed*4 + SEdir2D] *
	                                     v_.Give_data_edge_middle(id,i-1 + Mort);

	//----------
	//          Zelle
	//    S     u- - - v       u=WS     v=ES ^ Ed
	//          |      |
	//          |      |
	//    N      - - - |                                v*4 + u
	//  Kante:  W      E
	sum = sum + loc_m_exterior_edges[id][(i-1+Wd)*8 +  Ed*4 + SWdir2D] *
	                                     v_.Give_data_edge_middle(id,i-1 + Wort);

	return sum;
}


template <>
template <>
inline complex<double> Expr2D_Local_stiffness_matrix2D< complex<double>,  complex<double> >::Give_data<pointEl> ( params2D_in ) const {
	int num;
	int pos_SE, pos_NW, pos_NE;
	dir2D_sons ce;
	complex<double> sum;
	sum = 0.0;


// test Diff
//sum = v_.Give_data_point(id);
//sum = 0.0;
//return sum;




	int number_neighbor_rec =
	    Give_blockgrid()->Give_unstructured_grid()->Give_point(id)->give_number_neighbors_rec();

	for ( num = 0;num < number_neighbor_rec;++num ) {
	        PointEl* poi = Give_blockgrid()->Give_unstructured_grid()->Give_point(id);
		pos_SE = poi->give_local_pointId_rec(num,SErec);
		pos_NW = poi->give_local_pointId_rec(num,NWrec);
		pos_NE = poi->give_local_pointId_rec(num,NErec);
		ce = NEdir2D; //cell
		sum = sum +
		      loc_m_neighbor_points[id][num*4+SWdir2D] *  // SW
		      v_.Give_data_point(id ) +

		      loc_m_neighbor_points[id][num*4+SEdir2D] *  // SE
		      v_.Give_data_neighbor_point(id, pos_SE ) +

		      loc_m_neighbor_points[id][num*4+NWdir2D] *  // NW
		      v_.Give_data_neighbor_point(id, pos_NW ) +

		      loc_m_neighbor_points[id][num*4+NEdir2D] *  // NE
		      v_.Give_data_neighbor_point(id, pos_NE );
	}
	return sum;
}

/////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////


template <typename TYPE2>
Expr2D_Diag_Local_stiffness_matrix2D<TYPE2> Local_stiffness_matrix2D<TYPE2>::diag ()
  const {
  return Expr2D_Diag_Local_stiffness_matrix2D<TYPE2>(loc_m_rectangles,
						     loc_m_interior_edges,
                                                     loc_m_exterior_edges,
                                                     loc_m_neighbor_points,
                                                     blockgrid);
}


template <typename TYPE2>
Local_stiffness_matrix2D<TYPE2>::Local_stiffness_matrix2D(Blockgrid2D& bg_) {

  blockgrid = &bg_;
  Unstructured2DGrid* ug = blockgrid->Give_unstructured_grid();

  // for parallel
  my_rank = ug->Give_my_rank();
  comm    = ug->Give_MPI_comm();

  loc_m_rectangles = new TYPE2*[ug->Give_number_rectangle()];
  for(int id=0;id<ug->Give_number_rectangle();++id) {
      int Nx = blockgrid->Give_Nx_rectangle(id);
      int Ny = blockgrid->Give_Ny_rectangle(id);
      int N_total = Nx * Ny * 16;
      loc_m_rectangles[id] = new TYPE2[N_total];
      for(int i=0;i<N_total;++i) loc_m_rectangles[id][i] = (TYPE2)0;
  }

  
  loc_m_interior_edges = new TYPE2*[ug->Give_number_edges()];
  loc_m_exterior_edges = new TYPE2*[ug->Give_number_edges()];
  for(int id=0;id<ug->Give_number_edges();++id) {
    if(ug->Give_edge(id)->my_object(my_rank)) {
      int Nx = blockgrid->Give_Nx_edge(id);
      int N_total = Nx * 8;
      loc_m_interior_edges[id] = new TYPE2[N_total];
      for(int i=0;i<N_total;++i)
           loc_m_interior_edges[id][i] = (TYPE2)0;
      if(ug->Give_edge(id)->isAtBoundary()==false){
         loc_m_exterior_edges[id] = new TYPE2[N_total];
         for(int i=0;i<N_total;++i)
           loc_m_exterior_edges[id][i] = (TYPE2)0;
      }
      else loc_m_exterior_edges[id] = NULL;
    }
    else{
      loc_m_interior_edges[id] = NULL;
      loc_m_exterior_edges[id] = NULL;
    }
  }

  loc_m_neighbor_points = new TYPE2*[ug->Give_number_points()];
  for(int id=0;id<ug->Give_number_points();++id) {
    if(ug->Give_point(id)->my_object(my_rank)) {
      int N_total = ug->Give_point(id)->give_number_neighbors_rec() * 4;
      loc_m_neighbor_points[id] = new TYPE2[N_total];
      for(int num=0;num<N_total;++num)
               loc_m_neighbor_points[id][num] = (TYPE2)0;
    }
    else loc_m_neighbor_points[id] = NULL;
  }
}

template <typename TYPE2>
Local_stiffness_matrix2D<TYPE2>::~Local_stiffness_matrix2D() {
  if(loc_m_rectangles != NULL){
    for(int id=0;id<blockgrid->Give_unstructured_grid()->Give_number_rectangle();++id)
       if(loc_m_rectangles[id] != NULL)
          delete [] loc_m_rectangles[id];
    delete [] loc_m_rectangles;
  }
  loc_m_rectangles = NULL;

 
  if(loc_m_interior_edges != NULL){
    for(int id=0;id<blockgrid->Give_unstructured_grid()->Give_number_edges();++id)
       if(loc_m_interior_edges[id] != NULL)
          delete [] loc_m_interior_edges[id];
    delete [] loc_m_interior_edges;
  }
  loc_m_interior_edges = NULL;

  if(loc_m_exterior_edges != NULL){
    for(int id=0;id<blockgrid->Give_unstructured_grid()->Give_number_edges();++id)
       if(loc_m_exterior_edges[id] != NULL)
          delete [] loc_m_exterior_edges[id];
    delete [] loc_m_exterior_edges;
  }
  loc_m_exterior_edges = NULL;


  if(loc_m_neighbor_points != NULL){
    for(int id=0;id<blockgrid->Give_unstructured_grid()->Give_number_points();++id)
      if(loc_m_neighbor_points[id] != NULL)
         delete [] loc_m_neighbor_points[id];
    delete [] loc_m_neighbor_points;
  }
  loc_m_neighbor_points = NULL;
}


#endif // DIFFOP2D_CC_H

