/**********************************************************************************
 * Copyright 2010 Christoph Pflaum 
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

#ifndef RESSTEN2D_CC_H_
#define RESSTEN2D_CC_H_

template <class DTyp, class Tc>
void Restrict2D_local_stiffness_matrix_2D(DTyp*  Stencil_coarse,
					DTyp** Stencils_fine,
					int* coarse_grid_functions) {
  int i,k,ii,jj,l,m;
  int* v_function;
  int* u_function;
  int u_f, v_f;

    // 1.Part: Set zero
    for (i = 0; i < 4; ++i) {
      for (k = 0; k < 4; ++k) {
        Stencil_coarse[k + 4 * i] = 0.0;
      }
    }
    // 2.Part: Iterate



    for (i = 0; i < 4; ++i) { // iteration for U over corners of coarse cell
      u_function = &(coarse_grid_functions[i * 9]);
      for (k = 0; k < 4; ++k) { // iteration for V over corners of coarse cell
        v_function = &(coarse_grid_functions[k * 9]);
        for (ii = 0; ii < 2; ++ii) { //iteration over fine cells in y-direction
          for (jj = 0; jj < 2; ++jj) { //iteration over fine cells in x-direction
            for (l = 0; l < 4; ++l) { //iteration over local basis functions for u
              for (m = 0; m < 4; ++m) { //iteration over local basis functions for v

                //get lokal index
                int ind_fine_u = 3 * (ii + l / 2) + l % 2 + jj;
                int ind_fine_v = 3 * (ii + m / 2) + m % 2 + jj;
                Stencil_coarse[4 * i + k] += 0.25 * u_function[ind_fine_u]
                    * 0.25 * v_function[ind_fine_v]
                    * Stencils_fine[2 * ii + jj][4 * l + m];
                //cout << ind_fine_u << "  " << ind_fine_v  << "  : " << u_function[ind_fine_u]*v_function[ind_fine_v] << "  " << Stencils_fine[2*ii+jj][4*l + m] <<  endl;
              }
            }
            //cout << ii << " " << jj << " " << i << " " << k << endl << endl << endl;
          }
        }
        //cout << Stencil_coarse[4 * i + k] << "  ";return;
      }
      //cout << endl;
    }

//    for (i = 0; i < 4; ++i) { // iteration for U over corners of coarse cell
//      for (k = 0; k < 4; ++k) {
//        cout << Stencil_coarse[4 * i + k] << "  ";
//      }
//      cout << endl;
//    }


//            for (l = 0; l < 4; ++l) { // iteration over corners of cell for U
//                u_f = u_function[trans_cellP_9((dir2D_sons) k, (dir2D_sons) l)];
//                if (u_f != 0)
//                  for (m = 0; m < 4; ++m) { // iteration over corners of cell for V
//                      v_f = v_function[trans_cellP_9((dir2D_sons) k,
//                          (dir2D_sons) m)];
//                      if (v_f != 0)
//
//                        for (ii = 0; ii < 2; ++ii)
//                          for (jj = 0; jj < 2; ++jj)
//                            Stencil_coarse[i][j] += 0.25 *
//                            Stencil_coarse[Tc::index(jj, j & 11)
//                                + 8 * Tc::index(ii, i & 1)] =
//                                Stencil_coarse[Tc::index(jj, j & 1)
//                                    + 8 * Tc::index(ii, i & 1)]
//                                    + (double) u_f * (double) v_f * 0.25
//                                        * Stencils_fine[k][Tc::index(jj, m & 1)
//                                            + 8 * Tc::index(ii, l & 1)];
//                    }
//              }
}


template <class DTyp, class Tc>
void Restrict2D_local_stiffness_matrix_1D(DTyp*  Stencil_coarse,
					DTyp** Stencils_fine) {
  int i,j,ii,jj,iii,jjj,k,l,m;
  int* v_function;
  int* u_function;
  int u_f, v_f;

  int coarse_grid_functions[2][3];

  coarse_grid_functions[0][0] = 2;
  coarse_grid_functions[0][1] = 1;
  coarse_grid_functions[0][2] = 0;

  coarse_grid_functions[1][2] = 2;
  coarse_grid_functions[1][1] = 1;
  coarse_grid_functions[1][0] = 0;

  // 1.Part: Set zero
  for(i=0;i<8;++i) for(j=0;j<8;++j) Stencil_coarse[j+8*i] = 0.0;

  // 2.Part: Iterate
  for(i=0;i<2;++i) { // iteration for U
    u_function = coarse_grid_functions[i];
    for(j=0;j<2;++j) { // iteration for V
      v_function = coarse_grid_functions[j];

      for(l=0;l<2;++l) { // iteration over corners of cell for U
    	  u_f = u_function[trans_cellP_3((dir1d)k,(dir1d)l)];
    	  if(u_f != 0)
    		  for(m=0;m<2;++m) { // iteration over corners of cell for V
    			  v_f = v_function[trans_cellP_3((dir1d)k,(dir1d)m)];
    			  if(v_f!=0)
    				  for(ii=0;ii<2;++ii)
    						  for(jj=0;jj<2;++jj)
    								  Stencil_coarse[Tc::index(j,jj)+8*Tc::index(i,ii)] =
    								  Stencil_coarse[Tc::index(j,jj)+8*Tc::index(i,ii)] +
    								  (double)u_f * (double)v_f * 0.25 *
    								  Stencils_fine[k][Tc::index(m,jj)+8*Tc::index(l,ii)];
    		  }
      }
    }
  }
}
#endif
