/**********************************************************************************
 * Copyright 2020 Christoph Pflaum
 * 		Department Informatik Lehrstuhl 10 - Systemsimulation
 *		Friedrich-Alexander Universität Erlangen-Nürnberg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **********************************************************************************/

//#define _USE_MATH_DEFINES
#include <cmath>
#include <complex>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
using std::vector;


#ifndef GMRESSYSTEM_H
#define GMRESSYSTEM_H

/**
 GMRES Löser für \f$ A x = b \f$
 **/
template<class DTyp, class Vari, class Mat>
  class GmResSystem : public IterativeSolverSystemAbstract <DTyp,Vari,Mat> {

  public:
    /** 
     * @param blockgrid Gitter auf dem die Gleichung gelösst werden soll
     **/
    GmResSystem(VariableVector<DTyp,Vari>* res);
    //~IterativeSolverAbstract() { };
    ~GmResSystem();

    /** 
     * set maximum size of Kryloc-space
     * @param numberRestart_ important GMRES parameter, which is default 100, changing this parameter changes memory allocation
     **/
    void setRestartNumber(int numberRestart_);
    
    /**
     * solves the linear equation system \f$ A x = b \f$
     * class Mat must contain member function   Mat::apply(x,b)
     **/  
    void solve(Mat& A, VariableVector<DTyp,Vari>& b, VariableVector<DTyp,Vari>& x);

    bool isResiduumCalculated() { return true; }    
  private:

    std::vector< VariableVector<DTyp,Vari>* > Q;

    int Restart;

    DTyp** H;     // size [Restart + 1][Restart]
    DTyp* C;      // size [Restart]
    DTyp* S;      // size [Restart]
    DTyp* y;      // size [Restart]    
    DTyp* gamma;  // size [Restart+1]

    static void GivensRotation(const complex<double>& x1, complex<double>& x2, complex<double>& c, complex<double>& s);    
    static void GivensRotation(const double& x1, double& x2, double& c, double& s);        
};


 
  
#endif
